@@ -1,27 +1,44 @@
 """Collection of utilities to detect properties of the underlying architecture."""
 
+from functools import cached_property
 from subprocess import PIPE, Popen, DEVNULL, run
-
-from cached_property import cached_property
-import cpuinfo
+from pathlib import Path
 import ctypes
-import numpy as np
-import psutil
 import re
 import os
 import sys
+import json
+
+import cpuinfo
+import numpy as np
+import psutil
 
 from devito.logger import warning
 from devito.tools import as_tuple, all_equal, memoized_func
 
 __all__ = ['platform_registry', 'get_cpu_info', 'get_gpu_info', 'get_nvidia_cc',
-           'get_cuda_path', 'check_cuda_runtime', 'get_m1_llvm_path',
-           'Platform', 'Cpu64', 'Intel64', 'Amd', 'Arm', 'Power', 'Device',
-           'NvidiaDevice', 'AmdDevice', 'IntelDevice',
-           'INTEL64', 'SNB', 'IVB', 'HSW', 'BDW', 'SKX', 'KNL', 'KNL7210',  # Intel
-           'AMD', 'ARM', 'M1',  # ARM
-           'POWER8', 'POWER9',  # Other loosely supported CPU architectures
-           'AMDGPUX', 'NVIDIAX', 'INTELGPUX']  # GPUs
+           'get_cuda_path', 'get_hip_path', 'check_cuda_runtime', 'get_m1_llvm_path',
+           'get_advisor_path', 'Platform', 'Cpu64', 'Intel64', 'IntelSkylake', 'Amd',
+           'Arm', 'Power', 'Device', 'NvidiaDevice', 'AmdDevice', 'IntelDevice',
+           # Brand-agnostic
+           'ANYCPU', 'ANYGPU',
+           # Intel CPUs
+           'INTEL64', 'SNB', 'IVB', 'HSW', 'BDW', 'KNL', 'KNL7210',
+           'SKX', 'KLX', 'CLX', 'CLK', 'SPR',
+           # AMD CPUs
+           'AMD',
+           # ARM CPUs
+           'ARM', 'AppleArm', 'M1', 'M2', 'M3',
+           'Graviton', 'GRAVITON2', 'GRAVITON3', 'GRAVITON4',
+           'Cortex', 'NvidiaArm', 'GRACE',
+           # Other legacy CPUs
+           'POWER8', 'POWER9',
+           # Generic GPUs
+           'AMDGPUX', 'NVIDIAX', 'INTELGPUX',
+           # Nvidia GPUs
+           'VOLTA', 'AMPERE', 'HOPPER', 'BLACKWELL',
+           # Intel GPUs
+           'PVC', 'INTELGPUMAX', 'MAX1100', 'MAX1550']
 
 
 @memoized_func
@@ -162,7 +179,6 @@ def get_gpu_info():
         homogeneous, otherwise None.
         """
         if gpu_infos == []:
-            warning('No graphics cards detected')
             return {}
 
         # Check must ignore physical IDs as they may differ
@@ -175,7 +191,7 @@ def get_gpu_info():
 
         warning('Different models of graphics cards detected')
 
-        return {}
+        return {'ncards': len(gpu_infos)}
 
     # Parse textual gpu info into a dict
 
@@ -213,7 +229,7 @@ def get_gpu_info():
         for i in ['total', 'free', 'used']:
             def make_cbk(i):
                 def cbk(deviceid=0):
-                    info_cmd = ['nvidia-smi', '--query-gpu=memory.%s' % i, '--format=csv']
+                    info_cmd = ['nvidia-smi', f'--query-gpu=memory.{i}', '--format=csv']
                     proc = Popen(info_cmd, stdout=PIPE, stderr=DEVNULL)
                     raw_info = str(proc.stdout.read())
 
@@ -235,14 +251,152 @@ def get_gpu_info():
 
                 return cbk
 
-            gpu_info['mem.%s' % i] = make_cbk(i)
+            gpu_info[f'mem.{i}'] = make_cbk(i)
 
         return gpu_info
 
     except OSError:
         pass
 
-    # *** Second try: `lshw`
+    # *** Second try: `rocm-smi`, clearly only works with AMD cards
+    try:
+        gpu_infos = {}
+
+        # Base gpu info
+        info_cmd = ['rocm-smi', '--showproductname']
+        proc = Popen(info_cmd, stdout=PIPE, stderr=DEVNULL)
+        raw_info = str(proc.stdout.read())
+
+        lines = raw_info.replace('\\n', '\n').replace('b\'', '').replace('\\t', '')
+        lines = lines.splitlines()
+
+        for line in lines:
+            if 'GPU' in line:
+                # Product
+                pattern = r'GPU\[(\d+)\].*?Card [sS]eries:\s*(.*?)\s*$'
+                match1 = re.match(pattern, line)
+
+                if match1:
+                    gid = match1.group(1)
+                    gpu_infos.setdefault(gid, dict())
+                    gpu_infos[gid]['physicalid'] = gid
+                    gpu_infos[gid]['product'] = match1.group(2)
+
+                # Model
+                pattern = r'GPU\[(\d+)\].*?Card [mM]odel:\s*(.*?)\s*$'
+                match2 = re.match(pattern, line)
+
+                if match2:
+                    gid = match2.group(1)
+                    gpu_infos.setdefault(gid, dict())
+                    gpu_infos[gid]['physicalid'] = match2.group(1)
+                    gpu_infos[gid]['model'] = match2.group(2)
+
+        gpu_info = homogenise_gpus(list(gpu_infos.values()))
+
+        # Also attach callbacks to retrieve instantaneous memory info
+        info_cmd = ['rocm-smi', '--showmeminfo', 'vram', '--json']
+        proc = Popen(info_cmd, stdout=PIPE, stderr=DEVNULL)
+        raw_info = str(proc.stdout.read())
+        lines = raw_info.replace('\\n', '').replace('b\'', '').replace('\'', '')
+        info = json.loads(lines)
+
+        for i in ['total', 'free', 'used']:
+            def make_cbk(i):
+                def cbk(deviceid=0):
+                    try:
+                        # Should only contain Used and total
+                        assert len(info[f'card{deviceid}']) == 2
+                        used = [int(v) for k, v in info[f'card{deviceid}'].items()
+                                if 'Used' in k][0]
+                        total = [int(v) for k, v in info[f'card{deviceid}'].items()
+                                 if 'Used' not in k][0]
+                        free = total - used
+                        return {'total': total, 'free': free, 'used': used}[i]
+                    except:
+                        # We shouldn't really end up here, unless nvidia-smi changes
+                        # the output format (though we still have tests in place that
+                        # will catch this)
+                        return None
+
+                return cbk
+
+            gpu_info[f'mem.{i}'] = make_cbk(i)
+
+        gpu_info['architecture'] = 'unspecified'
+        gpu_info['vendor'] = 'AMD'
+
+        return gpu_info
+
+    except (json.JSONDecodeError, OSError):
+        pass
+
+    # *** Third try: `sycl-ls`, clearly only works with Intel cards
+    try:
+        gpu_infos = {}
+
+        # sycl-ls sometimes finds gpu twice with opencl and without so
+        # we need to make sure we don't get duplicates
+        selected_platform = None
+        platform_block = ""
+
+        proc = Popen(["sycl-ls", "--verbose"], stdout=PIPE, stderr=DEVNULL, text=True)
+        sycl_output, _ = proc.communicate()
+
+        # Extract platform blocks
+        platforms = re.findall(r"Platform \[#(\d+)\]:([\s\S]*?)(?=Platform \[#\d+\]:|$)",
+                               sycl_output)
+
+        # Select Level-Zero if available, otherwise use OpenCL
+        for platform_id, platform_content in platforms:
+            if "Intel(R) Level-Zero" in platform_content:
+                selected_platform = platform_id
+                platform_block = platform_content
+                break
+            elif "Intel(R) OpenCL Graphics" in platform_content and \
+                    selected_platform is None:
+                selected_platform = platform_id
+                platform_block = platform_content
+
+        # Extract GPU devices from the selected platform
+        devices = re.findall(r"Device \[#(\d+)\]:([\s\S]*?)(?=Device \[#\d+\]:|$)",
+                             platform_block)
+
+        for device_id, device_block in devices:
+            if re.search(r"^\s*Type\s*:\s*gpu", device_block, re.MULTILINE):
+                name_match = re.search(r"^\s*Name\s*:\s*(.+)", device_block, re.MULTILINE)
+
+                if name_match:
+                    name = name_match.group(1).strip()
+
+                    # Store GPU info with correct physical ID
+                    gpu_infos[device_id] = {
+                        "physicalid": device_id,
+                        "product": name
+                    }
+
+        gpu_info = homogenise_gpus(list(gpu_infos.values()))
+
+        # Also attach callbacks to retrieve instantaneous memory info
+        # Now this should be done using xpu-smi but for some reason
+        # it throws a lot of weird errors in docker so skipping for now
+        for i in ['total', 'free', 'used']:
+            def make_cbk(i):
+                def cbk(deviceid=0):
+                    return None
+                return cbk
+
+            gpu_info['mem.%s' % i] = make_cbk(i)
+
+        gpu_info['architecture'] = 'unspecified'
+        gpu_info['vendor'] = 'INTEL'
+
+        return gpu_info
+
+    except OSError:
+        pass
+
+    # *** Fourth try: `lshw`
     try:
         info_cmd = ['lshw', '-C', 'video']
         proc = Popen(info_cmd, stdout=PIPE, stderr=DEVNULL)
@@ -287,7 +441,7 @@ def get_gpu_info():
     except OSError:
         pass
 
-    # Third try: `lspci`, which is more readable but less detailed than `lshw`
+    # Fifth try: `lspci`, which is more readable but less detailed than `lshw`
     try:
         info_cmd = ['lspci']
         proc = Popen(info_cmd, stdout=PIPE, stderr=DEVNULL)
@@ -303,7 +457,7 @@ def get_gpu_info():
         gpu_infos = []
         for line in lines:
             # Graphics cards are listed as VGA or 3D controllers in lspci
-            if 'VGA' in line or '3D' in line:
+            if any(i in line for i in ('VGA', '3D', 'Display')):
                 gpu_info = {}
                 # Lines produced by lspci command are of the form:
                 #   xxxx:xx:xx.x Device Type: Name
@@ -377,6 +531,38 @@ def get_cuda_path():
     return None
 
 
+@memoized_func
+def get_advisor_path():
+    """
+    Detect if Intel Advisor is installed on the machine and return
+    its location if it is.
+    """
+    path = None
+
+    env_path = os.environ["PATH"]
+    env_path_dirs = env_path.split(":")
+
+    for env_path_dir in env_path_dirs:
+        # intel/oneapi/advisor is the directory for Intel oneAPI
+        if "intel/advisor" in env_path_dir or "intel/oneapi/advisor" in env_path_dir:
+            path = Path(env_path_dir)
+            if path.name.startswith('bin'):
+                return path.parent
+
+    return path
+
+
+@memoized_func
+def get_hip_path():
+    # *** First try: via commonly used environment variables
+    for i in ['HIP_HOME']:
+        hip_home = os.environ.get(i)
+        if hip_home:
+            return hip_home
+
+    return None
+
+
 @memoized_func
 def get_m1_llvm_path(language):
     # Check if Apple's llvm is installed (installable via Homebrew), which supports
@@ -483,7 +669,7 @@ def get_platform():
             if 'phi' in brand:
                 # Intel Xeon Phi?
                 return platform_registry['knl']
-            # Unknown Xeon ? May happen on some virtualizes systems...
+            # Unknown Xeon ? May happen on some virtualized systems...
             return platform_registry['intel64']
         elif 'intel' in brand:
             # Most likely a desktop i3/i5/i7
@@ -502,14 +688,32 @@ def get_platform():
         pass
 
     # Unable to detect platform. Stick to default...
-    return CPU64
+    return ANYCPU
 
 
-class Platform(object):
+class Platform:
+
+    registry = {}
+    """
+    The Platform registry.
+
+    Each new Platform instance is automatically added to the registry.
+    """
+
+    max_mem_trans_nbytes = None
+    """Maximum memory transaction size in bytes."""
 
     def __init__(self, name):
         self.name = name
 
+        self.registry[name] = self
+
+    def __eq__(self, other):
+        return isinstance(other, Platform) and self.name == other.name
+
+    def __hash__(self):
+        return hash(self.name)
+
     @classmethod
     def _mro(cls):
         return [Platform]
@@ -526,19 +730,20 @@ class Platform(object):
     def _detect_isa(self):
         return 'unknown'
 
+    @property
+    def numa_domains(self):
+        """
+        Number of NUMA domains, or None if unknown.
+        """
+        return 1
+
     @property
     def threads_per_core(self):
         return self.cores_logical // self.cores_physical
 
     @property
-    def simd_reg_size(self):
-        """Size in bytes of a SIMD register."""
-        return isa_registry.get(self.isa, 0)
-
-    def simd_items_per_reg(self, dtype):
-        """Number of items of type ``dtype`` that can fit in a SIMD register."""
-        assert self.simd_reg_size % np.dtype(dtype).itemsize == 0
-        return int(self.simd_reg_size / np.dtype(dtype).itemsize)
+    def cores_physical_per_numa_domain(self):
+        return self.cores_physical // self.numa_domains
 
     @property
     def memtotal(self):
@@ -549,9 +754,46 @@ class Platform(object):
         """Available physical memory in bytes, or None if unknown."""
         return None
 
+    def max_mem_trans_size(self, dtype):
+        """
+        Number of items of type `dtype` that can be transferred in a single
+        memory transaction.
+        """
+        itemsize = np.dtype(dtype).itemsize
+
+        # NOTE: This method conservatively uses the node's `max_mem_trans_size`,
+        # instead of self's, so that we always pad by a compatible amount should
+        # the user switch target platforms dynamically
+        mmtb = node_max_mem_trans_nbytes(self)
+        assert mmtb % itemsize == 0
+
+        return int(mmtb / itemsize)
+
+    def limits(self, compiler=None, language=None):
+        """
+        Return the architecture-specific limits for the given compiler and
+        language.
+        """
+        return {
+            'max-par-dims': sys.maxsize,
+            'max-block-dims': sys.maxsize,
+        }
+
+    def supports(self, query, language=None):
+        """
+        Return True if the platform supports a given feature, False otherwise.
+        """
+        return False
+
 
 class Cpu64(Platform):
 
+    # The vast majority of CPUs have a 64-byte cache line size
+    max_mem_trans_nbytes = 64
+
+    # The known ISAs are to be provided by the subclasses
+    known_isas = ()
+
     def __init__(self, name, cores_logical=None, cores_physical=None, isa=None):
         super().__init__(name)
 
@@ -561,9 +803,6 @@ class Cpu64(Platform):
         self.cores_physical = cores_physical or cpu_info['physical']
         self.isa = isa or self._detect_isa()
 
-    # The known ISAs are to be provided by the subclasses
-    known_isas = ()
-
     @classmethod
     def _mro(cls):
         # Retain only the CPU Platforms
@@ -583,6 +822,28 @@ class Cpu64(Platform):
                 return i
         return 'cpp'
 
+    @property
+    def simd_reg_nbytes(self):
+        """
+        Size in bytes of a SIMD register.
+        """
+        return isa_registry.get(self.isa, 0)
+
+    def simd_items_per_reg(self, dtype):
+        """
+        Number of items of type `dtype` that fit in a SIMD register.
+        """
+        assert self.simd_reg_nbytes % np.dtype(dtype).itemsize == 0
+        return int(self.simd_reg_nbytes / np.dtype(dtype).itemsize)
+
+    @cached_property
+    def numa_domains(self):
+        try:
+            return int(lscpu()['NUMA node(s)'])
+        except (ValueError, TypeError, KeyError):
+            warning("NUMA domain count autodetection failed, assuming 1")
+            return 1
+
     @cached_property
     def memtotal(self):
         return psutil.virtual_memory().total
@@ -596,11 +857,71 @@ class Intel64(Cpu64):
     known_isas = ('cpp', 'sse', 'avx', 'avx2', 'avx512')
 
 
+class IntelSkylake(Intel64):
+    pass
+
+
+class IntelGoldenCove(Intel64):
+    pass
+
+
 class Arm(Cpu64):
 
     known_isas = ('fp', 'asimd', 'asimdrdm')
 
 
+class AppleArm(Arm):
+
+    @cached_property
+    def march(self):
+        sysinfo = run(["sysctl", "-n", "machdep.cpu.brand_string"],
+                      stdout=PIPE, stderr=DEVNULL).stdout.decode("utf-8")
+        mx = sysinfo.split(' ')[1].lower()
+        # Currently clang only supports up to m2
+        return min(mx, 'm2')
+
+
+class Graviton(Arm):
+
+    @property
+    def version(self):
+        return int(self.name.split('graviton')[-1])
+
+    @cached_property
+    def march(self):
+        if self.version >= 4:
+            return 'neoverse-v2'
+        elif self.version == 3:
+            return 'neoverse-v1'
+        else:
+            return 'neoverse-n1'
+
+
+class Cortex(Arm):
+
+    @property
+    def version(self):
+        return int(self.name.split('cortexa')[-1])
+
+    @cached_property
+    def march(self):
+        return 'armv8-a+crc+simd'
+
+    @cached_property
+    def mtune(self):
+        return f'cortex-a{self.version}'
+
+
+class NvidiaArm(Arm):
+
+    @cached_property
+    def march(self):
+        if self.name == 'grace':
+            return 'neoverse-v2'
+        else:
+            return 'native'
+
+
 class Amd(Cpu64):
 
     known_isas = ('cpp', 'sse', 'avx', 'avx2')
@@ -614,13 +935,24 @@ class Power(Cpu64):
 
 class Device(Platform):
 
-    def __init__(self, name, cores_logical=1, cores_physical=1, isa='cpp'):
+    def __init__(self, name, cores_logical=None, cores_physical=None, isa='cpp',
+                 max_threads_per_block=1024, max_threads_dimx=1024,
+                 max_threads_dimy=1024, max_threads_dimz=64,
+                 max_thread_block_cluster_size=8):
         super().__init__(name)
 
-        self.cores_logical = cores_logical
-        self.cores_physical = cores_physical
+        cpu_info = get_cpu_info()
+
+        self.cores_logical = cores_logical or cpu_info['logical']
+        self.cores_physical = cores_physical or cpu_info['physical']
         self.isa = isa
 
+        self.max_threads_per_block = max_threads_per_block
+        self.max_threads_dimx = max_threads_dimx
+        self.max_threads_dimy = max_threads_dimy
+        self.max_threads_dimz = max_threads_dimz
+        self.max_thread_block_cluster_size = max_thread_block_cluster_size
+
     @classmethod
     def _mro(cls):
         # Retain only the Device Platforms
@@ -632,10 +964,18 @@ class Device(Platform):
                 break
         return retval
 
-    @cached_property
+    @property
     def march(self):
         return None
 
+    @cached_property
+    def numa_domains(self):
+        info = get_gpu_info()
+        try:
+            return info['ncards']
+        except KeyError:
+            return None
+
     @cached_property
     def memtotal(self):
         info = get_gpu_info()
@@ -654,16 +994,31 @@ class Device(Platform):
         except (AttributeError, KeyError):
             return None
 
+    def limits(self, compiler=None, language=None):
+        return {
+            'max-par-dims': 3,
+            'max-block-dims': 3,
+        }
+
 
 class IntelDevice(Device):
 
-    @cached_property
+    max_mem_trans_nbytes = 64
+
+    def __init__(self, *args, sub_group_size=32, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        self.sub_group_size = sub_group_size
+
+    @property
     def march(self):
         return ''
 
 
 class NvidiaDevice(Device):
 
+    max_mem_trans_nbytes = 128
+
     @cached_property
     def march(self):
         info = get_gpu_info()
@@ -673,9 +1028,60 @@ class NvidiaDevice(Device):
                 return 'tesla'
         return None
 
+    def supports(self, query, language=None):
+        if language != 'cuda':
+            return False
+
+        cc = get_nvidia_cc()
+        if cc is None:
+            # Something off with `get_nvidia_cc` on this system
+            warning(f"Couldn't establish if `query={query}` is supported on this "
+                    "system. Assuming it is not.")
+            return False
+        elif query == 'async-loads' and cc >= 80:
+            # Asynchronous pipeline loads -- introduced in Ampere
+            return True
+        elif query in ('tma', 'thread-block-cluster') and cc >= 90:
+            # Tensor Memory Accelerator -- introduced in Hopper
+            return True
+        else:
+            return False
+
+
+class Volta(NvidiaDevice):
+    pass
+
+
+class Ampere(Volta):
+
+    def supports(self, query, language=None):
+        if query == 'async-loads':
+            return True
+        else:
+            return super().supports(query, language)
+
+
+class Hopper(Ampere):
+
+    def __init__(self, *args, **kwargs):
+        kwargs.setdefault('max_thread_block_cluster_size', 16)
+        super().__init__(*args, **kwargs)
+
+    def supports(self, query, language=None):
+        if query in ('tma', 'thread-block-cluster'):
+            return True
+        else:
+            return super().supports(query, language)
+
+
+class Blackwell(Hopper):
+    pass
+
 
 class AmdDevice(Device):
 
+    max_mem_trans_nbytes = 256
+
     @cached_property
     def march(cls):
         # TODO: this corresponds to Vega, which acts as the fallback `march`
@@ -690,8 +1096,12 @@ class AmdDevice(Device):
         #     mygpu will only print values accepted by cuda clang in
         #     the clang argument --cuda-gpu-arch.
         try:
-            p1 = Popen(['mygpu', '-d', 'gfx900'], stdout=PIPE, stderr=PIPE)
+            p1 = Popen(['offload-arch'], stdout=PIPE, stderr=PIPE)
         except OSError:
+            try:
+                p1 = Popen(['mygpu', '-d', fallback], stdout=PIPE, stderr=PIPE)
+            except OSError:
+                pass
             return fallback
 
         output, _ = p1.communicate()
@@ -701,23 +1111,69 @@ class AmdDevice(Device):
             return fallback
 
 
+@memoized_func
+def node_max_mem_trans_nbytes(platform):
+    """
+    Return the maximum memory transaction size in bytes for the underlying
+    node, which, as such, takes into account all available platforms.
+    """
+    mmtb0 = platform.max_mem_trans_nbytes
+
+    if isinstance(platform, Cpu64):
+        gpu_info = get_gpu_info()
+        if not gpu_info:
+            # This node may simply not have a GPU
+            return mmtb0
+
+        mapper = {
+            'NVIDIA': NvidiaDevice,
+            'AMD': AmdDevice,
+            'INTEL': IntelDevice,
+        }
+        try:
+            mmtb1 = mapper[gpu_info['vendor']].max_mem_trans_nbytes
+            return max(mmtb0, mmtb1)
+        except KeyError:
+            # Fallback -- act even more conservatively
+            mmtb1 = max(p.max_mem_trans_nbytes for p in
+                        [NvidiaDevice, AmdDevice, IntelDevice])
+            mmtb = max(mmtb0, mmtb1)
+            warning("Unable to determine GPU type, assuming a maximum memory "
+                    f"transaction size of {mmtb} bytes")
+            return mmtb
+    elif isinstance(platform, Device):
+        return max(Cpu64.max_mem_trans_nbytes, mmtb0)
+    else:
+        assert False, f"Unknown platform type: {type(platform)}"
+
+
 # CPUs
-CPU64 = Cpu64('cpu64')
+ANYCPU = Cpu64('cpu64')
 CPU64_DUMMY = Intel64('cpu64-dummy', cores_logical=2, cores_physical=1, isa='sse')
 
 INTEL64 = Intel64('intel64')
-SNB = Intel64('snb')
-IVB = Intel64('ivb')
-HSW = Intel64('hsw')
-BDW = Intel64('bdw', isa='avx2')
-SKX = Intel64('skx')
-KLX = Intel64('klx')
-CLX = Intel64('clx')
-KNL = Intel64('knl')
-KNL7210 = Intel64('knl', cores_logical=256, cores_physical=64, isa='avx512')
+SNB = Intel64('snb')  # Sandy Bridge
+IVB = Intel64('ivb')  # Ivy Bridge
+HSW = Intel64('hsw')  # Haswell
+BDW = Intel64('bdw', isa='avx2')  # Broadwell
+KNL = Intel64('knl')  # Knights Landing
+KNL7210 = Intel64('knl7210', cores_logical=256, cores_physical=64, isa='avx512')
+SKX = IntelSkylake('skx')  # Skylake
+KLX = IntelSkylake('klx')  # Kaby Lake
+CLX = IntelSkylake('clx')  # Coffee Lake
+CLK = IntelSkylake('clk')  # Cascade Lake
+SPR = IntelGoldenCove('spr')  # Sapphire Rapids
 
 ARM = Arm('arm')
-M1 = Arm('m1')
+GRAVITON2 = Graviton('graviton2')
+GRAVITON3 = Graviton('graviton3')
+GRAVITON4 = Graviton('graviton4')
+M1 = AppleArm('m1')
+M2 = AppleArm('m2')
+M3 = AppleArm('m3')
+CORTEX = Cortex('cortex')
+CORTEXA76 = Cortex('cortexa76')
+GRACE = NvidiaArm('grace')
 
 AMD = Amd('amd')
 
@@ -725,39 +1181,25 @@ POWER8 = Power('power8')
 POWER9 = Power('power9')
 
 # Devices
+ANYGPU = Cpu64('gpu')
+
 NVIDIAX = NvidiaDevice('nvidiaX')
+VOLTA = Volta('volta')
+AMPERE = Ampere('ampere')
+HOPPER = Hopper('hopper')
+BLACKWELL = Blackwell('blackwell')
+
 AMDGPUX = AmdDevice('amdgpuX')
+
 INTELGPUX = IntelDevice('intelgpuX')
+PVC = IntelDevice('pvc')  # Legacy codename for MAX GPUs
+INTELGPUMAX = IntelDevice('intelgpuMAX')
+MAX1100 = IntelDevice('max1100')
+MAX1550 = IntelDevice('max1550')
 
-
-platform_registry = {
-    'cpu64-dummy': CPU64_DUMMY,
-    'intel64': INTEL64,
-    'snb': SNB,  # Sandy Bridge
-    'ivb': IVB,  # Ivy Bridge
-    'hsw': HSW,  # Haswell
-    'bdw': BDW,  # Broadwell
-    'skx': SKX,  # Skylake
-    'klx': KLX,  # Kaby Lake
-    'clx': CLX,  # Coffee Lake
-    'knl': KNL,
-    'knl7210': KNL7210,
-    'arm': ARM,  # Generic ARM CPU
-    'm1': M1,
-    'amd': AMD,  # Generic AMD CPU
-    'power8': POWER8,
-    'power9': POWER9,
-    'nvidiaX': NVIDIAX,  # Generic NVidia GPU
-    'amdgpuX': AMDGPUX,   # Generic AMD GPU
-    'intelgpuX': INTELGPUX   # Generic Intel GPU
-}
-"""
-Registry dict for deriving Platform classes according to the environment variable
-DEVITO_PLATFORM. Developers should add new platform classes here.
-"""
+platform_registry = Platform.registry
 platform_registry['cpu64'] = get_platform  # Autodetection
 
-
 isa_registry = {
     'cpp': 16,
     'sse': 16,
