@@ -4,68 +4,35 @@ import numpy as np
 
 from devito.core.operator import CoreOperator, CustomOperator, ParTile
 from devito.exceptions import InvalidOperator
+from devito.operator.operator import rcompile
+from devito.passes import is_on_device, stream_dimensions
 from devito.passes.equations import collect_derivatives
-from devito.passes.clusters import (Lift, Streaming, Tasker, blocking, buffering,
-                                    cire, cse, factorize, fission, fuse,
+from devito.passes.clusters import (Lift, tasking, memcpy_prefetch, blocking,
+                                    buffering, cire, cse, factorize, fission, fuse,
                                     optimize_pows)
-from devito.passes.iet import (DeviceOmpTarget, DeviceAccTarget, mpiize, hoist_prodders,
-                               is_on_device, linearize, pthreadify, relax_incr_dimensions)
+from devito.passes.iet import (DeviceOmpTarget, DeviceAccTarget, DeviceCXXOmpTarget,
+                               mpiize, hoist_prodders, linearize, pthreadify,
+                               relax_incr_dimensions, check_stability)
 from devito.tools import as_tuple, timed_pass
 
 __all__ = ['DeviceNoopOperator', 'DeviceAdvOperator', 'DeviceCustomOperator',
            'DeviceNoopOmpOperator', 'DeviceAdvOmpOperator', 'DeviceFsgOmpOperator',
            'DeviceCustomOmpOperator', 'DeviceNoopAccOperator', 'DeviceAdvAccOperator',
-           'DeviceFsgAccOperator', 'DeviceCustomAccOperator']
+           'DeviceFsgAccOperator', 'DeviceCustomAccOperator', 'DeviceNoopCXXOmpOperator',
+           'DeviceAdvCXXOmpOperator', 'DeviceFsgCXXOmpOperator',
+           'DeviceCustomCXXOmpOperator']
 
 
-class DeviceOperatorMixin(object):
+class DeviceOperatorMixin:
 
     BLOCK_LEVELS = 0
-    """
-    Loop blocking depth. So, 1 => "blocks", 2 => "blocks" and "sub-blocks",
-    3 => "blocks", "sub-blocks", and "sub-sub-blocks", ...
-    """
-
-    BLOCK_EAGER = True
-    """
-    Apply loop blocking as early as possible, and in particular prior to CIRE.
-    """
-
-    BLOCK_RELAX = False
-    """
-    If set to True, bypass the compiler heuristics that prevent loop blocking in
-    situations where the performance impact might be detrimental.
-    """
-
-    CIRE_MINGAIN = 10
-    """
-    Minimum operation count reduction for a redundant expression to be optimized
-    away. Higher (lower) values make a redundant expression less (more) likely to
-    be optimized away.
-    """
-
-    CIRE_SCHEDULE = 'automatic'
-    """
-    Strategy used to schedule derivatives across loops. This impacts the operational
-    intensity of the generated kernel.
-    """
-
-    PAR_CHUNK_NONAFFINE = 3
-    """
-    Coefficient to adjust the chunk size in non-affine parallel loops.
-    """
+    MPI_MODES = (True, 'basic',)
 
     GPU_FIT = 'all-fallback'
     """
     Assuming all functions fit into the gpu memory.
     """
 
-    MAPIFY_REDUCE = False
-    """
-    Vector-expand all scalar reductions to turn them into explicit map-reductions,
-    which may be easier to parallelize for certain backends.
-    """
-
     @classmethod
     def _normalize_kwargs(cls, **kwargs):
         o = {}
@@ -77,10 +44,16 @@ class DeviceOperatorMixin(object):
 
         # Buffering
         o['buf-async-degree'] = oo.pop('buf-async-degree', None)
+        o['buf-reuse'] = oo.pop('buf-reuse', None)
 
         # Fusion
         o['fuse-tasks'] = oo.pop('fuse-tasks', False)
 
+        # Flops minimization
+        o['cse-min-cost'] = oo.pop('cse-min-cost', cls.CSE_MIN_COST)
+        o['cse-algo'] = oo.pop('cse-algo', cls.CSE_ALGO)
+        o['fact-schedule'] = oo.pop('fact-schedule', cls.FACT_SCHEDULE)
+
         # Blocking
         o['blockinner'] = oo.pop('blockinner', True)
         o['blocklevels'] = oo.pop('blocklevels', cls.BLOCK_LEVELS)
@@ -98,19 +71,34 @@ class DeviceOperatorMixin(object):
         o['cire-schedule'] = oo.pop('cire-schedule', cls.CIRE_SCHEDULE)
 
         # GPU parallelism
-        o['par-tile'] = ParTile(oo.pop('par-tile', False), default=(32, 4))
+        o['par-tile'] = ParTile(oo.pop('par-tile', False), default=(32, 4, 4),
+                                sparse=oo.pop('par-tile-sparse', None),
+                                reduce=oo.pop('par-tile-reduce', None))
         o['par-collapse-ncores'] = 1  # Always collapse (meaningful if `par-tile=False`)
         o['par-collapse-work'] = 1  # Always collapse (meaningful if `par-tile=False`)
         o['par-chunk-nonaffine'] = oo.pop('par-chunk-nonaffine', cls.PAR_CHUNK_NONAFFINE)
         o['par-dynamic-work'] = np.inf  # Always use static scheduling
         o['par-nested'] = np.inf  # Never use nested parallelism
         o['par-disabled'] = oo.pop('par-disabled', True)  # No host parallelism by default
-        o['gpu-fit'] = as_tuple(oo.pop('gpu-fit', cls._normalize_gpu_fit(**kwargs)))
+        o['gpu-fit'] = cls._normalize_gpu_fit(oo, **kwargs)
+        o['gpu-create'] = as_tuple(oo.pop('gpu-create', ()))
+
+        # Distributed parallelism
+        o['dist-drop-unwritten'] = oo.pop('dist-drop-unwritten', cls.DIST_DROP_UNWRITTEN)
+
+        # Code generation options for derivatives
+        o['expand'] = oo.pop('expand', cls.EXPAND)
+        o['deriv-schedule'] = oo.pop('deriv-schedule', cls.DERIV_SCHEDULE)
+        o['deriv-unroll'] = oo.pop('deriv-unroll', False)
 
         # Misc
-        o['optcomms'] = oo.pop('optcomms', True)
+        o['opt-comms'] = oo.pop('opt-comms', True)
         o['linearize'] = oo.pop('linearize', False)
         o['mapify-reduce'] = oo.pop('mapify-reduce', cls.MAPIFY_REDUCE)
+        o['index-mode'] = oo.pop('index-mode', cls.INDEX_MODE)
+        o['place-transfers'] = oo.pop('place-transfers', True)
+        o['errctl'] = oo.pop('errctl', cls.ERRCTL)
+        o['scalar-min-type'] = oo.pop('scalar-min-type', cls.SCALAR_MIN_TYPE)
 
         if oo:
             raise InvalidOperator("Unsupported optimization options: [%s]"
@@ -121,11 +109,52 @@ class DeviceOperatorMixin(object):
         return kwargs
 
     @classmethod
-    def _normalize_gpu_fit(cls, **kwargs):
-        if any(i in kwargs['mode'] for i in ['tasking', 'streaming']):
-            return None
-        else:
-            return cls.GPU_FIT
+    def _normalize_gpu_fit(cls, oo, **kwargs):
+        try:
+            gfit = as_tuple(oo.pop('gpu-fit'))
+            gfit = set().union(*[f.values() if f.is_AbstractTensor else [f]
+                                 for f in gfit])
+            return tuple(gfit)
+        except KeyError:
+            if any(i in kwargs['mode'] for i in ['tasking', 'streaming']):
+                return (None,)
+            else:
+                return as_tuple(cls.GPU_FIT)
+
+    @classmethod
+    def _rcompile_wrapper(cls, **kwargs0):
+        options0 = kwargs0.pop('options')
+
+        def wrapper(expressions, mode='default', options=None, **kwargs1):
+            kwargs = {**kwargs0, **kwargs1}
+            options = options or {}
+
+            if mode == 'host':
+                target = {
+                    'platform': 'cpu64',
+                    'language': 'C' if options0['par-disabled'] else 'openmp',
+                    'compiler': 'custom'
+                }
+            else:
+                # Always use the default `par-tile` for recursive compilation
+                # unless the caller explicitly overrides it so that if the user
+                # supplies a multi par-tile there is no need to worry about the
+                # small kernels typically generated by recursive compilation
+                par_tile0 = options0['par-tile']
+                par_tile = options.get('par-tile')
+                if par_tile0 and par_tile:
+                    options = {**options0, **options, 'par-tile': par_tile}
+                elif par_tile0:
+                    par_tile = ParTile(par_tile0.default, default=par_tile0.default)
+                    options = {**options0, **options, 'par-tile': par_tile}
+                else:
+                    options = {**options0, **options}
+
+                target = None
+
+            return rcompile(expressions, kwargs, options, target=target)
+
+        return wrapper
 
 # Mode level
 
@@ -149,7 +178,7 @@ class DeviceNoopOperator(DeviceOperatorMixin, CoreOperator):
         parizer.initialize(graph, options=options)
 
         # Symbol definitions
-        cls._Target.DataManager(sregistry, options).process(graph)
+        cls._Target.DataManager(**kwargs).process(graph)
 
         return graph
 
@@ -186,14 +215,14 @@ class DeviceAdvOperator(DeviceOperatorMixin, CoreOperator):
 
         # Reduce flops
         clusters = cire(clusters, 'sops', sregistry, options, platform)
-        clusters = factorize(clusters)
+        clusters = factorize(clusters, **kwargs)
         clusters = optimize_pows(clusters)
 
         # The previous passes may have created fusion opportunities
         clusters = fuse(clusters)
 
         # Reduce flops
-        clusters = cse(clusters, sregistry)
+        clusters = cse(clusters, **kwargs)
 
         # Blocking to define thread blocks
         if options['blocklazy']:
@@ -213,7 +242,7 @@ class DeviceAdvOperator(DeviceOperatorMixin, CoreOperator):
         mpiize(graph, **kwargs)
 
         # Lower BlockDimensions so that blocks of arbitrary shape may be used
-        relax_incr_dimensions(graph)
+        relax_incr_dimensions(graph, **kwargs)
 
         # GPU parallelism
         parizer = cls._Target.Parizer(sregistry, options, platform, compiler)
@@ -223,11 +252,14 @@ class DeviceAdvOperator(DeviceOperatorMixin, CoreOperator):
         # Misc optimizations
         hoist_prodders(graph)
 
+        # Perform error checking
+        check_stability(graph, **kwargs)
+
         # Symbol definitions
-        cls._Target.DataManager(sregistry, options).process(graph)
+        cls._Target.DataManager(**kwargs).process(graph)
 
         # Linearize n-dimensional Indexeds
-        linearize(graph, mode=options['linearize'], sregistry=sregistry)
+        linearize(graph, **kwargs)
 
         return graph
 
@@ -257,28 +289,21 @@ class DeviceCustomOperator(DeviceOperatorMixin, CustomOperator):
         platform = kwargs['platform']
         sregistry = kwargs['sregistry']
 
-        # Callbacks used by `Tasking` and `Streaming`
-        runs_on_host, reads_if_on_host = make_callbacks(options)
-
-        # Callback used by `buffering`
-        def callback(f):
-            if not is_on_device(f, options['gpu-fit']):
-                return [f.time_dim]
-            else:
-                return None
+        callback = lambda f: not is_on_device(f, options['gpu-fit'])
+        stream_key = stream_wrap(callback)
 
         return {
-            'buffering': lambda i: buffering(i, callback, sregistry, options),
             'blocking': lambda i: blocking(i, sregistry, options),
-            'tasking': Tasker(runs_on_host, sregistry).process,
-            'streaming': Streaming(reads_if_on_host, sregistry).process,
+            'buffering': lambda i: buffering(i, stream_key, sregistry, options),
+            'tasking': lambda i: tasking(i, stream_key, sregistry),
+            'streaming': lambda i: memcpy_prefetch(i, stream_key, sregistry),
             'factorize': factorize,
             'fission': fission,
             'fuse': lambda i: fuse(i, options=options),
             'lift': lambda i: Lift().process(cire(i, 'invariants', sregistry,
                                                   options, platform)),
             'cire-sops': lambda i: cire(i, 'sops', sregistry, options, platform),
-            'cse': lambda i: cse(i, sregistry),
+            'cse': lambda i: cse(i, sregistry, options),
             'opt-pows': optimize_pows,
             'topofuse': lambda i: fuse(i, toposort=True, options=options)
         }
@@ -291,15 +316,14 @@ class DeviceCustomOperator(DeviceOperatorMixin, CustomOperator):
         sregistry = kwargs['sregistry']
 
         parizer = cls._Target.Parizer(sregistry, options, platform, compiler)
-        orchestrator = cls._Target.Orchestrator(sregistry)
+        orchestrator = cls._Target.Orchestrator(**kwargs)
 
         return {
             'parallel': parizer.make_parallel,
             'orchestrate': partial(orchestrator.process),
             'pthreadify': partial(pthreadify, sregistry=sregistry),
             'mpi': partial(mpiize, **kwargs),
-            'linearize': partial(linearize, mode=options['linearize'],
-                                 sregistry=sregistry),
+            'linearize': partial(linearize, **kwargs),
             'prodders': partial(hoist_prodders),
             'init': partial(parizer.initialize, options=options)
         }
@@ -323,7 +347,7 @@ class DeviceCustomOperator(DeviceOperatorMixin, CustomOperator):
 
 # OpenMP
 
-class DeviceOmpOperatorMixin(object):
+class DeviceOmpOperatorMixin:
 
     _Target = DeviceOmpTarget
 
@@ -344,19 +368,41 @@ class DeviceOmpOperatorMixin(object):
 
         return kwargs
 
+    @classmethod
+    def _check_kwargs(cls, **kwargs):
+        oo = kwargs['options']
+
+        if len(oo['gpu-create']):
+            raise InvalidOperator("Unsupported gpu-create option for omp operators")
+
 
 class DeviceNoopOmpOperator(DeviceOmpOperatorMixin, DeviceNoopOperator):
     pass
 
 
+class DeviceNoopCXXOmpOperator(DeviceNoopOmpOperator):
+    _Target = DeviceCXXOmpTarget
+    LINEARIZE = True
+
+
 class DeviceAdvOmpOperator(DeviceOmpOperatorMixin, DeviceAdvOperator):
     pass
 
 
+class DeviceAdvCXXOmpOperator(DeviceAdvOmpOperator):
+    _Target = DeviceCXXOmpTarget
+    LINEARIZE = True
+
+
 class DeviceFsgOmpOperator(DeviceOmpOperatorMixin, DeviceFsgOperator):
     pass
 
 
+class DeviceFsgCXXOmpOperator(DeviceFsgOmpOperator):
+    _Target = DeviceCXXOmpTarget
+    LINEARIZE = True
+
+
 class DeviceCustomOmpOperator(DeviceOmpOperatorMixin, DeviceCustomOperator):
 
     _known_passes = DeviceCustomOperator._known_passes + ('openmp',)
@@ -369,9 +415,14 @@ class DeviceCustomOmpOperator(DeviceOmpOperatorMixin, DeviceCustomOperator):
         return mapper
 
 
+class DeviceCustomCXXOmpOperator(DeviceCustomOmpOperator):
+    _Target = DeviceCXXOmpTarget
+    LINEARIZE = True
+
+
 # OpenACC
 
-class DeviceAccOperatorMixin(object):
+class DeviceAccOperatorMixin:
 
     _Target = DeviceAccTarget
 
@@ -410,26 +461,21 @@ class DeviceCustomAccOperator(DeviceAccOperatorMixin, DeviceCustomOperator):
     assert not (set(_known_passes) & set(DeviceCustomOperator._known_passes_disabled))
 
 
-# Utils
+# *** Utils
 
-
-def make_callbacks(options):
-    """
-    Options-dependent callbacks used by various compiler passes.
-    """
-
-    def is_on_host(f):
-        return not is_on_device(f, options['gpu-fit'])
-
-    def runs_on_host(c):
-        # The only situation in which a Cluster doesn't get offloaded to
-        # the device is when it writes to a host Function
-        return any(is_on_host(f) for f in c.scope.writes)
-
-    def reads_if_on_host(c):
-        if not runs_on_host(c):
-            return [f for f in c.scope.reads if is_on_host(f)]
+def stream_wrap(callback):
+    def stream_key(items, *args):
+        """
+        Given one or more Functions `f(d_1, ...d_n)`, return the Dimensions
+        `(d_i, ..., d_n)` requiring data streaming.
+        """
+        found = [f for f in as_tuple(items) if callback(f)]
+        retval = {stream_dimensions(f) for f in found}
+        if len(retval) > 1:
+            raise ValueError("Cannot determine homogenous stream Dimensions")
+        elif len(retval) == 1:
+            return retval.pop()
         else:
-            return []
+            return None
 
-    return runs_on_host, reads_if_on_host
+    return stream_key
