@@ -1,38 +1,80 @@
 import abc
-from functools import reduce
-from operator import mul
+import ctypes
+from ctypes.util import find_library
 import mmap
 import os
 import sys
 
 import numpy as np
-import ctypes
-from ctypes.util import find_library
 
 from devito.logger import logger
 from devito.parameters import configuration
-from devito.tools import dtype_to_ctype
+from devito.tools import is_integer, infer_datasize
 
-__all__ = ['ALLOC_FLAT', 'ALLOC_NUMA_LOCAL', 'ALLOC_NUMA_ANY',
+__all__ = ['ALLOC_ALIGNED', 'ALLOC_NUMA_LOCAL', 'ALLOC_NUMA_ANY',
            'ALLOC_KNL_MCDRAM', 'ALLOC_KNL_DRAM', 'ALLOC_GUARD',
            'default_allocator']
 
 
-class MemoryAllocator(object):
+class AbstractMemoryAllocator:
 
-    """Abstract class defining the interface to memory allocators."""
+    """
+    The MemoryAllocator interface.
+    """
 
     __metaclass__ = abc.ABCMeta
 
-    is_Posix = False
-    is_Numa = False
+    guaranteed_alignment = 64
+    """Guaranteed data alignment."""
+
+    @abc.abstractmethod
+    def alloc(self, shape, dtype, padding=0):
+        """
+        Allocate memory.
+
+        Parameters
+        ----------
+        shape : tuple of ints
+            Shape of the allocated array.
+        dtype : numpy.dtype
+            The data type of the raw data.
+        padding : int or 2-tuple of ints, optional
+            The number of points that are allocated before and after the data,
+            that is in addition to the requested shape. Defaults to 0.
+
+        Returns
+        -------
+        ndarray, memfree_args
+            The first element of the tuple is a numpy array that uses the
+            allocated memory underneath. The second element is an opaque
+            object that is needed only for the "memfree" call.
+        """
+        return
+
+    @abc.abstractmethod
+    def free(self, *args):
+        """
+        Free memory previously allocated with `self.alloc`.
+
+        Arguments are provided exactly as returned in the second element of the
+        tuple returned by `alloc`.
+        """
+        return
+
+
+class MemoryAllocator(AbstractMemoryAllocator):
+
+    """
+    A memory allocator implementing the alloc method by resorting to a C-level
+    memory allocation function, to be specified by subclasses.
+
+    This is still an abstract class, and subclasses are expected to implement the
+    `_alloc_C_libcall` and `free` methods.
+    """
 
     _attempted_init = False
     lib = None
 
-    guaranteed_alignment = 64
-    """Guaranteed data alignment."""
-
     @classmethod
     def available(cls):
         if cls._attempted_init is False:
@@ -44,50 +86,43 @@ class MemoryAllocator(object):
     def initialize(cls):
         """
         Initialize the MemoryAllocator.
-
-        Notes
-        -----
-        This method must be implemented by all subclasses of MemoryAllocator.
         """
         return
 
-    def alloc(self, shape, dtype):
-        """
-        Allocate memory.
+    def alloc(self, shape, dtype, padding=0):
+        ctype, datasize = infer_datasize(dtype, shape)
 
-        Parameters
-        ----------
-        shape : tuple of ints
-            Shape of the allocated array.
-        dtype : numpy.dtype
-            The data type of the raw data.
+        # Add padding, if any
+        try:
+            padleft, padright = padding
+        except TypeError:
+            padleft, padright = padding, padding
+        if not is_integer(padleft) and not is_integer(padright):
+            raise TypeError("padding must be an int or a 2-tuple of ints")
+        size = datasize + padleft + padright
 
-        Returns
-        -------
-        pointer, memfree_args
-            The first element of the tuple is the reference that can be used to
-            access the data as a ctypes object. The second element is an opaque
-            object that is needed only for the "memfree" call.
-        """
-        size = int(reduce(mul, shape))
-        ctype = dtype_to_ctype(dtype)
+        padleft_pointer, memfree_args = self._alloc_C_libcall(size, ctype)
+        if padleft_pointer is None:
+            raise RuntimeError(f"Unable to allocate {size} elements in memory")
 
-        c_pointer, memfree_args = self._alloc_C_libcall(size, ctype)
-        if c_pointer is None:
-            raise RuntimeError("Unable to allocate %d elements in memory", str(size))
+        # Compute the pointer to the user data
+        padleft_bytes = padleft * ctypes.sizeof(ctype)
+        c_pointer = ctypes.c_void_p(padleft_pointer.value + padleft_bytes)
 
-        # cast to 1D array of the specified size
-        ctype_1d = ctype * size
+        # Cast to 1D array of the specified `datasize`
+        ctype_1d = ctype * datasize
         buf = ctypes.cast(c_pointer, ctypes.POINTER(ctype_1d)).contents
-        pointer = np.frombuffer(buf, dtype=dtype)
-        # pointer.reshape should not be used here because it may introduce a copy
-        # From https://docs.scipy.org/doc/numpy/reference/generated/numpy.reshape.html:
-        # It is not always possible to change the shape of an array without copying the
-        # data. If you want an error to be raised when the data is copied, you should
-        # assign the new shape to the shape attribute of the array:
-        pointer.shape = shape
+        array = np.frombuffer(buf, dtype=dtype)
 
-        return (pointer, memfree_args)
+        # `array.reshape` should not be used here because it may introduce
+        # a copy. From `docs.scipy.org/doc/numpy/reference/generated/numpy.reshape`:
+        #   It is not always possible to change the shape of an array without
+        #   copying the data. If you want an error to be raised when the data
+        #   is copied, you should assign the new shape to the shape attribute
+        #   of the array:
+        array.shape = shape
+
+        return (array, memfree_args)
 
     @abc.abstractmethod
     def _alloc_C_libcall(self, size, ctype):
@@ -102,20 +137,6 @@ class MemoryAllocator(object):
         """
         return
 
-    @abc.abstractmethod
-    def free(self, *args):
-        """
-        Free memory previously allocated with ``self.alloc``.
-
-        Arguments are provided exactly as returned in the second element of the
-        tuple returned by _alloc_C_libcall
-
-        Notes
-        -----
-        This method must be implemented by all subclasses of MemoryAllocator.
-        """
-        return
-
 
 class PosixAllocator(MemoryAllocator):
 
@@ -124,8 +145,6 @@ class PosixAllocator(MemoryAllocator):
     aligned to page boundaries.
     """
 
-    is_Posix = True
-
     @classmethod
     def initialize(cls):
         handle = find_library('c')
@@ -162,7 +181,7 @@ class PosixAllocator(MemoryAllocator):
 class GuardAllocator(PosixAllocator):
 
     """
-    Memory allocator based on ``posix`` functions. The allocated memory is
+    Memory allocator based on `posix` functions. The allocated memory is
     aligned to page boundaries.  Additionally, it allocates extra memory
     before and after the data, and configures it so that an SEGV is thrown
     immediately if an out-of-bounds access occurs.
@@ -195,20 +214,20 @@ class GuardAllocator(PosixAllocator):
         if ret != 0:
             return None, None
 
-        # generate pointers to the left padding, the user data, and the right pad
+        # Generate pointers to the left padding, the user data, and the right pad
         padleft_pointer = c_pointer
         c_pointer = ctypes.c_void_p(c_pointer.value + self.padding_bytes)
         padright_pointer = ctypes.c_void_p(c_pointer.value + npages_user * pagesize)
 
-        # and set the permissions on the pad memory to 0 (no access)
-        # if these fail, don't worry about failing the entire allocation
+        # And set the permissions on the pad memory to 0 (no access)
+        # If these fail, don't worry about failing the entire allocation
         c_padsize = ctypes.c_ulong(self.padding_bytes)
         if self.lib.mprotect(padleft_pointer, c_padsize, ctypes.c_int(0)):
             logger.warning("couldn't protect memory")
         if self.lib.mprotect(padright_pointer, c_padsize, ctypes.c_int(0)):
             logger.warning("couldn't protect memory")
 
-        # if there is a multiple of 4 bytes left, use the code below to poison
+        # If there is a multiple of 4 bytes left, use the code below to poison
         # the memory
         if nbytes_user % 4 == 0:
             poison_size = npages_user*pagesize - nbytes_user
@@ -216,16 +235,16 @@ class GuardAllocator(PosixAllocator):
             poison_ptr = ctypes.cast(ctypes.c_void_p(c_pointer.value + nbytes_user),
                                      intp_type)
 
-            # for both float32 and float64, a sequence of -100 int32s represents NaNs,
-            # at least on little-endian architectures.  It shouldn't matter what we
-            # put in there, anyway
+            # For both float32 and float64, a sequence of -100 int32s
+            # represents NaNs, at least on little-endian architectures;
+            # it shouldn't matter what we put in there, anyway
             for i in range(poison_size // 4):
                 poison_ptr[i] = -100
 
         return c_pointer, (padleft_pointer, c_bytesize)
 
     def free(self, c_pointer, total_size):
-        # unprotect it, since free() accesses it, I think...
+        # Unprotect it, since free() accesses it, I think...
         self.lib.mprotect(c_pointer, total_size,
                           ctypes.c_int(mmap.PROT_READ | mmap.PROT_WRITE))
         self.lib.free(c_pointer)
@@ -247,8 +266,6 @@ class NumaAllocator(MemoryAllocator):
         ("allocate on any NUMA node with sufficient free memory") are accepted.
     """
 
-    is_Numa = True
-
     @classmethod
     def initialize(cls):
         handle = find_library('numa')
@@ -268,7 +285,7 @@ class NumaAllocator(MemoryAllocator):
         cls.lib = lib
 
     def __init__(self, node):
-        super(NumaAllocator, self).__init__()
+        super().__init__()
         self._node = node
 
     def _alloc_C_libcall(self, size, ctype):
@@ -317,10 +334,10 @@ class NumaAllocator(MemoryAllocator):
         return self._node == 'local'
 
 
-class ExternalAllocator(MemoryAllocator):
+class DataReference(MemoryAllocator):
 
     """
-    An ExternalAllocator is used to assign pre-existing user data to Functions.
+    A DataReference is used to assign pre-existing user data to Functions.
     Thus, Devito does not allocate any memory.
 
     Parameters
@@ -330,23 +347,23 @@ class ExternalAllocator(MemoryAllocator):
 
     Notes
     -------
-    * Use ExternalAllocator and pass a reference to the external memory when
+    * Use DataReference and pass a reference to the external memory when
       creating a Function. This Function will now use this memory as its f.data.
 
-    * If the data present in this external memory is valuable, provide a noop
-      initialiser, or else Devito will reset it to 0.
+    * This can be used to pass one Function's data to another to avoid copying
+      during Function rebuilds (this should only be used internally).
 
     Example
     --------
     >>> from devito import Grid, Function
-    >>> from devito.data.allocators import ExternalAllocator
+    >>> from devito.data.allocators import DataReference
     >>> import numpy as np
     >>> shape = (2, 2)
     >>> numpy_array = np.ones(shape, dtype=np.float32)
     >>> g = Grid(shape)
     >>> space_order = 0
     >>> f = Function(name='f', grid=g, space_order=space_order,
-    ...      allocator=ExternalAllocator(numpy_array), initializer=lambda x: None)
+    ...      allocator=DataReference(numpy_array))
     >>> f.data[0, 1] = 2
     >>> numpy_array
     array([[1., 2.],
@@ -356,25 +373,30 @@ class ExternalAllocator(MemoryAllocator):
     def __init__(self, numpy_array):
         self.numpy_array = numpy_array
 
-    def alloc(self, shape, dtype):
+    def alloc(self, shape, dtype, padding=0):
         assert shape == self.numpy_array.shape, \
-            "Provided array has shape %s. Expected %s" %\
-            (str(self.numpy_array.shape), str(shape))
+            (f"Provided array has shape {str(self.numpy_array.shape)}. "
+             f"Expected {str(shape)}")
         assert dtype == self.numpy_array.dtype, \
-            "Provided array has dtype %s. Expected %s" %\
-            (str(self.numpy_array.dtype), str(dtype))
+            (f"Provided array has dtype {str(self.numpy_array.dtype)}. "
+             f"Expected {str(dtype)}")
 
         return (self.numpy_array, None)
 
 
+# For backward compatibility
+ExternalAllocator = DataReference
+
 ALLOC_GUARD = GuardAllocator(1048576)
-ALLOC_FLAT = PosixAllocator()
+ALLOC_ALIGNED = PosixAllocator()
 ALLOC_KNL_DRAM = NumaAllocator(0)
 ALLOC_KNL_MCDRAM = NumaAllocator(1)
 ALLOC_NUMA_ANY = NumaAllocator('any')
 ALLOC_NUMA_LOCAL = NumaAllocator('local')
 
-custom_allocators = {}
+custom_allocators = {
+    'fallback': ALLOC_ALIGNED,
+}
 """User-defined allocators."""
 
 
@@ -383,11 +405,11 @@ def register_allocator(name, allocator):
     Register a custom MemoryAllocator.
     """
     if not isinstance(name, str):
-        raise TypeError("name must be a str, not `%s`" % type(name))
+        raise TypeError(f"name must be a str, not `{type(name)}`)")
     if name in custom_allocators:
-        raise ValueError("A MemoryAllocator for `%s` already exists" % name)
-    if not isinstance(allocator, MemoryAllocator):
-        raise TypeError("Expected a MemoryAllocator, not `%s`" % type(allocator))
+        raise ValueError(f"A MemoryAllocator for `{name}` already exists")
+    if not isinstance(allocator, AbstractMemoryAllocator):
+        raise TypeError(f"Expected a MemoryAllocator, not `{type(allocator)}`")
 
     custom_allocators[name] = allocator
 
@@ -403,8 +425,8 @@ def default_allocator(name=None):
 
         * ALLOC_GUARD: Only used in so-called "develop mode", to trigger SIGSEGV as
                        soon as OOB accesses are performed.
-        * ALLOC_FLAT: Align memory to page boundaries using the posix function
-                      `posix_memalign`.
+        * ALLOC_ALIGNED: Align memory to page boundaries using the function
+                         `posix_memalign`.
         * ALLOC_NUMA_LOCAL: Allocate memory in the "closest" NUMA node. This only
                             makes sense on a NUMA architecture. Falls back to
                             allocation in an arbitrary NUMA node if there isn't
@@ -424,10 +446,9 @@ def default_allocator(name=None):
 
     if configuration['develop-mode']:
         return ALLOC_GUARD
-    elif NumaAllocator.available():
-        if configuration['platform'].name == 'knl' and infer_knl_mode() == 'flat':
-            return ALLOC_KNL_MCDRAM
-        else:
-            return ALLOC_NUMA_LOCAL
+    elif (NumaAllocator.available() and
+          configuration['platform'].name == 'knl' and
+          infer_knl_mode() == 'flat'):
+        return ALLOC_KNL_MCDRAM
     else:
-        return ALLOC_FLAT
+        return custom_allocators.get('default', custom_allocators['fallback'])
