@@ -3,7 +3,7 @@ from functools import wraps
 
 import numpy as np
 
-from devito.data.allocators import ALLOC_FLAT
+from devito.data.allocators import ALLOC_ALIGNED
 from devito.data.utils import *
 from devito.logger import warning
 from devito.parameters import configuration
@@ -26,28 +26,32 @@ class Data(np.ndarray):
     decomposition : tuple of Decomposition, optional
         The data decomposition, for each dimension.
     modulo : tuple of bool, optional
-        If the i-th entry is True, then the i-th array dimension uses modulo indexing.
+        If the i-th entry is True, then the i-th array dimension uses modulo
+        indexing.
     allocator : MemoryAllocator, optional
-        Used to allocate memory. Defaults to ``ALLOC_FLAT``.
+        Used to allocate memory. Defaults to `ALLOC_ALIGNED`.
     distributor : Distributor, optional
-        The distributor from which the original decomposition was produced. Note that
-        the decomposition Parameter above may be different to distributor.decomposition.
+        The distributor from which the original decomposition was produced.
+        Note that `decomposition` may differ from `distributor.decomposition`.
+    padding : int or 2-tuple of ints, optional
+        The number of points that are allocated before and after the data,
+        that is in addition to the requested shape. Defaults to 0.
 
     Notes
     -----
     NumPy array subclassing is described at: ::
 
-        https://docs.scipy.org/doc/numpy-1.13.0/user/basics.subclassing.html
+        https://numpy.org/doc/stable/user/basics.subclassing.html
 
     Any view or copy created from ``self``, for instance via a slice operation
     or a universal function ("ufunc" in NumPy jargon), will still be of type
-    Data.
+    `Data`.
     """
 
-    def __new__(cls, shape, dtype, decomposition=None, modulo=None, allocator=ALLOC_FLAT,
-                distributor=None):
+    def __new__(cls, shape, dtype, decomposition=None, modulo=None,
+                allocator=ALLOC_ALIGNED, distributor=None, padding=0):
         assert len(shape) == len(modulo)
-        ndarray, memfree_args = allocator.alloc(shape, dtype)
+        ndarray, memfree_args = allocator.alloc(shape, dtype, padding=padding)
         obj = ndarray.view(cls)
         obj._allocator = allocator
         obj._memfree_args = memfree_args
@@ -87,7 +91,7 @@ class Data(np.ndarray):
 
     def __reduce__(self):
         warning("Pickling of `Data` objects is not supported. Casting to `numpy.ndarray`")
-        return np.array(self).__reduce__()
+        return self.view(np.ndarray).__reduce__()
 
     def __array_finalize__(self, obj):
         # `self` is the newly created object
@@ -109,9 +113,9 @@ class Data(np.ndarray):
             self._is_distributed = False
             self._modulo = tuple(False for i in range(self.ndim))
             self._decomposition = (None,)*self.ndim
+            self._allocator = ALLOC_ALIGNED
         elif obj._index_stash is not None:
             # From `__getitem__`
-            self._is_distributed = obj._is_distributed
             self._distributor = obj._distributor
             glb_idx = obj._normalize_index(obj._index_stash)
             self._modulo = tuple(m for i, m in zip(glb_idx, obj._modulo)
@@ -125,9 +129,13 @@ class Data(np.ndarray):
                 else:
                     decomposition.append(dec.reshape(i))
             self._decomposition = tuple(decomposition)
+            self._allocator = obj._allocator
+            decomp = any(i is not None for i in self._decomposition)
+            self._is_distributed = decomp and obj._is_distributed
         else:
-            self._is_distributed = obj._is_distributed
             self._distributor = obj._distributor
+            self._allocator = obj._allocator
+            self._is_distributed = obj._is_distributed
             if self.ndim == obj.ndim:
                 # E.g., from a ufunc, such as `np.add`
                 self._modulo = obj._modulo
@@ -146,30 +154,40 @@ class Data(np.ndarray):
 
     def _global(self, glb_idx, decomposition):
         """A "global" view of ``self`` over a given Decomposition."""
-        if self._is_distributed:
+        if self._is_decomposed:
             raise ValueError("Cannot derive a decomposed view from a decomposed Data")
         if len(decomposition) != self.ndim:
-            raise ValueError("`decomposition` should have ndim=%d entries" % self.ndim)
+            raise ValueError(f"`decomposition` should have ndim={self.ndim} entries")
         ret = self[glb_idx]
         ret._decomposition = decomposition
         ret._is_distributed = any(i is not None for i in decomposition)
         return ret
 
+    def _prune_shape(self, shape):
+        # Reduce distributed MPI `Data`'s shape to that of an equivalently
+        # sliced numpy array.
+        decomposition = tuple(d for d in self._decomposition if d.size > 1)
+        retval = self.reshape(shape)
+        retval._decomposition = decomposition
+        return retval
+
     def _check_idx(func):
         """Check if __getitem__/__setitem__ may require communication across MPI ranks."""
         @wraps(func)
         def wrapper(data, *args, **kwargs):
             glb_idx = args[0]
-            if len(args) > 1 and isinstance(args[1], Data) \
-                    and args[1]._is_mpi_distributed:
+            is_gather = isinstance(kwargs.get('gather_rank'), int)
+            if is_gather and all(i == slice(None, None, 1) for i in glb_idx):
+                comm_type = gather
+            elif len(args) > 1 and isinstance(args[1], Data) and args[1]._is_decomposed:
                 comm_type = index_by_index
-            elif data._is_mpi_distributed:
+            elif data._is_decomposed:
                 for i in as_tuple(glb_idx):
                     if isinstance(i, slice) and i.step is not None and i.step < 0:
                         comm_type = index_by_index
                         break
-                    else:
-                        comm_type = serial
+                else:
+                    comm_type = serial
             else:
                 comm_type = serial
             kwargs['comm_type'] = comm_type
@@ -177,8 +195,9 @@ class Data(np.ndarray):
         return wrapper
 
     @property
-    def _is_mpi_distributed(self):
-        return self._is_distributed and configuration['mpi']
+    def _is_decomposed(self):
+        return self._is_distributed and configuration['mpi'] and \
+            self._distributor.is_parallel
 
     def __repr__(self):
         return super(Data, self._local).__repr__()
@@ -189,12 +208,51 @@ class Data(np.ndarray):
     @_check_idx
     def __getitem__(self, glb_idx, comm_type, gather_rank=None):
         loc_idx = self._index_glb_to_loc(glb_idx)
-        is_gather = True if isinstance(gather_rank, int) else False
-        if comm_type is index_by_index or is_gather:
-            # Retrieve the pertinent local data prior to mpi send/receive operations
+        is_gather = isinstance(gather_rank, int)
+        if is_gather and comm_type is gather:
+            comm = self._distributor.comm
+            rank = comm.Get_rank()
+
+            sendbuf = self.flat[:]
+            sendcounts = np.array(comm.gather(len(sendbuf), gather_rank))
+
+            if rank == gather_rank:
+                recvbuf = np.zeros(sum(sendcounts), dtype=self.dtype.type)
+            else:
+                recvbuf = None
+            comm.Gatherv(sendbuf=sendbuf, recvbuf=(recvbuf, sendcounts), root=gather_rank)
+
+            # Reshape the gathered data to produce the output
+            if rank == gather_rank:
+                if len(self.shape) > len(self._distributor.glb_shape):
+                    glb_shape = list(self._distributor.glb_shape)
+                    for i in range(len(self.shape) - len(self._distributor.glb_shape)):
+                        glb_shape.insert(i, self.shape[i])
+                else:
+                    glb_shape = self._distributor.glb_shape
+                retval = np.zeros(glb_shape, dtype=self.dtype.type)
+                start, stop, step = 0, 0, 1
+                for i, _ in enumerate(sendcounts):
+                    if i > 0:
+                        start += sendcounts[i-1]
+                    stop += sendcounts[i]
+                    data_slice = recvbuf[slice(start, stop, step)]
+                    shape = [r.stop-r.start for r in self._distributor.all_ranges[i]]
+                    idx = [slice(r.start - d.glb_min, r.stop - d.glb_min, r.step)
+                           for r, d in zip(self._distributor.all_ranges[i],
+                                           self._distributor.decomposition)]
+                    for j in range(len(self.shape) - len(self._distributor.glb_shape)):
+                        shape.insert(j, glb_shape[j])
+                        idx.insert(j, slice(0, glb_shape[j]+1, 1))
+                    retval[tuple(idx)] = data_slice.reshape(tuple(shape))
+                return retval
+            else:
+                return None
+        elif comm_type is index_by_index or is_gather:
+            # Retrieve the pertinent local data prior to MPI send/receive operations
             data_idx = loc_data_idx(loc_idx)
             self._index_stash = flip_idx(glb_idx, self._decomposition)
-            local_val = super(Data, self).__getitem__(data_idx)
+            local_val = super().__getitem__(data_idx)
             self._index_stash = None
 
             comm = self._distributor.comm
@@ -208,11 +266,12 @@ class Data(np.ndarray):
             if not is_gather:
                 retval = Data(local_val.shape, local_val.dtype.type,
                               decomposition=local_val._decomposition,
-                              modulo=(False,)*len(local_val.shape))
+                              modulo=(False,)*len(local_val.shape),
+                              distributor=local_val._distributor)
             elif rank == gather_rank:
                 retval = np.zeros(it.shape)
             else:
-                retval = np.empty((0, )*len(it.shape))
+                retval = None
             # Iterate over each element of data
             while not it.finished:
                 index = it.multi_index
@@ -249,10 +308,13 @@ class Data(np.ndarray):
                 it.iternext()
             # Check if dimensions of the view should now be reduced to
             # be consistent with those of an equivalent NumPy serial view
-            reshape = tuple([s for s, i in zip(retval.shape, loc_idx)
-                             if type(i) is not np.int64])
-            if reshape and (0 not in reshape) and (reshape != retval.shape):
-                return retval.reshape(reshape)
+            if not is_gather:
+                newshape = tuple(s for s, i in zip(retval.shape, loc_idx)
+                                 if type(i) is not np.int64)
+            else:
+                newshape = ()
+            if newshape and (0 not in newshape) and (newshape != retval.shape):
+                return retval._prune_shape(newshape)
             else:
                 return retval
         elif loc_idx is NONLOCAL:
@@ -261,13 +323,14 @@ class Data(np.ndarray):
             return None
         else:
             self._index_stash = glb_idx
-            retval = super(Data, self).__getitem__(loc_idx)
+            retval = super().__getitem__(loc_idx)
             self._index_stash = None
             return retval
 
     @_check_idx
     def __setitem__(self, glb_idx, val, comm_type):
         loc_idx = self._index_glb_to_loc(glb_idx)
+
         if loc_idx is NONLOCAL:
             # no-op
             return
@@ -275,10 +338,10 @@ class Data(np.ndarray):
             if index_is_basic(loc_idx):
                 # Won't go through `__getitem__` as it's basic indexing mode,
                 # so we should just propage `loc_idx`
-                super(Data, self).__setitem__(loc_idx, val)
+                super().__setitem__(loc_idx, val)
             else:
-                super(Data, self).__setitem__(glb_idx, val)
-        elif isinstance(val, Data) and val._is_distributed:
+                super().__setitem__(glb_idx, val)
+        elif isinstance(val, Data) and val._is_decomposed:
             if comm_type is index_by_index:
                 glb_idx, val = self._process_args(glb_idx, val)
                 val_idx = as_tuple([slice(i.glb_min, i.glb_max+1, 1) for
@@ -298,14 +361,14 @@ class Data(np.ndarray):
                         or data_global[j].size == 0
                     if not skip:
                         self.__setitem__(idx_global[j], data_global[j])
-            elif self._is_distributed:
+            elif self._is_decomposed:
                 # `val` is decomposed, `self` is decomposed -> local set
-                super(Data, self).__setitem__(glb_idx, val)
+                super().__setitem__(glb_idx, val)
             else:
                 # `val` is decomposed, `self` is replicated -> gatherall-like
                 raise NotImplementedError
         elif isinstance(val, np.ndarray):
-            if self._is_distributed:
+            if self._is_decomposed:
                 # `val` is replicated, `self` is decomposed -> `val` gets decomposed
                 glb_idx = self._normalize_index(glb_idx)
                 glb_idx, val = self._process_args(glb_idx, val)
@@ -336,15 +399,15 @@ class Data(np.ndarray):
             else:
                 # `val` is replicated`, `self` is replicated -> plain ndarray.__setitem__
                 pass
-            super(Data, self).__setitem__(glb_idx, val)
+            super().__setitem__(glb_idx, val)
         elif isinstance(val, Iterable):
-            if self._is_mpi_distributed:
+            if self._is_decomposed:
                 raise NotImplementedError("With MPI, data can only be set "
                                           "via scalars, numpy arrays or "
                                           "other data ")
-            super(Data, self).__setitem__(glb_idx, val)
+            super().__setitem__(glb_idx, val)
         else:
-            raise ValueError("Cannot insert obj of type `%s` into a Data" % type(val))
+            raise ValueError(f"Cannot insert obj of type `{type(val)}` into a Data")
 
     def _normalize_index(self, idx):
         if isinstance(idx, np.ndarray):
@@ -366,6 +429,11 @@ class Data(np.ndarray):
 
     def _process_args(self, idx, val):
         """If comm_type is parallel we need to first retrieve local unflipped data."""
+        if (len(as_tuple(idx)) < len(val.shape)) and (len(val.shape) <= len(self.shape)):
+            idx_processed = as_list(idx)
+            for _ in range(len(val.shape)-len(as_tuple(idx))):
+                idx_processed.append(slice(None, None, 1))
+            idx = as_tuple(idx_processed)
         if any(isinstance(i, slice) and i.step is not None and i.step < 0
                for i in as_tuple(idx)):
             processed = []
@@ -410,7 +478,7 @@ class Data(np.ndarray):
         if len(glb_idx) > self.ndim:
             # Maybe user code is trying to add a new axis (see np.newaxis),
             # so the resulting array will be higher dimensional than `self`
-            if self._is_mpi_distributed:
+            if self._is_decomposed:
                 raise ValueError("Cannot increase dimensionality of MPI-distributed Data")
             else:
                 # As by specification, we are forced to ignore modulo indexing
@@ -426,7 +494,7 @@ class Data(np.ndarray):
                 try:
                     v = convert_index(i, dec, mode='glb_to_loc')
                 except TypeError:
-                    if self._is_mpi_distributed:
+                    if self._is_decomposed:
                         raise NotImplementedError("Unsupported advanced indexing with "
                                                   "MPI-distributed Data")
                     v = i
@@ -532,4 +600,5 @@ class Data(np.ndarray):
 class CommType(Tag):
     pass
 index_by_index = CommType('index_by_index')  # noqa
-serial = CommType('serial')
+serial = CommType('serial')  # noqa
+gather = CommType('gather')  # noqa
