@@ -1,7 +1,7 @@
 from collections.abc import Iterable
+from functools import cached_property
 
 import numpy as np
-from cached_property import cached_property
 
 from devito.data.meta import LEFT
 from devito.tools import is_integer, as_tuple
@@ -59,7 +59,7 @@ class Decomposition(tuple):
             raise TypeError("Illegal Decomposition element type")
         if not is_integer(local) and (0 <= local < len(items)):
             raise ValueError("`local` must be an index in ``items``.")
-        obj = super(Decomposition, cls).__new__(cls, [np.array(i) for i in items])
+        obj = super().__new__(cls, [np.array(i) for i in items])
         obj._local = local
         return obj
 
@@ -204,9 +204,26 @@ class Decomposition(tuple):
         >>> d.index_glb_to_loc((1, 6), rel=False)
         (5, 6)
         """
+        # Offset the loc_abs_min, loc_abs_max, glb_min, and glb_max
+        # In the case of a Function defined on a SubDomain, the global indices
+        # for accessing the data associated with this Function will run from
+        # `ltkn` to `x_M-rtkn-1`. However, the indices for accessing this array
+        # will run from `0` to `x_M-ltkn-rtkn-1`. As such, the global minimum
+        # (`ltkn`) should be subtracted for the purpose of indexing into the local
+        # array.
+        if not self.loc_empty:
+            loc_abs_min = self.loc_abs_min - self.glb_min
+            loc_abs_max = self.loc_abs_max - self.glb_min
+            glb_max = self.glb_max - self.glb_min
+        else:
+            loc_abs_min = self.loc_abs_min
+            loc_abs_max = self.loc_abs_max
+            glb_max = self.glb_max
 
-        base = self.loc_abs_min if rel is True else 0
-        top = self.loc_abs_max
+        glb_min = 0
+
+        base = loc_abs_min if rel else 0
+        top = loc_abs_max
 
         if len(args) == 1:
             glb_idx = args[0]
@@ -217,11 +234,11 @@ class Decomposition(tuple):
                     return None
                 # -> Handle negative index
                 if glb_idx < 0:
-                    glb_idx = self.glb_max + glb_idx + 1
+                    glb_idx = glb_max + glb_idx + 1
                 # -> Do the actual conversion
-                if self.loc_abs_min <= glb_idx <= self.loc_abs_max:
+                if loc_abs_min <= glb_idx <= loc_abs_max:
                     return glb_idx - base
-                elif self.glb_min <= glb_idx <= self.glb_max:
+                elif glb_min <= glb_idx <= glb_max:
                     return None
                 else:
                     # This should raise an exception when used to access a numpy.array
@@ -239,30 +256,32 @@ class Decomposition(tuple):
                 elif isinstance(glb_idx, slice):
                     if self.loc_empty:
                         return slice(-1, -3)
-                    if glb_idx.step >= 0 and glb_idx.stop == self.glb_min:
-                        glb_idx_min = self.glb_min if glb_idx.start is None \
+                    if glb_idx.step >= 0 and glb_idx.stop == glb_min:
+                        glb_idx_min = glb_min if glb_idx.start is None \
                             else glb_idx.start
-                        glb_idx_max = self.glb_min
+                        glb_idx_max = glb_min
                         retfunc = lambda a, b: slice(a, b, glb_idx.step)
                     elif glb_idx.step >= 0:
-                        glb_idx_min = self.glb_min if glb_idx.start is None \
+                        glb_idx_min = glb_min if glb_idx.start is None \
                             else glb_idx.start
-                        glb_idx_max = self.glb_max if glb_idx.stop is None \
+                        glb_idx_max = glb_max \
+                            if glb_idx.stop is None \
                             else glb_idx.stop-1
                         retfunc = lambda a, b: slice(a, b + 1, glb_idx.step)
                     else:
-                        glb_idx_min = self.glb_min if glb_idx.stop is None \
+                        glb_idx_min = glb_min if glb_idx.stop is None \
                             else glb_idx.stop+1
-                        glb_idx_max = self.glb_max if glb_idx.start is None \
+                        glb_idx_max = glb_max if glb_idx.start is None \
                             else glb_idx.start
                         retfunc = lambda a, b: slice(b, a - 1, glb_idx.step)
                 else:
                     raise TypeError("Cannot convert index from `%s`" % type(glb_idx))
                 # -> Handle negative min/max
                 if glb_idx_min is not None and glb_idx_min < 0:
-                    glb_idx_min = self.glb_max + glb_idx_min + 1
+                    glb_idx_min = glb_max + glb_idx_min + 1
                 if glb_idx_max is not None and glb_idx_max < 0:
-                    glb_idx_max = self.glb_max + glb_idx_max + 1
+                    glb_idx_max = glb_max + glb_idx_max + 1
+
                 # -> Do the actual conversion
                 # Compute loc_min. For a slice with step > 0 this will be
                 # used to produce slice.start and for a slice with step < 0 slice.stop.
@@ -271,19 +290,19 @@ class Decomposition(tuple):
                 # coincide with loc_abs_min.
                 if isinstance(glb_idx, slice) and glb_idx.step is not None \
                         and glb_idx.step > 1:
-                    if glb_idx_min > self.loc_abs_max:
+                    if glb_idx_min > loc_abs_max:
                         return retfunc(-1, -3)
                     elif glb_idx.start is None:  # glb start is zero.
-                        loc_min = self.loc_abs_min - base \
+                        loc_min = loc_abs_min - base \
                             + np.mod(glb_idx.step - np.mod(base, glb_idx.step),
                                      glb_idx.step)
                     else:  # glb start is given explicitly
-                        loc_min = self.loc_abs_min - base \
+                        loc_min = loc_abs_min - base \
                             + np.mod(glb_idx.step - np.mod(base - glb_idx.start,
                                                            glb_idx.step), glb_idx.step)
-                elif glb_idx_min is None or glb_idx_min < self.loc_abs_min:
-                    loc_min = self.loc_abs_min - base
-                elif glb_idx_min > self.loc_abs_max:
+                elif glb_idx_min is None or glb_idx_min < loc_abs_min:
+                    loc_min = loc_abs_min - base
+                elif glb_idx_min > loc_abs_max:
                     return retfunc(-1, -3)
                 else:
                     loc_min = glb_idx_min - base
@@ -294,19 +313,19 @@ class Decomposition(tuple):
                 # coincide with loc_abs_max.
                 if isinstance(glb_idx, slice) and glb_idx.step is not None \
                         and glb_idx.step < -1:
-                    if glb_idx_max < self.loc_abs_min:
+                    if glb_idx_max < loc_abs_min:
                         return retfunc(-1, -3)
                     elif glb_idx.start is None:
                         loc_max = top - base \
-                            + np.mod(glb_idx.step - np.mod(top - self.glb_max,
+                            + np.mod(glb_idx.step - np.mod(top - glb_max,
                                                            glb_idx.step), glb_idx.step)
                     else:
                         loc_max = top - base \
                             + np.mod(glb_idx.step - np.mod(top - glb_idx.start,
                                                            glb_idx.step), glb_idx.step)
-                elif glb_idx_max is None or glb_idx_max > self.loc_abs_max:
-                    loc_max = self.loc_abs_max - base
-                elif glb_idx_max < self.loc_abs_min:
+                elif glb_idx_max is None or glb_idx_max > loc_abs_max:
+                    loc_max = loc_abs_max - base
+                elif glb_idx_max < loc_abs_min:
                     return retfunc(-1, -3)
                 else:
                     loc_max = glb_idx_max - base
@@ -320,16 +339,20 @@ class Decomposition(tuple):
             if self.loc_empty:
                 return None
             abs_ofs, side = args
-            if side is LEFT:
-                rel_ofs = self.glb_min + abs_ofs - base
+            if side == LEFT:
+                rel_ofs = glb_min + abs_ofs - base
                 if abs_ofs >= base and abs_ofs <= top:
                     return rel_ofs
+                elif abs_ofs > top:
+                    return top + 1
                 else:
                     return None
             else:
-                rel_ofs = abs_ofs - (self.glb_max - top)
-                if abs_ofs >= self.glb_max - top and abs_ofs <= self.glb_max - base:
+                rel_ofs = abs_ofs - (glb_max - top)
+                if abs_ofs >= glb_max - top and abs_ofs <= glb_max - base:
                     return rel_ofs
+                elif abs_ofs > glb_max - base:
+                    return glb_max - base + 1
                 else:
                     return None
         else:
