@@ -1,6 +1,6 @@
 from collections import ChainMap
 from itertools import product
-from functools import singledispatch
+from functools import singledispatch, cached_property
 
 import numpy as np
 import sympy
@@ -8,14 +8,21 @@ from sympy.core.add import _addsort
 from sympy.core.mul import _keep_coeff, _mulsort
 from sympy.core.decorators import call_highest_priority
 from sympy.core.evalf import evalf_table
+try:
+    from sympy.core.core import ordering_of_classes
+except ImportError:
+    # Moved in 1.13
+    from sympy.core.basic import ordering_of_classes
 
-from cached_property import cached_property
-from devito.finite_differences.tools import make_shift_x0
+from devito.finite_differences.tools import make_shift_x0, coeff_priority
 from devito.logger import warning
-from devito.tools import as_tuple, filter_ordered, flatten, is_integer, split
+from devito.tools import (as_tuple, filter_ordered, flatten, frozendict,
+                          infer_dtype, extract_dtype, is_integer, split, is_number)
 from devito.types import Array, DimensionTuple, Evaluable, StencilDimension
+from devito.types.basic import AbstractFunction
 
-__all__ = ['Differentiable', 'IndexDerivative', 'EvalDerivative']
+__all__ = ['Differentiable', 'DiffDerivative', 'IndexDerivative', 'EvalDerivative',
+           'Weights', 'Real', 'Imag', 'Conj']
 
 
 class Differentiable(sympy.Expr, Evaluable):
@@ -29,7 +36,7 @@ class Differentiable(sympy.Expr, Evaluable):
     # operators to be used
     _op_priority = sympy.Expr._op_priority + 1.
 
-    _state = ('space_order', 'time_order', 'indices')
+    __rkwargs__ = ('space_order', 'time_order', 'indices')
 
     @cached_property
     def _functions(self):
@@ -56,6 +63,7 @@ class Differentiable(sympy.Expr, Evaluable):
     @cached_property
     def grid(self):
         grids = {getattr(i, 'grid', None) for i in self._args_diff} - {None}
+        grids = {g.root for g in grids}
         if len(grids) > 1:
             warning("Expression contains multiple grids, returning first found")
         try:
@@ -63,6 +71,11 @@ class Differentiable(sympy.Expr, Evaluable):
         except KeyError:
             return None
 
+    @cached_property
+    def dtype(self):
+        dtypes = {f.dtype for f in self._functions} - {None}
+        return infer_dtype(dtypes)
+
     @cached_property
     def indices(self):
         return tuple(filter_ordered(flatten(getattr(i, 'indices', ())
@@ -73,6 +86,11 @@ class Differentiable(sympy.Expr, Evaluable):
         return tuple(filter_ordered(flatten(getattr(i, 'dimensions', ())
                                             for i in self._args_diff)))
 
+    @cached_property
+    def root_dimensions(self):
+        """Tuple of root Dimensions of the physical space Dimensions."""
+        return tuple(d.root for d in self.dimensions if d.is_Space)
+
     @property
     def indices_ref(self):
         """The reference indices of the object (indices at first creation)."""
@@ -97,16 +115,40 @@ class Differentiable(sympy.Expr, Evaluable):
 
     @cached_property
     def _fd(self):
-        return dict(ChainMap(*[getattr(i, '_fd', {}) for i in self._args_diff]))
+        # Filter out all args with fd order too high
+        fd_args = []
+        for f in self._args_diff:
+            try:
+                if f.space_order <= self.space_order and \
+                        (not f.is_TimeDependent or f.time_order <= self.time_order):
+                    fd_args.append(f)
+            except AttributeError:
+                pass
+        return dict(ChainMap(*[getattr(i, '_fd', {}) for i in fd_args]))
 
     @cached_property
     def _symbolic_functions(self):
         return frozenset([i for i in self._functions if i.coefficients == 'symbolic'])
 
+    @cached_property
+    def function(self):
+        if len(self._functions) == 1:
+            return set(self._functions).pop()
+        else:
+            return None
+
     @cached_property
     def _uses_symbolic_coefficients(self):
         return bool(self._symbolic_functions)
 
+    @cached_property
+    def coefficients(self):
+        coefficients = {f.coefficients for f in self._functions}
+        # If there is multiple ones, we have to revert to the highest priority
+        # i.e we have to remove symbolic
+        key = lambda x: coeff_priority.get(x, -1)
+        return sorted(coefficients, key=key, reverse=True)[0]
+
     def _eval_at(self, func):
         if not func.is_Staggered:
             # Cartesian grid, do no waste time
@@ -114,9 +156,9 @@ class Differentiable(sympy.Expr, Evaluable):
         return self.func(*[getattr(a, '_eval_at', lambda x: a)(func) for a in self.args])
 
     def _subs(self, old, new, **hints):
-        if old is self:
+        if old == self:
             return new
-        if old is new:
+        if old == new:
             return self
         args = list(self.args)
         for i, arg in enumerate(args):
@@ -135,7 +177,7 @@ class Differentiable(sympy.Expr, Evaluable):
         return .75 if self.is_TimeDependent else .5
 
     def __hash__(self):
-        return super(Differentiable, self).__hash__()
+        return super().__hash__()
 
     def __getattr__(self, name):
         """
@@ -147,7 +189,7 @@ class Differentiable(sympy.Expr, Evaluable):
         """
         if name in self._fd:
             return self._fd[name][0](self)
-        raise AttributeError("%r object has no attribute %r" % (self.__class__, name))
+        raise AttributeError(f"{self.__class__!r} object has no attribute {name!r}")
 
     # Override SymPy arithmetic operators
     @call_highest_priority('__radd__')
@@ -221,8 +263,21 @@ class Differentiable(sympy.Expr, Evaluable):
         return Mul(sympy.S.NegativeOne, self)
 
     def __eq__(self, other):
-        return super(Differentiable, self).__eq__(other) and\
-            all(getattr(self, i, None) == getattr(other, i, None) for i in self._state)
+        ret = super().__eq__(other)
+        if ret is NotImplemented or not ret:
+            # Non comparable or not equal as sympy objects
+            return False
+
+        return all(getattr(self, i, None) == getattr(other, i, None)
+                   for i in self.__rkwargs__)
+
+    def _hashable_content(self):
+        # SymPy computes the hash of all Basic objects as:
+        # `hash((type(self).__name__,) + self._hashable_content())`
+        # However, our subclasses will be named after the main SymPy classes,
+        # for example sympy.Add -> differentiable.Add, so we need to override
+        # the hashable content to specify it's our own subclasses
+        return super()._hashable_content() + ('differentiable',)
 
     @property
     def name(self):
@@ -241,24 +296,95 @@ class Differentiable(sympy.Expr, Evaluable):
         Generates a symbolic expression for the Laplacian, the second
         derivative w.r.t all spatial Dimensions.
         """
-        space_dims = [d for d in self.dimensions if d.is_Space]
-        derivs = tuple('d%s2' % d.name for d in space_dims)
-        return Add(*[getattr(self, d) for d in derivs])
+        return self.laplacian()
 
-    def div(self, shift=None):
-        space_dims = [d for d in self.dimensions if d.is_Space]
+    def laplacian(self, shift=None, order=None, method='FD', **kwargs):
+        """
+        Laplacian of the Differentiable with shifted derivatives and custom
+        FD order.
+
+        Each second derivative is left-right (i.e D^T D with D the first derivative ):
+        `(self.dx(x0=dim+shift*dim.spacing,
+                  fd_order=order)).dx(x0=dim-shift*dim.spacing, fd_order=order)`
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite differences.
+        """
+        w = kwargs.get('weights', kwargs.get('w'))
+        order = order or self.space_order
+        space_dims = self.root_dimensions
         shift_x0 = make_shift_x0(shift, (len(space_dims),))
-        return Add(*[getattr(self, 'd%s' % d.name)(x0=shift_x0(shift, d, None, i))
+        derivs = tuple(f'd{d.name}2' for d in space_dims)
+        return Add(*[getattr(self, d)(x0=shift_x0(shift, space_dims[i], None, i),
+                                      method=method, fd_order=order, w=w)
+                     for i, d in enumerate(derivs)])
+
+    def div(self, shift=None, order=None, method='FD', **kwargs):
+        """
+        Divergence of the input Function.
+
+        Parameters
+        ----------
+        func : Function or TensorFunction
+            Function to take the divergence of
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite difference coefficients.
+        """
+        w = kwargs.get('weights', kwargs.get('w'))
+        space_dims = self.root_dimensions
+        shift_x0 = make_shift_x0(shift, (len(space_dims),))
+        order = order or self.space_order
+        return Add(*[getattr(self, f'd{d.name}')(x0=shift_x0(shift, d, None, i),
+                                                 fd_order=order, method=method, w=w)
                      for i, d in enumerate(space_dims)])
 
-    def grad(self, shift=None):
+    def grad(self, shift=None, order=None, method='FD', **kwargs):
+        """
+        Gradient of the input Function.
+
+        Parameters
+        ----------
+        func : Function or TensorFunction
+            Function to take the gradient of
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite
+        """
         from devito.types.tensor import VectorFunction, VectorTimeFunction
-        space_dims = [d for d in self.dimensions if d.is_Space]
+        space_dims = self.root_dimensions
         shift_x0 = make_shift_x0(shift, (len(space_dims),))
-        comps = [getattr(self, 'd%s' % d.name)(x0=shift_x0(shift, d, None, i))
+        order = order or self.space_order
+        w = kwargs.get('weights', kwargs.get('w'))
+        comps = [getattr(self, f'd{d.name}')(x0=shift_x0(shift, d, None, i),
+                                             fd_order=order, method=method, w=w)
                  for i, d in enumerate(space_dims)]
         vec_func = VectorTimeFunction if self.is_TimeDependent else VectorFunction
-        return vec_func(name='grad_%s' % self.name, time_order=self.time_order,
+        return vec_func(name=f'grad_{self.name}', time_order=self.time_order,
                         space_order=self.space_order, components=comps, grid=self.grid)
 
     def biharmonic(self, weight=1):
@@ -266,8 +392,8 @@ class Differentiable(sympy.Expr, Evaluable):
         Generates a symbolic expression for the weighted biharmonic operator w.r.t.
         all spatial Dimensions Laplace(weight * Laplace (self))
         """
-        space_dims = [d for d in self.dimensions if d.is_Space]
-        derivs = tuple('d%s2' % d.name for d in space_dims)
+        space_dims = self.root_dimensions
+        derivs = tuple(f'd{d.name}2' for d in space_dims)
         return Add(*[getattr(self.laplace * weight, d) for d in derivs])
 
     def diff(self, *symbols, **assumptions):
@@ -311,7 +437,11 @@ class Differentiable(sympy.Expr, Evaluable):
 
 
 def highest_priority(DiffOp):
-    prio = lambda x: getattr(x, '_fd_priority', 0)
+    # We want to get the object with highest priority
+    # We also need to make sure that the object with the largest
+    # set of dimensions is used when multiple ones with the same
+    # priority appear
+    prio = lambda x: (getattr(x, '_fd_priority', 0), len(x.dimensions))
     return sorted(DiffOp._args_diff, key=prio, reverse=True)[0]
 
 
@@ -391,6 +521,7 @@ class Add(DifferentiableOp, sympy.Add):
         # set of basic simplifications
 
         # (a+b)+c -> a+b+c (flattening)
+        # TODO: use symbolics.flatten_args; not using it to avoid a circular import
         nested, others = split(args, lambda e: isinstance(e, Add))
         args = flatten(e.args for e in nested) + list(others)
 
@@ -413,23 +544,23 @@ class Mul(DifferentiableOp, sympy.Mul):
         # to avoid generating functional, but ugly, code
 
         # (a*b)*c -> a*b*c (flattening)
+        # TODO: use symbolics.flatten_args; not using it to avoid a circular import
         nested, others = split(args, lambda e: isinstance(e, Mul))
         args = flatten(e.args for e in nested) + list(others)
 
+        # Gather all numbers and simplify
+        nums, others = split(args, lambda e: is_number(e))
+        scalar = sympy.Mul(*nums)
+
         # a*0 -> 0
-        if any(i == 0 for i in args):
+        if scalar == 0:
             return sympy.S.Zero
 
         # a*1 -> a
-        args = [i for i in args if i != 1]
-
-        # a*-1 -> a*-1
-        # a*-1*-1 -> a
-        # a*-1*-1*-1 -> a*-1
-        nminus = len([i for i in args if i == sympy.S.NegativeOne])
-        args = [i for i in args if i != sympy.S.NegativeOne]
-        if nminus % 2 == 1:
-            args.append(sympy.S.NegativeOne)
+        if scalar - 1 == 0:
+            args = others
+        else:
+            args = [scalar] + others
 
         # Reorder for homogeneity with pure SymPy types
         _mulsort(args)
@@ -444,7 +575,7 @@ class Mul(DifferentiableOp, sympy.Mul):
                 r, b = b.as_coeff_Mul()
                 if r is sympy.S.One and type(b) is Add:
                     return Add(*[_keep_coeff(a, bi) for bi in b.args], evaluate=False)
-        except ValueError:
+        except (AttributeError, ValueError):
             pass
 
         return super().__new__(cls, *args, **kwargs)
@@ -467,7 +598,7 @@ class Mul(DifferentiableOp, sympy.Mul):
 
         func_args = highest_priority(self)
         new_args = []
-        ref_inds = func_args.indices_ref._getters
+        ref_inds = func_args.indices_ref.getters
 
         for f in self.args:
             if f not in self._args_diff:
@@ -475,7 +606,7 @@ class Mul(DifferentiableOp, sympy.Mul):
             elif f is func_args or isinstance(f, DifferentiableFunction):
                 new_args.append(f)
             else:
-                ind_f = f.indices_ref._getters
+                ind_f = f.indices_ref.getters
                 mapper = {ind_f.get(d, d): ref_inds.get(d, d)
                           for d in self.dimensions
                           if ind_f.get(d, d) is not ref_inds.get(d, d)}
@@ -496,13 +627,72 @@ class Mod(DifferentiableOp, sympy.Mod):
     __sympy_class__ = sympy.Mod
 
 
-class IndexSum(DifferentiableOp):
+class SafeInv(Differentiable, sympy.core.function.Application):
+    _fd_priority = 0
+
+    @property
+    def base(self):
+        return self.args[1]
+
+    @property
+    def val(self):
+        return self.args[0]
+
+    def __str__(self):
+        return Pow(self.args[0], -1).__str__()
+
+    __repr__ = __str__
+
+
+class ComplexPart(Differentiable, sympy.core.function.Application):
+    """Abstract class for `Real`, `Imag`, or `Conj` of an expression"""
+    _name = None
+
+    def __new__(cls, *args, **kwargs):
+        if len(args) != 1:
+            raise ValueError(f"{cls.__name__} expects exactly one arg;"
+                             f" {len(args)} were supplied instead.")
+
+        return super().__new__(cls, *args, **kwargs)
+
+    def __str__(self):
+        return f"{self.__class__.__name__}({self.args[0]})"
+
+    __repr__ = __str__
+
+
+class RealComplexPart(ComplexPart):
+
+    @cached_property
+    def dtype(self):
+        dtype = extract_dtype(self)
+        return dtype(0).real.__class__
+
+
+class Real(RealComplexPart):
+    """Get the real part of an expression"""
+    _name = 'real'
+
+
+class Imag(RealComplexPart):
+    """Get the imaginary part of an expression"""
+    _name = 'imag'
+
+
+class Conj(ComplexPart):
+    """Get the complex conjugate of an expression"""
+    _name = 'conj'
+
+
+class IndexSum(sympy.Expr, Evaluable):
 
     """
     Represent the summation over a multiindex, that is a collection of
     Dimensions, of an indexed expression.
     """
 
+    __rargs__ = ('expr', 'dimensions')
+
     is_commutative = True
 
     def __new__(cls, expr, dimensions, **kwargs):
@@ -517,16 +707,21 @@ class IndexSum(DifferentiableOp):
                 pass
             raise ValueError("Expected Dimension with numeric size, "
                              "got `%s` instead" % d)
-        if not expr.has_free(*dimensions):
+
+        # TODO: `has_free` only available with SymPy v>=1.10
+        # We should start using `not expr.has_free(*dimensions)` once we drop
+        # support for SymPy 1.8<=v<1.0
+        if not all(d in expr.free_symbols for d in dimensions):
             raise ValueError("All Dimensions `%s` must appear in `expr` "
                              "as free variables" % str(dimensions))
+
         for i in expr.find(IndexSum):
             for d in dimensions:
                 if d in i.dimensions:
                     raise ValueError("Dimension `%s` already appears in a "
                                      "nested tensor contraction" % d)
 
-        obj = sympy.Expr.__new__(cls, expr, *dimensions)
+        obj = sympy.Expr.__new__(cls, expr)
         obj._expr = expr
         obj._dimensions = dimensions
 
@@ -538,6 +733,14 @@ class IndexSum(DifferentiableOp):
 
     __str__ = __repr__
 
+    def _sympystr(self, printer):
+        return str(self)
+
+    _latex = _sympystr
+
+    def _hashable_content(self):
+        return super()._hashable_content() + (self.dimensions,)
+
     @property
     def expr(self):
         return self._expr
@@ -550,7 +753,7 @@ class IndexSum(DifferentiableOp):
         expr = self.expr._evaluate(**kwargs)
 
         if not kwargs.get('expand', True):
-            return self.func(expr, self.dimensions)
+            return self._rebuild(expr)
 
         values = product(*[list(d.range) for d in self.dimensions])
         terms = []
@@ -563,6 +766,8 @@ class IndexSum(DifferentiableOp):
     def free_symbols(self):
         return super().free_symbols - set(self.dimensions)
 
+    func = DifferentiableOp._rebuild
+
 
 class Weights(Array):
 
@@ -579,38 +784,132 @@ class Weights(Array):
         assert isinstance(d, StencilDimension) and d.symbolic_size == len(weights)
         assert isinstance(weights, (list, tuple, np.ndarray))
 
-        kwargs['scope'] = 'static'
+        # Normalize `weights`
+        from devito.symbolics import pow_to_mul  # noqa, sigh
+        weights = tuple(pow_to_mul(sympy.sympify(i)) for i in weights)
+
+        kwargs['scope'] = kwargs.get('scope', 'stack')
+        kwargs['initvalue'] = weights
 
         super().__init_finalize__(*args, **kwargs)
 
+    @classmethod
+    def class_key(cls):
+        # Ensure Weights appear before any other AbstractFunction
+        p, v, _ = Array.class_key()
+        return p, v - 1, cls.__name__
+
+    def __eq__(self, other):
+        return (isinstance(other, Weights) and
+                self.name == other.name and
+                self.dimension == other.dimension and
+                self.indices == other.indices and
+                self.weights == other.weights)
+
+    __hash__ = sympy.Basic.__hash__
+
+    def _hashable_content(self):
+        return (self.name, self.dimension, str(self.weights), self.scope)
+
     @property
     def dimension(self):
         return self.dimensions[0]
 
     weights = Array.initvalue
 
+    def _xreplace(self, rule):
+        if self in rule:
+            return rule[self], True
+        elif not rule:
+            return self, False
+        else:
+            try:
+                weights, flags = zip(*[i._xreplace(rule) for i in self.weights])
+                if any(flags):
+                    return self.func(initvalue=weights, function=None), True
+            except AttributeError:
+                # `float` weights
+                pass
+            return super()._xreplace(rule)
+
+    @cached_property
+    def _npweights(self):
+        # NOTE: `self.weights` cannot just be an array or SymPy will fail
+        # internally at `__eq__` since numpy arrays requite .all, not ==,
+        # for equality comparison
+        return np.array(self.weights)
+
+    def value(self, idx):
+        try:
+            v = self.weights[idx]
+        except TypeError:
+            # E.g., `idx` is a tuple
+            v = self._npweights[idx]
+        if v.is_Number or v.is_Indexed:
+            return sympy.sympify(v)
+        else:
+            return self[idx]
+
 
 class IndexDerivative(IndexSum):
 
-    def __new__(cls, expr, dimensions, **kwargs):
-        if not (expr.is_Mul and len(expr.args) == 2):
-            raise ValueError("Expect expr*weights, got `%s` instead" % str(expr))
-        _, weights = expr.args
-        if not isinstance(weights, Weights):
-            # All of the SymPy versions we support end up placing the Weights
-            # array here, so if something changes we'll get an alarm through
-            # this exception
-            raise ValueError("Couldn't find weights array")
+    __rargs__ = ('expr', 'mapper')
+
+    def __new__(cls, expr, mapper, **kwargs):
+        dimensions = as_tuple(set(mapper.values()))
+
+        # Detect the Weights among the arguments
+        weightss = []
+        for a in expr.args:
+            try:
+                f = a.function
+            except AttributeError:
+                continue
+            if isinstance(f, Weights):
+                weightss.append(a)
+
+        # Sanity check
+        if not (expr.is_Mul and len(weightss) == 1):
+            raise ValueError("Expect `expr*weights`, got `%s` instead" % str(expr))
+        weights = weightss.pop()
 
         obj = super().__new__(cls, expr, dimensions)
         obj._weights = weights
+        obj._mapper = frozendict(mapper)
 
         return obj
 
+    def _hashable_content(self):
+        return super()._hashable_content() + (self.mapper,)
+
+    def compare(self, other):
+        if self is other:
+            return 0
+        n1 = self.__class__
+        n2 = other.__class__
+        if n1.__name__ == n2.__name__:
+            return (self.weights.compare(other.weights) or
+                    self.base.compare(other.base))
+        else:
+            return super().compare(other)
+
+    @cached_property
+    def base(self):
+        return self.expr.func(*[a for a in self.expr.args if a is not self.weights])
+
     @property
     def weights(self):
         return self._weights
 
+    @property
+    def mapper(self):
+        return self._mapper
+
+    @property
+    def depth(self):
+        iderivs = self.expr.find(IndexDerivative)
+        return 1 + max([i.depth for i in iderivs], default=0)
+
     def _evaluate(self, **kwargs):
         expr = super()._evaluate(**kwargs)
 
@@ -618,11 +917,21 @@ class IndexDerivative(IndexSum):
             return expr
 
         w = self.weights
+        f = w.function
         d = w.dimension
-        mapper = {w.subs(d, i): w.weights[n] for n, i in enumerate(d.range)}
+        mapper = {w.subs(d, i): f.weights[n] for n, i in enumerate(d.range)}
         expr = expr.xreplace(mapper)
 
-        return expr
+        return EvalDerivative(*expr.args, base=self.base)
+
+
+class DiffDerivative(IndexDerivative, DifferentiableOp):
+    pass
+
+
+# SymPy args ordering is the same for Derivatives and IndexDerivatives
+for i in ('DiffDerivative', 'IndexDerivative'):
+    ordering_of_classes.insert(ordering_of_classes.index('Derivative') + 1, i)
 
 
 class EvalDerivative(DifferentiableOp, sympy.Add):
@@ -656,12 +965,17 @@ class EvalDerivative(DifferentiableOp, sympy.Add):
 
     func = DifferentiableOp._rebuild
 
+    # Since obj.base = base, then Differentiable.__eq__ leads to infinite recursion
+    # as it checks obj.base == other.base
+    __eq__ = sympy.Add.__eq__
+    __hash__ = sympy.Add.__hash__
+
     def _new_rawargs(self, *args, **kwargs):
         kwargs.pop('is_commutative', None)
         return self.func(*args, **kwargs)
 
 
-class diffify(object):
+class diffify:
 
     """
     Helper class based on single dispatch to reconstruct all nodes in a sympy
@@ -734,6 +1048,12 @@ def diff2sympy(expr):
             ax, af = _diff2sympy(a)
             args.append(ax)
             flag |= af
+
+        # Handle special objects
+        if isinstance(obj, DiffDerivative):
+            return IndexDerivative(*args, obj.mapper), True
+
+        # Handle generic objects such as arithmetic operations
         try:
             return obj.__sympy_class__(*args, evaluate=False), True
         except AttributeError:
@@ -763,3 +1083,35 @@ def diff2sympy(expr):
 evalf_table[Add] = evalf_table[sympy.Add]
 evalf_table[Mul] = evalf_table[sympy.Mul]
 evalf_table[Pow] = evalf_table[sympy.Pow]
+
+
+# Interpolation for finite differences
+@singledispatch
+def interp_for_fd(expr, x0, **kwargs):
+    return expr
+
+
+@interp_for_fd.register(sympy.Derivative)
+def _(expr, x0, **kwargs):
+    x0_expr = {d: v for d, v in x0.items() if d not in expr.dims}
+    return expr.func(interp_for_fd(expr.expr, x0_expr, **kwargs))
+
+
+@interp_for_fd.register(sympy.Expr)
+def _(expr, x0, **kwargs):
+    if expr.args:
+        return expr.func(*[interp_for_fd(i, x0, **kwargs) for i in expr.args])
+    else:
+        return expr
+
+
+@interp_for_fd.register(AbstractFunction)
+def _(expr, x0, **kwargs):
+    from devito.finite_differences.derivative import Derivative
+    x0_expr = {d: v for d, v in x0.items() if v is not expr.indices_ref[d]}
+    if x0_expr:
+        dims = tuple((d, 0) for d in x0_expr)
+        fd_o = tuple([2]*len(dims))
+        return Derivative(expr, *dims, fd_order=fd_o, x0=x0_expr)
+    else:
+        return expr
