@@ -1,9 +1,8 @@
 from sympy import sympify
 
-from .differentiable import EvalDerivative, IndexDerivative, Weights
-from .tools import (numeric_weights, symbolic_weights, left, right,
-                    generate_indices, centered, direct, transpose,
-                    check_input, check_symbolic)
+from .differentiable import EvalDerivative, DiffDerivative, Weights
+from .tools import (left, right, generate_indices, centered, direct, transpose,
+                    check_input, fd_weights_registry, process_weights)
 
 __all__ = ['first_derivative', 'cross_derivative', 'generic_derivative',
            'left', 'right', 'centered', 'transpose', 'generate_indices']
@@ -14,87 +13,7 @@ _PRECISION = 9
 
 
 @check_input
-@check_symbolic
-def first_derivative(expr, dim, fd_order=None, side=centered, matvec=direct, x0=None,
-                     symbolic=False, expand=True):
-    """
-    First-order derivative of a given expression.
-
-    Parameters
-    ----------
-    expr : expr-like
-        Expression for which the first-order derivative is produced.
-    dim : Dimension
-        The Dimension w.r.t. which to differentiate.
-    fd_order : int, optional
-        Coefficient discretization order. Note: this impacts the width of
-        the resulting stencil. Defaults to `expr.space_order`.
-    side : Side, optional
-        Side of the finite difference location, centered (at x), left (at x - 1)
-        or right (at x +1). Defaults to `centered`.
-    matvec : Transpose, optional
-        Forward (matvec=direct) or transpose (matvec=transpose) mode of the
-        finite difference. Defaults to `direct`.
-    x0 : dict, optional
-        Origin of the finite-difference scheme as a map dim: origin_dim.
-    symbolic : bool, optional
-        Use default or custom coefficients (weights). Defaults to False.
-    expand : bool, optional
-        If True, the derivative is fully expanded as a sum of products,
-        otherwise an IndexSum is returned. Defaults to True.
-
-    Returns
-    -------
-    expr-like
-        First-order derivative of ``expr``.
-
-    Examples
-    --------
-    >>> from devito import Function, Grid, first_derivative, transpose
-    >>> grid = Grid(shape=(4, 4))
-    >>> x, _ = grid.dimensions
-    >>> f = Function(name='f', grid=grid)
-    >>> g = Function(name='g', grid=grid)
-    >>> first_derivative(f*g, dim=x)
-    -f(x, y)*g(x, y)/h_x + f(x + h_x, y)*g(x + h_x, y)/h_x
-
-    Semantically, this is equivalent to
-
-    >>> (f*g).dx
-    Derivative(f(x, y)*g(x, y), x)
-
-    The only difference is that in the latter case derivatives remain unevaluated.
-    The expanded form is obtained via ``evaluate``
-
-    >>> (f*g).dx.evaluate
-    -f(x, y)*g(x, y)/h_x + f(x + h_x, y)*g(x + h_x, y)/h_x
-
-    For the adjoint mode of the first derivative, pass ``matvec=transpose``
-
-    >>> g = Function(name='g', grid=grid)
-    >>> first_derivative(f*g, dim=x, matvec=transpose)
-    -f(x, y)*g(x, y)/h_x + f(x - h_x, y)*g(x - h_x, y)/h_x
-
-    This is also accessible via the .T shortcut
-
-    >>> (f*g).dx.T.evaluate
-    -f(x, y)*g(x, y)/h_x + f(x - h_x, y)*g(x - h_x, y)/h_x
-
-    Finally the x0 argument allows to choose the origin of the finite-difference
-
-    >>> first_derivative(f, dim=x, x0={x: 1})
-    -f(1, y)/h_x + f(h_x + 1, y)/h_x
-    """
-    fd_order = fd_order or expr.space_order
-    deriv_order = 1
-
-    return make_derivative(expr, dim, fd_order, deriv_order, side,
-                           matvec, x0, symbolic, expand)
-
-
-@check_input
-@check_symbolic
-def cross_derivative(expr, dims, fd_order, deriv_order, **kwargs):
+def cross_derivative(expr, dims, fd_order, deriv_order, x0=None, side=None, **kwargs):
     """
     Arbitrary-order cross derivative of a given expression.
 
@@ -104,21 +23,22 @@ def cross_derivative(expr, dims, fd_order, deriv_order, **kwargs):
         Expression for which the cross derivative is produced.
     dims : tuple of Dimension
         Dimensions w.r.t. which to differentiate.
-    fd_order : tuple of ints
+    fd_order : int, optional, default=expr.space_order
         Coefficient discretization order. Note: this impacts the width of
         the resulting stencil.
-    deriv_order : tuple of ints
-        Derivative order, e.g. 2 for a second-order derivative.
-    matvec : Transpose, optional
+    side : Side, optional, default=centered
+        Side of the finite difference location, centered (at x), left (at x - 1)
+        or right (at x +1).
+    matvec : Transpose, optional, default=direct
         Forward (matvec=direct) or transpose (matvec=transpose) mode of the
-        finite difference. Defaults to `direct`.
-    x0 : dict, optional
+        finite difference.
+    x0 : dict, optional, default=None
         Origin of the finite-difference scheme as a map dim: origin_dim.
-    symbolic : bool, optional
-        Use default or custom coefficients (weights). Defaults to False.
-    expand : bool, optional
+    coefficients : string, optional, default='taylor'
+        Use taylor or custom coefficients (weights).
+    expand : bool, optional, default=True
         If True, the derivative is fully expanded as a sum of products,
-        otherwise an IndexSum is returned. Defaults to True.
+        otherwise an IndexSum is returned.
 
     Returns
     -------
@@ -151,21 +71,23 @@ def cross_derivative(expr, dims, fd_order, deriv_order, **kwargs):
     Finally the x0 argument allows to choose the origin of the finite-difference
 
     >>> cross_derivative(f*g, dims=(x, y), fd_order=(2, 2), deriv_order=(1, 1), \
-    x0={x: 1, y: 2})
-    (-1/h_y)*(-f(1, 2)*g(1, 2)/h_x + f(h_x + 1, 2)*g(h_x + 1, 2)/h_x) + (-f(1, h_y + 2)*\
-g(1, h_y + 2)/h_x + f(h_x + 1, h_y + 2)*g(h_x + 1, h_y + 2)/h_x)/h_y
+    x0={x: x + x.spacing, y: y + y.spacing})
+    (-1/h_y)*(-f(x + h_x, y + h_y)*g(x + h_x, y + h_y)/h_x + \
+f(x + 2*h_x, y + h_y)*g(x + 2*h_x, y + h_y)/h_x) + \
+(-f(x + h_x, y + 2*h_y)*g(x + h_x, y + 2*h_y)/h_x + \
+f(x + 2*h_x, y + 2*h_y)*g(x + 2*h_x, y + 2*h_y)/h_x)/h_y
     """
-    x0 = kwargs.get('x0', {})
+    x0 = x0 or {}
     for d, fd, dim in zip(deriv_order, fd_order, dims):
-        expr = generic_derivative(expr, dim=dim, fd_order=fd, deriv_order=d, x0=x0)
+        expr = generic_derivative(expr, dim=dim, fd_order=fd, deriv_order=d, x0=x0,
+                                  side=side, **kwargs)
 
     return expr
 
 
 @check_input
-@check_symbolic
 def generic_derivative(expr, dim, fd_order, deriv_order, matvec=direct, x0=None,
-                       symbolic=False, expand=True):
+                       coefficients='taylor', expand=True, weights=None, side=None):
     """
     Arbitrary-order derivative of a given expression.
 
@@ -175,57 +97,91 @@ def generic_derivative(expr, dim, fd_order, deriv_order, matvec=direct, x0=None,
         Expression for which the derivative is produced.
     dim : Dimension
         The Dimension w.r.t. which to differentiate.
-    fd_order : int
+    fd_order : int, optional, default=expr.space_order
         Coefficient discretization order. Note: this impacts the width of
         the resulting stencil.
-    deriv_order : int
-        Derivative order, e.g. 2 for a second-order derivative.
-    matvec : Transpose, optional
+    side : Side, optional, default=centered
+        Side of the finite difference location, centered (at x), left (at x - 1)
+        or right (at x +1).
+    matvec : Transpose, optional, default=direct
         Forward (matvec=direct) or transpose (matvec=transpose) mode of the
-        finite difference. Defaults to `direct`.
-    x0 : dict, optional
+        finite difference.
+    x0 : dict, optional, default=None
         Origin of the finite-difference scheme as a map dim: origin_dim.
-    symbolic : bool, optional
-        Use default or custom coefficients (weights). Defaults to False.
-    expand : bool, optional
+    coefficients : string, optional, default='taylor'
+        Use taylor or custom coefficients (weights).
+    expand : bool, optional, default=True
         If True, the derivative is fully expanded as a sum of products,
-        otherwise an IndexSum is returned. Defaults to True.
+        otherwise an IndexSum is returned.
 
     Returns
     -------
     expr-like
         ``deriv-order`` derivative of ``expr``.
     """
-    side = None
-    # First order derivative with 2nd order FD is highly non-recommended so taking
+    # First order derivative with 2nd order FD is strongly discouraged so taking
     # first order fd that is a lot better
-    if deriv_order == 1 and fd_order == 2 and not symbolic:
+    if deriv_order == 1 and fd_order == 2 and side is None:
         fd_order = 1
 
-    return make_derivative(expr, dim, fd_order, deriv_order, side,
-                           matvec, x0, symbolic, expand)
+    # Zeroth order derivative is just the expression itself if not shifted
+    if deriv_order == 0 and not x0:
+        return expr
 
-
-def make_derivative(expr, dim, fd_order, deriv_order, side, matvec, x0, symbolic, expand):
-    # The stencil indices
-    indices, x0 = generate_indices(expr, dim, fd_order, side=side, matvec=matvec, x0=x0)
-
-    # Finite difference weights from Taylor approximation given these positions
-    if symbolic:
-        weights = symbolic_weights(expr, deriv_order, indices, x0)
+    # Enforce stable time coefficients
+    if dim.is_Time:
+        coefficients = 'taylor'
     else:
-        weights = numeric_weights(deriv_order, indices, x0)
+        coefficients = expr.coefficients
+
+    return make_derivative(expr, dim, fd_order, deriv_order, side,
+                           matvec, x0, coefficients, expand, weights)
+
+
+# Backward compatibility
+def first_derivative(expr, dim, fd_order, **kwargs):
+    return generic_derivative(expr, dim, fd_order, 1, **kwargs)
+
+
+def make_derivative(expr, dim, fd_order, deriv_order, side, matvec, x0, coefficients,
+                    expand, weights=None):
+    # Always expand time derivatives to avoid issue with buffering and streaming.
+    # Time derivative are almost always short stencils and won't benefit from
+    # unexpansion in the rare case the derivative is not evaluated for time stepping.
+    expand = True if dim.is_Time else expand
+
+    # The stencil indices
+    nweights, wdim, scale = process_weights(weights, expr, dim)
+    indices, x0 = generate_indices(expr, dim, fd_order, side=side, matvec=matvec,
+                                   x0=x0, nweights=nweights)
+    # Finite difference weights corresponding to the indices. Computed via the
+    # `coefficients` method (`taylor` or `symbolic`)
+    if weights is None:
+        weights = fd_weights_registry[coefficients](expr, deriv_order, indices, x0)
+    # Did fd_weights_registry return a new Function/Expression instead of a values?
+    _, wdim, _ = process_weights(weights, expr, dim)
+    if wdim is not None:
+        weights = [weights._subs(wdim, i) for i in range(len(indices))]
 
     # Enforce fixed precision FD coefficients to avoid variations in results
-    weights = [sympify(w).evalf(_PRECISION) for w in weights]
+    if scale:
+        scale = dim.spacing**(-deriv_order)
+    else:
+        scale = 1
+    weights = [sympify(scale * w).evalf(_PRECISION) for w in weights]
 
     # Transpose the FD, if necessary
-    if matvec:
-        indices = indices.scale(matvec.val)
+    if matvec == transpose:
+        weights = weights[::-1]
+        indices = indices.transpose()
 
     # Shift index due to staggering, if any
     indices = indices.shift(-(expr.indices_ref[dim] - dim))
 
+    # The user may wish to restrict expansion to selected derivatives
+    if callable(expand):
+        expand = expand(dim)
+
     if not expand and indices.expr is not None:
         weights = Weights(name='w', dimensions=indices.free_dim, initvalue=weights)
 
@@ -240,7 +196,7 @@ def make_derivative(expr, dim, fd_order, deriv_order, side, matvec, x0, symbolic
             # Pure number
             pass
 
-        deriv = IndexDerivative(expr*weights, weights.dimension)
+        deriv = DiffDerivative(expr*weights, {dim: indices.free_dim})
     else:
         terms = []
         for i, c in zip(indices, weights):
