@@ -5,15 +5,22 @@ from itertools import groupby
 import numpy as np
 import sympy
 
-from devito.exceptions import InvalidOperator
-from devito.ir.support import Any, Backward, Forward, IterationSpace, PARALLEL_IF_ATOMIC
+from devito.exceptions import CompilationError
+from devito.finite_differences.elementary import Max, Min
+from devito.ir.support import (Any, Backward, Forward, IterationSpace, erange,
+                               pull_dims)
+from devito.ir.equations import OpMin, OpMax, identity_mapper
 from devito.ir.clusters.analysis import analyze
 from devito.ir.clusters.cluster import Cluster, ClusterGroup
-from devito.ir.clusters.visitors import Queue, QueueStateful, cluster_pass
-from devito.symbolics import retrieve_indexed, uxreplace, xreplace_indices
+from devito.ir.clusters.visitors import Queue, cluster_pass
+from devito.ir.support import Scope
+from devito.mpi.halo_scheme import HaloScheme, HaloTouch
+from devito.mpi.reduction_scheme import DistReduce
+from devito.symbolics import (limits_mapper, retrieve_indexed, uxreplace,
+                              xreplace_indices)
 from devito.tools import (DefaultOrderedDict, Stamp, as_mapper, flatten,
-                          is_integer, timed_pass)
-from devito.types import Array, Eq, Inc, Symbol
+                          is_integer, split, timed_pass, toposort)
+from devito.types import Array, Eq, Symbol
 from devito.types.dimension import BOTTOM, ModuloDimension
 
 __all__ = ['clusterize']
@@ -27,10 +34,11 @@ def clusterize(exprs, **kwargs):
     clusters = [Cluster(e, e.ispace) for e in exprs]
 
     # Setup the IterationSpaces based on data dependence analysis
+    clusters = impose_total_ordering(clusters)
     clusters = Schedule().process(clusters)
 
     # Handle SteppingDimensions
-    clusters = Stepper().process(clusters)
+    clusters = Stepper(**kwargs).process(clusters)
 
     # Handle ConditionalDimensions
     clusters = guard(clusters)
@@ -38,13 +46,39 @@ def clusterize(exprs, **kwargs):
     # Determine relevant computational properties (e.g., parallelism)
     clusters = analyze(clusters)
 
-    # Input normalization (e.g., SSA)
+    # Input normalization
     clusters = normalize(clusters, **kwargs)
 
+    # Derive the necessary communications for distributed-memory parallelism
+    clusters = communications(clusters)
+
     return ClusterGroup(clusters)
 
 
-class Schedule(QueueStateful):
+def impose_total_ordering(clusters):
+    """
+    Create a new sequence of Clusters whose IterationSpaces are totally ordered
+    according to a global set of relations.
+    """
+    global_relations = set().union(*[c.ispace.relations for c in clusters])
+    ordering = toposort(global_relations)
+
+    processed = []
+    for c in clusters:
+        key = lambda d: ordering.index(d)
+        try:
+            relations = {tuple(sorted(c.ispace.itdims, key=key))}
+        except ValueError:
+            # See issue #2204
+            relations = c.ispace.relations
+        ispace = c.ispace.reorder(relations=relations, mode='total')
+
+        processed.append(c.rebuild(ispace=ispace))
+
+    return processed
+
+
+class Schedule(Queue):
 
     """
     This special Queue produces a new sequence of "scheduled" Clusters, which
@@ -102,7 +136,7 @@ class Schedule(QueueStateful):
         # `clusters` are supposed to share it
         candidates = prefix[-1].dim._defines
 
-        scope = self._fetch_scope(clusters)
+        scope = Scope(flatten(c.exprs for c in clusters))
 
         # Handle the nastiest case -- ambiguity due to the presence of both a
         # flow- and an anti-dependence.
@@ -116,10 +150,12 @@ class Schedule(QueueStateful):
             require_break = scope.d_flow.cause & maybe_break
             if require_break:
                 backlog = [clusters[-1]] + backlog
-                # Try with increasingly smaller ClusterGroups until the ambiguity is gone
+                # Try with increasingly smaller ClusterGroups until the
+                # ambiguity is gone
                 return self.callback(clusters[:-1], prefix, backlog, require_break)
 
-        # Schedule Clusters over different IterationSpaces if this increases parallelism
+        # Schedule Clusters over different IterationSpaces if this increases
+        # parallelism
         for i in range(1, len(clusters)):
             if self._break_for_parallelism(scope, candidates, i):
                 return self.callback(clusters[:i], prefix, clusters[i:] + backlog,
@@ -141,8 +177,8 @@ class Schedule(QueueStateful):
         if not backlog:
             return processed
 
-        # Handle the backlog -- the Clusters characterized by flow- and anti-dependences
-        # along one or more Dimensions
+        # Handle the backlog -- the Clusters characterized by flow- and
+        # anti-dependences along one or more Dimensions
         idir = {d: Any for d in known_break}
         stamp = Stamp()
         for i, c in enumerate(list(backlog)):
@@ -198,8 +234,17 @@ def guard(clusters):
                 else:
                     d = cd.parent
 
+                # If `cd` uses, as condition, an arbitrary SymPy expression, then
+                # we must ensure to nest it inside the last of the Dimensions
+                # appearing in `cd.condition`
+                if cd._factor is not None:
+                    k = d
+                else:
+                    dims = pull_dims(cd.condition, flag=False)
+                    k = max(dims, default=d, key=lambda i: c.ispace.index(i))
+
                 # Pull `cd` from any expr
-                condition = guards.setdefault(d, [])
+                condition = guards.setdefault(k, [])
                 for e in exprs:
                     try:
                         condition.append(e.conditionals[cd])
@@ -229,8 +274,8 @@ class Stepper(Queue):
     sub-iterators induced by a SteppingDimension.
     """
 
-    def process(self, clusters):
-        return self._process_fdta(clusters, 1)
+    def __init__(self, sregistry=None, **kwargs):
+        self.sregistry = sregistry
 
     def callback(self, clusters, prefix):
         if not prefix:
@@ -262,12 +307,16 @@ class Stepper(Queue):
                 elif len(sis) == 1:
                     si = sis.pop()
                 else:
-                    raise InvalidOperator("Cannot use multiple SteppingDimensions "
-                                          "to index into a Function")
+                    raise CompilationError("Cannot use multiple SteppingDimensions "
+                                           "to index into a Function")
                 size = i.function.shape_allocated[d]
                 assert is_integer(size)
 
-                mapper[size][si].add(iaf)
+                # Resolve StencilDimensions in case of unexpanded expressions
+                # E.g. `i0 + t` -> `(t - 1, t, t + 1)`
+                iafs = erange(iaf)
+
+                mapper[size][si].update(iafs)
 
         # Construct the ModuloDimensions
         mds = []
@@ -277,10 +326,11 @@ class Stepper(Queue):
                 # SymPy's index ordering (t, t-1, t+1) afer modulo replacement so
                 # that associativity errors are consistent. This corresponds to
                 # sorting offsets {-1, 0, 1} as {0, -1, 1} assigning -inf to 0
-                siafs = sorted(iafs, key=lambda i: -np.inf if i - si == 0 else (i - si))
+                key = lambda i: -np.inf if i - si == 0 else (i - si)
+                siafs = sorted(iafs, key=key)
 
                 for iaf in siafs:
-                    name = '%s%d' % (si.name, len(mds))
+                    name = self.sregistry.make_name(prefix='t')
                     offset = uxreplace(iaf, {si: d.root})
                     mds.append(ModuloDimension(name, si, offset, size, origin=iaf))
 
@@ -294,6 +344,11 @@ class Stepper(Queue):
         # Reconstruct the Clusters
         processed = []
         for c in clusters:
+            exprs = c.exprs
+
+            sub_iterators = dict(c.ispace.sub_iterators)
+            sub_iterators[d] = [i for i in sub_iterators[d] if i not in subiters]
+
             # Apply substitutions to expressions
             # Note: In an expression, there could be `u[t+1, ...]` and `v[t+1,
             # ...]`, where `u` and `v` are TimeFunction with circular time
@@ -301,18 +356,15 @@ class Stepper(Queue):
             # indices above are therefore conceptually different, so they will
             # be replaced with the proper ModuloDimension through two different
             # calls to `xreplace_indices`
-            exprs = c.exprs
             groups = as_mapper(mds, lambda d: d.modulo)
             for size, v in groups.items():
-                mapper = {md.origin: md for md in v}
+                key = partial(rule, size)
+                subs = {md.origin: md for md in v}
+                sub_iterators[d].extend(v)
 
-                func = partial(xreplace_indices, mapper=mapper, key=partial(rule, size))
+                func = partial(xreplace_indices, mapper=subs, key=key)
                 exprs = [e.apply(func) for e in exprs]
 
-            # Augment IterationSpace
-            sub_iterators = dict(c.ispace.sub_iterators)
-            sub_iterators[d] = tuple(i for i in sub_iterators[d] + tuple(mds)
-                                     if i not in subiters)
             ispace = IterationSpace(c.ispace.intervals, sub_iterators,
                                     c.ispace.directions)
 
@@ -321,12 +373,144 @@ class Stepper(Queue):
         return processed
 
 
-def normalize(clusters, **kwargs):
-    options = kwargs['options']
-    sregistry = kwargs['sregistry']
+@timed_pass(name='communications')
+def communications(clusters):
+    """
+    Enrich a sequence of Clusters by adding special Clusters representing data
+    communications for distributed parallelism.
+    """
+    clusters = HaloComms().process(clusters)
+    clusters = reduction_comms(clusters)
 
+    return clusters
+
+
+class HaloComms(Queue):
+
+    """
+    Inject Clusters representing halo exchanges for distributed-memory parallelism.
+    """
+
+    _q_guards_in_key = True
+    _q_properties_in_key = True
+
+    B = Symbol(name='⊥')
+
+    def process(self, clusters):
+        return self._process_fatd(clusters, 1, seen=set())
+
+    def callback(self, clusters, prefix, seen=None):
+        if not prefix:
+            return clusters
+
+        d = prefix[-1].dim
+
+        # Construct a representation of the halo accesses
+        processed = []
+        for c in clusters:
+            if c.properties.is_sequential(d) or \
+               c in seen:
+                continue
+
+            hs = HaloScheme(c.exprs, c.ispace)
+            if hs.is_void or \
+               not d._defines & hs.distributed_aindices:
+                continue
+
+            points = set()
+            for f in hs.fmapper:
+                for a in c.scope.getreads(f):
+                    points.add(a.access)
+
+            # We also add all written symbols to ultimately create mock WARs
+            # with `c`, which will prevent the newly created HaloTouch to ever
+            # be rescheduled after `c` upon topological sorting
+            points.update(a.access for a in c.scope.accesses if a.is_write)
+
+            # Sort for determinism
+            # NOTE: not sorting might impact code generation. The order of
+            # the args is important because that's what search functions honor!
+            points = sorted(points, key=str)
+
+            # Construct the HaloTouch Cluster
+            expr = Eq(self.B, HaloTouch(*points, halo_scheme=hs))
+
+            key = lambda i: i in prefix[:-1] or i in hs.loc_indices
+            ispace = c.ispace.project(key)
+            # HaloTouches are not parallel
+            properties = c.properties.sequentialize()
+
+            halo_touch = c.rebuild(exprs=expr, ispace=ispace, properties=properties)
+
+            processed.append(halo_touch)
+            seen.update({halo_touch, c})
+
+        processed.extend(clusters)
+
+        return processed
+
+
+def reduction_comms(clusters):
+    processed = []
+    fifo = []
+
+    def _update(reductions):
+        for _, reds in groupby(reductions, key=lambda r: r.ispace):
+            reds = list(reds)
+            exprs = flatten([dr.exprs for dr in reds])
+            processed.append(reds[0].rebuild(exprs=exprs))
+
+    for c in clusters:
+        # Schedule the global distributed reductions encountered before `c`,
+        # if `c`'s IterationSpace is such that the reduction can be carried out
+        found, fifo = split(fifo, lambda dr: dr.ispace.is_subset(c.ispace))
+        _update(found)
+
+        # Detect the global distributed reductions in `c`
+        for e in c.exprs:
+            op = e.operation
+            if op is None or c.is_sparse:
+                continue
+
+            var = e.lhs
+            grid = c.grid
+            if grid is None:
+                continue
+
+            # Is Inc/Max/Min/... actually used for a reduction?
+            ispace = c.ispace.project(lambda d: d in var.free_symbols)
+            if ispace.itdims == c.ispace.itdims:
+                continue
+
+            # The reduced Dimensions
+            rdims = set(c.ispace.itdims) - set(ispace.itdims)
+
+            # The reduced Dimensions inducing a global distributed reduction
+            grdims = {d for d in rdims if d._defines & c.dist_dimensions}
+            if not grdims:
+                continue
+
+            # The IterationSpace within which the global distributed reduction
+            # must be carried out
+            ispace = c.ispace.prefix(lambda d: d in var.free_symbols)
+            expr = [Eq(var, DistReduce(var, op=op, grid=grid, ispace=ispace))]
+            fifo.append(c.rebuild(exprs=expr, ispace=ispace))
+
+        processed.append(c)
+
+    # Leftover reductions are placed at the very end
+    _update(fifo)
+
+    return processed
+
+
+def normalize(clusters, sregistry=None, options=None, platform=None, **kwargs):
     clusters = normalize_nested_indexeds(clusters, sregistry)
-    clusters = normalize_reductions(clusters, sregistry, options)
+    if options['mapify-reduce']:
+        clusters = normalize_reductions_dense(clusters, sregistry, platform)
+    else:
+        clusters = normalize_reductions_minmax(clusters)
+    clusters = normalize_reductions_sparse(clusters, sregistry)
 
     return clusters
 
@@ -368,21 +552,146 @@ def normalize_nested_indexeds(cluster, sregistry):
     return cluster.rebuild(processed)
 
 
-@cluster_pass(mode='all')
-def normalize_reductions(cluster, sregistry, options):
+@cluster_pass(mode='dense')
+def normalize_reductions_minmax(cluster):
     """
-    Extract the right-hand sides of reduction Eq's in to temporaries.
+    Initialize the reduction variables to their neutral element and use them
+    to compute the reduction.
     """
-    opt_mapify_reduce = options['mapify-reduce']
-
-    dims = [d for d, v in cluster.properties.items() if PARALLEL_IF_ATOMIC in v]
-
+    dims = [d for d in cluster.ispace.itdims
+            if cluster.properties.is_parallel_atomic(d)]
     if not dims:
         return cluster
 
+    init = []
     processed = []
     for e in cluster.exprs:
-        if e.is_Reduction and e.lhs.is_Indexed and cluster.is_sparse:
+        lhs, rhs = e.args
+        f = lhs.function
+
+        if e.operation is OpMin:
+            if not f.is_Input:
+                expr = Eq(lhs, limits_mapper[lhs.dtype].max)
+                ispace = cluster.ispace.project(lambda i: i not in dims)
+                init.append(cluster.rebuild(exprs=expr, ispace=ispace))
+
+            processed.append(e.func(lhs, Min(lhs, rhs)))
+
+        elif e.operation is OpMax:
+            if not f.is_Input:
+                expr = Eq(lhs, limits_mapper[lhs.dtype].min)
+                ispace = cluster.ispace.project(lambda i: i not in dims)
+                init.append(cluster.rebuild(exprs=expr, ispace=ispace))
+
+            processed.append(e.func(lhs, Max(lhs, rhs)))
+
+        else:
+            processed.append(e)
+
+    return init + [cluster.rebuild(processed)]
+
+
+def normalize_reductions_dense(cluster, sregistry, platform):
+    """
+    Extract the right-hand sides of reduction Eq's in to temporaries.
+    """
+    return _normalize_reductions_dense(cluster, {}, sregistry, platform)
+
+
+@cluster_pass(mode='dense')
+def _normalize_reductions_dense(cluster, mapper, sregistry, platform):
+    """
+    Transform augmented expressions whose left-hand side is a scalar into
+    map-reduces.
+
+    Examples
+    --------
+    Given an increment expression such as
+
+        s += f(u[x], v[x], ...)
+
+    Turn it into
+
+        r[x] = f(u[x], v[x], ...)
+        s += r[x]
+    """
+    # The candidate Dimensions along which to perform the map part
+    candidates = [d for d in cluster.ispace.itdims
+                  if cluster.properties.is_parallel_atomic(d)]
+    if not candidates:
+        return cluster
+
+    # If there are more parallel dimensions than the maximum allowed by the
+    # target platform, we must restrain the number of candidates
+    max_par_dims = platform.limits()['max-par-dims']
+    dims = candidates[-max_par_dims:]
+
+    # All other dimensions must be sequentialized because the output Array
+    # is constrained to `dims`
+    sequentialize = candidates[:-max_par_dims]
+
+    processed = []
+    properties = cluster.properties
+    for e in cluster.exprs:
+        if e.is_Reduction:
+            lhs, rhs = e.args
+
+            try:
+                f = rhs.function
+            except AttributeError:
+                f = None
+
+            if lhs.function.is_Array:
+                # Probably a compiler-generated reduction, e.g. via
+                # recursive compilation; it's an Array already, so nothing to do
+                processed.append(e)
+            elif rhs in mapper:
+                # Seen this RHS already, so reuse the Array that was created for it
+                processed.append(e.func(lhs, mapper[rhs].indexify()))
+            elif f and f.is_Array and sum(flatten(f._size_nodomain)) == 0:
+                # Special case: the RHS is an Array with no halo/padding, meaning
+                # that the written data values are contiguous in memory, hence
+                # we can simply reuse the Array itself as we're already in the
+                # desired memory layout
+                processed.append(e)
+            else:
+                name = sregistry.make_name()
+                try:
+                    grid = cluster.grid
+                except ValueError:
+                    grid = None
+                a = mapper[rhs] = Array(name=name, dtype=e.dtype, dimensions=dims,
+                                        grid=grid)
+
+                # Populate the Array (the "map" part)
+                processed.append(e.func(a.indexify(), rhs, operation=None))
+
+                # Set all untouched entried to the identity value if necessary
+                if e.conditionals:
+                    nc = {d: sympy.Not(v) for d, v in e.conditionals.items()}
+                    v = identity_mapper[e.lhs.dtype][e.operation]
+                    processed.append(
+                        e.func(a.indexify(), v, operation=None, conditionals=nc)
+                    )
+
+                processed.append(e.func(lhs, a.indexify()))
+
+                for d in sequentialize:
+                    properties = properties.sequentialize(d)
+        else:
+            processed.append(e)
+
+    return cluster.rebuild(exprs=processed, properties=properties)
+
+
+@cluster_pass(mode='sparse')
+def normalize_reductions_sparse(cluster, sregistry):
+    """
+    Extract the right-hand sides of reduction Eq's in to temporaries.
+    """
+    processed = []
+    for e in cluster.exprs:
+        if e.is_Reduction and e.lhs.is_Indexed:
             # Transform `e` such that we reduce into a scalar (ultimately via
             # atomic ops, though this part is carried out by a much later pass)
             # For example, given `i = m[p_src]` (i.e., indirection array), turn:
@@ -394,21 +703,6 @@ def normalize_reductions(cluster, sregistry, options):
             v = Symbol(name=name, dtype=e.dtype)
             processed.extend([e.func(v, e.rhs, operation=None),
                               e.func(e.lhs, v)])
-
-        elif e.is_Reduction and e.lhs.is_Symbol and opt_mapify_reduce:
-            # Transform `e` into what is in essence an explicit map-reduce
-            # For example, turn:
-            # `s += f(u[x], v[x], ...)`
-            # into
-            # `r[x] = f(u[x], v[x], ...)`
-            # `s += r[x]`
-            # This makes it much easier to parallelize the map part regardless
-            # of the target backend
-            name = sregistry.make_name()
-            a = Array(name=name, dtype=e.dtype, dimensions=dims)
-            processed.extend([Eq(a.indexify(), e.rhs),
-                              Inc(e.lhs, a.indexify())])
-
         else:
             processed.append(e)
 
