@@ -1,32 +1,39 @@
-from devito.ir.clusters.visitors import QueueStateful
+from devito.ir.clusters.cluster import Cluster
+from devito.ir.clusters.visitors import Queue
 from devito.ir.support import (AFFINE, PARALLEL, PARALLEL_INDEP, PARALLEL_IF_ATOMIC,
-                               ROUNDABLE, SEQUENTIAL, Forward)
+                               SEQUENTIAL, Property, Scope)
+from devito.ir.support.space import IterationSpace
 from devito.tools import as_tuple, flatten, timed_pass
+from devito.types.dimension import Dimension
 
 __all__ = ['analyze']
 
 
+# Describes properties fetched by a `Detector`
+Properties = dict[Cluster, dict[Dimension, set[Property]]]
+
+
 @timed_pass()
 def analyze(clusters):
-    state = QueueStateful.State()
+    properties: Properties = {}
 
     # Collect properties
-    clusters = Parallelism(state).process(clusters)
-    clusters = Affiness(state).process(clusters)
-    clusters = Rounding(state).process(clusters)
+    clusters = Parallelism().process(clusters, properties=properties)
+    clusters = Affiness().process(clusters, properties=properties)
 
     # Reconstruct Clusters attaching the discovered properties
-    processed = [c.rebuild(properties=state.properties.get(c)) for c in clusters]
+    processed = [c.rebuild(properties=properties.get(c)) for c in clusters]
 
     return processed
 
 
-class Detector(QueueStateful):
+class Detector(Queue):
 
-    def process(self, elements):
-        return self._process_fatd(elements, 1)
+    def process(self, clusters: list[Cluster], properties: Properties) -> list[Cluster]:
+        return self._process_fatd(clusters, 1, properties=properties)
 
-    def callback(self, clusters, prefix):
+    def callback(self, clusters: list[Cluster], prefix: IterationSpace | None,
+                 properties: Properties) -> list[Cluster]:
         if not prefix:
             return clusters
 
@@ -42,11 +49,19 @@ class Detector(QueueStateful):
         # Update `self.state`
         if retval:
             for c in clusters:
-                properties = self.state.properties.setdefault(c, {})
-                properties.setdefault(d, set()).update(retval)
+                c_properties = properties.setdefault(c, {})
+                c_properties.setdefault(d, set()).update(retval)
 
         return clusters
 
+    def _callback(self, clusters: list[Cluster], dim: Dimension,
+                  prefix: IterationSpace | None) -> set[Property]:
+        """
+        Callback to be implemented by subclasses. It should return a set of
+        properties for the given dimension.
+        """
+        raise NotImplementedError()
+
 
 class Parallelism(Detector):
 
@@ -73,10 +88,10 @@ class Parallelism(Detector):
             the 'write' is known to be an associative and commutative increment
     """
 
-    def _callback(self, clusters, d, prefix):
+    def _callback(self, clusters, dim, prefix):
         # Rule out if non-unitary increment Dimension (e.g., `t0=(time+1)%2`)
-        if any(c.sub_iterators[d] for c in clusters):
-            return SEQUENTIAL
+        if any(c.sub_iterators[dim] for c in clusters):
+            return {SEQUENTIAL}
 
         # All Dimensions up to and including `i-1`
         prev = flatten(i.dim._defines for i in prefix[:-1])
@@ -84,16 +99,16 @@ class Parallelism(Detector):
         is_parallel_indep = True
         is_parallel_atomic = False
 
-        scope = self._fetch_scope(clusters)
+        scope = Scope(flatten(c.exprs for c in clusters))
         for dep in scope.d_all_gen():
-            test00 = dep.is_indep(d) and not dep.is_storage_related(d)
+            test00 = dep.is_indep(dim) and not dep.is_storage_related(dim)
             test01 = all(dep.is_reduce_atmost(i) for i in prev)
             if test00 and test01:
                 continue
 
             test1 = len(prev) > 0 and any(dep.is_carried(i) for i in prev)
             if test1:
-                is_parallel_indep &= (dep.distance_mapper.get(d.root) == 0)
+                is_parallel_indep &= (dep.distance_mapper.get(dim.root) == 0)
                 continue
 
             if dep.function in scope.initialized:
@@ -104,49 +119,14 @@ class Parallelism(Detector):
                 is_parallel_atomic = True
                 continue
 
-            return SEQUENTIAL
+            return {SEQUENTIAL}
 
         if is_parallel_atomic:
-            return PARALLEL_IF_ATOMIC
+            return {PARALLEL_IF_ATOMIC}
         elif is_parallel_indep:
             return {PARALLEL, PARALLEL_INDEP}
         else:
-            return PARALLEL
-
-
-class Rounding(Detector):
-
-    def _callback(self, clusters, d, prefix):
-        itinterval = prefix[-1]
-
-        # The iteration direction must be Forward -- ROUNDABLE is for rounding *up*
-        if itinterval.direction is not Forward:
-            return
-
-        properties = self._fetch_properties(clusters, prefix)
-        if PARALLEL not in properties[d]:
-            return
-
-        scope = self._fetch_scope(clusters)
-
-        # All accessed Functions must have enough room in the PADDING region
-        # so that `i`'s trip count can safely be rounded up
-        # Note: autopadding guarantees that the padding size along the
-        # Fastest Varying Dimension is a multiple of the SIMD vector length
-        functions = [f for f in scope.functions if f.is_AbstractFunction]
-        try:
-            if any(not f._honors_autopadding for f in functions):
-                return
-        except ValueError:
-            # E.g., lazily allocated Functions don't have an accessible
-            # `.shape` until after the first call to `f._arg_values`
-            return
-
-        # Mixed data types (e.g., float and double) is unsupported
-        if len({f.dtype for f in functions}) > 1:
-            return
-
-        return ROUNDABLE
+            return {PARALLEL}
 
 
 class Affiness(Detector):
@@ -155,8 +135,11 @@ class Affiness(Detector):
     Detect the AFFINE Dimensions.
     """
 
-    def _callback(self, clusters, d, prefix):
-        scope = self._fetch_scope(clusters)
+    def _callback(self, clusters, dim, prefix):
+        scope = Scope(flatten(c.exprs for c in clusters))
         accesses = [a for a in scope.accesses if not a.is_scalar]
-        if all(a.is_regular and a.affine_if_present(d._defines) for a in accesses):
-            return AFFINE
+
+        if all(a.is_regular and a.affine_if_present(dim._defines) for a in accesses):
+            return {AFFINE}
+
+        return set()
