@@ -1,20 +1,26 @@
 from itertools import chain
+from functools import cached_property
 
 import numpy as np
-from cached_property import cached_property
 
 from devito.ir.equations import ClusterizedEq
-from devito.ir.support import (PARALLEL, PARALLEL_IF_PVT, BaseGuardBoundNext, Forward,
-                               Interval, IntervalGroup, IterationSpace, DataSpace, Scope,
-                               detect_accesses, detect_io, normalize_properties,
-                               normalize_syncs)
+from devito.ir.support import (
+    PARALLEL, PARALLEL_IF_PVT, BaseGuardBoundNext, Forward, Interval, IntervalGroup,
+    IterationSpace, DataSpace, Guards, Properties, Scope, WaitLock, WithLock,
+    PrefetchUpdate, detect_accesses, detect_io, normalize_properties,
+    tailor_properties, update_properties, normalize_syncs, minimum, maximum,
+    null_ispace
+)
+from devito.mpi.halo_scheme import HaloScheme, HaloTouch
+from devito.mpi.reduction_scheme import DistReduce
 from devito.symbolics import estimate_cost
-from devito.tools import as_tuple, flatten, frozendict
+from devito.tools import as_tuple, filter_ordered, flatten, infer_dtype
+from devito.types import Fence, WeakFence, CriticalRegion
 
 __all__ = ["Cluster", "ClusterGroup"]
 
 
-class Cluster(object):
+class Cluster:
 
     """
     A Cluster is an ordered sequence of expressions in an IterationSpace.
@@ -24,7 +30,7 @@ class Cluster(object):
     exprs : expr-like or list of expr-like
         An ordered sequence of expressions computing a tensor.
     ispace : IterationSpace, optional
-        The cluster iteration space.
+        The Cluster iteration space.
     guards : dict, optional
         Mapper from Dimensions to expr-like, representing the conditions under
         which the Cluster should be computed.
@@ -35,19 +41,22 @@ class Cluster(object):
         Mapper from Dimensions to lists of SyncOps, that is ordered sequences of
         synchronization operations that must be performed in order to compute the
         Cluster asynchronously.
+    halo_scheme : HaloScheme, optional
+        The halo exchanges required by the Cluster.
     """
 
-    def __init__(self, exprs, ispace=None, guards=None, properties=None, syncs=None):
-        ispace = ispace or IterationSpace([])
-
+    def __init__(self, exprs, ispace=null_ispace, guards=None, properties=None,
+                 syncs=None, halo_scheme=None):
         self._exprs = tuple(ClusterizedEq(e, ispace=ispace) for e in as_tuple(exprs))
         self._ispace = ispace
-        self._guards = frozendict(guards or {})
-        self._syncs = frozendict(syncs or {})
+        self._guards = Guards(guards or {})
+        self._syncs = normalize_syncs(syncs or {})
 
-        properties = dict(properties or {})
-        properties.update({i.dim: properties.get(i.dim, set()) for i in ispace.intervals})
-        self._properties = frozendict(properties)
+        properties = Properties(properties or {})
+        properties = tailor_properties(properties, ispace)
+        self._properties = update_properties(properties, self.exprs)
+
+        self._halo_scheme = halo_scheme
 
     def __repr__(self):
         return "Cluster([%s])" % ('\n' + ' '*9).join('%s' % i for i in self.exprs)
@@ -72,10 +81,7 @@ class Cluster(object):
 
         guards = root.guards
 
-        properties = {}
-        for c in clusters:
-            for d, v in c.properties.items():
-                properties[d] = normalize_properties(properties.get(d, v), v)
+        properties = reduce_properties(clusters)
 
         try:
             syncs = normalize_syncs(*[c.syncs for c in clusters])
@@ -83,7 +89,9 @@ class Cluster(object):
             raise ValueError("Cannot build a Cluster from Clusters with "
                              "non-compatible synchronization operations")
 
-        return Cluster(exprs, ispace, guards, properties, syncs)
+        halo_scheme = HaloScheme.union([c.halo_scheme for c in clusters])
+
+        return Cluster(exprs, ispace, guards, properties, syncs, halo_scheme)
 
     def rebuild(self, *args, **kwargs):
         """
@@ -98,11 +106,12 @@ class Cluster(object):
                 raise ValueError("`exprs` provided both as arg and kwarg")
             kwargs['exprs'] = args[0]
 
-        return Cluster(exprs=kwargs.get('exprs', self.exprs),
-                       ispace=kwargs.get('ispace', self.ispace),
-                       guards=kwargs.get('guards', self.guards),
-                       properties=kwargs.get('properties', self.properties),
-                       syncs=kwargs.get('syncs', self.syncs))
+        return self.__class__(exprs=kwargs.get('exprs', self.exprs),
+                              ispace=kwargs.get('ispace', self.ispace),
+                              guards=kwargs.get('guards', self.guards),
+                              properties=kwargs.get('properties', self.properties),
+                              syncs=kwargs.get('syncs', self.syncs),
+                              halo_scheme=kwargs.get('halo_scheme', self.halo_scheme))
 
     @property
     def exprs(self):
@@ -136,6 +145,10 @@ class Cluster(object):
     def syncs(self):
         return self._syncs
 
+    @property
+    def halo_scheme(self):
+        return self._halo_scheme
+
     @cached_property
     def free_symbols(self):
         return set().union(*[e.free_symbols for e in self.exprs])
@@ -152,7 +165,21 @@ class Cluster(object):
         example, reduction or redundant (i.e., invariant) Dimensions won't
         appear in an expression.
         """
-        return {i for i in self.free_symbols if i.is_Dimension}
+        idims = set.union(*[set(e.implicit_dims) for e in self.exprs])
+        return {i for i in self.free_symbols if i.is_Dimension} | idims
+
+    @cached_property
+    def dist_dimensions(self):
+        """
+        The Cluster's distributed Dimensions.
+        """
+        ret = set()
+        for f in self.functions:
+            try:
+                ret.update(f._dist_dimensions)
+            except AttributeError:
+                pass
+        return frozenset(ret)
 
     @cached_property
     def scope(self):
@@ -163,52 +190,105 @@ class Cluster(object):
         return self.scope.functions
 
     @cached_property
-    def has_increments(self):
-        return any(e.is_Increment for e in self.exprs)
+    def grid(self):
+        grids = set(f.grid for f in self.functions if f.is_AbstractFunction)
+        grids.discard(None)
+        grids = {g.root for g in grids}
+        if len(grids) == 0:
+            return None
+        elif len(grids) == 1:
+            return grids.pop()
+        else:
+            raise ValueError("Cluster has no unique Grid")
 
     @cached_property
     def is_scalar(self):
         return not any(f.is_Function for f in self.scope.writes)
 
-    @cached_property
-    def grid(self):
-        grids = set(f.grid for f in self.functions if f.is_DiscreteFunction) - {None}
-        if len(grids) == 1:
-            return grids.pop()
-        else:
-            raise ValueError("Cluster has no unique Grid")
-
     @cached_property
     def is_dense(self):
         """
-        A Cluster is dense if at least one of the following conditions is True:
+        True if at least one of the following conditions are True:
 
             * It is defined over a unique Grid and all of the Grid Dimensions
               are PARALLEL.
             * Only DiscreteFunctions are written and only affine index functions
               are used (e.g., `a[x+1, y-2]` is OK, while `a[b[x], y-2]` is not)
+
+        False in all other cases.
         """
         # Hopefully it's got a unique Grid and all Dimensions are PARALLEL (or
         # at most PARALLEL_IF_PVT). This is a quick and easy check so we try it first
         try:
             pset = {PARALLEL, PARALLEL_IF_PVT}
-            grid = self.grid
-            for d in grid.dimensions:
-                if not any(pset & v for k, v in self.properties.items()
-                           if d in k._defines):
-                    raise ValueError
-            return True
-        except ValueError:
+            target = set(self.grid.dimensions)
+            dims = {d for d in self.properties if d._defines & target}
+            if any(pset & self.properties[d] for d in dims):
+                return True
+        except (AttributeError, ValueError):
             pass
 
         # Fallback to legacy is_dense checks
         return (not any(e.conditionals for e in self.exprs) and
                 not any(f.is_SparseFunction for f in self.functions) and
+                not self.is_wild and
                 all(a.is_regular for a in self.scope.accesses))
 
-    @property
+    @cached_property
     def is_sparse(self):
-        return not self.is_dense
+        """
+        True if it represents a sparse operation, i.e iff there's at least
+        one irregular access, False otherwise.
+        """
+        return any(a.is_irregular for a in self.scope.accesses)
+
+    @cached_property
+    def is_wild(self):
+        """
+        True if encoding a non-mathematical operation, False otherwise.
+        """
+        return (self.is_halo_touch or
+                self.is_dist_reduce or
+                self.is_weak_fence or
+                self.is_critical_region)
+
+    @cached_property
+    def is_halo_touch(self):
+        return self.exprs and all(isinstance(e.rhs, HaloTouch) for e in self.exprs)
+
+    @cached_property
+    def is_dist_reduce(self):
+        return self.exprs and all(isinstance(e.rhs, DistReduce) for e in self.exprs)
+
+    @cached_property
+    def is_fence(self):
+        return (self.exprs and all(isinstance(e.rhs, Fence) for e in self.exprs) or
+                self.is_critical_region)
+
+    @cached_property
+    def is_weak_fence(self):
+        return self.exprs and all(isinstance(e.rhs, WeakFence) for e in self.exprs)
+
+    @cached_property
+    def is_critical_region(self):
+        return self.exprs and all(isinstance(e.rhs, CriticalRegion) for e in self.exprs)
+
+    @cached_property
+    def is_async(self):
+        """
+        True if an asynchronous Cluster, False otherwise.
+        """
+        return any(isinstance(s, (WithLock, PrefetchUpdate))
+                   for s in flatten(self.syncs.values()))
+
+    @cached_property
+    def is_wait(self):
+        """
+        True if a Cluster waiting on a lock (that is a special synchronization
+        operation), False otherwise.
+        """
+        return any(isinstance(s, WaitLock)
+                   for s in flatten(self.syncs.values()))
 
     @cached_property
     def dtype(self):
@@ -230,17 +310,8 @@ class Cluster(object):
             except TypeError:
                 # E.g. `i.dtype` is a ctypes pointer, which has no dtype equivalent
                 pass
-        fdtypes = {i for i in dtypes if np.issubdtype(i, np.floating)}
 
-        if len(fdtypes) > 1:
-            return max(fdtypes, key=lambda i: np.dtype(i).itemsize)
-        elif len(fdtypes) == 1:
-            return fdtypes.pop()
-        elif len(dtypes) == 1:
-            return dtypes.pop()
-        else:
-            # E.g., mixed integer arithmetic
-            return max(dtypes, key=lambda i: np.dtype(i).itemsize, default=None)
+        return infer_dtype(dtypes)
 
     @cached_property
     def dspace(self):
@@ -257,7 +328,10 @@ class Cluster(object):
             if f is None:
                 continue
 
-            intervals = [Interval(d, min(offs), max(offs)) for d, offs in v.items()]
+            intervals = [Interval(d,
+                                  min([minimum(i, ispace=self.ispace) for i in o]),
+                                  max([maximum(i, ispace=self.ispace) for i in o]))
+                         for d, o in v.items()]
             intervals = IntervalGroup(intervals)
 
             # Factor in the IterationSpace -- if the min/max points aren't zero,
@@ -278,13 +352,17 @@ class Cluster(object):
                 if len(ret) != 1:
                     continue
                 if ret.pop().direction is Forward:
-                    intervals = intervals.translate(d, v1=-1)
+                    intervals = intervals.translate(d._defines, v1=-1)
                 else:
-                    intervals = intervals.translate(d, 1)
+                    intervals = intervals.translate(d._defines, 1)
+            for d in self.properties:
+                if self.properties.is_inbound(d):
+                    intervals = intervals.zero(d._defines)
 
             # Special case: if the factor of a ConditionalDimension has value 1,
             # then we can safely resort to the parent's Interval
-            intervals = intervals.promote(lambda d: d.is_Conditional and d.factor == 1)
+            key = lambda d: d.is_Conditional and d.condition is None and d.factor == 1
+            intervals = intervals.promote(key)
 
             parts[f] = intervals
 
@@ -311,6 +389,7 @@ class Cluster(object):
         # Construct the `intervals` of the DataSpace, that is a global,
         # Dimension-centric view of the data space
         intervals = IntervalGroup.generate('union', *parts.values())
+
         # E.g., `db0 -> time`, but `xi NOT-> x`
         intervals = intervals.promote(lambda d: not d.is_Sub)
         intervals = intervals.zero(set(intervals.dimensions) - oobs)
@@ -334,6 +413,11 @@ class Cluster(object):
         """
         reads, writes = detect_io(self.exprs, relax=True)
         accesses = [(i, 'r') for i in reads] + [(i, 'w') for i in writes]
+
+        # Ordering isn't important at this point, so returning an unordered
+        # collection makes the caller's life easier
+        uispace = self.ispace.reorder(mode='unordered')
+
         ret = {}
         for i, mode in accesses:
             if not i.is_AbstractFunction:
@@ -341,12 +425,23 @@ class Cluster(object):
             elif i in self.dspace.parts:
                 # Stencils extend the data spaces beyond the iteration spaces
                 intervals = self.dspace.parts[i]
+
                 # Assume that invariant dimensions always cause new loads/stores
-                invariants = self.ispace.intervals.drop(intervals.dimensions)
+                invariants = uispace.intervals.drop(intervals.dimensions)
+
                 intervals = intervals.generate('union', invariants, intervals)
-                ret[(i, mode)] = intervals
+
+                # Bundles impact depends on the number of components
+                try:
+                    v = len(i.components)
+                except AttributeError:
+                    v = 1
+
+                for n in range(v):
+                    ret[(i, mode, n)] = intervals
             else:
-                ret[(i, mode)] = self.ispace.intervals
+                ret[(i, mode, 0)] = uispace.intervals
+
         return ret
 
 
@@ -359,13 +454,13 @@ class ClusterGroup(tuple):
     ----------
     clusters : tuple of Clusters
         Input elements.
-    itintervals : tuple of IterationIntervals, optional
-        The region of iteration space shared by the ``clusters``.
+    ispace : IterationSpace, optional
+        The IterationSpace section shared by all `clusters`.
     """
 
-    def __new__(cls, clusters, itintervals=None):
-        obj = super(ClusterGroup, cls).__new__(cls, flatten(as_tuple(clusters)))
-        obj._itintervals = itintervals
+    def __new__(cls, clusters, ispace=None):
+        obj = super().__new__(cls, flatten(as_tuple(clusters)))
+        obj._ispace = ispace
         return obj
 
     @classmethod
@@ -381,25 +476,36 @@ class ClusterGroup(tuple):
         return Scope(exprs=self.exprs)
 
     @cached_property
-    def itintervals(self):
-        """The prefix IterationIntervals common to all Clusters in self."""
-        return self._itintervals
+    def ispace(self):
+        return self._ispace
+
+    @cached_property
+    def properties(self):
+        return tailor_properties(reduce_properties(self), self.ispace)
 
     @cached_property
     def guards(self):
-        """The guards of each Cluster in self."""
-        return tuple(i.guards for i in self)
+        """
+        A view of the ClusterGroup's guards.
+        """
+        return tuple(filter_ordered(i.guards for i in self))
 
     @cached_property
     def syncs(self):
-        """The synchronization operations of each Cluster in self."""
-        return tuple(i.syncs for i in self)
+        """
+        A view of the ClusterGroup's synchronization operations.
+        """
+        return normalize_syncs(*[c.syncs for c in self], strict=False)
 
     @cached_property
     def dspace(self):
         """Return the DataSpace of this ClusterGroup."""
         return DataSpace.union(*[i.dspace.reset() for i in self])
 
+    @property
+    def is_halo_touch(self):
+        return all(i.is_halo_touch for i in self)
+
     @cached_property
     def dtype(self):
         """
@@ -411,17 +517,9 @@ class ClusterGroup(tuple):
         If two Clusters perform calculations with different precision, the
         data type with highest precision is returned.
         """
-        dtypes = {i.dtype for i in self}
-        fdtypes = {i for i in dtypes if np.issubdtype(i, np.floating)}
-        if len(fdtypes) > 1:
-            return max(fdtypes, key=lambda i: np.dtype(i).itemsize)
-        elif len(fdtypes) == 1:
-            return fdtypes.pop()
-        elif len(dtypes) == 1:
-            return dtypes.pop()
-        else:
-            # E.g., mixed integer arithmetic
-            return max(dtypes, key=lambda i: np.dtype(i).itemsize)
+        dtypes = {i.dtype for i in self} - {None}
+
+        return infer_dtype(dtypes)
 
     @cached_property
     def meta(self):
@@ -432,3 +530,18 @@ class ClusterGroup(tuple):
             The data type and the data space of the ClusterGroup.
         """
         return (self.dtype, self.dspace)
+
+
+# *** Utils
+
+def reduce_properties(clusters):
+    """
+    The normalized intersection (basically, a reduction) of the Properties in
+    `clusters`.
+    """
+    properties = {}
+    for c in clusters:
+        for d, v in c.properties.items():
+            properties[d] = normalize_properties(properties.get(d, v), v)
+
+    return Properties(properties)
