@@ -1,8 +1,9 @@
 """The Iteration/Expression Tree (IET) hierarchy."""
 
 import abc
+import ctypes
 import inspect
-from cached_property import cached_property
+from functools import cached_property
 from collections import OrderedDict, namedtuple
 from collections.abc import Iterable
 
@@ -10,23 +11,26 @@ import cgen as c
 from sympy import IndexedBase, sympify
 
 from devito.data import FULL
+from devito.ir.cgen import ccode
 from devito.ir.equations import DummyEq, OpInc, OpMin, OpMax
-from devito.ir.support import (SEQUENTIAL, PARALLEL, PARALLEL_IF_ATOMIC,
-                               PARALLEL_IF_PVT, VECTORIZED, AFFINE, COLLAPSED,
-                               Property, Forward, detect_io)
-from devito.symbolics import ListInitializer, CallFromPointer, ccode
-from devito.tools import Signer, Tag, as_tuple, filter_ordered, filter_sorted, flatten
-from devito.types.basic import AbstractFunction, AbstractSymbol
-from devito.types.object import AbstractObject
-from devito.types import Indexed, Symbol
+from devito.ir.support import (INBOUND, SEQUENTIAL, PARALLEL, PARALLEL_IF_ATOMIC,
+                               PARALLEL_IF_PVT, VECTORIZED, AFFINE, Property,
+                               Forward, WithLock, PrefetchUpdate, detect_io)
+from devito.symbolics import ListInitializer, CallFromPointer
+from devito.tools import (Signer, as_tuple, filter_ordered, filter_sorted, flatten,
+                          ctypes_to_cstr)
+from devito.types.basic import (AbstractFunction, AbstractSymbol, Basic, Indexed,
+                                Symbol)
+from devito.types.object import AbstractObject, LocalObject
 
-__all__ = ['Node', 'Block', 'Expression', 'Callable', 'Call',
-           'Conditional', 'Iteration', 'List', 'Section', 'TimedList', 'Prodder',
-           'MetaCall', 'PointerCast', 'HaloSpot', 'Definition', 'ExpressionBundle',
-           'AugmentedExpression', 'Increment', 'Return', 'While',
-           'ParallelIteration', 'ParallelBlock', 'Dereference', 'Lambda',
-           'SyncSpot', 'Pragma', 'DummyExpr', 'BlankLine', 'ParallelTree',
-           'BusyWait', 'CallableBody', 'Transfer', 'HPtr', 'DPtr']
+__all__ = ['Node', 'MultiTraversable', 'Block', 'Expression', 'Callable',
+           'Call', 'ExprStmt', 'Conditional', 'Iteration', 'List', 'Section',
+           'TimedList', 'Prodder', 'MetaCall', 'PointerCast', 'HaloSpot',
+           'Definition', 'ExpressionBundle', 'AugmentedExpression', 'Break',
+           'Increment', 'Return', 'While', 'ListMajor', 'ParallelIteration',
+           'ParallelBlock', 'Dereference', 'Lambda', 'SyncSpot', 'Pragma',
+           'DummyExpr', 'BlankLine', 'ParallelTree', 'BusyWait', 'UsingNamespace',
+           'Using', 'CallableBody', 'Transfer', 'EmptyList']
 
 # First-class IET nodes
 
@@ -61,7 +65,7 @@ class Node(Signer):
     """
 
     def __new__(cls, *args, **kwargs):
-        obj = super(Node, cls).__new__(cls)
+        obj = super().__new__(cls)
         argnames, _, _, defaultvalues, _, _, _ = inspect.getfullargspec(cls.__init__)
         try:
             defaults = dict(zip(argnames[-len(defaultvalues):], defaultvalues))
@@ -98,10 +102,16 @@ class Node(Signer):
 
     @property
     def view(self):
-        """A representation of the IET rooted in ``self``."""
+        """A high-level representation of the IET rooted in `self`."""
         from devito.ir.iet.visitors import printAST
         return printAST(self)
 
+    @property
+    def view_cir(self):
+        from devito.ir.iet.visitors import CGen
+        from devito.passes.iet.languages.CIR import CIRPrinter
+        return str(CGen(printer=CIRPrinter).visit(self))
+
     @property
     def children(self):
         """Return the traversable children."""
@@ -125,12 +135,12 @@ class Node(Signer):
 
     @property
     def functions(self):
-        """All AbstractFunction objects used by this node."""
+        """All AbstractFunctions and AbstractObjects used by this node."""
         return ()
 
     @property
     def expr_symbols(self):
-        """All symbols appearing in an expression within this node."""
+        """All symbols appearing within this node."""
         return ()
 
     @property
@@ -144,10 +154,10 @@ class Node(Signer):
         return ()
 
     def _signature_items(self):
-        return (str(self.ccode),)
+        return (self.view_cir,)
 
 
-class ExprStmt(object):
+class ExprStmt:
 
     """
     A mixin for Nodes that represent C expression statements, which are expressions
@@ -168,6 +178,15 @@ class ExprStmt(object):
     pass
 
 
+class MultiTraversable(Node):
+
+    """
+    An abstract base class for Nodes comprising more than one traversable children.
+    """
+
+    pass
+
+
 class List(Node):
 
     """A sequence of Nodes."""
@@ -178,7 +197,7 @@ class List(Node):
 
     def __init__(self, header=None, body=None, footer=None):
         body = as_tuple(body)
-        if len(body) == 1 and all(type(i) == List for i in [self, body[0]]):
+        if len(body) == 1 and all(type(i) is List for i in [self, body[0]]):
             # De-nest Lists
             #
             # Note: to avoid disgusting metaclass voodoo (due to
@@ -199,6 +218,11 @@ class List(Node):
                                       len(self.body), len(self.footer))
 
 
+class EmptyList(List):
+    """A plain List node without a body, such as a header/footer only"""
+    _traversable = []
+
+
 class Block(List):
 
     """A sequence of Nodes, wrapped in a block {...}."""
@@ -233,21 +257,17 @@ class Call(ExprStmt, Node):
         Defaults to False.
     writes : list, optional
         The AbstractFunctions that will be written to by the called function.
-        Explicitly tagging these AbstractFunctions is useful in the case of external
-        calls, that is whenever the compiler would be unable to retrieve that
-        information by analysis of the IET graph.
-    types : tuple, optional
-        An N-tuple of CallArgTypes for an N-arguments Call. It may be used to
-        describe the input arguments types when it's neither implicit in the
-        argument (e.g., a generic SymPy expression, host pointer vs device
-        pointer) nor evident from the context (e.g., the Call represents a call
-        to an external routine rather than to a local Callable).
+        Explicitly tagging these AbstractFunctions is useful in the case of
+        external calls, that is whenever the compiler would be unable to
+        retrieve that information by analysis of the IET graph.
+    templates : list of Basic, optional
+        The template arguments of the Call.
     """
 
     is_Call = True
 
     def __init__(self, name, arguments=None, retobj=None, is_indirect=False,
-                 cast=False, writes=None, types=None):
+                 cast=False, writes=None, templates=None):
         if isinstance(name, CallFromPointer):
             self.base = name.base
         else:
@@ -258,15 +278,22 @@ class Call(ExprStmt, Node):
         self.is_indirect = is_indirect
         self.cast = cast
         self._writes = as_tuple(writes)
-        self.types = as_tuple(types)
-
-        # Sanity check
-        assert not self.types or len(self.types) == len(self.arguments)
+        self.templates = as_tuple(templates)
 
     def __repr__(self):
         ret = "" if self.retobj is None else "%s = " % self.retobj
         return "%sCall::\n\t%s(...)" % (ret, self.name)
 
+    def _rebuild(self, *args, **kwargs):
+        if args:
+            # Not elegant, but basically it handles the fact that a Call might
+            # have nested Calls/Lambdas among its `arguments`, and these might
+            # change, and we are in such a case *if and only if* we have `args`
+            assert len(args) == len(self.children)
+            mapper = dict(zip(self.children, args))
+            kwargs['arguments'] = [mapper.get(i, i) for i in self.arguments]
+        return super()._rebuild(**kwargs)
+
     @property
     def children(self):
         return tuple(i for i in self.arguments if isinstance(i, (Call, Lambda)))
@@ -291,10 +318,13 @@ class Call(ExprStmt, Node):
                             retval.append(s.function)
                     except AttributeError:
                         continue
+
         if self.base is not None:
             retval.append(self.base.function)
+
         if self.retobj is not None:
             retval.append(self.retobj.function)
+
         return tuple(filter_ordered(retval))
 
     @cached_property
@@ -304,7 +334,7 @@ class Call(ExprStmt, Node):
             if isinstance(i, AbstractFunction):
                 continue
             elif isinstance(i, (Indexed, IndexedBase, AbstractObject, Symbol)):
-                retval.append(i)
+                retval.extend(i.free_symbols)
             elif isinstance(i, Call):
                 retval.extend(i.expr_symbols)
             else:
@@ -312,18 +342,21 @@ class Call(ExprStmt, Node):
                     retval.extend(i.free_symbols)
                 except AttributeError:
                     pass
+
         if self.base is not None:
             retval.append(self.base)
-        if self.retobj is not None:
+
+        if isinstance(self.retobj, Indexed):
+            retval.extend(self.retobj.free_symbols)
+        elif self.retobj is not None:
             retval.append(self.retobj)
+
         return tuple(filter_ordered(retval))
 
     @property
     def defines(self):
         ret = ()
-        if self.base is not None:
-            ret += (self.base,)
-        if self.retobj is not None:
+        if isinstance(self.retobj, Basic):
             ret += (self.retobj,)
         return ret
 
@@ -358,8 +391,9 @@ class Expression(ExprStmt, Node):
         self.operation = operation
 
     def __repr__(self):
-        return "<%s::%s>" % (self.__class__.__name__,
-                             filter_ordered([f.func for f in self.functions]))
+        return "<%s::%s=%s>" % (self.__class__.__name__,
+                                type(self.write),
+                                ','.join('%s' % type(f) for f in self.functions))
 
     @property
     def dtype(self):
@@ -388,7 +422,7 @@ class Expression(ExprStmt, Node):
     @property
     def is_scalar(self):
         """True if the LHS is a scalar, False otherwise."""
-        return isinstance(self.expr.lhs, (AbstractSymbol, IndexedBase))
+        return isinstance(self.expr.lhs, (AbstractSymbol, IndexedBase, LocalObject))
 
     @property
     def is_tensor(self):
@@ -405,8 +439,8 @@ class Expression(ExprStmt, Node):
         """
         True if it can be an initializing assignment, False otherwise.
         """
-        return ((self.is_scalar and not self.is_reduction) or
-                (self.is_tensor and isinstance(self.expr.rhs, ListInitializer)))
+        return (((self.is_scalar and not self.is_reduction) or
+                 (self.is_tensor and isinstance(self.expr.rhs, ListInitializer))))
 
     @property
     def defines(self):
@@ -557,11 +591,8 @@ class Iteration(Node):
         return VECTORIZED in self.properties
 
     @property
-    def ncollapsed(self):
-        for i in self.properties:
-            if i.name == 'collapsed':
-                return i.val
-        return 0
+    def is_Inbound(self):
+        return INBOUND in self.properties
 
     @property
     def symbolic_bounds(self):
@@ -593,6 +624,14 @@ class Iteration(Node):
 
         return (_min, _max)
 
+    @property
+    def start(self):
+        """The start value."""
+        if self.direction is Forward:
+            return self.dim.symbolic_min
+        else:
+            return self.dim.symbolic_max
+
     @property
     def step(self):
         """The step value."""
@@ -621,7 +660,31 @@ class Iteration(Node):
         return self.dimensions
 
 
-class While(Node):
+class DoIf(Node):
+
+    """
+    An abstract Node to express computation subjected to a condition.
+    """
+
+    def __init__(self, condition):
+        self.condition = condition
+
+    @property
+    def functions(self):
+        ret = []
+        for i in self.condition.free_symbols:
+            try:
+                ret.append(i.function)
+            except AttributeError:
+                pass
+        return tuple(ret)
+
+    @property
+    def expr_symbols(self):
+        return tuple(self.condition.free_symbols)
+
+
+class While(DoIf):
 
     """
     Implement a while-loop.
@@ -639,7 +702,7 @@ class While(Node):
     _traversable = ['body']
 
     def __init__(self, condition, body=None):
-        self.condition = condition
+        super().__init__(condition)
         self.body = as_tuple(body)
 
     def __repr__(self):
@@ -658,18 +721,25 @@ class Callable(Node):
     body : Node or list of Node
         The Callable body.
     retval : str
-        The return type of Callable.
+        The return type of the Callable.
     parameters : list of Basic, optional
         The objects in input to the Callable.
     prefix : list of str, optional
-        Qualifiers to prepend to the Callable signature. None by defaults.
+        Qualifiers to prepend to the Callable signature. None by default.
+    templates : list of Basic, optional
+        The template parameters of the Callable.
+    attributes : list of str, optional
+        Additional attributes to append to the Callable signature. An
+        attributes is one or more keywords that appear in between the
+        return type and the function name. None by default.
     """
 
     is_Callable = True
 
     _traversable = ['body']
 
-    def __init__(self, name, body, retval, parameters=None, prefix=None):
+    def __init__(self, name, body, retval, parameters=None, prefix=None,
+                 templates=None, attributes=None):
         self.name = name
         if not isinstance(body, CallableBody):
             self.body = CallableBody(body)
@@ -678,22 +748,29 @@ class Callable(Node):
         self.retval = retval
         self.prefix = as_tuple(prefix)
         self.parameters = as_tuple(parameters)
+        self.templates = as_tuple(templates)
+        self.attributes = as_tuple(attributes)
 
     def __repr__(self):
+        param_types = [ctypes_to_cstr(i._C_ctype) for i in self.parameters]
         return "%s[%s]<%s; %s>" % (self.__class__.__name__, self.name, self.retval,
-                                   ",".join([i._C_typename for i in self.parameters]))
+                                   ",".join(param_types))
+
+    @property
+    def all_parameters(self):
+        return self.parameters + self.templates
 
     @property
     def functions(self):
-        return tuple(i.function for i in self.parameters
-                     if isinstance(i.function, AbstractFunction))
+        return tuple(i.function for i in self.all_parameters
+                     if isinstance(i.function, (AbstractFunction, AbstractObject)))
 
     @property
     def defines(self):
-        return self.parameters
+        return self.all_parameters
 
 
-class CallableBody(Node):
+class CallableBody(MultiTraversable):
 
     """
     The immediate child of a Callable.
@@ -707,40 +784,71 @@ class CallableBody(Node):
     init : Node, optional
         A piece of IET to perform some initialization relevant for `body`
         (e.g., to initialize the target language runtime).
+    standalones : list of Definitions, optional
+        Object definitions for `body`. Instantiating these objects does not
+        require passing any arguments to their constructors, so these
+        Definitions can be scheduled safely right after `init`. They may or may
+        not be required by some of the subsequent nodes (e.g., `allocs`,
+        `maps`).
     allocs : list of Nodes, optional
         Data definitions and allocations for `body`.
+    stacks : list of Definitions, optional
+        Definitions for the stack-scoped objects appearing in `body`.
     casts : list of PointerCasts, optional
         Sequence of PointerCasts required by the `body`.
+    bundles : list of Nodes, optional
+        Data bundling for `body`. Used to initialize data subjected to layout
+        transformation (w.r.t. how it arrives from Python), such as vector types.
     maps : Transfer or list of Transfer, optional
         Data maps for `body` (a data map may e.g. trigger a data transfer from
         host to device).
+    strides : list of Nodes, optional
+        Statements defining symbols used to access linearized arrays.
     objs : list of Definitions, optional
-        Object definitions for `body`.
+        Object definitions for `body`. Instantiating these objects may or may
+        not require some of the symbols defined in the previous nodes (e.g.,
+        `allocs`, `maps`).
     unmaps : Transfer or list of Transfer, optional
         Data unmaps for `body`.
+    unbundles : list of Nodes, optional
+        Data unbundling for `body`.
     frees : list of Calls, optional
         Data deallocations for `body`.
+    errors : list of Nodes, optional
+        Error handling for `body`.
+    retstmt : Node, optional
+        The return statement for `body`.
     """
 
     is_CallableBody = True
 
-    _traversable = ['unpacks', 'init', 'allocs', 'casts', 'maps', 'objs',
-                    'body', 'unmaps', 'frees']
+    _traversable = ['unpacks', 'init', 'standalones', 'allocs', 'stacks',
+                    'casts', 'bundles', 'maps', 'strides', 'objs', 'body',
+                    'unmaps', 'unbundles', 'frees', 'errors', 'retstmt']
 
-    def __init__(self, body, init=None, unpacks=None, allocs=None, casts=None,
-                 objs=None, maps=None, unmaps=None, frees=None):
+    def __init__(self, body, init=(), standalones=(), unpacks=(), strides=(),
+                 allocs=(), stacks=(), casts=(), bundles=(), objs=(), maps=(),
+                 unmaps=(), unbundles=(), frees=(), errors=(), retstmt=()):
         # Sanity check
         assert not isinstance(body, CallableBody), "CallableBody's cannot be nested"
 
         self.body = as_tuple(body)
-        self.init = as_tuple(init)
+
         self.unpacks = as_tuple(unpacks)
+        self.init = as_tuple(init)
+        self.standalones = as_tuple(standalones)
         self.allocs = as_tuple(allocs)
+        self.stacks = as_tuple(stacks)
         self.casts = as_tuple(casts)
+        self.strides = as_tuple(strides)
+        self.bundles = as_tuple(bundles)
         self.maps = as_tuple(maps)
         self.objs = as_tuple(objs)
         self.unmaps = as_tuple(unmaps)
+        self.unbundles = as_tuple(unbundles)
         self.frees = as_tuple(frees)
+        self.errors = as_tuple(errors)
+        self.retstmt = as_tuple(retstmt)
 
     def __repr__(self):
         return ("<CallableBody <unpacks=%d, allocs=%d, casts=%d, maps=%d, "
@@ -750,7 +858,7 @@ class CallableBody(Node):
                  len(self.frees)))
 
 
-class Conditional(Node):
+class Conditional(DoIf):
 
     """
     A node to express if-then-else blocks.
@@ -770,7 +878,7 @@ class Conditional(Node):
     _traversable = ['then_body', 'else_body']
 
     def __init__(self, condition, then_body, else_body=None):
-        self.condition = condition
+        super().__init__(condition)
         self.then_body = as_tuple(then_body)
         self.else_body = as_tuple(else_body)
 
@@ -781,20 +889,6 @@ class Conditional(Node):
         else:
             return "<[%s] ? [%s]" % (ccode(self.condition), repr(self.then_body))
 
-    @property
-    def functions(self):
-        ret = []
-        for i in self.condition.free_symbols:
-            try:
-                ret.append(i.function)
-            except AttributeError:
-                pass
-        return tuple(ret)
-
-    @property
-    def expr_symbols(self):
-        return tuple(self.condition.free_symbols)
-
 
 # Second level IET nodes
 
@@ -817,20 +911,20 @@ class TimedList(List):
         self._name = lname
         self._timer = timer
 
-        super().__init__(header=c.Line('START_TIMER(%s)' % lname),
+        super().__init__(header=c.Line('START(%s)' % lname),
                          body=body,
-                         footer=c.Line('STOP_TIMER(%s,%s)' % (lname, timer.name)))
+                         footer=c.Line('STOP(%s,%s)' % (lname, timer.name)))
 
     @classmethod
     def _start_timer_header(cls):
-        return ('START_TIMER(S)', ('struct timeval start_ ## S , end_ ## S ; '
-                                   'gettimeofday(&start_ ## S , NULL);'))
+        return ('START(S)', ('struct timeval start_ ## S , end_ ## S ; '
+                             'gettimeofday(&start_ ## S , NULL);'))
 
     @classmethod
     def _stop_timer_header(cls):
-        return ('STOP_TIMER(S,T)', ('gettimeofday(&end_ ## S, NULL); T->S += (double)'
-                                    '(end_ ## S .tv_sec-start_ ## S.tv_sec)+(double)'
-                                    '(end_ ## S .tv_usec-start_ ## S .tv_usec)/1000000;'))
+        return ('STOP(S,T)', ('gettimeofday(&end_ ## S, NULL); T->S += (double)'
+                              '(end_ ## S .tv_sec-start_ ## S.tv_sec)+(double)'
+                              '(end_ ## S .tv_usec-start_ ## S .tv_usec)/1000000;'))
 
     @property
     def name(self):
@@ -849,35 +943,54 @@ class Definition(ExprStmt, Node):
 
     """
     A node encapsulating a variable definition.
-
-    If `shape` is given, then the Definition implements an array allocated
-    on the stack, otherwise it implements an uninitialized pointer.
     """
 
     is_Definition = True
 
-    def __init__(self, function, shape=None, qualifier=None, initvalue=None,
-                 cargs=None):
+    def __init__(self, function):
         self.function = function
-        self.shape = shape
-        self.qualifier = qualifier
-        self.initvalue = initvalue
-        self.cargs = as_tuple(cargs)
 
     def __repr__(self):
         return "<Def(%s)>" % self.function
 
     @property
     def functions(self):
-        return (self.function,)
+        ret = [self.function]
+        for i in self.expr_symbols:
+            f = i.function
+            if f.is_AbstractFunction or f.is_AbstractObject:
+                ret.append(i.function)
+        return tuple(ret)
 
     @property
     def defines(self):
-        if self.shape is not None:
+        if self.function._mem_stack:
             return (self.function.indexed,)
         else:
             return (self.function,)
 
+    @property
+    def expr_symbols(self):
+        f = self.function
+        if f.is_LocalObject:
+            ret = set(flatten(i.free_symbols for i in f.cargs))
+            try:
+                ret.update(f.initvalue.free_symbols)
+            except AttributeError:
+                pass
+            return tuple(ret)
+        elif f.is_Array and f.initvalue is not None:
+            # These are just a handful of values so it's OK to iterate them over
+            ret = set()
+            for i in f.initvalue:
+                try:
+                    ret.update(i.free_symbols)
+                except AttributeError:
+                    pass
+            return tuple(ret)
+        else:
+            return ()
+
 
 class PointerCast(ExprStmt, Node):
 
@@ -915,14 +1028,17 @@ class PointerCast(ExprStmt, Node):
     @property
     def expr_symbols(self):
         f = self.function
-        if f.is_ArrayBasic:
+        if not self.flat and f.is_ArrayBasic:
             return tuple(flatten(s.free_symbols for s in f.symbolic_shape[1:]))
         else:
             return ()
 
     @property
     def defines(self):
-        return (self.function.indexed,)
+        if isinstance(self.obj, IndexedBase):
+            return (self.obj,)
+        else:
+            return (self.function.indexed,)
 
 
 class Dereference(ExprStmt, Node):
@@ -934,6 +1050,8 @@ class Dereference(ExprStmt, Node):
         * `pointer` is a PointerArray or TempFunction, and `pointee` is an Array.
         * `pointer` is an ArrayObject representing a pointer to a C struct, and
           `pointee` is a field in `pointer`.
+        * `pointer` is a Symbol with its _C_ctype deriving from ct._Pointer, and
+          `pointee` is a Symbol representing the dereferenced value.
     """
 
     is_Dereference = True
@@ -952,13 +1070,18 @@ class Dereference(ExprStmt, Node):
 
     @property
     def expr_symbols(self):
-        ret = [self.pointer.indexed]
-        if self.pointer.is_PointerArray or self.pointer.is_TempFunction:
-            ret.append(self.pointee.indexed)
-            ret.extend(flatten(i.free_symbols for i in self.pointee.symbolic_shape[1:]))
+        ret = []
+        if self.pointer.is_Symbol:
+            assert issubclass(self.pointer._C_ctype, ctypes._Pointer), \
+                   "Scalar dereference must have a pointer ctype"
+            ret.extend([self.pointer._C_symbol, self.pointee._C_symbol])
+        elif self.pointer.is_PointerArray or self.pointer.is_TempFunction:
+            ret.extend([self.pointer.indexed, self.pointee.indexed])
+            ret.extend(flatten(i.free_symbols
+                               for i in self.pointee.symbolic_shape[1:]))
             ret.extend(self.pointer.free_symbols)
         else:
-            ret.append(self.pointee._C_symbol)
+            ret.extend([self.pointer.indexed, self.pointee._C_symbol])
         return tuple(filter_ordered(ret))
 
     @property
@@ -977,7 +1100,7 @@ class Lambda(Node):
     A callable C++ lambda function. Several syntaxes are possible; here we
     implement one of the common ones:
 
-        [captures](parameters){body}
+        [captures](parameters){body} SPECIAL [[attributes]]
 
     For more info about C++ lambda functions:
 
@@ -991,22 +1114,38 @@ class Lambda(Node):
         The captures of the lambda function.
     parameters : list of Basic or expr-like, optional
         The objects in input to the lambda function.
+    special : list of Basic, optional
+        Placeholder for custom lambdas, to add in e.g. macros.
+    attributes : list of str, optional
+        The attributes of the lambda function.
     """
 
     _traversable = ['body']
 
-    def __init__(self, body, captures=None, parameters=None):
+    def __init__(self, body, captures=None, parameters=None, special=None,
+                 attributes=None):
         self.body = as_tuple(body)
         self.captures = as_tuple(captures)
         self.parameters = as_tuple(parameters)
+        self.special = as_tuple(special)
+        self.attributes = as_tuple(attributes)
 
     def __repr__(self):
         return "Lambda[%s](%s)" % (self.captures, self.parameters)
 
+    @property
+    def functions(self):
+        return tuple(i.function for i in self.parameters
+                     if isinstance(i.function, AbstractFunction))
+
     @cached_property
     def expr_symbols(self):
         return tuple(self.parameters)
 
+    @property
+    def defines(self):
+        return tuple(self.parameters)
+
 
 class Section(List):
 
@@ -1023,7 +1162,7 @@ class Section(List):
     is_Section = True
 
     def __init__(self, name, body=None, is_subsection=False):
-        super(Section, self).__init__(body=body)
+        super().__init__(body=body)
         self.name = name
         self.is_subsection = is_subsection
 
@@ -1044,7 +1183,7 @@ class ExpressionBundle(List):
     is_ExpressionBundle = True
 
     def __init__(self, ispace, ops, traffic, body=None):
-        super(ExpressionBundle, self).__init__(body=body)
+        super().__init__(body=body)
         self.ispace = ispace
         self.ops = ops
         self.traffic = traffic
@@ -1089,27 +1228,57 @@ class Prodder(Call):
         return self._periodic
 
 
+class Using(Node):
+
+    """
+    A C++ using directive.
+    """
+
+    def __init__(self, name):
+        self.name = name
+
+    def __repr__(self):
+        return "<Using(%s)>" % self.name
+
+
+class UsingNamespace(Node):
+
+    """
+    A C++ using namespace directive.
+    """
+
+    def __init__(self, namespace):
+        self.namespace = namespace
+
+    def __repr__(self):
+        return "<UsingNamespace(%s)>" % self.namespace
+
+
 class Pragma(Node):
 
     """
-    One or more pragmas floating in the IET constructed through a callback.
+    One or more pragmas floating in the IET.
     """
 
-    def __init__(self, callback, arguments=None):
+    def __init__(self, pragma, arguments=None):
         super().__init__()
 
-        self.callback = callback
+        if not isinstance(pragma, str):
+            raise TypeError("Pragma name must be a string, not %s" % type(pragma))
+
+        self.pragma = pragma
         self.arguments = as_tuple(arguments)
 
     def __repr__(self):
-        return '<Pragmas>'
+        return '<Pragma>'
 
     @cached_property
-    def pragmas(self):
-        return as_tuple(self.callback(*self.arguments))
+    def _generate(self):
+        # Subclasses may override this property to customize the pragma generation
+        return self.pragma % self.arguments
 
 
-class Transfer(object):
+class Transfer:
 
     """
     An interface for Nodes that represent host-device data transfers.
@@ -1130,58 +1299,7 @@ class ParallelIteration(Iteration):
     Implement a parallel for-loop.
     """
 
-    def __init__(self, *args, **kwargs):
-        pragmas, kwargs, properties = self._make_header(**kwargs)
-        super().__init__(*args, pragmas=pragmas, properties=properties, **kwargs)
-
-    @classmethod
-    def _make_header(cls, **kwargs):
-        construct = cls._make_construct(**kwargs)
-        clauses = cls._make_clauses(**kwargs)
-        header = c.Pragma(' '.join([construct] + clauses))
-
-        # Extract the Iteration Properties
-        properties = cls._process_properties(**kwargs)
-
-        # Drop the unrecognised or unused kwargs
-        kwargs = cls._process_kwargs(**kwargs)
-
-        return (header,), kwargs, properties
-
-    @classmethod
-    def _make_construct(cls, **kwargs):
-        # To be overridden by subclasses
-        raise NotImplementedError
-
-    @classmethod
-    def _make_clauses(cls, **kwargs):
-        return []
-
-    @classmethod
-    def _process_properties(cls, **kwargs):
-        properties = as_tuple(kwargs.get('properties'))
-        properties += (COLLAPSED(kwargs.get('ncollapse', 1)),)
-
-        return properties
-
-    @classmethod
-    def _process_kwargs(cls, **kwargs):
-        kwargs.pop('pragmas', None)
-        kwargs.pop('properties', None)
-
-        # Recognised clauses
-        kwargs.pop('ncollapse', None)
-        kwargs.pop('reduction', None)
-
-        return kwargs
-
-    @cached_property
-    def collapsed(self):
-        ret = [self]
-        for i in range(self.ncollapsed - 1):
-            ret.append(ret[i].nodes[0])
-        assert all(i.is_Iteration for i in ret)
-        return tuple(ret)
+    pass
 
 
 class ParallelTree(List):
@@ -1224,6 +1342,10 @@ class ParallelTree(List):
     def functions(self):
         return as_tuple(self.nthreads)
 
+    @property
+    def expr_symbols(self):
+        return as_tuple(self.nthreads)
+
     @property
     def root(self):
         return self.body[0]
@@ -1300,6 +1422,21 @@ class SyncSpot(List):
     def __repr__(self):
         return "<SyncSpot (%s)>" % ",".join(str(i) for i in self.sync_ops)
 
+    @property
+    def is_async_op(self):
+        """
+        True if the SyncSpot contains an asynchronous operation, False otherwise.
+        If False, the SyncSpot may for example represent a wait on a lock.
+        """
+        return any(isinstance(s, (WithLock, PrefetchUpdate))
+                   for s in self.sync_ops)
+
+    @property
+    def functions(self):
+        ret = [(s.lock, s.function, s.target) for s in self.sync_ops]
+        ret = tuple(filter_ordered(f for f in flatten(ret) if f is not None))
+        return ret
+
 
 class CBlankLine(List):
 
@@ -1310,12 +1447,20 @@ class CBlankLine(List):
         return ""
 
 
+class Break(Node):
+    pass
+
+
 class Return(Node):
 
     def __init__(self, value=None):
         self.value = value
 
 
+class ListMajor(List):
+    pass
+
+
 def DummyExpr(*args, init=False):
     return Expression(DummyEq(*args), init=init)
 
@@ -1337,9 +1482,9 @@ class HaloSpot(Node):
 
     _traversable = ['body']
 
-    def __init__(self, halo_scheme, body=None):
-        super(HaloSpot, self).__init__()
-        self._halo_scheme = halo_scheme
+    def __init__(self, body, halo_scheme):
+        super().__init__()
+
         if isinstance(body, Node):
             self._body = body
         elif isinstance(body, (list, tuple)) and len(body) == 1:
@@ -1349,6 +1494,8 @@ class HaloSpot(Node):
         else:
             raise ValueError("`body` is expected to be a single Node")
 
+        self._halo_scheme = halo_scheme
+
     def __repr__(self):
         functions = "(%s)" % ",".join(i.name for i in self.functions)
         return "<%s%s>" % (self.__class__.__name__, functions)
@@ -1374,8 +1521,8 @@ class HaloSpot(Node):
         return self.halo_scheme.arguments
 
     @property
-    def is_empty(self):
-        return len(self.halo_scheme) == 0
+    def is_void(self):
+        return self.halo_scheme.is_void
 
     @property
     def body(self):
@@ -1395,11 +1542,3 @@ Metadata for Callables. ``root`` is a pointer to the callable
 Iteration/Expression tree. ``local`` is a boolean indicating whether the
 definition of the callable is known or not.
 """
-
-
-class CallArgType(Tag):
-    pass
-
-
-HPtr = CallArgType('host pointer')
-DPtr = CallArgType('device pointer')
