@@ -1,26 +1,27 @@
 import abc
 from collections import OrderedDict
-from functools import reduce
-from itertools import chain
+from functools import reduce, cached_property
 from operator import mul
 
-from cached_property import cached_property
 from sympy import Expr
 
+from devito.ir.support.utils import minimum, maximum
 from devito.ir.support.vector import Vector, vmin, vmax
-from devito.tools import (PartialOrderTuple, Stamp, as_list, as_tuple, filter_ordered,
-                          flatten, frozendict, is_integer, toposort)
+from devito.tools import (Ordering, Stamp, as_list, as_set, as_tuple,
+                          filter_ordered, flatten, frozendict, is_integer,
+                          toposort)
 from devito.types import Dimension, ModuloDimension
 
 __all__ = ['NullInterval', 'Interval', 'IntervalGroup', 'IterationSpace',
-           'DataSpace', 'Forward', 'Backward', 'Any']
+           'IterationInterval', 'DataSpace', 'Forward', 'Backward', 'Any',
+           'null_ispace']
 
 
 # The default Stamp, used by all new Intervals
 S0 = Stamp()
 
 
-class AbstractInterval(object):
+class AbstractInterval:
 
     """
     An abstract representation of an iterated closed interval on Z.
@@ -36,7 +37,7 @@ class AbstractInterval(object):
         self.stamp = stamp
 
     def __eq__(self, o):
-        return (type(self) == type(o) and
+        return (type(self) is type(o) and
                 self.dim is o.dim and
                 self.stamp == o.stamp)
 
@@ -49,7 +50,8 @@ class AbstractInterval(object):
     def _rebuild(self):
         return
 
-    @abc.abstractproperty
+    @property
+    @abc.abstractmethod
     def relaxed(self):
         return
 
@@ -122,8 +124,8 @@ class Interval(AbstractInterval):
 
     is_Defined = True
 
-    def __init__(self, dim, lower, upper, stamp=S0):
-        super(Interval, self).__init__(dim, stamp)
+    def __init__(self, dim, lower=0, upper=0, stamp=S0):
+        super().__init__(dim, stamp)
 
         try:
             self.lower = int(lower)
@@ -146,7 +148,7 @@ class Interval(AbstractInterval):
         if self is o:
             return True
 
-        return (super(Interval, self).__eq__(o) and
+        return (super().__eq__(o) and
                 self.lower == o.lower and
                 self.upper == o.upper)
 
@@ -285,27 +287,41 @@ class Interval(AbstractInterval):
                 pass
         return self
 
+    def expand(self):
+        return Interval(
+            self.dim, minimum(self.lower), maximum(self.upper), self.stamp
+        )
 
-class IntervalGroup(PartialOrderTuple):
+
+class IntervalGroup(Ordering):
 
     """
-    A partially-ordered sequence of Intervals equipped with set-like
-    operations.
+    A sequence of Intervals equipped with set-like operations.
     """
 
     @classmethod
     def reorder(cls, items, relations):
         if not all(isinstance(i, AbstractInterval) for i in items):
-            raise ValueError("Cannot create an IntervalGroup from objects of type [%s]" %
+            raise ValueError("Cannot create IntervalGroup from objs of type [%s]" %
                              ', '.join(str(type(i)) for i in items))
-        # The relations are between dimensions, not intervals. So we take
-        # care of that here
-        ordering = filter_ordered(toposort(relations) + [i.dim for i in items])
+
+        if len(relations) == 1:
+            # Special case: avoid expensive topological sorting if possible
+            relation, = relations
+            ordering = filter_ordered(list(relation) + [i.dim for i in items])
+        else:
+            ordering = filter_ordered(toposort(relations) + [i.dim for i in items])
+
         return sorted(items, key=lambda i: ordering.index(i.dim))
 
+    @classmethod
+    def simplify_relations(cls, relations, items, mode):
+        if mode == 'total':
+            return [tuple(i.dim for i in items)]
+        else:
+            return super().simplify_relations(relations, items, mode)
+
     def __eq__(self, o):
-        # No need to look at the relations -- if the partial ordering is the same,
-        # then the IntervalGroups are considered equal
         return len(self) == len(o) and all(i == j for i, j in zip(self, o))
 
     def __contains__(self, d):
@@ -337,7 +353,7 @@ class IntervalGroup(PartialOrderTuple):
         return len(self.dimensions) == len(set(self.dimensions))
 
     @classmethod
-    def generate(self, op, *interval_groups):
+    def generate(cls, op, *interval_groups, relations=None):
         """
         Create a new IntervalGroup from the iterative application of an
         operation to some IntervalGroups.
@@ -349,6 +365,8 @@ class IntervalGroup(PartialOrderTuple):
             or 'union'.
         *interval_groups
             Input IntervalGroups.
+        relations : tuple, optional
+            Additional relations to build the new IntervalGroup.
 
         Examples
         --------
@@ -360,10 +378,17 @@ class IntervalGroup(PartialOrderTuple):
 
         >>> ig = IntervalGroup.generate('intersection', ig0, ig1, ig2)
         """
+        if op == 'intersection':
+            dims = set.intersection(*[set(ig.dimensions) for ig in interval_groups])
+        else:
+            dims = set().union(*[ig.dimensions for ig in interval_groups])
+
         mapper = {}
         for ig in interval_groups:
             for i in ig:
-                mapper.setdefault(i.dim, []).append(i)
+                if i.dim in dims:
+                    mapper.setdefault(i.dim, []).append(i)
+
         intervals = []
         for v in mapper.values():
             # Create a new Interval through the concatenation v0.key(v1).key(v2)...
@@ -371,13 +396,23 @@ class IntervalGroup(PartialOrderTuple):
             for i in v[1:]:
                 interval = getattr(interval, op)(i)
             intervals.append(interval)
-        relations = set().union(*[ig.relations for ig in interval_groups])
-        return IntervalGroup(intervals, relations=relations)
+
+        relations = as_set(relations)
+        relations.update(set().union(*[ig.relations for ig in interval_groups]))
+
+        modes = set(ig.mode for ig in interval_groups)
+        assert len(modes) <= 1
+        try:
+            mode = modes.pop()
+        except KeyError:
+            mode = 'total'
+
+        return IntervalGroup(intervals, relations=relations, mode=mode)
 
     def is_compatible(self, o):
         """
-        Two IntervalGroups are compatible iff they can be ordered according
-        to some common partial ordering.
+        Two IntervalGroups are compatible iff their elements are not subjected
+        to sets of relations that would prevent a reordering.
         """
         if set(self) != set(o):
             return False
@@ -388,7 +423,7 @@ class IntervalGroup(PartialOrderTuple):
             self.add(o)
             return True
         except ValueError:
-            # Cyclic dependence detected, there is no common partial ordering
+            # Cyclic dependence detected
             return False
 
     def _normalize(func):
@@ -399,67 +434,85 @@ class IntervalGroup(PartialOrderTuple):
         def wrapper(self, o):
             if not isinstance(o, IntervalGroup):
                 o = IntervalGroup(as_tuple(o))
-            return func(self, o)
+            if self.mode != o.mode:
+                raise ValueError("Incompatible ordering modes")
+            relations = self.relations | o.relations
+            return func(self, o, relations, self.mode)
         return wrapper
 
     @_normalize
-    def intersection(self, o):
+    def intersection(self, o, relations, mode):
         mapper = OrderedDict([(i.dim, i) for i in o])
         intervals = [i.intersection(mapper.get(i.dim, i)) for i in self]
-        return IntervalGroup(intervals, relations=(self.relations | o.relations))
+        return IntervalGroup(intervals, relations=relations, mode=mode)
 
     @_normalize
-    def add(self, o):
+    def add(self, o, relations, mode):
         mapper = OrderedDict([(i.dim, i) for i in o])
         intervals = [i.add(mapper.get(i.dim, NullInterval(i.dim))) for i in self]
-        return IntervalGroup(intervals, relations=(self.relations | o.relations))
+        return IntervalGroup(intervals, relations=relations, mode=mode)
 
     @_normalize
-    def subtract(self, o):
+    def subtract(self, o, relations, mode):
         mapper = OrderedDict([(i.dim, i) for i in o])
         intervals = [i.subtract(mapper.get(i.dim, NullInterval(i.dim))) for i in self]
-        return IntervalGroup(intervals, relations=(self.relations | o.relations))
+        return IntervalGroup(intervals, relations=relations, mode=mode)
 
     def relaxed(self, d=None):
         d = set(self.dimensions if d is None else as_tuple(d))
         intervals = [i.relaxed if i.dim in d else i for i in self]
-        return IntervalGroup(intervals, relations=self.relations)
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
-    def promote(self, cond):
-        intervals = IntervalGroup([i.promote(cond) for i in self],
-                                  relations=self.relations)
+    def promote(self, cond, mode='unordered'):
+        intervals = [i.promote(cond) for i in self]
+        intervals = IntervalGroup(intervals, relations=None, mode=mode)
 
-        # There could be duplicate Dimensions at this point, so we sum up the Intervals
-        # defined over the same Dimension to produce a well-defined IntervalGroup
+        # There could be duplicate Dimensions at this point, so we sum up the
+        # Intervals defined over the same Dimension to produce a well-defined
+        # IntervalGroup
         intervals = IntervalGroup.generate('add', intervals)
 
         return intervals
 
-    def drop(self, d):
+    def drop(self, maybe_callable, strict=False):
+        if callable(maybe_callable):
+            dims = {i.dim for i in self if maybe_callable(i.dim)}
+        else:
+            dims = as_set(maybe_callable)
+
         # Dropping
-        dims = set().union(*[i._defines for i in as_tuple(d)])
-        intervals = [i._rebuild() for i in self if not i.dim._defines & dims]
+        if strict:
+            intervals = [i._rebuild() for i in self if i.dim not in dims]
+        else:
+            dims = set().union(*[i._defines for i in dims])
+            intervals = [i._rebuild() for i in self if not i.dim._defines & dims]
 
         # Clean up relations
-        relations = [tuple(i for i in r if i in intervals) for r in self.relations]
+        relations = [tuple(i for i in r if i not in dims) for r in self.relations]
 
-        return IntervalGroup(intervals, relations=relations)
+        return IntervalGroup(intervals, relations=relations, mode=self.mode)
 
     def negate(self):
-        return IntervalGroup([i.negate() for i in self], relations=self.relations)
+        intervals = [i.negate() for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
     def zero(self, d=None):
         d = self.dimensions if d is None else as_tuple(d)
-        return IntervalGroup([i.zero() if i.dim in d else i for i in self],
-                             relations=self.relations)
+        intervals = [i.zero() if i.dim in d else i for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
     def lift(self, d=None, v=None):
         d = set(self.dimensions if d is None else as_tuple(d))
-        return IntervalGroup([i.lift(v) if i.dim._defines & d else i for i in self],
-                             relations=self.relations)
+        intervals = [i.lift(v) if i.dim._defines & d else i for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
     def reset(self):
-        return IntervalGroup([i.reset() for i in self], relations=self.relations)
+        intervals = [i.reset() for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
     def switch(self, d0, d1):
         intervals = [i.switch(d1) if i.dim is d0 else i for i in self]
@@ -467,25 +520,34 @@ class IntervalGroup(PartialOrderTuple):
         # Update relations too
         relations = {tuple(d1 if i is d0 else i for i in r) for r in self.relations}
 
-        return IntervalGroup(intervals, relations=relations)
+        return IntervalGroup(intervals, relations=relations, mode=self.mode)
 
     def translate(self, d, v0=0, v1=None):
-        intervals = [i.translate(v0, v1) if i.dim in as_tuple(d) else i for i in self]
-        return IntervalGroup(intervals, relations=self.relations)
+        dims = as_tuple(d)
+        intervals = [i.translate(v0, v1) if i.dim in dims else i for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
+
+    def expand(self, d=None):
+        if d is None:
+            d = self.dimensions
+        intervals = [i.expand() if i.dim in as_tuple(d) else i for i in self]
+
+        return IntervalGroup(intervals, relations=self.relations, mode=self.mode)
 
     def index(self, key):
         if isinstance(key, Interval):
-            return super(IntervalGroup, self).index(key)
+            return super().index(key)
         elif isinstance(key, Dimension):
-            return super(IntervalGroup, self).index(self[key])
+            return super().index(self[key])
         raise ValueError("Expected Interval or Dimension, got `%s`" % type(key))
 
     def __getitem__(self, key):
         if is_integer(key):
-            return super(IntervalGroup, self).__getitem__(key)
+            return super().__getitem__(key)
         elif isinstance(key, slice):
-            retval = super(IntervalGroup, self).__getitem__(key)
-            return IntervalGroup(retval, relations=self.relations)
+            retval = super().__getitem__(key)
+            return IntervalGroup(retval, relations=self.relations, mode=self.mode)
 
         if not self.is_well_defined:
             raise ValueError("Cannot fetch Interval from ill defined Space")
@@ -494,17 +556,21 @@ class IntervalGroup(PartialOrderTuple):
             return NullInterval(key)
 
         for i in self:
+            # Obvious case
             if i.dim is key:
                 return i
+
+            # There are a few special cases for which we can support
+            # derived-Dimension-based indexing
             if key.is_NonlinearDerived and i.dim in key._defines:
-                # NonlinearDerived Dimensions cannot appear in iteration Intervals,
-                # but their parent can
+                return i
+            if i.dim.is_Custom and key in i.dim._defines:
                 return i
 
         return NullInterval(key)
 
 
-class IterationDirection(object):
+class IterationDirection:
 
     """
     A representation of the direction in which an iteration space is traversed.
@@ -533,46 +599,34 @@ Any = IterationDirection('*')
 """Wildcard direction (both '++' and '--' would be OK)."""
 
 
-class IterationInterval(object):
+class IterationInterval(Interval):
 
     """
-    An Interval associated with an IterationDirection.
+    An Interval associated with metadata.
     """
 
-    def __init__(self, interval, sub_iterators, direction):
-        self.interval = interval
+    def __init__(self, interval, sub_iterators=(), direction=Forward):
+        super().__init__(interval.dim, *interval.offsets, stamp=interval.stamp)
         self.sub_iterators = sub_iterators
         self.direction = direction
 
     def __repr__(self):
-        return "%s%s" % (self.interval, self.direction)
+        return "%s%s" % (super().__repr__(), self.direction)
 
     def __eq__(self, other):
         if not isinstance(other, IterationInterval):
             return False
-        return self.direction is other.direction and self.interval == other.interval
+        return self.direction is other.direction and super().__eq__(other)
 
     def __hash__(self):
-        return hash((self.interval, self.direction))
+        return hash((self.dim, self.offsets, self.direction))
 
     @property
     def args(self):
-        return (self.interval, self.sub_iterators, self.direction)
-
-    @property
-    def dim(self):
-        return self.interval.dim
-
-    @property
-    def offsets(self):
-        return self.interval.offsets
-
-    @property
-    def size(self):
-        return self.interval.size
+        return (self, self.sub_iterators, self.direction)
 
 
-class Space(object):
+class Space:
 
     """
     A compact N-dimensional space defined by N Intervals.
@@ -618,6 +672,14 @@ class Space(object):
     def size(self):
         return self.intervals.size
 
+    @property
+    def mode(self):
+        return self.intervals.mode
+
+    @property
+    def relations(self):
+        return self.intervals.relations
+
     @property
     def dimension_map(self):
         """
@@ -641,15 +703,22 @@ class DataSpace(Space):
     """
 
     def __init__(self, intervals, parts=None):
-        super(DataSpace, self).__init__(intervals)
-        self._parts = frozendict(parts or {})
+        super().__init__(intervals)
+
+        # Normalize parts
+        parts = {k: v.expand() for k, v in (parts or {}).items()}
+        for k, v in list(parts.items()):
+            dims = set().union(*[d._defines for d in k.dimensions])
+            parts[k] = v.drop(lambda d: d not in dims)
+        self._parts = frozendict(parts)
 
     def __eq__(self, other):
-        return isinstance(other, DataSpace) and\
-            self.intervals == other.intervals and self.parts == other.parts
+        return (isinstance(other, DataSpace) and
+                self.intervals == other.intervals and
+                self.parts == other.parts)
 
     def __hash__(self):
-        return hash((super(DataSpace, self).__hash__(), self.parts))
+        return hash((super().__hash__(), self.parts))
 
     @classmethod
     def union(cls, *others):
@@ -703,12 +772,14 @@ class IterationSpace(Space):
     """
 
     def __init__(self, intervals, sub_iterators=None, directions=None):
-        super(IterationSpace, self).__init__(intervals)
+        super().__init__(intervals)
 
         # Normalize sub-iterators
-        sub_iterators = sub_iterators or {}
-        self._sub_iterators = frozendict([(k, tuple(filter_ordered(as_tuple(v))))
-                                          for k, v in sub_iterators.items()])
+        sub_iterators = dict([(k, tuple(filter_ordered(as_tuple(v))))
+                              for k, v in (sub_iterators or {}).items()])
+        sub_iterators.update({i.dim: () for i in self.intervals
+                              if i.dim not in sub_iterators})
+        self._sub_iterators = frozendict(sub_iterators)
 
         # Normalize directions
         if directions is None:
@@ -736,24 +807,33 @@ class IterationSpace(Space):
         return len(self.itintervals) < len(other.itintervals)
 
     def __hash__(self):
-        return hash((super(IterationSpace, self).__hash__(), self.sub_iterators,
+        return hash((super().__hash__(), self.sub_iterators,
                      self.directions))
 
+    def __contains__(self, d):
+        try:
+            self[d]
+            return True
+        except KeyError:
+            return False
+
     def __getitem__(self, key):
-        retval = self.intervals[key]
+        v = self.intervals[key]
         if isinstance(key, slice):
-            return self.project(lambda d: d in retval.dimensions)
+            return self.project(lambda d: d in v.dimensions)
         else:
-            return retval
+            d = v.dim
+            return IterationInterval(v, self.sub_iterators[d], self.directions[d])
 
     @classmethod
-    def union(cls, *others):
+    def generate(self, op, *others, relations=None):
         if not others:
             return IterationSpace(IntervalGroup())
-        elif len(others) == 1:
+        elif len(others) == 1 and not relations:
             return others[0]
 
-        intervals = IntervalGroup.generate('union', *[i.intervals for i in others])
+        intervals = [i.intervals for i in others]
+        intervals = IntervalGroup.generate(op, *intervals, relations=relations)
 
         directions = {}
         for i in others:
@@ -763,8 +843,8 @@ class IterationSpace(Space):
                     directions[k] = v
                 elif v is not Any:
                     # Clash detected
-                    raise ValueError("Cannot compute the union of `IterationSpace`s "
-                                     "with incompatible directions")
+                    raise ValueError("Cannot compute %s of `IterationSpace`s "
+                                     "with incompatible directions" % op)
 
         sub_iterators = {}
         for i in others:
@@ -774,6 +854,14 @@ class IterationSpace(Space):
 
         return IterationSpace(intervals, sub_iterators, directions)
 
+    @classmethod
+    def union(cls, *others, relations=None):
+        return cls.generate('union', *others, relations=relations)
+
+    @classmethod
+    def intersection(cls, *others, relations=None):
+        return cls.generate('intersection', *others, relations=relations)
+
     def index(self, key):
         return self.intervals.index(key)
 
@@ -796,35 +884,48 @@ class IterationSpace(Space):
 
         return IterationSpace(self.intervals, items, self.directions)
 
-    def switch(self, d0, d1):
+    def switch(self, d0, d1, direction=None):
         intervals = self.intervals.switch(d0, d1)
-        sub_iterators = {d1 if k is d0 else k: v
-                         for k, v in self.sub_iterators.items()}
-        directions = {d1 if k is d0 else k: v
-                      for k, v in self.directions.items()}
+
+        sub_iterators = dict(self.sub_iterators)
+        sub_iterators.pop(d0, None)
+        sub_iterators[d1] = ()
+
+        directions = dict(self.directions)
+        v = directions.pop(d0, None)
+        directions[d1] = direction or v
 
         return IterationSpace(intervals, sub_iterators, directions)
 
+    def translate(self, d, v0=0, v1=None):
+        intervals = self.intervals.translate(d, v0, v1)
+        return IterationSpace(intervals, self.sub_iterators, self.directions)
+
     def reset(self):
         return IterationSpace(self.intervals.reset(), self.sub_iterators, self.directions)
 
-    def project(self, cond):
+    def project(self, cond, strict=True):
         """
-        Create a new IterationSpace in which only some Dimensions
-        in ``self`` are retained. In particular, a Dimension ``d`` in ``self`` is
-        retained if:
+        Create a new IterationSpace retaining only some of the Dimensions in
+        `self`. In particular, a Dimension `d` in `self` is retained if:
 
-            * either ``cond(d)`` is true (``cond`` is a callable),
-            * or ``d in cond`` is true (``cond`` is an iterable)
+            * either `cond(d)` is true (`cond` is a callable),
+            * or `d in cond` is true (`cond` is an iterable)
         """
         if callable(cond):
             func = cond
         else:
             func = lambda i: i in cond
 
-        intervals = [i for i in self.intervals if func(i.dim)]
-        sub_iterators = {k: v for k, v in self.sub_iterators.items() if func(k)}
-        directions = {k: v for k, v in self.directions.items() if func(k)}
+        dims = [i.dim for i in self if not func(i.dim)]
+        intervals = self.intervals.drop(dims, strict=strict)
+
+        sub_iterators = {}
+        directions = {}
+        for i in intervals:
+            d = i.dim
+            sub_iterators[d] = self.sub_iterators[d]
+            directions[d] = self.directions[d]
 
         return IterationSpace(intervals, sub_iterators, directions)
 
@@ -860,8 +961,8 @@ class IterationSpace(Space):
 
         return IterationSpace(intervals, sub_iterators, directions)
 
-    def promote(self, cond):
-        intervals = self.intervals.promote(cond)
+    def promote(self, cond, mode='unordered'):
+        intervals = self.intervals.promote(cond, mode=mode)
         sub_iterators = {i.promote(cond).dim: self.sub_iterators[i.dim]
                          for i in self.intervals}
         directions = {i.promote(cond).dim: self.directions[i.dim]
@@ -869,6 +970,71 @@ class IterationSpace(Space):
 
         return IterationSpace(intervals, sub_iterators, directions)
 
+    def prefix(self, key):
+        """
+        Return the IterationSpace up to and including the last Interval
+        such that `key(interval)` is True.
+        """
+        try:
+            i = self.project(key)[-1]
+        except IndexError:
+            return null_ispace
+
+        return self[:self.index(i.dim) + 1]
+
+    def split(self, d):
+        """
+        Split the IterationSpace into two IterationSpaces, the first
+        containing all Intervals up to and including the one defined
+        over Dimension `d`, and the second containing all Intervals starting
+        from the one defined over Dimension `d`.
+        """
+        if isinstance(d, IterationInterval):
+            d = d.dim
+
+        if d is None:
+            return IterationSpace([]), self
+
+        try:
+            n = self.index(d) + 1
+            return self[:n], self[n:]
+        except ValueError:
+            return self, IterationSpace([])
+
+    def insert(self, d, dimensions, sub_iterators=None, directions=None):
+        """
+        Insert new Dimensions into the IterationSpace after Dimension `d`.
+        """
+        dimensions = as_tuple(dimensions)
+
+        intervals = IntervalGroup([Interval(i) for i in dimensions])
+        ispace = IterationSpace(intervals, sub_iterators, directions)
+
+        ispace0, ispace1 = self.split(d)
+        relations = (ispace0.itdims + dimensions, dimensions + ispace1.itdims)
+
+        return IterationSpace.union(ispace0, ispace, ispace1, relations=relations)
+
+    def reorder(self, relations=None, mode=None):
+        relations = relations or self.relations
+        mode = mode or self.mode
+        intervals = IntervalGroup(self.intervals, relations=relations, mode=mode)
+
+        return IterationSpace(intervals, self.sub_iterators, self.directions)
+
+    def is_subset(self, other):
+        """
+        True if `self` is included within `other`, False otherwise.
+        """
+        if not self:
+            return True
+
+        d = self[-1].dim
+        try:
+            return self == other[:other.index(d) + 1]
+        except ValueError:
+            return False
+
     def is_compatible(self, other):
         """
         A relaxed version of ``__eq__``, in which only non-derived dimensions
@@ -877,20 +1043,10 @@ class IterationSpace(Space):
         return (self.intervals.is_compatible(other.intervals) and
                 self.nonderived_directions == other.nonderived_directions)
 
-    def is_forward(self, dim):
-        return self.directions[dim] is Forward
-
-    def is_sub_iterator(self, dim):
-        return dim in chain(*self.sub_iterators.values())
-
     @property
-    def itdimensions(self):
+    def itdims(self):
         return self.intervals.dimensions
 
-    @property
-    def relations(self):
-        return self.intervals.relations
-
     @property
     def sub_iterators(self):
         return self._sub_iterators
@@ -899,17 +1055,31 @@ class IterationSpace(Space):
     def directions(self):
         return self._directions
 
+    @property
+    def outermost(self):
+        return self[0]
+
+    @property
+    def innermost(self):
+        return self[-1]
+
+    @cached_property
+    def concrete(self):
+        return self.project(lambda d: not d.is_Virtual)
+
     @cached_property
     def itintervals(self):
-        return tuple(IterationInterval(
-            i, self.sub_iterators.get(i.dim), self.directions[i.dim]
-        ) for i in self.intervals)
+        return tuple(self[d] for d in self.itdims)
 
     @cached_property
     def dimensions(self):
         sub_dims = flatten(i._defines for v in self.sub_iterators.values() for i in v)
-        return tuple(filter_ordered(self.itdimensions + tuple(sub_dims)))
+        r_dims = flatten(i._defines for v in self.relations for i in v)
+        return tuple(filter_ordered(self.itdims + tuple(sub_dims) + tuple(r_dims)))
 
     @cached_property
     def nonderived_directions(self):
         return {k: v for k, v in self.directions.items() if not k.is_Derived}
+
+
+null_ispace = IterationSpace([])
