@@ -4,8 +4,8 @@ import logging
 import sys
 from contextlib import contextmanager
 
-__all__ = ('set_log_level', 'set_log_noperf', 'is_log_enabled_for',
-           'log', 'warning', 'error', 'perf', 'perf_adv',
+__all__ = ('set_log_level', 'set_log_noperf', 'is_log_enabled_for', 'switch_log_level',
+           'log', 'warning', 'error', 'perf', 'hint',
            'RED', 'GREEN', 'BLUE')
 
 
@@ -22,6 +22,10 @@ ERROR = logging.ERROR
 CRITICAL = logging.CRITICAL
 
 logging.addLevelName(PERF, "PERF")
+# Note: Do not use
+# logging.captureWarnings(True)
+# here as it will swallow all warnings (not just Devito warnings)
+# Instead use the `devito.warnings` module to log warnings
 
 logger_registry = {
     'DEBUG': DEBUG,
@@ -52,10 +56,12 @@ def _set_log_level(level):
     Set the level of the Devito logger.
     """
     if level not in logger_registry:
-        raise ValueError("Illegal logging level %s" % level)
+        raise ValueError(f"Illegal logging level {level}")
 
     logger.setLevel(level)
 
+    return level
+
 
 def set_log_level(level, comm=None):
     """
@@ -69,21 +75,42 @@ def set_log_level(level, comm=None):
     comm : MPI communicator, optional
         An MPI communicator the logger should be collective over. If provided, only
         rank-0 on that communicator will write to the registered handlers, other
-        ranks will use a `logging.NullHandler`.  By default, ``comm`` is set
-        to ``None``, so all ranks will use the default handlers.  This could be
+        ranks will use a `logging.NullHandler`. By default, ``comm`` is set
+        to ``None``, so all ranks will use the default handlers. This could be
         used, for example, if one wants to log to one file per rank.
     """
     from devito import configuration
 
-    if comm is not None:
+    if comm is not None and configuration['mpi']:
         if comm.rank != 0:
             logger.removeHandler(stream_handler)
             logger.addHandler(logging.NullHandler())
+    else:
+        logger.addHandler(stream_handler)
 
     # Triggers a callback to `_set_log_level`
     configuration['log-level'] = level
 
 
+class switch_log_level:
+    """
+    A context manager to temporarily change MPI logging.
+    """
+
+    def __init__(self, comm):
+
+        from devito import configuration
+        self.level = configuration['log-level']
+        self.comm = comm
+
+    def __enter__(self):
+        # Limit logging to rank 0
+        set_log_level(self.level, self.comm)
+
+    def __exit__(self, *args):
+        set_log_level(self.level)
+
+
 def set_log_noperf():
     """Do not print performance-related messages."""
     logger.setLevel(WARNING)
@@ -122,8 +149,8 @@ def perf(msg, *args, **kwargs):
     log(msg, PERF, *args, **kwargs)
 
 
-def perf_adv(msg, *args, **kwargs):
-    log("Potential optimisation missed: %s" % msg, PERF, *args, **kwargs)
+def hint(msg, *args, **kwargs):
+    log(f"Hint: {msg}", PERF, *args, **kwargs)
 
 
 def warning(msg, *args, **kwargs):
