@@ -8,30 +8,34 @@ from operator import mul
 from sympy import Integer
 
 from devito.data import OWNED, HALO, NOPAD, LEFT, CENTER, RIGHT
-from devito.ir.equations import DummyEq
-from devito.ir.iet import (Call, Callable, Conditional, Expression, ExpressionBundle,
-                           AugmentedExpression, Iteration, List, Prodder, Return,
-                           make_efunc, FindNodes, Transformer)
+from devito.ir.equations import DummyEq, OpInc, OpMin, OpMax
+from devito.ir.iet import (Call, Callable, Conditional, ElementalFunction,
+                           Expression, ExpressionBundle, AugmentedExpression,
+                           Iteration, List, Prodder, Return, make_efunc, FindNodes,
+                           Transformer, ElementalCall, CommCallable)
 from devito.mpi import MPI
 from devito.symbolics import (Byref, CondNe, FieldFromPointer, FieldFromComposite,
-                              IndexedPointer, Macro, cast_mapper, subs_op_args)
-from devito.tools import dtype_to_mpitype, dtype_to_ctype, flatten, generator
-from devito.types import Array, Dimension, Eq, Symbol, LocalObject, CompositeObject
+                              IndexedPointer, Macro, cast, subs_op_args)
+from devito.tools import (as_mapper, dtype_to_mpitype, dtype_len, infer_datasize,
+                          flatten, generator, is_integer)
+from devito.types import (Array, Bag, BundleView, Dimension, Eq, Symbol,
+                          LocalObject, CompositeObject, CustomDimension)
 
-__all__ = ['HaloExchangeBuilder', 'mpi_registry']
+__all__ = ['HaloExchangeBuilder', 'ReductionBuilder', 'mpi_registry']
 
 
-class HaloExchangeBuilder(object):
+class HaloExchangeBuilder:
 
     """
-    Build IET-based routines to implement MPI halo exchange.
+    Build IET routines to generate MPI halo exchanges.
     """
 
-    def __new__(cls, mpimode, generators=None, lower=None, **kwargs):
+    def __new__(cls, mpimode, generators=None, rcompile=None, sregistry=None,
+                **kwargs):
         obj = object.__new__(mpi_registry[mpimode])
 
-        obj._lower = lower
-        obj._kwargs = kwargs
+        obj.rcompile = rcompile
+        obj.sregistry = sregistry
 
         # Unique name generators
         generators = generators or {}
@@ -45,10 +49,6 @@ class HaloExchangeBuilder(object):
 
         return obj
 
-    @property
-    def sregistry(self):
-        return self._kwargs['sregistry']
-
     @property
     def efuncs(self):
         return self._efuncs
@@ -69,6 +69,11 @@ class HaloExchangeBuilder(object):
         # Sanity check
         assert all(f.is_Function and f.grid is not None for f in hs.fmapper)
 
+        # Pack-up Functions into Bundles. Worst case scenario, we have Bundles
+        # with just one components each. This is to maximize the likelihood of
+        # packed sends/recvs
+        hs = self._make_bundles(hs)
+
         mapper = {}
         for f, hse in hs.fmapper.items():
             # Build an MPIMsg, a data structure to be propagated across the
@@ -104,7 +109,8 @@ class HaloExchangeBuilder(object):
         if isinstance(remainder, Callable):
             self._efuncs.append(remainder)
 
-        # Now build up the HaloSpot body, with explicit Calls to the constructed Callables
+        # Now build up the HaloSpot body, with explicit Calls to the constructed
+        # Callables
         haloupdates = []
         halowaits = []
         for i, (f, hse) in enumerate(hs.fmapper.items()):
@@ -113,15 +119,18 @@ class HaloExchangeBuilder(object):
             haloupdates.append(self._call_haloupdate(haloupdate.name, f, hse, msg))
             if halowait is not None:
                 halowaits.append(self._call_halowait(halowait.name, f, hse, msg))
-        body = []
-        body.append(HaloUpdateList(body=haloupdates))
-        if callcompute is not None:
-            body.append(callcompute)
-        body.append(HaloWaitList(body=halowaits))
-        if remainder is not None:
-            body.append(self._call_remainder(remainder))
 
-        return List(body=body)
+        body = self._make_body(callcompute, remainder, haloupdates, halowaits)
+
+        return body
+
+    @abc.abstractmethod
+    def _make_bundles(self, hs):
+        """
+        Create Bundles from the Functions in `hs` to minimize the total number
+        of communications.
+        """
+        return
 
     @abc.abstractmethod
     def _make_region(self, hs, key):
@@ -258,6 +267,14 @@ class HaloExchangeBuilder(object):
         """
         return
 
+    @abc.abstractmethod
+    def _make_body(self, callcompute, remainder, haloupdates, halowaits):
+        """
+        Chain together the `compute`, `remainder`, `haloupdate`, and `halowait`
+        Calls.
+        """
+        return
+
 
 class BasicHaloExchangeBuilder(HaloExchangeBuilder):
 
@@ -270,6 +287,41 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         compute()
     """
 
+    def _make_bundles(self, hs):
+        halo_scheme = hs.halo_scheme
+
+        mapper = as_mapper(halo_scheme.fmapper, lambda i: halo_scheme.fmapper[i])
+        for hse, components in mapper.items():
+            halo_scheme = halo_scheme.drop(components)
+
+            # Existing Bundles are preserved
+            if hse.bundle:
+                if set(components) == set(hse.bundle.components):
+                    halo_scheme = halo_scheme.add(hse.bundle, hse)
+                else:
+                    name = f'bundleview_{hse.bundle.name}'
+                    bundle_view = BundleView(
+                        name=name, components=components, parent=hse.bundle
+                    )
+                    halo_scheme = halo_scheme.add(bundle_view, hse)
+                continue
+
+            # We recast everything else as Bags for simplicity -- worst case
+            # scenario all Bags only have one component.
+            try:
+                name = "bag_%s" % "".join(f.name for f in components)
+                bag = Bag(name=name, components=components)
+                halo_scheme = halo_scheme.add(bag, hse)
+            except ValueError:
+                for i in components:
+                    name = "bag_%s" % i.name
+                    bag = Bag(name=name, components=i)
+                    halo_scheme = halo_scheme.add(bag, hse)
+
+        hs = hs._rebuild(halo_scheme=halo_scheme)
+
+        return hs
+
     def _make_msg(self, f, hse, key):
         return
 
@@ -296,32 +348,48 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
 
     def _make_copy(self, f, hse, key, swap=False):
         dims = [d.root for d in f.dimensions if d not in hse.loc_indices]
-        buf = Array(name='buf', dimensions=dims, dtype=f.dtype, padding=0)
+        ofs = [Symbol(name='o%s' % d.root, is_const=True) for d in f.dimensions]
 
-        f_offsets = []
-        f_indices = []
-        for d, h in zip(f.dimensions, f._size_nodomain.left):
-            offset = Symbol(name='o%s' % d.root, is_const=True)
-            f_offsets.append(offset)
-            offset_nohalo = offset - h
-            f_indices.append(offset_nohalo + (d.root if d not in hse.loc_indices else 0))
-
-        if swap is False:
-            eq = Eq(buf[dims], f[f_indices])
-            name = 'gather%s' % key
-        else:
-            eq = Eq(f[f_indices], buf[dims])
-            name = 'scatter%s' % key
+        bshape = [Symbol(name='b%s' % d.symbolic_size) for d in dims]
+        bdims = [CustomDimension(name=d.name, parent=d, symbolic_size=s)
+                 for d, s in zip(dims, bshape)]
 
         eqns = []
-        eqns.extend([Eq(d.symbolic_min, 0) for d in dims])
-        eqns.extend([Eq(d.symbolic_max, d.symbolic_size - 1) for d in dims])
-        eqns.append(eq)
+        eqns.extend([Eq(d.symbolic_min, 0) for d in bdims])
+        eqns.extend([Eq(d.symbolic_max, d.symbolic_size - 1) for d in bdims])
+
+        vd = CustomDimension(name='vd', symbolic_size=f.ncomp)
+        buf = Array(name='buf', dimensions=[vd] + bdims, dtype=f.c0.dtype,
+                    padding=0)
+
+        mapper = dict(zip(dims, bdims))
+        findices = [o - h + mapper.get(d.root, 0)
+                    for d, o, h in zip(f.dimensions, ofs, f._size_nodomain.left)]
+
+        if swap is False:
+            swap = lambda i, j: (i, j)
+            name = 'gather%s' % key
+        else:
+            swap = lambda i, j: (j, i)
+            name = 'scatter%s' % key
+
+        if isinstance(f, Bag):
+            for i, c in enumerate(f.components):
+                eqns.append(Eq(*swap(buf[[i] + bdims], c[findices])))
+        elif isinstance(f, BundleView):
+            assert f.parent is hse.bundle
+            for i, c in enumerate(f.components):
+                indices = [f.parent.components.index(c), *findices]
+                eqns.append(Eq(*swap(buf[[i] + bdims], f.parent[indices])))
+        else:
+            assert f.is_Bundle
+            for i in range(f.ncomp):
+                eqns.append(Eq(*swap(buf[[i] + bdims], f[[i] + findices])))
 
         # Compile `eqns` into an IET via recursive compilation
-        irs, _ = self._lower(eqns, **self._kwargs)
+        irs, _ = self.rcompile(eqns, options={'mpi': False})
 
-        parameters = [buf] + list(buf.shape) + [f] + f_offsets
+        parameters = [buf] + bshape + list(f.handles) + ofs
 
         return CopyBuffer(name, irs.uiet, parameters)
 
@@ -329,9 +397,11 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         comm = f.grid.distributor._obj_comm
 
         dims = [d.root for d in f.dimensions if d not in hse.loc_indices]
-        bufg = Array(name='bufg', dimensions=dims, dtype=f.dtype,
+        bdims = [CustomDimension(name='vd', symbolic_size=f.ncomp)] + dims
+
+        bufg = Array(name='bufg', dimensions=bdims, dtype=f.c0.dtype,
                      padding=0, liveness='eager')
-        bufs = Array(name='bufs', dimensions=dims, dtype=f.dtype,
+        bufs = Array(name='bufs', dimensions=bdims, dtype=f.c0.dtype,
                      padding=0, liveness='eager')
 
         ofsg = [Symbol(name='og%s' % d.root) for d in f.dimensions]
@@ -340,8 +410,12 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         fromrank = Symbol(name='fromrank')
         torank = Symbol(name='torank')
 
-        gather = Call('gather%s' % key, [bufg] + list(bufg.shape) + [f] + ofsg)
-        scatter = Call('scatter%s' % key, [bufs] + list(bufs.shape) + [f] + ofss)
+        shape = [d.symbolic_size for d in dims]
+
+        arguments = [bufg] + shape + list(f.handles) + ofsg
+        gather = Gather('gather%s' % key, arguments)
+        arguments = [bufs] + shape + list(f.handles) + ofss
+        scatter = Scatter('scatter%s' % key, arguments)
 
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
@@ -361,11 +435,15 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
         waitsend = Call('MPI_Wait', [Byref(rsend), Macro('MPI_STATUS_IGNORE')])
 
         iet = List(body=[recv, gather, send, waitsend, waitrecv, scatter])
-        parameters = ([f] + list(bufs.shape) + ofsg + ofss + [fromrank, torank, comm])
+
+        parameters = (list(f.handles) + shape + ofsg + ofss +
+                      [fromrank, torank, comm])
+
         return SendRecv('sendrecv%s' % key, iet, parameters, bufg, bufs)
 
     def _call_sendrecv(self, name, *args, **kwargs):
-        return Call(name, flatten(args))
+        args = list(args[0].handles) + flatten(args[1:])
+        return Call(name, args)
 
     def _make_haloupdate(self, f, hse, key, sendrecv, **kwargs):
         distributor = f.grid.distributor
@@ -417,13 +495,15 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
                 body.append(self._call_sendrecv(sendrecv.name, *args, **kwargs))
 
         iet = List(body=body)
-        parameters = [f, comm, nb] + list(fixed.values())
+
+        parameters = list(f.handles) + [comm, nb] + list(fixed.values())
+
         return HaloUpdate('haloupdate%s' % key, iet, parameters)
 
     def _call_haloupdate(self, name, f, hse, *args):
         comm = f.grid.distributor._obj_comm
         nb = f.grid.distributor._obj_neighborhood
-        args = [f, comm, nb] + list(hse.loc_indices.values())
+        args = list(f.handles) + [comm, nb] + list(hse.loc_indices.values())
         return HaloUpdateCall(name, flatten(args))
 
     def _make_compute(self, *args):
@@ -450,6 +530,18 @@ class BasicHaloExchangeBuilder(HaloExchangeBuilder):
     def _call_remainder(self, *args):
         return
 
+    def _make_body(self, callcompute, remainder, haloupdates, halowaits):
+        body = []
+
+        body.append(HaloUpdateList(body=haloupdates))
+        if callcompute is not None:
+            body.append(callcompute)
+        body.append(HaloWaitList(body=halowaits))
+        if remainder is not None:
+            body.append(self._call_remainder(remainder))
+
+        return List(body=body)
+
 
 class DiagHaloExchangeBuilder(BasicHaloExchangeBuilder):
 
@@ -495,10 +587,20 @@ class DiagHaloExchangeBuilder(BasicHaloExchangeBuilder):
                                             fromrank, torank, comm, **kwargs))
 
         iet = List(body=body)
-        parameters = [f, comm, nb] + list(fixed.values())
+
+        parameters = list(f.handles) + [comm, nb] + list(fixed.values())
+
         return HaloUpdate('haloupdate%s' % key, iet, parameters)
 
 
+class ComputeCall(ElementalCall):
+    pass
+
+
+class ComputeFunction(ElementalFunction):
+    _Call_cls = ComputeCall
+
+
 class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
 
     """
@@ -519,6 +621,7 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         return MPIMsg('msg%d' % key, f, halos)
 
     def _make_sendrecv(self, f, hse, key, msg=None):
+        fcast = cast(f.c0.dtype, '*')
         comm = f.grid.distributor._obj_comm
 
         bufg = FieldFromPointer(msg._C_field_bufg, msg)
@@ -532,11 +635,12 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         sizes = [FieldFromPointer('%s[%d]' % (msg._C_field_sizes, i), msg)
                  for i in range(len(f._dist_dimensions))]
 
-        gather = Call('gather%s' % key, [bufg] + sizes + [f] + ofsg)
+        arguments = [fcast(bufg)] + sizes + list(f.handles) + ofsg
+        gather = Gather('gather%s' % key, arguments)
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
 
-        count = reduce(mul, sizes, 1)
+        count = reduce(mul, sizes, 1)*dtype_len(f.dtype)
         rrecv = Byref(FieldFromPointer(msg._C_field_rrecv, msg))
         rsend = Byref(FieldFromPointer(msg._C_field_rsend, msg))
         recv = IrecvCall([bufs, count, Macro(dtype_to_mpitype(f.dtype)),
@@ -545,7 +649,9 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
                          torank, Integer(13), comm, rsend])
 
         iet = List(body=[recv, gather, send])
-        parameters = ([f] + ofsg + [fromrank, torank, comm, msg])
+
+        parameters = list(f.handles) + ofsg + [fromrank, torank, comm, msg]
+
         return SendRecv('sendrecv%s' % key, iet, parameters, bufg, bufs)
 
     def _call_sendrecv(self, name, *args, msg=None, haloid=None):
@@ -554,7 +660,7 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         # to collect and scatter the result of an MPI_Irecv
         f, _, ofsg, _, fromrank, torank, comm = args
         msg = Byref(IndexedPointer(msg, haloid))
-        return Call(name, [f] + ofsg + [fromrank, torank, comm, msg])
+        return Call(name, list(f.handles) + ofsg + [fromrank, torank, comm, msg])
 
     def _make_haloupdate(self, f, hse, key, sendrecv, msg=None):
         iet = super()._make_haloupdate(f, hse, key, sendrecv, msg=msg)
@@ -570,7 +676,8 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         if hs.body.is_Call:
             return None
         else:
-            return make_efunc('compute%d' % key, hs.body, hs.arguments)
+            return make_efunc('compute%d' % key, hs.body, hs.arguments,
+                              efunc_type=ComputeFunction)
 
     def _call_compute(self, hs, compute, *args):
         if compute is None:
@@ -580,6 +687,8 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
             return compute.make_call(dynamic_args_mapper=hs.omapper.core)
 
     def _make_wait(self, f, hse, key, msg=None):
+        fcast = cast(f.c0.dtype, '*')
+
         bufs = FieldFromPointer(msg._C_field_bufs, msg)
 
         ofss = [Symbol(name='os%s' % d.root) for d in f.dimensions]
@@ -588,7 +697,8 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
 
         sizes = [FieldFromPointer('%s[%d]' % (msg._C_field_sizes, i), msg)
                  for i in range(len(f._dist_dimensions))]
-        scatter = Call('scatter%s' % key, [bufs] + sizes + [f] + ofss)
+        arguments = [fcast(bufs)] + sizes + list(f.handles) + ofss
+        scatter = Scatter('scatter%s' % key, arguments)
 
         # The `scatter` must be guarded as we must not alter the halo values along
         # the domain boundary, where the sender is actually MPI.PROC_NULL
@@ -600,7 +710,9 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
         waitsend = Call('MPI_Wait', [rsend, Macro('MPI_STATUS_IGNORE')])
 
         iet = List(body=[waitsend, waitrecv, scatter])
-        parameters = ([f] + ofss + [fromrank, msg])
+
+        parameters = (list(f.handles) + ofss + [fromrank, msg])
+
         return Callable('wait_%s' % key, iet, 'void', parameters, ('static',))
 
     def _make_halowait(self, f, hse, key, wait, msg=None):
@@ -621,20 +733,24 @@ class OverlapHaloExchangeBuilder(DiagHaloExchangeBuilder):
 
             msgi = Byref(IndexedPointer(msg, len(body)))
 
-            body.append(Call(wait.name, [f] + ofss + [fromrank, msgi]))
+            arguments = list(f.handles) + ofss + [fromrank, msgi]
+            body.append(Call(wait.name, arguments))
 
         iet = List(body=body)
-        parameters = [f] + list(fixed.values()) + [nb, msg]
-        return Callable('halowait%d' % key, iet, 'void', parameters, ('static',))
+
+        parameters = list(f.handles) + list(fixed.values()) + [nb, msg]
+
+        return HaloWait(f'halowait{key}', iet, parameters)
 
     def _call_halowait(self, name, f, hse, msg):
         nb = f.grid.distributor._obj_neighborhood
-        return HaloWaitCall(name, [f] + list(hse.loc_indices.values()) + [nb, msg])
+        arguments = list(f.handles) + list(hse.loc_indices.values()) + [nb, msg]
+        return HaloWaitCall(name, arguments)
 
     def _make_remainder(self, hs, key, callcompute, *args):
         assert callcompute.is_Call
         body = [callcompute._rebuild(dynamic_args_mapper=i) for _, i in hs.omapper.owned]
-        return make_efunc('remainder%d' % key, body)
+        return Remainder.make('remainder%d' % key, body)
 
     def _call_remainder(self, remainder):
         efunc = remainder.make_call()
@@ -663,7 +779,7 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
     def _make_msg(self, f, hse, key):
         # Only retain the halos required by the Diag scheme
         halos = sorted(i for i in hse.halos if isinstance(i.dim, tuple))
-        return MPIMsgEnriched('msg%d' % key, f, halos)
+        return MPIMsgEnriched(f'msg{key}', f, halos)
 
     def _make_sendrecv(self, *args, **kwargs):
         return
@@ -672,7 +788,7 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         return
 
     def _make_haloupdate(self, f, hse, key, *args, msg=None):
-        cast = cast_mapper[(f.dtype, '*')]
+        fcast = cast(f.c0.dtype, '*')
         comm = f.grid.distributor._obj_comm
 
         fixed = {d: Symbol(name="o%s" % d.root) for d in hse.loc_indices}
@@ -694,12 +810,12 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         ofsg = [fixed.get(d) or ofsg.pop(0) for d in f.dimensions]
 
         # The `gather` is unnecessary if sending to MPI.PROC_NULL
-
-        gather = Call('gather%s' % key, [cast(bufg)] + sizes + [f] + ofsg)
+        arguments = [fcast(bufg)] + sizes + list(f.handles) + ofsg
+        gather = Gather('gather%s' % key, arguments)
         gather = Conditional(CondNe(torank, Macro('MPI_PROC_NULL')), gather)
 
         # Make Irecv/Isend
-        count = reduce(mul, sizes, 1)
+        count = reduce(mul, sizes, 1)*dtype_len(f.dtype)
         rrecv = Byref(FieldFromComposite(msg._C_field_rrecv, msgi))
         rsend = Byref(FieldFromComposite(msg._C_field_rsend, msgi))
         recv = IrecvCall([bufs, count, Macro(dtype_to_mpitype(f.dtype)),
@@ -710,16 +826,16 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         # The -1 below is because an Iteration, by default, generates <=
         ncomms = Symbol(name='ncomms')
         iet = Iteration([recv, gather, send], dim, ncomms - 1)
-        parameters = ([f, comm, msg, ncomms]) + list(fixed.values())
+        parameters = f.handles + (comm, msg, ncomms) + tuple(fixed.values())
         return HaloUpdate('haloupdate%s' % key, iet, parameters)
 
     def _call_haloupdate(self, name, f, hse, msg):
         comm = f.grid.distributor._obj_comm
-        args = [f, comm, msg, msg.npeers] + list(hse.loc_indices.values())
+        args = f.handles + (comm, msg, msg.npeers) + tuple(hse.loc_indices.values())
         return HaloUpdateCall(name, args)
 
     def _make_halowait(self, f, hse, key, *args, msg=None):
-        cast = cast_mapper[(f.dtype, '*')]
+        fcast = cast(f.c0.dtype, '*')
 
         fixed = {d: Symbol(name="o%s" % d.root) for d in hse.loc_indices}
 
@@ -739,7 +855,8 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
 
         # The `scatter` must be guarded as we must not alter the halo values along
         # the domain boundary, where the sender is actually MPI.PROC_NULL
-        scatter = Call('scatter%s' % key, [cast(bufs)] + sizes + [f] + ofss)
+        arguments = [fcast(bufs)] + sizes + list(f.handles) + ofss
+        scatter = Scatter('scatter%s' % key, arguments)
         scatter = Conditional(CondNe(fromrank, Macro('MPI_PROC_NULL')), scatter)
 
         rrecv = Byref(FieldFromComposite(msg._C_field_rrecv, msgi))
@@ -750,11 +867,11 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         # The -1 below is because an Iteration, by default, generates <=
         ncomms = Symbol(name='ncomms')
         iet = Iteration([waitsend, waitrecv, scatter], dim, ncomms - 1)
-        parameters = ([f] + list(fixed.values()) + [msg, ncomms])
-        return Callable('halowait%d' % key, iet, 'void', parameters, ('static',))
+        parameters = f.handles + tuple(fixed.values()) + (msg, ncomms)
+        return HaloWait(f'halowait{key}', iet, parameters)
 
     def _call_halowait(self, name, f, hse, msg):
-        args = [f] + list(hse.loc_indices.values()) + [msg, msg.npeers]
+        args = f.handles + tuple(hse.loc_indices.values()) + (msg, msg.npeers)
         return HaloWaitCall(name, args)
 
     def _make_wait(self, *args, **kwargs):
@@ -781,7 +898,7 @@ class Overlap2HaloExchangeBuilder(OverlapHaloExchangeBuilder):
         # The -1 below is because an Iteration, by default, generates <=
         iet = Iteration(iet, dim, region.nregions - 1)
 
-        return make_efunc('remainder%d' % key, iet)
+        return Remainder.make('remainder%d' % key, iet)
 
 
 class Diag2HaloExchangeBuilder(Overlap2HaloExchangeBuilder):
@@ -797,6 +914,9 @@ class Diag2HaloExchangeBuilder(Overlap2HaloExchangeBuilder):
         compute()
     """
 
+    def _make_region(self, hs, key):
+        return
+
     def _make_compute(self, hs, key, *args):
         return
 
@@ -811,6 +931,36 @@ class Diag2HaloExchangeBuilder(Overlap2HaloExchangeBuilder):
         return remainder
 
 
+class DualHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
+
+    """
+    "Dual" of Overlap2HaloExchangeBuilder, as the "remainder" is now the first
+    thing getting computed.
+
+    Generates:
+
+        remainder()
+        haloupdate()
+        compute_core()
+        halowait()
+    """
+
+    def _make_body(self, callcompute, remainder, haloupdates, halowaits):
+        body = []
+
+        assert remainder is not None
+        body.append(self._call_remainder(remainder))
+
+        body.append(HaloUpdateList(body=haloupdates))
+
+        assert callcompute is not None
+        body.append(callcompute)
+
+        body.append(HaloWaitList(body=halowaits))
+
+        return List(body=body)
+
+
 class FullHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
 
     """
@@ -832,7 +982,8 @@ class FullHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
             mapper = {i: List(body=[callpoke, i]) for i in
                       FindNodes(ExpressionBundle).visit(hs.body)}
             iet = Transformer(mapper).visit(hs.body)
-            return make_efunc('compute%d' % key, iet, hs.arguments)
+            return make_efunc('compute%d' % key, iet, hs.arguments,
+                              efunc_type=ComputeFunction)
 
     def _make_poke(self, hs, key, msgs):
         lflag = Symbol(name='lflag')
@@ -867,23 +1018,23 @@ class FullHaloExchangeBuilder(Overlap2HaloExchangeBuilder):
 
 
 mpi_registry = {
-    True: BasicHaloExchangeBuilder,
     'basic': BasicHaloExchangeBuilder,
     'diag': DiagHaloExchangeBuilder,
     'diag2': Diag2HaloExchangeBuilder,
     'overlap': OverlapHaloExchangeBuilder,
     'overlap2': Overlap2HaloExchangeBuilder,
-    'full': FullHaloExchangeBuilder
+    'full': FullHaloExchangeBuilder,
+    'dual': DualHaloExchangeBuilder
 }
 
 
 # Callable sub-hierarchy
 
 
-class MPICallable(Callable):
+class MPICallable(CommCallable):
 
     def __init__(self, name, body, parameters):
-        super(MPICallable, self).__init__(name, body, 'void', parameters, ('static',))
+        super().__init__(name, body, 'void', parameters, ('static',))
 
 
 class CopyBuffer(MPICallable):
@@ -893,33 +1044,53 @@ class CopyBuffer(MPICallable):
 class SendRecv(MPICallable):
 
     def __init__(self, name, body, parameters, bufg, bufs):
-        super(SendRecv, self).__init__(name, body, parameters)
+        super().__init__(name, body, parameters)
         self.bufg = bufg
         self.bufs = bufs
 
 
 class HaloUpdate(MPICallable):
+    pass
 
-    def __init__(self, name, body, parameters):
-        super(HaloUpdate, self).__init__(name, body, parameters)
+
+class HaloWait(MPICallable):
+    pass
+
+
+class Remainder(ElementalFunction):
+    pass
 
 
 # Call sub-hierarchy
 
+class Gather(Call):
+    pass
+
+
+class Scatter(Call):
+    pass
+
+
 class IsendCall(Call):
 
-    def __init__(self, arguments):
+    def __init__(self, arguments, **kwargs):
         super().__init__('MPI_Isend', arguments)
 
 
 class IrecvCall(Call):
 
-    def __init__(self, arguments):
+    def __init__(self, arguments, **kwargs):
         super().__init__('MPI_Irecv', arguments)
 
 
 class MPICall(Call):
-    pass
+
+    @property
+    def ncomps(self):
+        """
+        The number of components this MPICall was constructed for.
+        """
+        return len([f for f in self.functions if f.is_DiscreteFunction])
 
 
 class HaloUpdateCall(MPICall):
@@ -1019,42 +1190,69 @@ class MPIMsg(CompositeObject):
     def npeers(self):
         return len(self._halos)
 
-    def _arg_defaults(self, allocator, alias=None):
-        # Lazy initialization if `allocator` is necessary as the `allocator`
-        # type isn't really known until an Operator is constructed
-        self._allocator = allocator
+    def _as_number(self, v, args):
+        """
+        Turn a sympy.Symbol into a number. In doing so, perform a number of
+        sanity checks to ensure we get a Symbol iff the Msg is for an Array.
+        """
+        if is_integer(v):
+            return int(v)
+        else:
+            assert self.target.c0.is_Array
+            assert args is not None
+            return int(subs_op_args(v, args))
 
-        target = alias or self.target
+    def _arg_defaults(self, alias, args=None):
+        # `allocator` is lazily initialized as it is not known until an Operator
+        # is constructed
+        self._allocator = allocator = args.allocator
+
+        f = alias or self.target.c0
         for i, halo in enumerate(self.halos):
             entry = self.value[i]
 
-            # Buffer size for this peer
+            # Buffer shape for this peer
             shape = []
             for dim, side in zip(*halo):
                 try:
-                    shape.append(getattr(target._size_owned[dim], side.name))
+                    shape.append(getattr(f._size_owned[dim], side.name))
                 except AttributeError:
-                    assert side is CENTER
-                    shape.append(target._size_domain[dim])
+                    assert side == CENTER
+                    shape.append(self._as_number(f._size_domain[dim], args))
             entry.sizes = (c_int*len(shape))(*shape)
 
             # Allocate the send/recv buffers
-            size = reduce(mul, shape)
-            ctype = dtype_to_ctype(target.dtype)
+            ctype, datasize = infer_datasize(f.dtype, shape)
+            size = datasize * dtype_len(self.target.dtype)
             entry.bufg, bufg_memfree_args = allocator._alloc_C_libcall(size, ctype)
             entry.bufs, bufs_memfree_args = allocator._alloc_C_libcall(size, ctype)
 
-            # The `memfree_args` will be used to deallocate the buffer upon returning
-            # from C-land
+            # The `memfree_args` will be used to deallocate the buffer upon
+            # returning from C-land
             self._memfree_args.extend([bufg_memfree_args, bufs_memfree_args])
 
         return {self.name: self.value}
 
     def _arg_values(self, args=None, **kwargs):
-        return self._arg_defaults(
-            args.allocator,
-            alias=kwargs.get(self.target.name, self.target)
-        )
+        # Since the components are by construction homogeneous, any will do
+        v = self.target
+        if isinstance(v, BundleView):
+            # NOTE: This is a tricky one: think about a BundleView with just K
+            # components, while its parent has N (clearly N > K). There could be
+            # an override for one of the N-K parent's components, but not for any
+            # of the K BundleView's components. So we need to check the parent
+            # for the override, and not the BundleView itself
+            v = v.parent
+        for f in v.components:
+            try:
+                alias = kwargs[f.name]
+                break
+            except KeyError:
+                pass
+        else:
+            alias = f
+
+        return self._arg_defaults(alias=alias, args=args)
 
     def _arg_apply(self, *args, **kwargs):
         self._C_memfree()
@@ -1066,45 +1264,63 @@ class MPIMsgEnriched(MPIMsg):
     _C_field_ofsg = 'ofsg'
     _C_field_from = 'fromrank'
     _C_field_to = 'torank'
+    _C_field_components = 'components'
 
     fields = MPIMsg.fields + [
         (_C_field_ofss, POINTER(c_int)),
         (_C_field_ofsg, POINTER(c_int)),
         (_C_field_from, c_int),
-        (_C_field_to, c_int)
+        (_C_field_to, c_int),
+        (_C_field_components, POINTER(c_int)),
     ]
 
-    def _arg_defaults(self, allocator, alias=None):
-        super()._arg_defaults(allocator, alias)
+    def _arg_defaults(self, alias=None, args=None):
+        super()._arg_defaults(alias, args=args)
+
+        f = alias or self.target.c0
+        neighborhood = args.grid.distributor.neighborhood
 
-        function = alias or self.function
-        neighborhood = function.grid.distributor.neighborhood
         for i, halo in enumerate(self.halos):
             entry = self.value[i]
+
             # `torank` peer + gather offsets
             entry.torank = neighborhood[halo.side]
             ofsg = []
             for dim, side in zip(*halo):
                 try:
-                    ofsg.append(getattr(function._offset_owned[dim], side.name))
+                    v = getattr(f._offset_owned[dim], side.name)
+                    ofsg.append(self._as_number(v, args))
                 except AttributeError:
-                    assert side is CENTER
-                    ofsg.append(function._offset_owned[dim].left)
+                    assert side == CENTER
+                    ofsg.append(f._offset_owned[dim].left)
             entry.ofsg = (c_int*len(ofsg))(*ofsg)
+
             # `fromrank` peer + scatter offsets
             entry.fromrank = neighborhood[tuple(i.flip() for i in halo.side)]
             ofss = []
             for dim, side in zip(*halo):
                 try:
-                    ofss.append(getattr(function._offset_halo[dim], side.flip().name))
+                    v = getattr(f._offset_halo[dim], side.flip().name)
+                    ofss.append(self._as_number(v, args))
                 except AttributeError:
-                    assert side is CENTER
-                    # Note `_offset_owned`, and not `_offset_halo`, is *not* a bug here.
-                    # If it's the CENTER we need, we can't use `_offset_halo[d].left`
-                    # as otherwise we would be picking the corner
-                    ofss.append(function._offset_owned[dim].left)
+                    assert side == CENTER
+                    # Note `_offset_owned`, and not `_offset_halo`, is *not* a bug
+                    # here. If it's the CENTER we need, we can't just use
+                    # `_offset_halo[d].left` as otherwise we would pick the corner
+                    ofss.append(f._offset_owned[dim].left)
             entry.ofss = (c_int*len(ofss))(*ofss)
 
+            # Track the component accesses for packing/unpacking as numbers
+            # representing the field being accessed (that is: .x -> 0, .y -> 1,
+            # .z -> 2, .w -> 3), if any
+            if isinstance(self.target, BundleView):
+                ncomp = self.target.ncomp
+                component_indices = self.target.component_indices
+                entry.components = (c_int*ncomp)(*component_indices)
+            elif self.target.is_Bundle:
+                ncomp = self.target.ncomp
+                entry.components = (c_int*ncomp)(*range(ncomp))
+
         return {self.name: self.value}
 
 
@@ -1131,7 +1347,7 @@ class MPIRegion(CompositeObject):
             else:
                 fields.append((i.name, c_int))
 
-        super(MPIRegion, self).__init__(name, pname, fields)
+        super().__init__(name, pname, fields)
 
     def __value_setup__(self, dtype, value):
         # We eventually produce an array of `struct region` that is as big as
@@ -1174,3 +1390,38 @@ class MPIRegion(CompositeObject):
                     except AttributeError:
                         setattr(entry, a.name, mapper[a][0])
         return values
+
+
+class AllreduceCall(Call):
+
+    def __init__(self, arguments, **kwargs):
+        super().__init__('MPI_Allreduce', arguments)
+
+
+class ReductionBuilder(object):
+
+    """
+    Build IET routines performing MPI reductions.
+    """
+
+    mapper = {
+        OpInc: 'MPI_SUM',
+        OpMax: 'MPI_MAX',
+        OpMin: 'MPI_MIN',
+    }
+
+    def make(self, dr):
+        """
+        Construct Callables and Calls implementing distributed-memory reductions.
+        """
+        f = dr.var
+        comm = dr.grid.distributor._obj_comm
+
+        inplace = Macro('MPI_IN_PLACE')
+        mpitype = Macro(dtype_to_mpitype(f.dtype))
+        op = self.mapper[dr.op]
+
+        arguments = [inplace, Byref(f), Integer(1), mpitype, op, comm]
+        allreduce = AllreduceCall(arguments)
+
+        return allreduce
