@@ -1,22 +1,19 @@
-from collections import OrderedDict, namedtuple
+from collections import OrderedDict, defaultdict, namedtuple
 from contextlib import contextmanager
 from functools import reduce
 from operator import mul
-from pathlib import Path
-from subprocess import DEVNULL, PIPE, run
 from time import time as seq_time
-import os
 
 import cgen as c
 import numpy as np
 from sympy import S
 
-from devito.ir.iet import (BusyWait, ExpressionBundle, List, TimedList, Section,
+from devito.arch import get_advisor_path
+from devito.ir.iet import (ExpressionBundle, List, TimedList, Section,
                            Iteration, FindNodes, Transformer)
 from devito.ir.support import IntervalGroup
-from devito.logger import warning, error
+from devito.logger import warning
 from devito.mpi import MPI
-from devito.mpi.routines import MPICall, MPIList, RemainderCall
 from devito.parameters import configuration
 from devito.symbolics import subs_op_args
 from devito.tools import DefaultOrderedDict, flatten
@@ -30,13 +27,20 @@ PerfInput = namedtuple('PerfInput', 'time ops points traffic sops itershapes')
 PerfEntry = namedtuple('PerfEntry', 'time gflopss gpointss oi ops itershapes')
 
 
-class Profiler(object):
+class Profiler:
 
     _default_includes = []
     _default_libs = []
     _ext_calls = []
 
-    """Metadata for a profiled code section."""
+    _include_dirs = []
+    _lib_dirs = []
+
+    _supports_async_sections = False
+
+    _verbosity = 0
+
+    _attempted_init = False
 
     def __init__(self, name):
         self.name = name
@@ -51,7 +55,7 @@ class Profiler(object):
         # Python-level timers
         self.py_timers = OrderedDict()
 
-        self.initialized = True
+        self._attempted_init = True
 
     def analyze(self, iet):
         """
@@ -68,7 +72,11 @@ class Profiler(object):
             ops = sum(i.ops*i.ispace.size for i in bundles)
 
             # Operation count at each section iteration
-            sops = sum(i.ops for i in bundles)
+            # NOTE: for practical reasons, it makes much more sense to "flatten"
+            # the StencilDimensions, because one expects to see the number of
+            # operations per time- or space-Dimension
+            sops = sum(i.ops*max(i.ispace.project(lambda d: d.is_Stencil).size, 1)
+                       for i in bundles)
 
             # Total memory traffic
             mapper = {}
@@ -90,7 +98,9 @@ class Profiler(object):
             points = set()
             for i in bundles:
                 if any(e.write.is_TimeFunction for e in i.exprs):
-                    points.add(i.size)
+                    # The `ispace` is zero-ed because we don't care if a point
+                    # is touched redundantly
+                    points.add(i.ispace.zero().size)
             points = sum(points, S.Zero)
 
             self._sections[s.name] = SectionData(ops, sops, points, traffic, itermaps)
@@ -99,6 +109,25 @@ class Profiler(object):
         v = self._subsections.setdefault(sname, OrderedDict())
         v[name] = SectionData(S.Zero, S.Zero, S.Zero, S.Zero, [])
 
+    def group_as_subsections(self, sname, sections):
+        ops = sum(self._sections[i].ops for i in sections)
+        points = sum(self._sections[i].points for i in sections)
+        traffic = sum(self._sections[i].traffic for i in sections)
+        sectiondata = SectionData(ops, S.Zero, points, traffic, [])
+
+        v = self._subsections.setdefault(sname, OrderedDict())
+        v.update({i: self._sections[i] for i in sections})
+
+        new_sections = OrderedDict()
+        for k, v in self._sections.items():
+            try:
+                if sections.index(k) == len(sections) - 1:
+                    new_sections[sname] = sectiondata
+            except ValueError:
+                new_sections[k] = v
+        self._sections.clear()
+        self._sections.update(new_sections)
+
     def instrument(self, iet, timer):
         """
         Instrument the given IET for C-level performance profiling.
@@ -150,10 +179,6 @@ class Profiler(object):
     def all_sections(self):
         return list(self._sections) + flatten(self._subsections.values())
 
-    @property
-    def trackable_subsections(self):
-        return ()
-
     def summary(self, args, dtype, reduce_over=None):
         """
         Return a PerformanceSummary of the profiled sections.
@@ -188,21 +213,69 @@ class Profiler(object):
 
 
 class ProfilerVerbose1(Profiler):
-
-    @property
-    def trackable_subsections(self):
-        return (MPIList, RemainderCall, BusyWait)
+    _verbosity = 1
 
 
 class ProfilerVerbose2(Profiler):
-
-    @property
-    def trackable_subsections(self):
-        return (MPICall, BusyWait)
+    _verbosity = 2
 
 
 class AdvancedProfiler(Profiler):
 
+    _supports_async_sections = True
+
+    def _evaluate_section(self, name, data, args, dtype):
+        # Time to run the section
+        time = max(getattr(args[self.name]._obj, name), 10e-7)
+
+        # Number of FLOPs performed
+        try:
+            ops = int(subs_op_args(data.ops, args))
+        except (AttributeError, TypeError):
+            # E.g., a section comprising just function calls, or at least
+            # a sequence of unrecognized or non-conventional expr statements
+            ops = np.nan
+
+        try:
+            # Number of grid points computed
+            points = int(subs_op_args(data.points, args))
+
+            # Compulsory traffic
+            traffic = float(subs_op_args(data.traffic, args)*dtype().itemsize)
+        except (AttributeError, TypeError):
+            # E.g., the section has a dynamic loop size
+            points = np.nan
+
+            traffic = np.nan
+
+        # Nmber of FLOPs performed at each iteration
+        sops = data.sops
+
+        # Runtime itermaps/itershapes
+        try:
+            itermaps = [OrderedDict([(k, int(subs_op_args(v, args)))
+                                     for k, v in i.items()])
+                        for i in data.itermaps]
+            itershapes = tuple(tuple(i.values()) for i in itermaps)
+        except TypeError:
+            # E.g., a section comprising just function calls, or at least
+            # a sequence of unrecognized or non-conventional expr statements
+            itershapes = ()
+
+        return time, ops, points, traffic, sops, itershapes
+
+    def _allgather_from_comm(self, comm, time, ops, points, traffic, sops, itershapes):
+        times = comm.allgather(time)
+        assert comm.size == len(times)
+
+        opss = comm.allgather(ops)
+        pointss = comm.allgather(points)
+        traffics = comm.allgather(traffic)
+        sops = [sops]*comm.size
+        itershapess = comm.allgather(itershapes)
+
+        return list(zip(times, opss, pointss, traffics, sops, itershapess))
+
     # Override basic summary so that arguments other than runtime are computed.
     def summary(self, args, dtype, reduce_over=None):
         grid = args.grid
@@ -211,66 +284,45 @@ class AdvancedProfiler(Profiler):
         # Produce sections summary
         summary = PerformanceSummary()
         for name, data in self._sections.items():
-            # Time to run the section
-            time = max(getattr(args[self.name]._obj, name), 10e-7)
-
-            # Number of FLOPs performed
-            ops = int(subs_op_args(data.ops, args))
-
-            try:
-                # Number of grid points computed
-                points = int(subs_op_args(data.points, args))
-
-                # Compulsory traffic
-                traffic = float(subs_op_args(data.traffic, args)*dtype().itemsize)
-            except TypeError:
-                # E.g., the section has a dynamic loop size
-                points = np.nan
-
-                traffic = np.nan
-
-            # Runtime itermaps/itershapes
-            itermaps = [OrderedDict([(k, int(subs_op_args(v, args)))
-                                     for k, v in i.items()])
-                        for i in data.itermaps]
-            itershapes = tuple(tuple(i.values()) for i in itermaps)
+            items = self._evaluate_section(name, data, args, dtype)
 
             # Add local performance data
             if comm is not MPI.COMM_NULL:
                 # With MPI enabled, we add one entry per section per rank
-                times = comm.allgather(time)
-                assert comm.size == len(times)
-                opss = comm.allgather(ops)
-                pointss = comm.allgather(points)
-                traffics = comm.allgather(traffic)
-                sops = [data.sops]*comm.size
-                itershapess = comm.allgather(itershapes)
-                items = list(zip(times, opss, pointss, traffics, sops, itershapess))
+                items = self._allgather_from_comm(comm, *items)
                 for rank in range(comm.size):
                     summary.add(name, rank, *items[rank])
             else:
-                summary.add(name, None, time, ops, points, traffic, data.sops, itershapes)
+                summary.add(name, None, *items)
 
         # Enrich summary with subsections data
         for sname, v in self._subsections.items():
             for name, data in v.items():
-                # Time to run the section
-                time = max(getattr(args[self.name]._obj, name), 10e-7)
+                items = self._evaluate_section(name, data, args, dtype)
 
                 # Add local performance data
                 if comm is not MPI.COMM_NULL:
                     # With MPI enabled, we add one entry per section per rank
-                    times = comm.allgather(time)
-                    assert comm.size == len(times)
+                    items = self._allgather_from_comm(comm, *items)
                     for rank in range(comm.size):
-                        summary.add_subsection(sname, name, rank, time)
+                        summary.add_subsection(sname, name, rank, *items[rank])
                 else:
-                    summary.add_subsection(sname, name, None, time)
+                    summary.add_subsection(sname, name, None, *items)
 
         # Add global performance data
         if reduce_over is not None:
             # Vanilla metrics
-            summary.add_glb_vanilla(self.py_timers[reduce_over])
+            summary.add_glb_vanilla('vanilla', reduce_over)
+
+            # Same as above but without setup overheads (e.g., host-device
+            # data transfers)
+            mapper = defaultdict(list)
+            for (name, rank), v in summary.items():
+                mapper[name].append(v.time)
+            reduce_over_nosetup = sum(max(i) for i in mapper.values())
+            if reduce_over_nosetup == 0:
+                reduce_over_nosetup = reduce_over
+            summary.add_glb_vanilla('vanilla-nosetup', reduce_over_nosetup)
 
             # Typical finite difference benchmark metrics
             if grid is not None:
@@ -280,32 +332,31 @@ class AdvancedProfiler(Profiler):
                     min_t = args[grid.time_dim.min_name] or 0
                     nt = max_t - min_t + 1
                     points = reduce(mul, (nt,) + grid.shape)
-                    summary.add_glb_fdlike(points, self.py_timers[reduce_over])
+                    summary.add_glb_fdlike('fdlike', points, reduce_over)
+
+                    # Same as above but without setup overheads (e.g., host-device
+                    # data transfers)
+                    summary.add_glb_fdlike('fdlike-nosetup', points, reduce_over_nosetup)
 
         return summary
 
 
-class AdvancedProfilerVerbose1(AdvancedProfiler):
-
-    @property
-    def trackable_subsections(self):
-        return (MPIList, RemainderCall, BusyWait)
+class AdvancedProfilerVerbose(AdvancedProfiler):
+    pass
 
 
-class AdvancedProfilerVerbose2(AdvancedProfiler):
+class AdvancedProfilerVerbose1(AdvancedProfilerVerbose):
+    _verbosity = 1
 
-    @property
-    def trackable_subsections(self):
-        return (MPICall, BusyWait)
+
+class AdvancedProfilerVerbose2(AdvancedProfilerVerbose):
+    _verbosity = 2
 
 
 class AdvisorProfiler(AdvancedProfiler):
 
     """
-    Rely on Intel Advisor ``v >= 2020`` for performance profiling.
-    Tested versions of Intel Advisor:
-    - As contained in Intel Parallel Studio 2020 v 2020 Update 2
-    - As contained in Intel oneAPI 2021 beta08
+    Rely on Intel Advisor 2025.1 for performance profiling.
     """
 
     _api_resume = '__itt_resume'
@@ -316,21 +367,23 @@ class AdvisorProfiler(AdvancedProfiler):
     _ext_calls = [_api_resume, _api_pause]
 
     def __init__(self, name):
-        self.path = locate_intel_advisor()
-        if self.path is None:
-            self.initialized = False
+        if self._attempted_init:
+            return
+
+        super().__init__(name)
+
+        path = get_advisor_path()
+        if path:
+            self._include_dirs.append(path.joinpath('include').as_posix())
+            self._lib_dirs.append(path.joinpath('lib64').as_posix())
+            self._attempted_init = True
         else:
-            super(AdvisorProfiler, self).__init__(name)
-            # Make sure future compilations will get the proper header and
-            # shared object files
-            compiler = configuration['compiler']
-            compiler.add_include_dirs(self.path.joinpath('include').as_posix())
-            compiler.add_libraries(self._default_libs)
-            libdir = self.path.joinpath('lib64').as_posix()
-            compiler.add_library_dirs(libdir)
-            compiler.add_ldflags('-Wl,-rpath,%s' % libdir)
+            self._attempted_init = False
 
     def analyze(self, iet):
+        """
+        A no-op, as the Advisor profiler does not need to analyze the IET.
+        """
         return
 
     def instrument(self, iet, timer):
@@ -340,9 +393,9 @@ class AdvisorProfiler(AdvancedProfiler):
             if i.dim.is_Time:
                 # The calls to Advisor's Collection Control API are only for Operators
                 # with a time loop
-                mapper[i] = List(header=c.Statement('%s()' % self._api_resume),
+                mapper[i] = List(header=c.Statement(f'{self._api_resume}()'),
                                  body=i,
-                                 footer=c.Statement('%s()' % self._api_pause))
+                                 footer=c.Statement(f'{self._api_pause}()'))
                 return Transformer(mapper).visit(iet)
 
         # Return the IET intact if no time loop is found
@@ -352,7 +405,7 @@ class AdvisorProfiler(AdvancedProfiler):
 class PerformanceSummary(OrderedDict):
 
     def __init__(self, *args, **kwargs):
-        super(PerformanceSummary, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.subsections = DefaultOrderedDict(lambda: OrderedDict())
         self.input = OrderedDict()
         self.globals = {}
@@ -364,16 +417,12 @@ class PerformanceSummary(OrderedDict):
         performance data is local, that is "per-rank".
         """
         # Do not show unexecuted Sections (i.e., loop trip count was 0)
-        if ops == 0 or traffic == 0:
-            return
-        # Do not show dynamic Sections (i.e., loop trip counts varies dynamically)
-        if traffic is not None and np.isnan(traffic):
-            assert np.isnan(points)
+        if traffic == 0:
             return
 
         k = PerfKey(name, rank)
 
-        if ops is None:
+        if not ops or any(not np.isfinite(i) for i in [ops, points, traffic]):
             self[k] = PerfEntry(time, 0.0, 0.0, 0.0, 0, [])
         else:
             gflops = float(ops)/10**9
@@ -386,13 +435,13 @@ class PerformanceSummary(OrderedDict):
 
         self.input[k] = PerfInput(time, ops, points, traffic, sops, itershapes)
 
-    def add_subsection(self, sname, name, rank, time):
+    def add_subsection(self, sname, name, rank, time, *args):
         k0 = PerfKey(sname, rank)
         assert k0 in self
 
-        self.subsections[sname][name] = time
+        self.subsections[sname][name] = PerfEntry(time, None, None, None, None, [])
 
-    def add_glb_vanilla(self, time):
+    def add_glb_vanilla(self, key, time):
         """
         Reduce the following performance data:
 
@@ -404,19 +453,20 @@ class PerformanceSummary(OrderedDict):
         if not self.input:
             return
 
-        ops = sum(v.ops for v in self.input.values())
+        ops = sum(v.ops for v in self.input.values() if not np.isnan(v.ops))
         traffic = sum(v.traffic for v in self.input.values())
 
-        if np.isnan(traffic):
-            return
-
         gflops = float(ops)/10**9
         gflopss = gflops/time
-        oi = float(ops/traffic)
 
-        self.globals['vanilla'] = PerfEntry(time, gflopss, None, oi, None, None)
+        if np.isnan(traffic) or traffic == 0:
+            oi = None
+        else:
+            oi = float(ops/traffic)
 
-    def add_glb_fdlike(self, points, time):
+        self.globals[key] = PerfEntry(time, gflopss, None, oi, None, None)
+
+    def add_glb_fdlike(self, key, points, time):
         """
         Add the typical GPoints/s finite-difference metric.
         """
@@ -426,7 +476,7 @@ class PerformanceSummary(OrderedDict):
         gpoints = float(points)/10**9
         gpointss = gpoints/time
 
-        self.globals['fdlike'] = PerfEntry(time, None, gpointss, None, None, None)
+        self.globals[key] = PerfEntry(time, None, gpointss, None, None, None)
 
     @property
     def globals_all(self):
@@ -434,6 +484,12 @@ class PerformanceSummary(OrderedDict):
         v1 = self.globals['fdlike']
         return PerfEntry(v0.time, v0.gflopss, v1.gpointss, v0.oi, None, None)
 
+    @property
+    def globals_nosetup_all(self):
+        v0 = self.globals['vanilla-nosetup']
+        v1 = self.globals['fdlike-nosetup']
+        return PerfEntry(v0.time, v0.gflopss, v1.gpointss, v0.oi, None, None)
+
     @property
     def gflopss(self):
         return OrderedDict([(k, v.gflopss) for k, v in self.items()])
@@ -457,13 +513,13 @@ def create_profile(name):
         level = configuration['profiling']
     profiler = profiler_registry[level](name)
 
-    if profiler.initialized:
+    if profiler._attempted_init:
         return profiler
     else:
-        warning("Couldn't set up `%s` profiler; reverting to `advanced`" % level)
-        profiler = profiler_registry['basic'](name)
+        warning(f"Couldn't set up `{level}` profiler; reverting to 'advanced'")
+        profiler = profiler_registry['advanced'](name)
         # We expect the `advanced` profiler to always initialize successfully
-        assert profiler.initialized
+        assert profiler._attempted_init
         return profiler
 
 
@@ -477,55 +533,3 @@ profiler_registry = {
     'advisor': AdvisorProfiler
 }
 """Profiling levels."""
-
-
-def locate_intel_advisor():
-    """
-    Detect if Intel Advisor is installed on the machine and return
-    its location if it is.
-
-    """
-    path = None
-
-    try:
-        # Check if the directory to Intel Advisor is specified
-        path = Path(os.environ['DEVITO_ADVISOR_DIR'])
-    except KeyError:
-        # Otherwise, 'sniff' the location of Advisor's directory
-        error_msg = 'Intel Advisor cannot be found on your system, consider if you'\
-                    ' have sourced its environment variables correctly. Information can'\
-                    ' be found at https://software.intel.com/content/www/us/en/develop/'\
-                    'documentation/advisor-user-guide/top/launch-the-intel-advisor/'\
-                    'intel-advisor-cli/setting-and-using-intel-advisor-environment'\
-                    '-variables.html'
-        try:
-            res = run(["advixe-cl", "--version"], stdout=PIPE, stderr=DEVNULL)
-            ver = res.stdout.decode("utf-8")
-            if not ver:
-                error(error_msg)
-                return None
-        except (UnicodeDecodeError, FileNotFoundError):
-            error(error_msg)
-            return None
-
-        env_path = os.environ["PATH"]
-        env_path_dirs = env_path.split(":")
-
-        for env_path_dir in env_path_dirs:
-            # intel/advisor is the advisor directory for Intel Parallel Studio,
-            # intel/oneapi/advisor is the directory for Intel oneAPI
-            if "intel/advisor" in env_path_dir or "intel/oneapi/advisor" in env_path_dir:
-                path = Path(env_path_dir)
-                if path.name.startswith('bin'):
-                    path = path.parent
-
-        if not path:
-            error(error_msg)
-            return None
-
-    if path.joinpath('bin64').joinpath('advixe-cl').is_file():
-        return path
-    else:
-        warning("Requested `advisor` profiler, but couldn't locate executable"
-                "in advisor directory")
-        return None
