@@ -26,7 +26,8 @@ class OperatorRegistry(OrderedDict, metaclass=Singleton):
     """
 
     _modes = ('noop', 'advanced', 'advanced-fsg')
-    _languages = ('C', 'openmp', 'openacc', 'cuda')
+    _languages = ('C', 'CXX', 'openmp', 'Copenmp', 'CXXopenmp',
+                  'openacc', 'cuda', 'hip', 'sycl')
     _accepted = _modes + tuple(product(_modes, _languages))
 
     def add(self, operator, platform, mode, language='C'):
