@@ -27,7 +27,7 @@ class Parameters(OrderedDict, Signer):
     """
 
     def __init__(self, name=None, **kwargs):
-        super(Parameters, self).__init__(**kwargs)
+        super().__init__(**kwargs)
         self._name = name
 
         self._accepted = {}
@@ -48,16 +48,16 @@ class Parameters(OrderedDict, Signer):
             if accepted is not None:
                 tocheck = list(value) if isinstance(value, dict) else [value]
                 if any(i not in accepted for i in tocheck):
-                    raise ValueError("Illegal configuration parameter (%s, %s). "
-                                     "Accepted: %s" % (key, value, str(accepted)))
+                    raise ValueError(f"Illegal configuration parameter ({key}, {value}). "
+                                     f"Accepted: {str(accepted)}")
             return func(self, key, value)
         return wrapper
 
     def _check_key_deprecation(func):
         def wrapper(self, key, value=None):
             if key in self._deprecated:
-                warning("Trying to access deprecated config `%s`. Using `%s` instead"
-                        % (key, self._deprecated[key]))
+                warning(f"Trying to access deprecated config `{key}`. "
+                        f"Using `{self._deprecated[key]}` instead")
                 key = self._deprecated[key]
             return func(self, key, value)
         return wrapper
@@ -72,25 +72,18 @@ class Parameters(OrderedDict, Signer):
         else:
             return value
 
-    def _updated(self, key, value):
-        """
-        Execute the callback associated to ``key``, if any.
-        """
-        if key in self._update_functions:
-            retval = self._update_functions[key](value)
-            if retval is not None:
-                super(Parameters, self).__setitem__(key, retval)
-
     @_check_key_deprecation
     def __getitem__(self, key, *args):
-        return super(Parameters, self).__getitem__(key)
+        return super().__getitem__(key)
 
     @_check_key_deprecation
     @_check_key_value
     def __setitem__(self, key, value):
         value = self._preprocess(key, value)
-        super(Parameters, self).__setitem__(key, value)
-        self._updated(key, value)
+        if key in self._update_functions:
+            value = self._update_functions[key](value)
+        if value is not None:
+            super().__setitem__(key, value)
 
     @_check_key_deprecation
     @_check_key_value
@@ -100,7 +93,7 @@ class Parameters(OrderedDict, Signer):
         ``self[key] = value`` as both preprocessor and callback, if any,
         are bypassed.
         """
-        super(Parameters, self).__setitem__(key, value)
+        super().__setitem__(key, value)
 
     def add(self, key, value, accepted=None, preprocessor=None, callback=None,
             impacts_jit=True, deprecate=None):
@@ -116,7 +109,7 @@ class Parameters(OrderedDict, Signer):
         that the parameter doesn't affect code generation, so it can be excluded
         from the construction of the hash key.
         """
-        super(Parameters, self).__setitem__(key, value)
+        super().__setitem__(key, value)
         self._accepted[key] = accepted
         self._defaults[key] = value
         self._impact_jit[key] = impacts_jit
@@ -153,6 +146,8 @@ env_vars_mapper = {
     'DEVITO_DEVELOP': 'develop-mode',
     'DEVITO_OPT': 'opt',
     'DEVITO_MPI': 'mpi',
+    'DEVITO_TOPOLOGY': 'topology',
+    'DEVITO_DEVICEID': 'deviceid',
     'DEVITO_LANGUAGE': 'language',
     'DEVITO_AUTOTUNING': 'autotuning',
     'DEVITO_LOGGING': 'log-level',
@@ -187,11 +182,11 @@ def init_configuration(configuration=configuration, env_vars_mapper=env_vars_map
         mapper = dict(queue)
         for k, (v, msg) in env_vars_deprecated.items():
             if environ.get(k):
-                warning("`%s` is deprecated. %s" % (k, msg))
+                warning(f"`{k}` is deprecated. {msg}")
                 if environ.get(v):
-                    warning("Both `%s` and `%s` set. Ignoring `%s`" % (k, v, k))
+                    warning(f"Both `{k}` and `{v}` set. Ignoring `{k}`")
                 else:
-                    warning("Setting `%s=%s`" % (v, environ[k]))
+                    warning(f"Setting `{v}={environ[k]}`")
                     unprocessed[mapper[v]] = environ[k]
     else:
         # Attempt reading from the specified configuration file
@@ -229,31 +224,38 @@ def init_configuration(configuration=configuration, env_vars_mapper=env_vars_map
     configuration.initialize()
 
 
-class switchconfig(object):
+class switchconfig:
 
     """
-    Decorator to temporarily change `configuration` parameters.
+    Decorator or context manager to temporarily change `configuration` parameters.
     """
 
-    def __init__(self, **params):
-        self.params = {k.replace('_', '-'): v for k, v in params.items()}
+    def __init__(self, condition=True, **params):
+        if condition:
+            self.params = {k.replace('_', '-'): v for k, v in params.items()}
+        else:
+            self.params = {}
+        self.previous = {}
+
+    def __enter__(self, condition=True, **params):
+        self.previous = {}
+        for k, v in self.params.items():
+            self.previous[k] = configuration[k]
+            configuration[k] = v
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        for k, v in self.params.items():
+            try:
+                configuration[k] = self.previous[k]
+            except ValueError:
+                # E.g., `platform` and `compiler` will end up here
+                super(Parameters, configuration).__setitem__(k, self.previous[k])
 
     def __call__(self, func, *args, **kwargs):
         @wraps(func)
         def wrapper(*args, **kwargs):
-            previous = {}
-            for k, v in self.params.items():
-                previous[k] = configuration[k]
-                configuration[k] = v
-            try:
+            with self:
                 result = func(*args, **kwargs)
-            finally:
-                for k, v in self.params.items():
-                    try:
-                        configuration[k] = previous[k]
-                    except ValueError:
-                        # E.g., `platform` and `compiler` will end up here
-                        configuration[k] = previous[k].name
             return result
         return wrapper
 
@@ -262,11 +264,10 @@ def print_defaults():
     """Print the environment variables accepted by Devito, their default value,
     as well as all of the accepted values."""
     for k, v in env_vars_mapper.items():
-        info('%s: %s. Default: %s' % (k, configuration._accepted[v],
-                                      configuration._defaults[v]))
+        info(f'{k}: {configuration._accepted[v]}. Default: {configuration._defaults[v]}')
 
 
 def print_state():
     """Print the current configuration state."""
     for k, v in configuration.items():
-        info('%s: %s' % (k, v))
+        info(f'{k}: {v}')
