@@ -1,9 +1,11 @@
 from .utils import *  # noqa
 from .buffering import *  # noqa
-from .aliases import *  # noqa
 from .cse import *  # noqa
+from .aliases import *  # noqa
 from .factorization import *  # noqa
 from .blocking import *  # noqa
 from .asynchrony import *  # noqa
 from .implicit import *  # noqa
 from .misc import *  # noqa
+from .derivatives import *  # noqa
+from .unevaluate import *  # noqa
