@@ -1,4 +1,5 @@
 from functools import singledispatch
+from itertools import takewhile
 from abc import ABC
 
 import cgen as c
@@ -6,13 +7,13 @@ import cgen as c
 from devito.data import FULL
 from devito.ir import (DummyExpr, Call, Conditional, Expression, List, Prodder,
                        ParallelIteration, ParallelBlock, PointerCast, EntryFunction,
-                       AsyncCallable, FindNodes, FindSymbols)
+                       AsyncCallable, FindNodes, FindSymbols, IsPerfectIteration)
 from devito.mpi.distributed import MPICommObject
+from devito.passes import is_on_device
 from devito.passes.iet.engine import iet_pass
-from devito.passes.iet.misc import is_on_device
-from devito.symbolics import Byref, CondNe
-from devito.tools import as_list, prod
-from devito.types import Symbol, Wildcard
+from devito.symbolics import Byref, CondNe, SizeOf
+from devito.tools import as_list, is_integer, prod
+from devito.types import Symbol, QueueID, Wildcard
 
 __all__ = ['LangBB', 'LangTransformer']
 
@@ -30,21 +31,30 @@ class LangMeta(type):
             raise NotImplementedError("Missing required mapping for `%s`" % k)
         return self.mapper[k]
 
+    def get(self, k, v=None):
+        return self.mapper.get(k, v)
 
-class LangBB(object, metaclass=LangMeta):
+
+class LangBB(metaclass=LangMeta):
 
     """
     Abstract base class for Language Building Blocks.
     """
 
-    # NOTE: a subclass may want to override the values below, which represent
-    # IET node types used in the various lowering and/or transformation passes
+    # NOTE: a subclass may want to override the attributes below to customize
+    # code generation
+    BackendCall = Call
     Region = ParallelBlock
     HostIteration = ParallelIteration
     DeviceIteration = ParallelIteration
     Prodder = Prodder
     PointerCast = PointerCast
 
+    AsyncQueue = QueueID
+    """
+    An object used by the language to enqueue asynchronous operations.
+    """
+
     @classmethod
     def _get_num_devices(cls):
         """
@@ -144,7 +154,7 @@ class LangTransformer(ABC):
     an IET for a certain target language (e.g., C, C+OpenMP).
     """
 
-    lang = LangBB
+    langbb = LangBB
     """
     The constructs of the target language. To be specialized by a subclass.
     """
@@ -154,7 +164,8 @@ class LangTransformer(ABC):
         Parameters
         ----------
         key : callable, optional
-            Return True if an Iteration can and should be parallelized, False otherwise.
+            Return True if an Iteration can and should be parallelized,
+            False otherwise.
         sregistry : SymbolRegistry
             The symbol registry, to access the symbols appearing in an IET.
         platform : Platform
@@ -194,22 +205,185 @@ class LangTransformer(ABC):
 
     @property
     def Region(self):
-        return self.lang.Region
+        return self.langbb.Region
 
     @property
     def HostIteration(self):
-        return self.lang.HostIteration
+        return self.langbb.HostIteration
 
     @property
     def DeviceIteration(self):
-        return self.lang.DeviceIteration
+        return self.langbb.DeviceIteration
 
     @property
     def Prodder(self):
-        return self.lang.Prodder
+        return self.langbb.Prodder
 
 
-class DeviceAwareMixin(object):
+class ShmTransformer(LangTransformer):
+
+    """
+    Abstract base class for LangTransformers that want to emit
+    shared-memory-parallel IETs for CPUs.
+    """
+
+    def __init__(self, key, sregistry, options, platform, compiler):
+        """
+        Parameters
+        ----------
+        key : callable, optional
+            Return True if an Iteration can and should be parallelized,
+            False otherwise.
+        sregistry : SymbolRegistry
+            The symbol registry, to access the symbols appearing in an IET.
+        options : dict
+             The optimization options.
+             Accepted: ['par-collapse-ncores', 'par-collapse-work',
+             'par-chunk-nonaffine', 'par-dynamic-work', 'par-nested']
+             * 'par-collapse-ncores': use a collapse clause if the number of
+               available physical cores is greater than this threshold.
+             * 'par-collapse-work': use a collapse clause if the trip count of the
+               collapsable Iterations is statically known to exceed this threshold.
+             * 'par-chunk-nonaffine': coefficient to adjust the chunk size in
+               non-affine parallel Iterations.
+             * 'par-dynamic-work': use dynamic scheduling if the operation count per
+               iteration exceeds this threshold. Otherwise, use static scheduling.
+             * 'par-nested': nested parallelism if the number of hyperthreads
+               per core is greater than this threshold.
+        platform : Platform
+            The underlying platform.
+        compiler : Compiler
+            The underlying JIT compiler.
+        """
+        super().__init__(key, sregistry, platform, compiler)
+
+        self.collapse_ncores = options['par-collapse-ncores']
+        self.collapse_work = options['par-collapse-work']
+        self.chunk_nonaffine = options['par-chunk-nonaffine']
+        self.dynamic_work = options['par-dynamic-work']
+        self.nested = options['par-nested']
+
+    @property
+    def ncores(self):
+        return self.platform.cores_physical
+
+    @property
+    def nhyperthreads(self):
+        return self.platform.threads_per_core
+
+    @property
+    def nthreads(self):
+        return self.sregistry.nthreads
+
+    @property
+    def nthreads_nested(self):
+        return self.sregistry.nthreads_nested
+
+    @property
+    def nthreads_nonaffine(self):
+        return self.sregistry.nthreads_nonaffine
+
+    @property
+    def threadid(self):
+        return self.sregistry.threadid
+
+    def _score_candidate(self, n0, root, collapsable=()):
+        """
+        The score of a collapsable nest depends on the number of fully-parallel
+        Iterations and their position in the nest (the outer, the better).
+        """
+        nest = [root] + list(collapsable)
+        n = len(nest)
+
+        # Number of fully-parallel collapsable Iterations
+        key = lambda i: i.is_ParallelNoAtomic
+        fp_iters = list(takewhile(key, nest))
+        n_fp_iters = len(fp_iters)
+
+        # Number of parallel-if-atomic collapsable Iterations
+        key = lambda i: i.is_ParallelAtomic
+        pia_iters = list(takewhile(key, nest))
+        n_pia_iters = len(pia_iters)
+
+        # Prioritize the Dimensions that are more likely to define larger
+        # iteration spaces
+        key = lambda d: (not d.is_Derived or
+                         (d.is_Custom and not is_integer(d.symbolic_size)) or
+                         (d.is_Block and d._depth == 1))
+
+        fpdims = [i.dim for i in fp_iters]
+        n_fp_iters_large = len([d for d in fpdims if key(d)])
+
+        piadims = [i.dim for i in pia_iters]
+        n_pia_iters_large = len([d for d in piadims if key(d)])
+
+        return (
+            int(n_fp_iters == n),  # Fully-parallel nest
+            n_fp_iters_large,
+            n_fp_iters,
+            n_pia_iters_large,
+            n_pia_iters,
+            -(n0 + 1),  # The outer, the better
+            n,
+        )
+
+    def _select_candidates(self, candidates):
+        assert candidates
+
+        if self.ncores < self.collapse_ncores:
+            return candidates[0], []
+
+        mapper = {}
+        for n0, root in enumerate(candidates):
+
+            # Score `root` in isolation
+            mapper[(root, ())] = self._score_candidate(n0, root)
+
+            collapsable = []
+            for n, i in enumerate(candidates[n0+1:], n0+1):
+                # The Iteration nest [root, ..., i] must be perfect
+                if not IsPerfectIteration(depth=i).visit(root):
+                    break
+
+                # Loops are collapsable only if none of the iteration variables
+                # appear in initializer expressions. For example, the following
+                # two loops cannot be collapsed
+                #
+                # for (i = ... )
+                #   for (j = i ...)
+                #     ...
+                #
+                # Here, we make sure this won't happen
+                if any(j.dim in i.symbolic_min.free_symbols for j in candidates[n0:n]):
+                    break
+
+                # Can't collapse SIMD-vectorized Iterations
+                if i.is_Vectorized:
+                    break
+
+                # Would there be enough work per parallel iteration?
+                nested = candidates[n+1:]
+                if nested:
+                    try:
+                        work = prod([int(j.dim.symbolic_size) for j in nested])
+                        if work < self.collapse_work:
+                            break
+                    except TypeError:
+                        pass
+
+                collapsable.append(i)
+
+                # Score `root + collapsable`
+                v = tuple(collapsable)
+                mapper[(root, v)] = self._score_candidate(n0, root, v)
+
+        # Retrieve the candidates with highest score
+        root, collapsable = max(mapper, key=mapper.get)
+
+        return root, list(collapsable)
+
+
+class DeviceAwareMixin:
 
     @property
     def deviceid(self):
@@ -233,6 +407,49 @@ class DeviceAwareMixin(object):
         is sufficient reuse to implement the logic as a single method.
         """
 
+        def _extract_objcomm(iet):
+            for i in iet.parameters:
+                if isinstance(i, MPICommObject):
+                    return i
+
+            # Fallback -- might end up here because the Operator has no
+            # halo exchanges, but we now need it nonetheless to perform
+            # the rank-GPU assignment
+            if options['mpi']:
+                for i in iet.parameters:
+                    try:
+                        return i.grid.distributor._obj_comm
+                    except AttributeError:
+                        pass
+
+        def _make_setdevice_seq(iet, nodes=()):
+            devicetype = as_list(self.langbb[self.platform])
+            deviceid = self.deviceid
+
+            return list(nodes) + [Conditional(
+                CondNe(deviceid, -1),
+                self.langbb['set-device']([deviceid] + devicetype)
+            )]
+
+        def _make_setdevice_mpi(iet, objcomm, nodes=()):
+            devicetype = as_list(self.langbb[self.platform])
+            deviceid = self.deviceid
+
+            rank = Symbol(name='rank')
+            rank_decl = DummyExpr(rank, 0)
+            rank_init = Call('MPI_Comm_rank', [objcomm, Byref(rank)])
+
+            ngpus, call_ngpus = self.langbb._get_num_devices(self.platform)
+
+            osdd_then = self.langbb['set-device']([deviceid] + devicetype)
+            osdd_else = self.langbb['set-device']([rank % ngpus] + devicetype)
+
+            return list(nodes) + [Conditional(
+                CondNe(deviceid, -1),
+                osdd_then,
+                List(body=[rank_decl, rank_init, call_ngpus, osdd_else]),
+            )]
+
         @singledispatch
         def _initialize(iet):
             return iet, {}
@@ -241,63 +458,26 @@ class DeviceAwareMixin(object):
         def _(iet):
             assert iet.body.is_CallableBody
 
-            # TODO: we need to pick the rank from `comm_shm`, not `comm`,
-            # so that we have nranks == ngpus (as long as the user has launched
-            # the right number of MPI processes per node given the available
-            # number of GPUs per node)
-
-            objcomm = None
-            for i in iet.parameters:
-                if isinstance(i, MPICommObject):
-                    objcomm = i
-                    break
-            if objcomm is None and options['mpi']:
-                # Time to inject `objcomm`. If it's not here, it simply means
-                # there's no halo exchanges in the Operator, but we now need it
-                # nonetheless to perform the rank-GPU assignment
-                for i in iet.parameters:
-                    try:
-                        objcomm = i.grid.distributor._obj_comm
-                        break
-                    except AttributeError:
-                        pass
-                assert objcomm is not None
-
-            devicetype = as_list(self.lang[self.platform])
-            deviceid = self.deviceid
+            devicetype = as_list(self.langbb[self.platform])
 
             try:
-                lang_init = [self.lang['init'](devicetype)]
+                lang_init = [self.langbb['init'](devicetype)]
             except TypeError:
                 # Not all target languages need to be explicitly initialized
                 lang_init = []
 
+            objcomm = _extract_objcomm(iet)
+
             if objcomm is not None:
-                rank = Symbol(name='rank')
-                rank_decl = DummyExpr(rank, 0)
-                rank_init = Call('MPI_Comm_rank', [objcomm, Byref(rank)])
+                body = _make_setdevice_mpi(iet, objcomm, nodes=lang_init)
 
-                ngpus, call_ngpus = self.lang._get_num_devices(self.platform)
-
-                osdd_then = self.lang['set-device']([deviceid] + devicetype)
-                osdd_else = self.lang['set-device']([rank % ngpus] + devicetype)
-
-                body = lang_init + [Conditional(
-                    CondNe(deviceid, -1),
-                    osdd_then,
-                    List(body=[rank_decl, rank_init, call_ngpus, osdd_else]),
-                )]
-
-                header = c.Comment('Begin of %s+MPI setup' % self.lang['name'])
-                footer = c.Comment('End of %s+MPI setup' % self.lang['name'])
+                header = c.Comment('Beginning of %s+MPI setup' % self.langbb['name'])
+                footer = c.Comment('End of %s+MPI setup' % self.langbb['name'])
             else:
-                body = lang_init + [Conditional(
-                    CondNe(deviceid, -1),
-                    self.lang['set-device']([deviceid] + devicetype)
-                )]
+                body = _make_setdevice_seq(iet, nodes=lang_init)
 
-                header = c.Comment('Begin of %s setup' % self.lang['name'])
-                footer = c.Comment('End of %s setup' % self.lang['name'])
+                header = c.Comment('Beginning of %s setup' % self.langbb['name'])
+                footer = c.Comment('End of %s setup' % self.langbb['name'])
 
             init = List(header=header, body=body, footer=footer)
             iet = iet._rebuild(body=iet.body._rebuild(init=init))
@@ -306,19 +486,23 @@ class DeviceAwareMixin(object):
 
         @_initialize.register(AsyncCallable)
         def _(iet):
-            devicetype = as_list(self.lang[self.platform])
-            deviceid = self.deviceid
+            objcomm = _extract_objcomm(iet)
+            if objcomm is not None:
+                init = _make_setdevice_mpi(iet, objcomm)
+            else:
+                init = _make_setdevice_seq(iet)
 
-            init = Conditional(
-                CondNe(deviceid, -1),
-                self.lang['set-device']([deviceid] + devicetype)
-            )
             iet = iet._rebuild(body=iet.body._rebuild(init=init))
 
             return iet, {}
 
         return _initialize(iet)
 
+    def _device_pointers(self, iet):
+        functions = FindSymbols().visit(iet)
+        devfuncs = [f for f in functions if f.is_Array and f._mem_local]
+        return set(devfuncs)
+
     def _is_offloadable(self, iet):
         """
         True if the IET computation is offloadable to device, False otherwise.
@@ -333,6 +517,23 @@ class DeviceAwareMixin(object):
         return not (buffers and hostfuncs)
 
 
+class Sections(tuple):
+
+    def __new__(cls, function, *args):
+        obj = super().__new__(cls, args)
+        obj.function = function
+
+        return obj
+
+    @property
+    def size(self):
+        return prod(s for _, s in self)
+
+    @property
+    def nbytes(self):
+        return self.size*SizeOf(self.function.indexed._C_typedata)
+
+
 def make_sections_from_imask(f, imask=None):
     if imask is None:
         imask = [FULL]*f.ndim
@@ -366,7 +567,7 @@ def make_sections_from_imask(f, imask=None):
         size *= remainder_size
         sections.append((start, size))
 
-    return sections
+    return Sections(f, *sections)
 
 
 @singledispatch
