@@ -1,25 +1,29 @@
+from functools import cached_property
 from packaging.version import Version
 
 import cgen as c
 from sympy import And, Ne, Not
 
-from devito.arch import AMDGPUX, NVIDIAX, INTELGPUX
-from devito.arch.compiler import GNUCompiler
-from devito.ir import (Call, Conditional, List, Prodder, ParallelIteration,
+from devito.arch import AMDGPUX, NVIDIAX, INTELGPUX, PVC
+from devito.arch.compiler import GNUCompiler, NvidiaCompiler
+from devito.ir import (Call, Conditional, DeviceCall, List, Pragma, Prodder,
                        ParallelBlock, PointerCast, While, FindSymbols)
-from devito.passes.iet.definitions import DataManager
+from devito.passes.iet.definitions import DataManager, DeviceAwareDataManager
+from devito.passes.iet.langbase import LangBB
 from devito.passes.iet.orchestration import Orchestrator
 from devito.passes.iet.parpragma import (PragmaSimdTransformer, PragmaShmTransformer,
                                          PragmaDeviceAwareTransformer, PragmaLangBB,
-                                         PragmaTransfer, PragmaDeviceAwareDataManager)
+                                         PragmaIteration, PragmaTransfer)
+from devito.passes.iet.languages.utils import joins
 from devito.passes.iet.languages.C import CBB
-from devito.passes.iet.languages.utils import make_clause_reduction
+from devito.passes.iet.languages.CXX import CXXBB
 from devito.symbolics import CondEq, DefFunction
 from devito.tools import filter_ordered
 
 __all__ = ['SimdOmpizer', 'Ompizer', 'OmpIteration', 'OmpRegion',
            'DeviceOmpizer', 'DeviceOmpIteration', 'DeviceOmpDataManager',
-           'OmpDataManager', 'OmpOrchestrator']
+           'OmpDataManager', 'OmpOrchestrator', 'DeviceOmpOrchestrator',
+           'CXXOmpDataManager', 'CXXOmpOrchestrator']
 
 
 class OmpRegion(ParallelBlock):
@@ -30,7 +34,7 @@ class OmpRegion(ParallelBlock):
         return c.Pragma('omp parallel num_threads(%s) %s' % (nthreads.name, private))
 
 
-class OmpIteration(ParallelIteration):
+class OmpIteration(PragmaIteration):
 
     @classmethod
     def _make_construct(cls, parallel=False, **kwargs):
@@ -40,11 +44,12 @@ class OmpIteration(ParallelIteration):
             return 'omp for'
 
     @classmethod
-    def _make_clauses(cls, ncollapse=None, chunk_size=None, nthreads=None,
+    def _make_clauses(cls, ncollapsed=0, chunk_size=None, nthreads=None,
                       reduction=None, schedule=None, **kwargs):
         clauses = []
 
-        clauses.append('collapse(%d)' % (ncollapse or 1))
+        if ncollapsed > 1:
+            clauses.append('collapse(%d)' % ncollapsed)
 
         if chunk_size is not False:
             clauses.append('schedule(%s,%s)' % (schedule or 'dynamic',
@@ -54,21 +59,10 @@ class OmpIteration(ParallelIteration):
             clauses.append('num_threads(%s)' % nthreads)
 
         if reduction:
-            clauses.append(make_clause_reduction(reduction))
+            clauses.append(cls._make_clause_reduction_from_imask(reduction))
 
         return clauses
 
-    @classmethod
-    def _process_kwargs(cls, **kwargs):
-        kwargs = super()._process_kwargs(**kwargs)
-
-        kwargs.pop('schedule', None)
-        kwargs.pop('parallel', False)
-        kwargs.pop('chunk_size', None)
-        kwargs.pop('nthreads', None)
-
-        return kwargs
-
 
 class DeviceOmpIteration(OmpIteration):
 
@@ -88,14 +82,6 @@ class DeviceOmpIteration(OmpIteration):
 
         return clauses
 
-    @classmethod
-    def _process_kwargs(cls, **kwargs):
-        kwargs = super()._process_kwargs(**kwargs)
-
-        kwargs.pop('gpu_fit', None)
-
-        return kwargs
-
 
 class ThreadedProdder(Conditional, Prodder):
 
@@ -103,12 +89,12 @@ class ThreadedProdder(Conditional, Prodder):
 
     def __init__(self, prodder, arguments=None):
         # Atomic-ize any single-thread Prodders in the parallel tree
-        condition = CondEq(DefFunction(Ompizer.lang['thread-num']().name), 0)
+        condition = CondEq(DefFunction(Ompizer.langbb['thread-num']().name), 0)
 
         # Prod within a while loop until all communications have completed
         # In other words, the thread delegated to prodding is entrapped for as long
         # as it's required
-        prod_until = Not(DefFunction(prodder.name, [i.name for i in prodder.arguments]))
+        prod_until = Not(DefFunction(prodder.name, prodder.arguments))
         then_body = List(header=c.Comment('Entrap thread until comms have completed'),
                          body=While(prod_until))
         Conditional.__init__(self, condition, then_body)
@@ -117,7 +103,17 @@ class ThreadedProdder(Conditional, Prodder):
         Prodder.__init__(self, prodder.name, arguments, periodic=prodder.periodic)
 
 
-class OmpBB(PragmaLangBB):
+class SimdForAligned(Pragma):
+
+    @cached_property
+    def _generate(self):
+        assert len(self.arguments) > 1
+        n = self.arguments[0]
+        items = self.arguments[1:]
+        return self.pragma % (joins(*items), n)
+
+
+class AbstractOmpBB(LangBB):
 
     mapper = {
         # Misc
@@ -127,36 +123,69 @@ class OmpBB(PragmaLangBB):
         AMDGPUX: None,
         NVIDIAX: None,
         INTELGPUX: None,
+        PVC: None,
         # Runtime library
         'init': None,
         'thread-num': lambda retobj=None:
             Call('omp_get_thread_num', retobj=retobj),
+        # Pragmas
+        'simd-for':
+            Pragma('omp simd'),
+        'simd-for-aligned': lambda n, *a:
+            SimdForAligned('omp simd aligned(%s:%d)', arguments=(n, *a)),
+        'atomic':
+            Pragma('omp atomic update')
+    }
+
+    Region = OmpRegion
+    HostIteration = OmpIteration
+    DeviceIteration = DeviceOmpIteration
+    Prodder = ThreadedProdder
+
+
+class OmpBB(AbstractOmpBB):
+    mapper = {**AbstractOmpBB.mapper, **CBB.mapper}
+
+
+class CXXOmpBB(AbstractOmpBB):
+    mapper = {**AbstractOmpBB.mapper, **CXXBB.mapper}
+
+
+class DeviceOmpBB(OmpBB, PragmaLangBB):
+
+    BackendCall = DeviceCall
+
+    mapper = dict(OmpBB.mapper)
+    mapper.update({
+        # Runtime library
         'num-devices': lambda args, retobj:
             Call('omp_get_num_devices', args, retobj=retobj),
         'set-device': lambda args:
             Call('omp_set_default_device', args),
         # Pragmas
-        'simd-for': c.Pragma('omp simd'),
-        'simd-for-aligned': lambda i, j: c.Pragma('omp simd aligned(%s:%d)' % (i, j)),
-        'atomic': c.Pragma('omp atomic update'),
-        'map-enter-to': lambda i, j:
-            c.Pragma('omp target enter data map(to: %s%s)' % (i, j)),
-        'map-enter-alloc': lambda i, j:
-            c.Pragma('omp target enter data map(alloc: %s%s)' % (i, j)),
-        'map-update': lambda i, j:
-            c.Pragma('omp target update from(%s%s)' % (i, j)),
-        'map-update-host': lambda i, j:
-            c.Pragma('omp target update from(%s%s)' % (i, j)),
-        'map-update-device': lambda i, j:
-            c.Pragma('omp target update to(%s%s)' % (i, j)),
-        'map-release': lambda i, j:
-            c.Pragma('omp target exit data map(release: %s%s)' % (i, j)),
-        'map-release-if': lambda i, j, k:
-            c.Pragma('omp target exit data map(release: %s%s) if(%s)' % (i, j, k)),
-        'map-exit-delete': lambda i, j:
-            c.Pragma('omp target exit data map(delete: %s%s)' % (i, j)),
-        'map-exit-delete-if': lambda i, j, k:
-            c.Pragma('omp target exit data map(delete: %s%s) if(%s)' % (i, j, k)),
+        'map-enter-to': lambda f, imask:
+            PragmaTransfer('omp target enter data map(to: %s%s)', f, imask=imask),
+        'map-enter-alloc': lambda f, imask:
+            PragmaTransfer('omp target enter data map(alloc: %s%s)',
+                           f, imask=imask),
+        'map-update': lambda f, imask:
+            PragmaTransfer('omp target update from(%s%s)', f, imask=imask),
+        'map-update-host': lambda f, imask:
+            PragmaTransfer('omp target update from(%s%s)', f, imask=imask),
+        'map-update-device': lambda f, imask:
+            PragmaTransfer('omp target update to(%s%s)', f, imask=imask),
+        'map-release': lambda f, imask:
+            PragmaTransfer('omp target exit data map(release: %s%s)',
+                           f, imask=imask),
+        'map-release-if': lambda f, imask, a:
+            PragmaTransfer('omp target exit data map(release: %s%s) if(%s)',
+                           f, imask=imask, arguments=a),
+        'map-exit-delete': lambda f, imask:
+            PragmaTransfer('omp target exit data map(delete: %s%s)',
+                           f, imask=imask),
+        'map-exit-delete-if': lambda f, imask, a:
+            PragmaTransfer('omp target exit data map(delete: %s%s) if(%s)',
+                           f, imask=imask, arguments=a),
         'memcpy-to-device': lambda i, j, k:
             Call('omp_target_memcpy', [i, j, k, 0, 0,
                                        DefFunction('omp_get_device_num'),
@@ -171,16 +200,7 @@ class OmpBB(PragmaLangBB):
             Call('omp_target_alloc', (i, j), retobj=retobj, cast=True),
         'device-free': lambda i, j:
             Call('omp_target_free', (i, j))
-    }
-    mapper.update(CBB.mapper)
-
-    Region = OmpRegion
-    HostIteration = OmpIteration
-    DeviceIteration = DeviceOmpIteration
-    Prodder = ThreadedProdder
-
-
-class DeviceOmpBB(OmpBB):
+    })
 
     # NOTE: Work around clang>=10 issue concerning offloading arrays declared
     # with an `__attribute__(aligned(...))` qualifier
@@ -195,16 +215,18 @@ class DeviceOmpBB(OmpBB):
         if devicerm is not None:
             items.append(devicerm)
         argument = And(*items)
-        return PragmaTransfer(cls.mapper['map-exit-delete-if'], f, imask, argument)
+        return cls.mapper['map-exit-delete-if'](f, imask, argument)
 
 
 class SimdOmpizer(PragmaSimdTransformer):
-    lang = OmpBB
+    langbb = OmpBB
 
 
-class Ompizer(PragmaShmTransformer):
+class CXXSimdOmpizer(PragmaSimdTransformer):
+    langbb = CXXOmpBB
 
-    lang = OmpBB
+
+class AbstractOmpizer(PragmaShmTransformer):
 
     @classmethod
     def _support_array_reduction(cls, compiler):
@@ -213,20 +235,44 @@ class Ompizer(PragmaShmTransformer):
         if isinstance(compiler, GNUCompiler) and \
            compiler.version < Version("6.0"):
             return False
-        return True
+        elif isinstance(compiler, NvidiaCompiler):
+            # NVC++ does not support array reduction and leads to segfault
+            return False
+        else:
+            return True
+
+
+class Ompizer(AbstractOmpizer):
+    langbb = OmpBB
+
+
+class CXXOmpizer(AbstractOmpizer):
+    langbb = CXXOmpBB
 
 
 class DeviceOmpizer(PragmaDeviceAwareTransformer):
-    lang = DeviceOmpBB
+    langbb = DeviceOmpBB
 
 
 class OmpDataManager(DataManager):
-    lang = OmpBB
+    langbb = OmpBB
 
 
-class DeviceOmpDataManager(PragmaDeviceAwareDataManager):
-    lang = DeviceOmpBB
+class CXXOmpDataManager(DataManager):
+    langbb = CXXOmpBB
+
+
+class DeviceOmpDataManager(DeviceAwareDataManager):
+    langbb = DeviceOmpBB
 
 
 class OmpOrchestrator(Orchestrator):
-    lang = DeviceOmpBB
+    langbb = OmpBB
+
+
+class CXXOmpOrchestrator(Orchestrator):
+    langbb = CXXOmpBB
+
+
+class DeviceOmpOrchestrator(Orchestrator):
+    langbb = DeviceOmpBB
