@@ -1,24 +1,25 @@
+from collections import defaultdict
+from functools import cached_property
+
 import numpy as np
 import cgen as c
-from cached_property import cached_property
 from sympy import And, Max, true
 
 from devito.data import FULL
-from devito.ir import (Conditional, DummyEq, Dereference, Expression, ExpressionBundle,
-                       FindSymbols, FindNodes, ParallelTree, Pragma, Prodder, Transfer,
-                       List, Transformer, IsPerfectIteration, OpInc, filter_iterations,
-                       retrieve_iteration_tree, VECTORIZED)
-from devito.passes.iet.definitions import DeviceAwareDataManager
+from devito.ir import (Conditional, DummyEq, Dereference, Expression,
+                       ExpressionBundle, FindSymbols, FindNodes, ParallelIteration,
+                       ParallelTree, Pragma, Prodder, Transfer, List, Transformer,
+                       IsPerfectIteration, OpInc, filter_iterations, ccode,
+                       retrieve_iteration_tree, IMask, VECTORIZED)
 from devito.passes.iet.engine import iet_pass
 from devito.passes.iet.langbase import (LangBB, LangTransformer, DeviceAwareMixin,
-                                        make_sections_from_imask)
-from devito.symbolics import INT, ccode
-from devito.tools import as_tuple, flatten, prod
+                                        ShmTransformer, make_sections_from_imask)
+from devito.symbolics import INT
+from devito.tools import as_tuple, flatten, is_integer, prod
 from devito.types import Symbol
 
 __all__ = ['PragmaSimdTransformer', 'PragmaShmTransformer',
-           'PragmaDeviceAwareTransformer', 'PragmaLangBB', 'PragmaTransfer',
-           'PragmaDeviceAwareDataManager']
+           'PragmaDeviceAwareTransformer', 'PragmaLangBB', 'PragmaTransfer']
 
 
 class PragmaTransformer(LangTransformer):
@@ -42,11 +43,24 @@ class PragmaSimdTransformer(PragmaTransformer):
         return True
 
     @property
-    def simd_reg_size(self):
-        return self.platform.simd_reg_size
+    def simd_reg_nbytes(self):
+        return self.platform.simd_reg_nbytes
 
-    @iet_pass
-    def make_simd(self, iet):
+    def _make_simd_pragma(self, iet):
+        indexeds = FindSymbols('indexeds').visit(iet)
+        aligned = {i.base for i in indexeds if i.function.is_DiscreteFunction}
+        if aligned:
+            simd = self.langbb['simd-for-aligned']
+            simd = as_tuple(simd(self.simd_reg_nbytes, *aligned))
+        else:
+            simd = as_tuple(self.langbb['simd-for'])
+
+        return simd
+
+    def _make_simd(self, iet):
+        """
+        Carry out the bulk of `make_simd`.
+        """
         mapper = {}
         for tree in retrieve_iteration_tree(iet):
             candidates = [i for i in tree if i.is_ParallelRelaxed]
@@ -101,13 +115,7 @@ class PragmaSimdTransformer(PragmaTransformer):
                     continue
 
             # Add SIMD pragma
-            indexeds = FindSymbols('indexeds').visit(candidate)
-            aligned = {i.name for i in indexeds if i.function.is_DiscreteFunction}
-            if aligned:
-                simd = self.lang['simd-for-aligned']
-                simd = as_tuple(simd(','.join(sorted(aligned)), self.simd_reg_size))
-            else:
-                simd = as_tuple(self.lang['simd-for'])
+            simd = self._make_simd_pragma(candidate)
             pragmas = candidate.pragmas + simd
 
             # Add VECTORIZED property
@@ -119,147 +127,119 @@ class PragmaSimdTransformer(PragmaTransformer):
 
         return iet, {}
 
+    @iet_pass
+    def make_simd(self, iet):
+        return self._make_simd(iet)
 
-class PragmaShmTransformer(PragmaSimdTransformer):
+
+class PragmaIteration(ParallelIteration):
+
+    def __init__(self, *args, parallel=None, schedule=None, chunk_size=None,
+                 nthreads=None, ncollapsed=None, reduction=None, tile=None,
+                 gpu_fit=None, **kwargs):
+
+        construct = self._make_construct(
+            parallel=parallel, ncollapsed=ncollapsed, tile=tile
+        )
+        clauses = self._make_clauses(
+            ncollapsed=ncollapsed, chunk_size=chunk_size, nthreads=nthreads,
+            reduction=reduction, schedule=schedule, tile=tile, gpu_fit=gpu_fit,
+            **kwargs
+        )
+        kwargs['pragmas'] = Pragma(' '.join([construct] + clauses))
+
+        super().__init__(*args, **kwargs)
+
+        self.parallel = parallel
+        self.schedule = schedule
+        self.chunk_size = chunk_size
+        self.nthreads = nthreads
+        self.ncollapsed = ncollapsed
+        self.reduction = reduction
+        self.tile = tile
+        self.gpu_fit = gpu_fit
+
+    @classmethod
+    def _make_construct(cls, **kwargs):
+        # To be overridden by subclasses
+        raise NotImplementedError
+
+    @classmethod
+    def _make_clauses(cls, **kwargs):
+        return []
+
+    @classmethod
+    def _make_clause_reduction_from_imask(cls, reductions):
+        """
+        Build a string representing of a reduction clause given a list of
+        2-tuples `(symbol, ir.Operation)`.
+        """
+        mapper = defaultdict(list)
+        for i, imask, r in reductions:
+            if i.is_Indexed:
+                f = i.function
+                bounds = []
+                for k, d in zip(imask, f.dimensions):
+                    if is_integer(k):
+                        bounds.append('[%s]' % k)
+                    elif k is FULL:
+                        # Lower FULL Dimensions into a range spanning the entire
+                        # Dimension space, e.g. `reduction(+:f[0:f_vec->size[1]])`
+                        bounds.append('[0:%s]' % f._C_get_field(FULL, d).size)
+                    else:
+                        assert isinstance(k, tuple) and len(k) == 2
+                        bounds.append('[%s:%s]' % k)
+                mapper[r.name].append('%s%s' % (i.name, ''.join(bounds)))
+            else:
+                mapper[r.name].append(str(i))
+
+        args = ['reduction(%s:%s)' % (k, ','.join(v)) for k, v in mapper.items()]
+
+        return ' '.join(args)
+
+    @cached_property
+    def collapsed(self):
+        ret = [self]
+        for i in range(self.ncollapsed - 1):
+            ret.append(ret[i].nodes[0])
+        assert all(i.is_Iteration for i in ret)
+        return tuple(ret)
+
+
+class PragmaShmTransformer(ShmTransformer, PragmaSimdTransformer):
 
     """
-    Abstract base class for PragmaTransformers capable of emitting SIMD-parallel
-    and shared-memory-parallel IETs.
+    PragmaTransformer capable of emitting SIMD-parallel and shared-memory-parallel
+    IETs for CPUs.
     """
 
     def __init__(self, sregistry, options, platform, compiler):
-        """
-        Parameters
-        ----------
-        sregistry : SymbolRegistry
-            The symbol registry, to access the symbols appearing in an IET.
-        options : dict
-             The optimization options. Accepted: ['par-collapse-ncores',
-             'par-collapse-work', 'par-chunk-nonaffine', 'par-dynamic-work', 'par-nested']
-             * 'par-collapse-ncores': use a collapse clause if the number of
-               available physical cores is greater than this threshold.
-             * 'par-collapse-work': use a collapse clause if the trip count of the
-               collapsable Iterations is statically known to exceed this threshold.
-             * 'par-chunk-nonaffine': coefficient to adjust the chunk size in
-               non-affine parallel Iterations.
-             * 'par-dynamic-work': use dynamic scheduling if the operation count per
-               iteration exceeds this threshold. Otherwise, use static scheduling.
-             * 'par-nested': nested parallelism if the number of hyperthreads per core
-               is greater than this threshold.
-        platform : Platform
-            The underlying platform.
-        compiler : Compiler
-            The underlying JIT compiler.
-        """
         key = lambda i: i.is_ParallelRelaxed and not i.is_Vectorized
-        super().__init__(key, sregistry, platform, compiler)
-
-        self.collapse_ncores = options['par-collapse-ncores']
-        self.collapse_work = options['par-collapse-work']
-        self.chunk_nonaffine = options['par-chunk-nonaffine']
-        self.dynamic_work = options['par-dynamic-work']
-        self.nested = options['par-nested']
-
-    @property
-    def ncores(self):
-        return self.platform.cores_physical
-
-    @property
-    def nhyperthreads(self):
-        return self.platform.threads_per_core
-
-    @property
-    def nthreads(self):
-        return self.sregistry.nthreads
-
-    @property
-    def nthreads_nested(self):
-        return self.sregistry.nthreads_nested
-
-    @property
-    def nthreads_nonaffine(self):
-        return self.sregistry.nthreads_nonaffine
-
-    @property
-    def threadid(self):
-        return self.sregistry.threadid
-
-    def _select_candidates(self, candidates):
-        assert candidates
-
-        if self.ncores < self.collapse_ncores:
-            return candidates[0], []
-
-        mapper = {}
-        for n0, root in enumerate(candidates):
-
-            collapsable = []
-            for n, i in enumerate(candidates[n0+1:], n0+1):
-                # The Iteration nest [root, ..., i] must be perfect
-                if not IsPerfectIteration(depth=i).visit(root):
-                    break
-
-                # Loops are collapsable only if none of the iteration variables appear
-                # in initializer expressions. For example, the following two loops
-                # cannot be collapsed
-                #
-                # for (i = ... )
-                #   for (j = i ...)
-                #     ...
-                #
-                # Here, we make sure this won't happen
-                if any(j.dim in i.symbolic_min.free_symbols for j in candidates[n0:n]):
-                    break
-
-                # Also, we do not want to collapse SIMD-vectorized Iterations
-                if i.is_Vectorized:
-                    break
-
-                # Would there be enough work per parallel iteration?
-                nested = candidates[n+1:]
-                if nested:
-                    try:
-                        work = prod([int(j.dim.symbolic_size) for j in nested])
-                        if work < self.collapse_work:
-                            break
-                    except TypeError:
-                        pass
-
-                collapsable.append(i)
-
-            # Give a score to this candidate, based on the number of fully-parallel
-            # Iterations and their position (i.e. outermost to innermost) in the nest
-            score = (
-                int(root.is_ParallelNoAtomic),
-                int(len([i for i in collapsable if i.is_ParallelNoAtomic]) >= 1),
-                int(len([i for i in collapsable if i.is_ParallelRelaxed]) >= 1),
-                -(n0 + 1)  # The outermost, the better
-            )
-
-            mapper[(root, tuple(collapsable))] = score
-
-        # Retrieve the candidates with highest score
-        root, collapsable = max(mapper, key=mapper.get)
-
-        return root, list(collapsable)
+        super().__init__(key, sregistry, options, platform, compiler)
 
     def _make_reductions(self, partree):
         if not any(i.is_ParallelAtomic for i in partree.collapsed):
             return partree
 
         exprs = [i for i in FindNodes(Expression).visit(partree) if i.is_reduction]
-        reductions = [(i.output, i.operation) for i in exprs]
 
-        test0 = all(not i.is_Indexed for i, _ in reductions)
+        reductions = []
+        for e in exprs:
+            f = e.write
+            items = [i if i.is_Number else FULL for i in e.output.indices]
+            imask = IMask(*items, getters=f.dimensions)
+            reductions.append((e.output, imask, e.operation))
+
+        test0 = all(not i.is_Indexed for i, _, _ in reductions)
         test1 = (self._support_array_reduction(self.compiler) and
                  all(i.is_Affine for i in partree.collapsed))
 
         if test0 or test1:
             # Implement reduction
             mapper = {partree.root: partree.root._rebuild(reduction=reductions)}
-        elif all(i is OpInc for _, i in reductions):
+        elif all(i is OpInc for _, _, i in reductions):
             # Use atomic increments
-            mapper = {i: i._rebuild(pragmas=self.lang['atomic']) for i in exprs}
+            mapper = {i: i._rebuild(pragmas=self.langbb['atomic']) for i in exprs}
         else:
             raise NotImplementedError
 
@@ -277,7 +257,7 @@ class PragmaShmTransformer(PragmaSimdTransformer):
 
         # Get the collapsable Iterations
         root, collapsable = self._select_candidates(candidates)
-        ncollapse = 1 + len(collapsable)
+        ncollapsed = 1 + len(collapsable)
 
         # Prepare to build a ParallelTree
         if all(i.is_Affine for i in candidates):
@@ -290,24 +270,29 @@ class PragmaShmTransformer(PragmaSimdTransformer):
             if nthreads is None:
                 # pragma ... for ... schedule(..., 1)
                 nthreads = self.nthreads
-                body = self.HostIteration(schedule=schedule, ncollapse=ncollapse,
+                body = self.HostIteration(schedule=schedule, ncollapsed=ncollapsed,
                                           **root.args)
             else:
                 # pragma ... parallel for ... schedule(..., 1)
                 body = self.HostIteration(schedule=schedule, parallel=True,
-                                          ncollapse=ncollapse, nthreads=nthreads,
+                                          ncollapsed=ncollapsed, nthreads=nthreads,
                                           **root.args)
             prefix = []
+        elif nthreads is not None:
+            body = self.HostIteration(schedule='static',
+                                      parallel=nthreads is not self.nthreads_nested,
+                                      ncollapsed=ncollapsed, nthreads=nthreads,
+                                      **root.args)
+            prefix = []
         else:
             # pragma ... for ... schedule(..., expr)
-            assert nthreads is None
             nthreads = self.nthreads_nonaffine
             chunk_size = Symbol(name='chunk_size')
-            body = self.HostIteration(ncollapse=ncollapse, chunk_size=chunk_size,
+            body = self.HostIteration(ncollapsed=ncollapsed, chunk_size=chunk_size,
                                       **root.args)
 
             niters = prod([root.symbolic_size] + [j.symbolic_size for j in collapsable])
-            value = INT(Max(niters / (nthreads*self.chunk_nonaffine), 1))
+            value = INT(Max(INT(niters / (nthreads*self.chunk_nonaffine)), 1))
             prefix = [Expression(DummyEq(chunk_size, value, dtype=np.int32))]
 
         # Create a ParallelTree
@@ -334,7 +319,7 @@ class PragmaShmTransformer(PragmaSimdTransformer):
             vexpandeds.append(VExpanded(i, pi))
 
         if vexpandeds:
-            init = self.lang['thread-num'](retobj=self.threadid)
+            init = self.langbb['thread-num'](retobj=self.threadid)
             prefix = List(body=[init] + vexpandeds + list(partree.prefix),
                           footer=c.Line())
             partree = partree._rebuild(prefix=prefix)
@@ -387,7 +372,8 @@ class PragmaShmTransformer(PragmaSimdTransformer):
 
         return partree
 
-    def _make_parallel(self, iet):
+    @iet_pass
+    def _make_parallel(self, iet, sync_mapper=None):
         mapper = {}
         parrays = {}
         for tree in retrieve_iteration_tree(iet, mode='superset'):
@@ -396,6 +382,17 @@ class PragmaShmTransformer(PragmaSimdTransformer):
             if not candidates:
                 continue
 
+            # Ignore if already a ParallelIteration (e.g., by-product of
+            # recursive compilation)
+            if any(isinstance(n, ParallelIteration) for n in candidates):
+                continue
+
+            # Ignore if already part of an asynchronous region of code
+            # (e.g., an Iteartion embedded within a SyncSpot defining an
+            # asynchronous operation)
+            if any(n in sync_mapper for n in candidates):
+                continue
+
             # Outer parallelism
             root, partree = self._make_partree(candidates)
             if partree is None or root in mapper:
@@ -420,21 +417,21 @@ class PragmaShmTransformer(PragmaSimdTransformer):
 
         iet = Transformer(mapper).visit(iet)
 
-        return iet, {'includes': [self.lang['header']]}
+        return iet, {'includes': [self.langbb['header']]}
 
-    @iet_pass
-    def make_parallel(self, iet):
-        return self._make_parallel(iet)
+    def make_parallel(self, graph):
+        return self._make_parallel(graph, sync_mapper=graph.sync_mapper)
 
 
 class PragmaTransfer(Pragma, Transfer):
 
     """
-    A data transfer between host and device expressed by means of one or more pragmas.
+    A data transfer between host and device expressed by means of one or
+    more pragmas.
     """
 
-    def __init__(self, callback, function, imask=None, arguments=None):
-        super().__init__(callback, arguments)
+    def __init__(self, pragma, function, imask=None, arguments=None):
+        super().__init__(pragma, arguments)
 
         self._function = function
         self._imask = imask
@@ -447,17 +444,12 @@ class PragmaTransfer(Pragma, Transfer):
     def imask(self):
         return self._imask
 
-    @cached_property
+    # TODO: cached_property here will break our pickling tests for reasons that
+    # are still mysterious after considerable investigation
+    @property
     def sections(self):
         return make_sections_from_imask(self.function, self.imask)
 
-    @cached_property
-    def pragmas(self):
-        # Stringify sections
-        sections = ''.join(['[%s:%s]' % (ccode(i), ccode(j)) for i, j in self.sections])
-        arguments = [ccode(i) for i in self.arguments]
-        return as_tuple(self.callback(self.function.name, sections, *arguments))
-
     @property
     def functions(self):
         return (self.function,)
@@ -472,6 +464,14 @@ class PragmaTransfer(Pragma, Transfer):
                 pass
         return tuple(retval)
 
+    @cached_property
+    def _generate(self):
+        # Stringify sections
+        sections = ''.join(['[%s:%s]' % (ccode(i), ccode(j))
+                            for i, j in self.sections])
+        arguments = [ccode(i) for i in self.arguments]
+        return self.pragma % (self.function.name, sections, *arguments)
+
 
 class PragmaDeviceAwareTransformer(DeviceAwareMixin, PragmaShmTransformer):
 
@@ -484,9 +484,17 @@ class PragmaDeviceAwareTransformer(DeviceAwareMixin, PragmaShmTransformer):
         super().__init__(sregistry, options, platform, compiler)
 
         self.gpu_fit = options['gpu-fit']
-        self.par_tile = options['par-tile']
+        # Need to reset the tile in case was already used and iter over by blocking
+        self.par_tile = options['par-tile'].reset()
         self.par_disabled = options['par-disabled']
 
+    def _score_candidate(self, n0, root, collapsable=()):
+        # `ndptrs`, the number of device pointers, part of the score too to
+        # ensure the outermost loop is offloaded
+        ndptrs = len(self._device_pointers(root))
+
+        return (ndptrs,) + super()._score_candidate(n0, root, collapsable)
+
     def _make_threaded_prodders(self, partree):
         if isinstance(partree.root, self.DeviceIteration):
             # no-op for now
@@ -494,7 +502,7 @@ class PragmaDeviceAwareTransformer(DeviceAwareMixin, PragmaShmTransformer):
         else:
             return super()._make_threaded_prodders(partree)
 
-    def _make_partree(self, candidates, nthreads=None):
+    def _make_partree(self, candidates, nthreads=None, index=None):
         """
         Parallelize the `candidates` Iterations. In particular:
 
@@ -512,7 +520,8 @@ class PragmaDeviceAwareTransformer(DeviceAwareMixin, PragmaShmTransformer):
 
         if self._is_offloadable(root):
             body = self.DeviceIteration(gpu_fit=self.gpu_fit,
-                                        ncollapse=len(collapsable) + 1,
+                                        ncollapsed=len(collapsable)+1,
+                                        tile=self.par_tile.nextitem(),
                                         **root.args)
             partree = ParallelTree([], body, nthreads=nthreads)
 
@@ -570,20 +579,6 @@ class PragmaDeviceAwareTransformer(DeviceAwareMixin, PragmaShmTransformer):
             return super()._make_nested_partree(partree)
 
 
-class PragmaDeviceAwareDataManager(DeviceAwareDataManager):
-
-    def process(self, graph):
-        super().process(graph)
-        self.place_devptr(graph)
-
-    @iet_pass
-    def place_devptr(self, iet, **kwargs):
-        """
-        Transform `iet` such that device pointers are used in DeviceCalls.
-        """
-        return iet, {}
-
-
 class PragmaLangBB(LangBB):
 
     @classmethod
@@ -594,13 +589,13 @@ class PragmaLangBB(LangBB):
 
     @classmethod
     def _map_to(cls, f, imask=None, qid=None):
-        return PragmaTransfer(cls.mapper['map-enter-to'], f, imask)
+        return cls.mapper['map-enter-to'](f, imask)
 
     _map_to_wait = _map_to
 
     @classmethod
     def _map_alloc(cls, f, imask=None):
-        return PragmaTransfer(cls.mapper['map-enter-alloc'], f, imask)
+        return cls.mapper['map-enter-alloc'](f, imask)
 
     @classmethod
     def _map_present(cls, f, imask=None):
@@ -611,26 +606,26 @@ class PragmaLangBB(LangBB):
 
     @classmethod
     def _map_update(cls, f, imask=None):
-        return PragmaTransfer(cls.mapper['map-update'], f, imask)
+        return cls.mapper['map-update'](f, imask)
 
     @classmethod
     def _map_update_host(cls, f, imask=None, qid=None):
-        return PragmaTransfer(cls.mapper['map-update-host'], f, imask)
+        return cls.mapper['map-update-host'](f, imask)
 
     _map_update_host_async = _map_update_host
 
     @classmethod
     def _map_update_device(cls, f, imask=None, qid=None):
-        return PragmaTransfer(cls.mapper['map-update-device'], f, imask)
+        return cls.mapper['map-update-device'](f, imask)
 
     _map_update_device_async = _map_update_device
 
     @classmethod
     def _map_release(cls, f, imask=None, devicerm=None):
         if devicerm:
-            return PragmaTransfer(cls.mapper['map-release-if'], f, imask, devicerm)
+            return cls.mapper['map-release-if'](f, imask, devicerm)
         else:
-            return PragmaTransfer(cls.mapper['map-release'], f, imask)
+            return cls.mapper['map-release'](f, imask)
 
 
 # Utils
