@@ -1,22 +1,28 @@
 """
 Extended SymPy hierarchy.
 """
+import re
 
 import numpy as np
 import sympy
-from sympy import Expr, Integer, Function, Number, Tuple, sympify
+from sympy import Expr, Function, Number, Tuple, cacheit, sympify
 from sympy.core.decorators import call_highest_priority
 
-from devito.symbolics.printer import ccode
-from devito.tools import Pickable, as_tuple, is_integer
+from devito.finite_differences.elementary import Min, Max
+from devito.tools import (Pickable, Bunch, as_tuple, is_integer, float2,  # noqa
+                          float3, float4, double2, double3, double4, int2, int3,
+                          int4, dtype_to_ctype, ctypes_to_cstr, ctypes_vector_mapper,
+                          ctypes_to_cstr)
 from devito.types import Symbol
+from devito.types.basic import Basic
 
-__all__ = ['CondEq', 'CondNe', 'IntDiv', 'CallFromPointer', 'FieldFromPointer',
-           'FieldFromComposite', 'ListInitializer', 'Byref', 'IndexedPointer', 'Cast',
-           'DefFunction', 'InlineIf', 'Keyword', 'String', 'Macro', 'MacroArgument',
-           'CustomType', 'Deref', 'INT', 'FLOAT', 'DOUBLE', 'VOID', 'CEIL',
-           'FLOOR', 'MAX', 'MIN', 'Null', 'SizeOf', 'rfunc', 'cast_mapper',
-           'BasicWrapperMixin']
+__all__ = ['CondEq', 'CondNe', 'IntDiv', 'CallFromPointer',  # noqa
+           'CallFromComposite', 'FieldFromPointer', 'FieldFromComposite',
+           'ListInitializer', 'Byref', 'IndexedPointer', 'Cast', 'DefFunction',
+           'MathFunction', 'InlineIf', 'ReservedWord', 'Keyword', 'String',
+           'Macro', 'Class', 'MacroArgument', 'Deref', 'Namespace',
+           'Rvalue', 'Null', 'SizeOf', 'rfunc', 'BasicWrapperMixin', 'ValueLimit',
+           'VectorAccess']
 
 
 class CondEq(sympy.Eq):
@@ -79,16 +85,16 @@ class IntDiv(sympy.Expr):
     def __new__(cls, lhs, rhs, params=None):
         if rhs == 0:
             raise ValueError("Cannot divide by 0")
-        elif rhs == 1:
+        elif rhs == 1 or rhs is None:
             return lhs
 
         if not is_integer(rhs):
             # Perhaps it's a symbolic RHS -- but we wanna be sure it's of type int
             if not hasattr(rhs, 'dtype'):
-                raise ValueError("Symbolic RHS `%s` lacks dtype" % rhs)
+                raise ValueError(f"Symbolic RHS `{rhs}` lacks dtype")
             if not issubclass(rhs.dtype, np.integer):
-                raise ValueError("Symbolic RHS `%s` must be of type `int`, found "
-                                 "`%s` instead" % (rhs, rhs.dtype))
+                raise ValueError(f"Symbolic RHS `{rhs}` must be of type `int`, found "
+                                 f"`{rhs.dtype}` instead")
         rhs = sympify(rhs)
 
         obj = sympy.Expr.__new__(cls, lhs, rhs)
@@ -99,7 +105,7 @@ class IntDiv(sympy.Expr):
         return obj
 
     def __str__(self):
-        return "IntDiv(%s, %s)" % (self.lhs, self.rhs)
+        return f"IntDiv({self.lhs}, {self.rhs})"
 
     __repr__ = __str__
 
@@ -111,7 +117,7 @@ class IntDiv(sympy.Expr):
         return super().__mul__(other)
 
 
-class BasicWrapperMixin(object):
+class BasicWrapperMixin:
 
     """
     Abstract mixin class for objects wrapping types.Basic objects.
@@ -138,6 +144,19 @@ class BasicWrapperMixin(object):
         """
         return self.function.dtype
 
+    @property
+    def is_commutative(self):
+        """
+        Overridden SymPy assumption -- now based on the wrapped object dtype.
+        """
+        try:
+            return issubclass(self.dtype, np.number)
+        except TypeError:
+            return self.dtype in ctypes_vector_mapper
+
+    def _sympystr(self, printer):
+        return str(self)
+
 
 class CallFromPointer(sympy.Expr, Pickable, BasicWrapperMixin):
 
@@ -153,11 +172,9 @@ class CallFromPointer(sympy.Expr, Pickable, BasicWrapperMixin):
             pointer = Symbol(pointer)
         if isinstance(call, str):
             call = Symbol(call)
-        elif not isinstance(call, (CallFromPointer, DefFunction, sympy.Symbol)):
-            # NOTE: we need `sympy.Symbol`, rather than just (devito) `Symbol`
-            # because otherwise it breaks upon certain reconstructions on SymPy-1.8,
-            # due to the way `bound_symbols` and `canonical_variables` interact
-            raise ValueError("`call` must be CallFromPointer, DefFunction, or Symbol")
+        elif not isinstance(call, Basic):
+            raise ValueError("`call` must be a `devito.Basic` or a type "
+                             "with compatible interface")
         _params = []
         for p in as_tuple(params):
             if isinstance(p, str):
@@ -179,14 +196,13 @@ class CallFromPointer(sympy.Expr, Pickable, BasicWrapperMixin):
         return obj
 
     def __str__(self):
-        return '%s->%s(%s)' % (self.pointer, self.call,
-                               ", ".join(str(i) for i in as_tuple(self.params)))
+        params = ", ".join(str(i) for i in as_tuple(self.params))
+        return f'{self.pointer}->{self.call}({params})'
 
     __repr__ = __str__
 
     def _hashable_content(self):
-        return super(CallFromPointer, self)._hashable_content() +\
-            (self.call, self.pointer) + self.params
+        return super()._hashable_content() + (self.call, self.pointer) + self.params
 
     @property
     def base(self):
@@ -204,10 +220,25 @@ class CallFromPointer(sympy.Expr, Pickable, BasicWrapperMixin):
     def free_symbols(self):
         return super().free_symbols - self.bound_symbols
 
+    is_commutative = BasicWrapperMixin.is_commutative
+
     # Pickling support
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
+class CallFromComposite(CallFromPointer, Pickable):
+
+    """
+    Symbolic representation of the C notation ``composite.call(params)``.
+    """
+
+    def __str__(self):
+        params = ", ".join(str(i) for i in as_tuple(self.params))
+        return f'{self.pointer}.{self.call}({params})'
+
+    __repr__ = __str__
+
+
 class FieldFromPointer(CallFromPointer, Pickable):
 
     """
@@ -220,7 +251,7 @@ class FieldFromPointer(CallFromPointer, Pickable):
         return CallFromPointer.__new__(cls, field, pointer)
 
     def __str__(self):
-        return '%s->%s' % (self.pointer, self.field)
+        return f'{self.pointer}->{self.field}'
 
     @property
     def field(self):
@@ -242,7 +273,7 @@ class FieldFromComposite(CallFromPointer, Pickable):
         return CallFromPointer.__new__(cls, field, composite)
 
     def __str__(self):
-        return '%s.%s' % (self.composite, self.field)
+        return f'{self.composite}.{self.field}'
 
     @property
     def field(self):
@@ -262,27 +293,31 @@ class ListInitializer(sympy.Expr, Pickable):
     """
 
     __rargs__ = ('params',)
+    __rkwargs__ = ('dtype',)
 
-    def __new__(cls, params):
+    def __new__(cls, params, dtype=None):
         args = []
         for p in as_tuple(params):
-            if isinstance(p, str):
-                args.append(Symbol(p))
-            elif is_integer(p):
-                args.append(Integer(p))
-            elif not isinstance(p, Expr):
-                raise ValueError("`params` must be an iterable of Expr or str")
-            else:
-                args.append(p)
+            try:
+                args.append(sympify(p))
+            except sympy.SympifyError:
+                raise ValueError(f"Illegal param `{p}`")
         obj = sympy.Expr.__new__(cls, *args)
+
         obj.params = tuple(args)
+        obj.dtype = dtype
+
         return obj
 
     def __str__(self):
-        return "{%s}" % ", ".join(str(i) for i in self.params)
+        return f"{{{', '.join(str(i) for i in self.params)}}}"
 
     __repr__ = __str__
 
+    @property
+    def is_numeric(self):
+        return all(i.is_Number for i in self.params)
+
     # Pickling support
     __reduce_ex__ = Pickable.__reduce_ex__
 
@@ -304,8 +339,9 @@ class UnaryOp(sympy.Expr, Pickable, BasicWrapperMixin):
         except AttributeError:
             if isinstance(base, str):
                 base = Symbol(base)
-            elif not isinstance(base, sympy.Basic):
-                raise ValueError("`base` must be sympy.Basic or str")
+            else:
+                # Fallback: go plain sympy
+                base = sympify(base)
 
         obj = sympy.Expr.__new__(cls, base)
         obj._base = base
@@ -322,9 +358,9 @@ class UnaryOp(sympy.Expr, Pickable, BasicWrapperMixin):
 
     def __str__(self):
         if self.base.is_Symbol:
-            return "%s%s" % (self._op, ccode(self.base))
+            return f'{self._op}{self.base}'
         else:
-            return "%s(%s)" % (self._op, ccode(self.base))
+            return f'{self._op}({self.base})'
 
     __repr__ = __str__
 
@@ -356,15 +392,40 @@ class Cast(UnaryOp):
     Symbolic representation of the C notation `(type)expr`.
     """
 
-    _base_typ = ''
+    __rargs__ = ('base', )
+    __rkwargs__ = ('dtype', 'stars', 'reinterpret')
 
-    __rkwargs__ = ('stars',)
+    def __new__(cls, base, dtype=None, stars=None, reinterpret=False, **kwargs):
+        try:
+            if issubclass(dtype, np.generic) and sympify(base).is_Number:
+                base = sympify(dtype(base))
+        except TypeError:
+            # E.g. void
+            pass
+
+        dtype, stars = cls._process_dtype(dtype, stars)
 
-    def __new__(cls, base, stars=None, **kwargs):
         obj = super().__new__(cls, base)
-        obj._stars = stars
+        obj._stars = stars or ''
+        obj._dtype = dtype
+        obj._reinterpret = reinterpret
+
         return obj
 
+    @classmethod
+    def _process_dtype(cls, dtype, stars):
+        if not isinstance(dtype, str) or stars is not None:
+            return dtype, stars
+
+        # String dtype, e.g. "float", "int*", "foo**"
+        match = re.fullmatch(r'(\w+)\s*(\*+)?', dtype)
+        if match:
+            dtype = match.group(1)
+            stars = match.group(2) or ''
+            return dtype, stars
+        else:
+            return dtype, stars
+
     def _hashable_content(self):
         return super()._hashable_content() + (self._stars,)
 
@@ -375,12 +436,28 @@ class Cast(UnaryOp):
         return self._stars
 
     @property
-    def typ(self):
-        return '%s%s' % (self._base_typ, self.stars or '')
+    def dtype(self):
+        return self._dtype
+
+    @property
+    def reinterpret(self):
+        return self._reinterpret
+
+    @property
+    def _C_ctype(self):
+        ctype = ctypes_vector_mapper.get(self.dtype, self.dtype)
+        try:
+            ctype = dtype_to_ctype(ctype)
+        except TypeError:
+            pass
+        return ctype
 
     @property
     def _op(self):
-        return '(%s)' % self.typ
+        return f'({ctypes_to_cstr(self._C_ctype)}{self.stars})'
+
+    def __str__(self):
+        return f"{self._op}{self.base}"
 
 
 class IndexedPointer(sympy.Expr, Pickable, BasicWrapperMixin):
@@ -402,9 +479,9 @@ class IndexedPointer(sympy.Expr, Pickable, BasicWrapperMixin):
             if not isinstance(base, sympy.Basic):
                 raise ValueError("`base` must be of type sympy.Basic")
 
-        index = tuple(sympify(i) for i in as_tuple(index))
+        index = Tuple(*[sympify(i) for i in as_tuple(index)])
 
-        obj = sympy.Expr.__new__(cls, base, *index)
+        obj = sympy.Expr.__new__(cls, base, index)
         obj._base = base
         obj._index = index
 
@@ -419,10 +496,13 @@ class IndexedPointer(sympy.Expr, Pickable, BasicWrapperMixin):
         return self._index
 
     def __str__(self):
-        return "%s%s" % (self.base, ''.join('[%s]' % i for i in self.index))
+        indices = ''.join(f'[{i}]' for i in self.index)
+        return f"{self.base}{indices}"
 
     __repr__ = __str__
 
+    is_commutative = BasicWrapperMixin.is_commutative
+
     # Pickling support
     __reduce_ex__ = Pickable.__reduce_ex__
 
@@ -444,7 +524,7 @@ class ReservedWord(sympy.Atom, Pickable):
 
     def __new__(cls, value, **kwargs):
         if not isinstance(value, str):
-            raise TypeError("Expected str, got `%s`" % type(value))
+            raise TypeError(f"Expected str, got `{type(value)}`")
         obj = sympy.Atom.__new__(cls, **kwargs)
         obj.value = value
 
@@ -458,6 +538,9 @@ class ReservedWord(sympy.Atom, Pickable):
     def _hashable_content(self):
         return (self.value,)
 
+    def _sympystr(self, printer):
+        return str(self)
+
     # Pickling support
     __reduce_ex__ = Pickable.__reduce_ex__
 
@@ -466,10 +549,6 @@ class Keyword(ReservedWord):
     pass
 
 
-class CustomType(ReservedWord):
-    pass
-
-
 class String(ReservedWord):
     pass
 
@@ -478,14 +557,32 @@ class Macro(ReservedWord):
     pass
 
 
+class Class(ReservedWord):
+
+    def __str__(self):
+        return f"class {self.value}"
+
+    __repr__ = __str__
+
+
 class MacroArgument(sympy.Symbol):
 
     def __str__(self):
-        return "(%s)" % self.name
+        return f"({self.name})"
 
     __repr__ = __str__
 
 
+class ValueLimit(ReservedWord, sympy.Expr):
+
+    """
+    Symbolic representation of the so called limits macros, which provide the
+    minimum and maximum limits for various types, such as INT_MIN, INT_MAX etc.
+    """
+
+    pass
+
+
 class DefFunction(Function, Pickable):
 
     """
@@ -494,10 +591,12 @@ class DefFunction(Function, Pickable):
         https://github.com/sympy/sympy/issues/4297
     """
 
-    __rargs__ = ('name',)
-    __rkwargs__ = ('arguments',)
+    __rargs__ = ('name', 'arguments', 'template')
+
+    def __new__(cls, name, arguments=None, template=None, **kwargs):
+        if isinstance(name, str):
+            name = Keyword(name)
 
-    def __new__(cls, name, arguments=None, **kwargs):
         _arguments = []
         for i in as_tuple(arguments):
             if isinstance(i, str):
@@ -507,14 +606,31 @@ class DefFunction(Function, Pickable):
                 _arguments.append(ReservedWord(i))
             else:
                 _arguments.append(i)
-        arguments = tuple(_arguments)
-        if isinstance(name, str):
-            name = Keyword(name)
-        obj = Function.__new__(cls, name, Tuple(*arguments))
+
+        _template = []
+        for i in as_tuple(template):
+            if isinstance(i, str):
+                # Same story as above
+                _template.append(ReservedWord(i))
+            else:
+                _template.append(i)
+
+        args = [name]
+        args.append(Tuple(*_arguments))
+        if _template:
+            args.append(Tuple(*_template))
+
+        obj = Function.__new__(cls, *args)
         obj._name = name
-        obj._arguments = arguments
+        obj._arguments = tuple(_arguments)
+        obj._template = tuple(_template)
+
         return obj
 
+    def _eval_is_commutative(self):
+        # DefFunction defaults to commutative
+        return True
+
     @property
     def name(self):
         return self._name
@@ -524,19 +640,34 @@ class DefFunction(Function, Pickable):
         return self._arguments
 
     @property
-    def free_symbols(self):
-        return set().union(*[i.free_symbols for i in self.arguments
-                             if isinstance(i, Expr)])
+    def template(self):
+        return self._template
 
     def __str__(self):
-        return "%s(%s)" % (self.name, ', '.join(str(i) for i in self.arguments))
+        if self.template:
+            template = f"<{','.join(str(i) for i in self.template)}>"
+        else:
+            template = ''
+        arguments = ', '.join(str(i) for i in self.arguments)
+        return f"{self.name}{template}({arguments})"
 
     __repr__ = __str__
 
+    def _sympystr(self, printer):
+        return str(self)
+
+    func = Pickable._rebuild
+
     # Pickling support
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
+class MathFunction(DefFunction):
+
+    # Supposed to involve real operands
+    is_commutative = True
+
+
 class InlineIf(sympy.Expr, Pickable):
 
     """
@@ -571,7 +702,7 @@ class InlineIf(sympy.Expr, Pickable):
         return self._false_expr
 
     def __str__(self):
-        return "(%s) ? %s : %s" % (self.cond, self.true_expr, self.false_expr)
+        return f"({self.cond}) ? {self.true_expr} : {self.false_expr}"
 
     __repr__ = __str__
 
@@ -579,71 +710,149 @@ class InlineIf(sympy.Expr, Pickable):
     __reduce_ex__ = Pickable.__reduce_ex__
 
 
-# Shortcuts (mostly for retrocompatibility)
+class Namespace(sympy.Expr, Pickable):
 
-class CHAR(Cast):
-    _base_typ = 'char'
+    """
+    Symbolic representation of a C++ namespace `ns0::ns1::...`.
+    """
+
+    __rargs__ = ('items',)
+
+    def __new__(cls, items, **kwargs):
+        normalized_items = []
+        for i in as_tuple(items):
+            if isinstance(i, str):
+                normalized_items.append(ReservedWord(i))
+            elif isinstance(i, ReservedWord):
+                normalized_items.append(i)
+            else:
+                raise ValueError("`items` must be iterable of str or ReservedWord")
+
+        obj = sympy.Expr.__new__(cls)
+        obj._items = tuple(items)
+
+        return obj
+
+    def _hashable_content(self):
+        return super()._hashable_content() + self.items
+
+    @property
+    def items(self):
+        return self._items
+
+    def __str__(self):
+        return "::".join(str(i) for i in self.items)
+
+    __repr__ = __str__
 
 
-class INT(Cast):
-    _base_typ = 'int'
+class Rvalue(sympy.Expr, Pickable):
+
+    """
+    A generic C++ rvalue, that is a value that occupies a temporary location in
+    memory.
+    """
+
+    __rargs__ = ('expr',)
+    __rkwargs__ = ('namespace', 'init')
+
+    def __new__(cls, expr, namespace=None, init=None):
+        args = [expr]
+        if namespace is not None:
+            args.append(namespace)
+        if init is not None:
+            args.append(init)
+
+        obj = sympy.Expr.__new__(cls, *args)
+
+        obj._expr = expr
+        obj._namespace = namespace
+        obj._init = init
+
+        return obj
+
+    @property
+    def expr(self):
+        return self._expr
+
+    @property
+    def namespace(self):
+        return self._namespace
+
+    @property
+    def init(self):
+        return self._init
+
+    def __str__(self):
+        rvalue = str(self.expr)
+        if self.namespace:
+            rvalue = f"{self.namespace}::{rvalue}"
+        if self.init:
+            rvalue = f"{rvalue}{self.init}"
+        return rvalue
+
+    __repr__ = __str__
 
 
-class LONG(Cast):
-    _base_typ = 'long'
+class VectorAccess(Expr, Pickable, BasicWrapperMixin):
 
+    """
+    Represent a vector access operation at high-level.
+    """
 
-class ULONG(Cast):
-    _base_typ = 'unsigned long'
+    def __new__(cls, *args, **kwargs):
+        return Expr.__new__(cls, *args)
 
+    def __str__(self):
+        return f"VL<{self.base}>"
 
-class FLOAT(Cast):
-    _base_typ = 'float'
+    __repr__ = __str__
 
+    func = Pickable._rebuild
 
-class DOUBLE(Cast):
-    _base_typ = 'double'
+    @property
+    def base(self):
+        return self.args[0]
 
+    @property
+    def indices(self):
+        return self.base.indices
 
-class VOID(Cast):
-    _base_typ = 'void'
-
-
-class CastStar(object):
-
-    base = None
-
-    def __new__(cls, base=''):
-        return cls.base(base, '*')
-
-
-class CHARP(CastStar):
-    base = CHAR
-
-
-class INTP(CastStar):
-    base = INT
-
-
-class FLOATP(CastStar):
-    base = FLOAT
-
-
-class DOUBLEP(CastStar):
-    base = DOUBLE
+    @cacheit
+    def sort_key(self, order=None):
+        # Ensure that the VectorAccess is sorted as the base
+        return self.base.sort_key(order=order)
 
 
 # Some other utility objects
-
-CEIL = Function('ceil')
-FLOOR = Function('floor')
-MAX = Function('MAX')
-MIN = Function('MIN')
-
 Null = Macro('NULL')
 
+
 # DefFunction, unlike sympy.Function, generates e.g. `sizeof(float)`, not `sizeof(float_)`
-SizeOf = lambda *args: DefFunction('sizeof', tuple(args))
+class SizeOf(DefFunction):
+
+    __rargs__ = ('intype', 'stars')
+
+    def __new__(cls, intype, stars=None, **kwargs):
+        stars = stars or ''
+        if not isinstance(intype, (str, ReservedWord)):
+            ctype = dtype_to_ctype(intype)
+            for k, v in ctypes_vector_mapper.items():
+                if ctype is v:
+                    intype = k
+                    break
+            else:
+                intype = ctypes_to_cstr(ctype)
+
+        newobj = super().__new__(cls, 'sizeof', arguments=f'{intype}{stars}', **kwargs)
+        newobj.stars = stars
+        newobj.intype = intype
+
+        return newobj
+
+    @property
+    def args(self):
+        return super().args[1]
 
 
 def rfunc(func, item, *args):
@@ -653,10 +862,10 @@ def rfunc(func, item, *args):
     Examples
     ----------
     >> rfunc(min, [a, b, c, d])
-    MIN(a, MIN(b, MIN(c, d)))
+    Min(a, Min(b, Min(c, d)))
 
     >> rfunc(max, [a, b, c, d])
-    MAX(a, MAX(b, MAX(c, d)))
+    Max(a, Max(b, Max(c, d)))
     """
 
     assert func in rfunc_mapper
@@ -665,31 +874,10 @@ def rfunc(func, item, *args):
     if len(args) == 0:
         return item
     else:
-        return rf(item, rfunc(func, *args))
+        return rf(item, rfunc(func, *args), evaluate=False)
 
 
-cast_mapper = {
-    np.int8: CHAR,
-    np.uint8: CHAR,
-    int: INT,
-    np.int32: INT,
-    np.int64: LONG,
-    np.uint64: ULONG,
-    np.float32: FLOAT,
-    float: DOUBLE,
-    np.float64: DOUBLE,
-
-    (np.int8, '*'): CHARP,
-    (np.uint8, '*'): CHARP,
-    (int, '*'): INTP,
-    (np.int32, '*'): INTP,
-    (np.int64, '*'): INTP,
-    (np.float32, '*'): FLOATP,
-    (float, '*'): DOUBLEP,
-    (np.float64, '*'): DOUBLEP
-}
-
 rfunc_mapper = {
-    min: MIN,
-    max: MAX,
+    min: Min,
+    max: Max,
 }
