@@ -1,17 +1,21 @@
-from collections import Counter
 from functools import singledispatch
 
 import numpy as np
-from sympy import Function, Indexed, Integer, Mul, Number, Pow, S, Symbol, Tuple
+from sympy import (Function, Indexed, Integer, Mul, Number,
+                   Pow, S, Symbol, Tuple)
+from sympy.core.numbers import ImaginaryUnit
 
+from devito.finite_differences import Derivative
+from devito.finite_differences.differentiable import IndexDerivative
 from devito.logger import warning
-from devito.symbolics.extended_sympy import (INT, CallFromPointer, Cast,
+from devito.symbolics.extended_dtypes import INT
+from devito.symbolics.extended_sympy import (CallFromPointer, Cast,
                                              DefFunction, ReservedWord)
 from devito.symbolics.queries import q_routine
-from devito.symbolics.search import search
-from devito.tools import as_tuple
+from devito.tools import as_tuple, prod
+from devito.tools.dtypes_lowering import infer_dtype
 
-__all__ = ['compare_ops', 'count', 'estimate_cost']
+__all__ = ['compare_ops', 'estimate_cost', 'has_integer_args', 'sympy_dtype']
 
 
 def compare_ops(e1, e2):
@@ -42,9 +46,14 @@ def compare_ops(e1, e2):
     >>> compare_ops(u[x] + u[x+1], u[x] + u[y+10])
     True
     """
-    if type(e1) == type(e2) and len(e1.args) == len(e2.args):
+    if type(e1) is type(e2) and len(e1.args) == len(e2.args):
         if e1.is_Atom:
             return True if e1 == e2 else False
+        elif isinstance(e1, IndexDerivative) and isinstance(e2, IndexDerivative):
+            if e1.mapper == e2.mapper:
+                return compare_ops(e1.expr, e2.expr)
+            else:
+                return False
         elif e1.is_Indexed and e2.is_Indexed:
             return True if e1.base == e2.base else False
         else:
@@ -56,17 +65,6 @@ def compare_ops(e1, e2):
         return False
 
 
-def count(exprs, query):
-    """
-    Return a mapper ``{(k, v)}`` where ``k`` is a sub-expression in ``exprs``
-    matching ``query`` and ``v`` is the number of its occurrences.
-    """
-    mapper = Counter()
-    for expr in as_tuple(exprs):
-        mapper.update(Counter(search(expr, query, 'all', 'bfs')))
-    return dict(mapper)
-
-
 def estimate_cost(exprs, estimate=False):
     """
     Estimate the operation count of an expression.
@@ -94,13 +92,21 @@ def estimate_cost(exprs, estimate=False):
         # We don't use SymPy's count_ops because we do not count integer arithmetic
         # (e.g., array index functions such as i+1 in A[i+1])
         # Also, the routine below is *much* faster than count_ops
+        seen = {}
         flops = 0
         for expr in as_tuple(exprs):
+            # TODO: this if-then should be part of singledispatch too, but because
+            # of the annoying symbolics/ vs types/ structuring, we can't do that
+            # because of circular imports...
             if expr.is_Equality:
                 e = expr.rhs
+                if expr.is_Reduction:
+                    flops += 1
             else:
                 e = expr
-            flops += _estimate_cost(e, estimate)[0]
+
+            flops += _estimate_cost(e, estimate, seen)[0]
+
         return flops
     except:
         warning("Cannot estimate cost of `%s`" % str(exprs))
@@ -110,15 +116,37 @@ def estimate_cost(exprs, estimate=False):
 estimate_values = {
     'elementary': 100,
     'pow': 50,
-    'div': 5
+    'SafeInv': 10,
+    'div': 5,
+    'Abs': 5,
+    'floor': 1,
+    'ceil': 1,
 }
 
 
+def dont_count_if_seen(func):
+    """
+    This decorator is used to avoid counting the same expression multiple
+    times. This is necessary because the same expression may appear multiple
+    times in the same expression tree or even across different expressions.
+    """
+    def wrapper(expr, estimate, seen):
+        try:
+            _, flags = seen[expr]
+            flops = 0
+        except KeyError:
+            flops, flags = seen[expr] = func(expr, estimate, seen)
+        return flops, flags
+    return wrapper
+
+
 @singledispatch
-def _estimate_cost(expr, estimate):
+def _estimate_cost(expr, estimate, seen):
     # Retval: flops (int), flag (bool)
     # The flag tells wether it's an integer expression (implying flops==0) or not
-    flops, flags = zip(*[_estimate_cost(a, estimate) for a in expr.args])
+    if not expr.args:
+        return 0, False
+    flops, flags = zip(*[_estimate_cost(a, estimate, seen) for a in expr.args])
     flops = sum(flops)
     if all(flags):
         # `expr` is an operation involving integer operands only
@@ -131,28 +159,29 @@ def _estimate_cost(expr, estimate):
 
 @_estimate_cost.register(Tuple)
 @_estimate_cost.register(CallFromPointer)
-def _(expr, estimate):
+def _(expr, estimate, seen):
     try:
-        flops, flags = zip(*[_estimate_cost(a, estimate) for a in expr.args])
+        flops, flags = zip(*[_estimate_cost(a, estimate, seen) for a in expr.args])
     except ValueError:
         flops, flags = [], []
     return sum(flops), all(flags)
 
 
 @_estimate_cost.register(Integer)
-def _(expr, estimate):
+def _(expr, estimate, seen):
     return 0, True
 
 
+@_estimate_cost.register(ImaginaryUnit)
 @_estimate_cost.register(Number)
 @_estimate_cost.register(ReservedWord)
-def _(expr, estimate):
+def _(expr, estimate, seen):
     return 0, False
 
 
 @_estimate_cost.register(Symbol)
 @_estimate_cost.register(Indexed)
-def _(expr, estimate):
+def _(expr, estimate, seen):
     try:
         if issubclass(expr.dtype, np.integer):
             return 0, True
@@ -162,33 +191,38 @@ def _(expr, estimate):
 
 
 @_estimate_cost.register(Mul)
-def _(expr, estimate):
-    flops, flags = _estimate_cost.registry[object](expr, estimate)
+def _(expr, estimate, seen):
+    flops, flags = _estimate_cost.registry[object](expr, estimate, seen)
     if {S.One, S.NegativeOne}.intersection(expr.args):
         flops -= 1
+    if ImaginaryUnit in expr.args:
+        flops *= 2
     return flops, flags
 
 
 @_estimate_cost.register(INT)
-def _(expr, estimate):
-    return _estimate_cost(expr.base, estimate)[0], True
+def _(expr, estimate, seen):
+    return _estimate_cost(expr.base, estimate, seen)[0], True
 
 
 @_estimate_cost.register(Cast)
-def _(expr, estimate):
-    return _estimate_cost(expr.base, estimate)[0], False
+def _(expr, estimate, seen):
+    return _estimate_cost(expr.base, estimate, seen)[0], False
 
 
 @_estimate_cost.register(Function)
-def _(expr, estimate):
+def _(expr, estimate, seen):
     if q_routine(expr):
-        flops, _ = zip(*[_estimate_cost(a, estimate) for a in expr.args])
+        flops, _ = zip(*[_estimate_cost(a, estimate, seen) for a in expr.args])
         flops = sum(flops)
         if isinstance(expr, DefFunction):
             # Bypass user-defined or language-specific functions
             flops += 0
         elif estimate:
-            flops += estimate_values['elementary']
+            try:
+                flops += estimate_values[type(expr).__name__]
+            except KeyError:
+                flops += estimate_values['elementary']
         else:
             flops += 1
     else:
@@ -197,8 +231,8 @@ def _(expr, estimate):
 
 
 @_estimate_cost.register(Pow)
-def _(expr, estimate):
-    flops, _ = zip(*[_estimate_cost(a, estimate) for a in expr.args])
+def _(expr, estimate, seen):
+    flops, _ = zip(*[_estimate_cost(a, estimate, seen) for a in expr.args])
     flops = sum(flops)
     if estimate:
         if expr.exp.is_Number:
@@ -216,3 +250,81 @@ def _(expr, estimate):
     else:
         flops += 1
     return flops, False
+
+
+@_estimate_cost.register(Derivative)
+@dont_count_if_seen
+def _(expr, estimate, seen):
+    return _estimate_cost(expr._evaluate(expand=False), estimate, seen)
+
+
+@_estimate_cost.register(IndexDerivative)
+@dont_count_if_seen
+def _(expr, estimate, seen):
+    flops, _ = _estimate_cost(expr.expr, estimate, seen)
+
+    # It's an increment
+    flops += 1
+
+    # To be multiplied by the number of points this index sum implicitly
+    # iterates over
+    flops *= prod(i._size for i in expr.dimensions)
+
+    return flops, False
+
+
+def has_integer_args(*args):
+    """
+    True if all `args` are of integer type, False otherwise.
+    """
+    if len(args) == 0:
+        return False
+
+    if len(args) == 1:
+        try:
+            return np.issubdtype(args[0].dtype, np.integer)
+        except AttributeError:
+            return args[0].is_integer
+
+    res = True
+    for a in args:
+        try:
+            if isinstance(a, INT):
+                res = res and True
+            elif len(a.args) > 0:
+                res = res and has_integer_args(*a.args)
+            else:
+                res = res and has_integer_args(a)
+        except AttributeError:
+            res = res and has_integer_args(a)
+    return res
+
+
+def sympy_dtype(expr, base=None, default=None, smin=None):
+    """
+    Infer the dtype of the expression.
+    """
+    if expr is None:
+        return default
+
+    dtypes = {base} - {None}
+    for i in expr.free_symbols:
+        try:
+            dtypes.add(i.dtype)
+        except AttributeError:
+            pass
+
+    dtype = infer_dtype(dtypes)
+
+    # Promote if we missed complex number, i.e f + I
+    is_im = np.issubdtype(dtype, np.complexfloating)
+    if expr.has(ImaginaryUnit) and not is_im:
+        if dtype is None:
+            dtype = default or np.complex64
+        else:
+            dtype = np.promote_types(dtype, np.complex64).type
+
+    if smin is not None and not np.issubdtype(dtype, np.integer):
+        dtype = np.promote_types(dtype, smin).type
+
+    return dtype
