@@ -1,27 +1,39 @@
-from collections import OrderedDict, namedtuple
+from collections import OrderedDict
 from collections.abc import Iterable
 from functools import singledispatch
 
-from sympy import LM, LC, Pow, Add, Mul, Min, Max
+import numpy as np
+from sympy import Pow, Add, Mul, Min, Max, S, SympifyError, Tuple, sympify
 from sympy.core.add import _addsort
 from sympy.core.mul import _mulsort
 
-from devito.symbolics.extended_sympy import rfunc
+from devito.finite_differences.differentiable import (
+    EvalDerivative, IndexDerivative
+)
+from devito.symbolics.extended_sympy import DefFunction, rfunc
+from devito.symbolics.extended_dtypes import LONG
 from devito.symbolics.queries import q_leaf
-from devito.symbolics.search import retrieve_indexed, retrieve_functions
+from devito.symbolics.search import (
+    retrieve_indexed, retrieve_functions, retrieve_symbols
+)
+from devito.symbolics.unevaluation import (
+    Add as UnevalAdd, Mul as UnevalMul, Pow as UnevalPow, UnevaluableMixin
+)
 from devito.tools import as_list, as_tuple, flatten, split, transitive_closure
-from devito.types.basic import Basic
+from devito.types.basic import Basic, Indexed
+from devito.types.array import ComponentAccess
 from devito.types.equation import Eq
 from devito.types.relational import Le, Lt, Gt, Ge
 
-__all__ = ['xreplace_indices', 'pow_to_mul', 'indexify', 'split_affine',
-           'subs_op_args', 'uxreplace', 'Uxmapper', 'reuse_if_untouched',
-           'evalrel']
+__all__ = ['xreplace_indices', 'pow_to_mul', 'indexify', 'subs_op_args',
+           'normalize_args', 'uxreplace', 'Uxmapper', 'subs_if_composite',
+           'reuse_if_untouched', 'evalrel', 'flatten_args', 'unevaluate',
+           'as_long']
 
 
 def uxreplace(expr, rule):
     """
-    An alternative to SymPy's ``xreplace`` for when the caller can guarantee
+    An alternative to SymPy's `xreplace` for when the caller can guarantee
     that no re-evaluations are necessary or when re-evaluations should indeed
     be avoided at all costs (e.g., to prevent SymPy from unpicking Devito
     transformations, such as factorization).
@@ -32,12 +44,15 @@ def uxreplace(expr, rule):
     By avoiding re-evaluations, this function is typically much quicker than
     SymPy's xreplace.
 
-    A further feature of ``uxreplace`` consists of enabling the substitution
-    of compound nodes. Consider the expression `a*b*c*d`; if one wants to replace
-    `b*c` with say `e*f`, then the following mapper may be passed:
-    `{a*b*c*d: {b: e*f, c: None}}`. This way, only the `b` and `c` pertaining to
-    `a*b*c*d` will be affected, and in particular `c` will be dropped, while `b`
-    will be replaced by `e*f`, thus obtaining `a*d*e*f`.
+    A further feature of `uxreplace` is the support for compound nodes.
+    Consider the expression `a*b*c*d`; if one wants to replace `b*c` with say
+    `e*f`, then the following mapper may be passed: `{a*b*c*d: {b: e*f, c: None}}`.
+    This way, only the `b` and `c` pertaining to `a*b*c*d` will be affected, and
+    in particular `c` will be dropped, while `b` will be replaced by `e*f`, thus
+    obtaining `a*d*e*f`.
+
+    Finally, `uxreplace` supports Reconstructable objects, that is, it searches
+    for replacement opportunities inside the Reconstructable's `__rkwargs__`.
     """
     return _uxreplace(expr, rule)[0]
 
@@ -50,6 +65,12 @@ def _uxreplace(expr, rule):
         args, eargs = split(expr.args, lambda i: i in v)
         args = [v[i] for i in args if v[i] is not None]
         changed = True
+    elif expr.__class__ in rule:
+        # Exact type -> type substitution
+        cls = rule[expr.__class__]
+        expr = cls(*expr.args)
+        args, eargs = [], expr.args
+        changed = True
     else:
         try:
             args, eargs = [], expr.args
@@ -59,67 +80,135 @@ def _uxreplace(expr, rule):
         changed = False
 
     if rule:
-        for a in eargs:
-            try:
-                ax, flag = _uxreplace(a, rule)
-                args.append(ax)
-                changed |= flag
-            except AttributeError:
-                # E.g., un-sympified numbers
-                args.append(a)
+        eargs, flag = _uxreplace_dispatch(eargs, rule)
+        args.extend(eargs)
+        changed |= flag
+
+        # If a Reconstructable object, we need to parse the kwargs as well
+        if _uxreplace_registry.dispatchable(expr):
+            v = {i: getattr(expr, i) for i in expr.__rkwargs__}
+            kwargs, flag = _uxreplace_dispatch(v, rule)
+        else:
+            kwargs, flag = {}, False
+        changed |= flag
+
         if changed:
-            return _uxreplace_handle(expr, args), True
+            return _uxreplace_handle(expr, args, kwargs), True
 
     return expr, False
 
 
 @singledispatch
-def _uxreplace_handle(expr, args):
-    return expr.func(*args)
+def _uxreplace_dispatch(unknown, rule):
+    return unknown, False
+
+
+@_uxreplace_dispatch.register(Basic)
+def _(expr, rule):
+    return _uxreplace(expr, rule)
+
+
+@_uxreplace_dispatch.register(np.ndarray)
+@_uxreplace_dispatch.register(tuple)
+@_uxreplace_dispatch.register(Tuple)
+@_uxreplace_dispatch.register(list)
+def _(iterable, rule):
+    ret = []
+    changed = False
+    for a in iterable:
+        ax, flag = _uxreplace(a, rule)
+        ret.append(ax)
+        changed |= flag
+    return iterable.__class__(ret), changed
+
+
+@_uxreplace_dispatch.register(dict)
+def _(mapper, rule):
+    ret = {}
+    changed = False
+    for k, v in mapper.items():
+        vx, flag = _uxreplace_dispatch(v, rule)
+        ret[k] = vx
+        changed |= flag
+    return ret, changed
+
+
+@singledispatch
+def _uxreplace_handle(expr, args, kwargs):
+    try:
+        return expr.func(*args, evaluate=False)
+    except TypeError:
+        return expr.func(*args)
 
 
 @_uxreplace_handle.register(Min)
 @_uxreplace_handle.register(Max)
 @_uxreplace_handle.register(Pow)
-def _(expr, args):
+def _(expr, args, kwargs):
     evaluate = all(i.is_Number for i in args)
     return expr.func(*args, evaluate=evaluate)
 
 
 @_uxreplace_handle.register(Add)
-def _(expr, args):
+def _(expr, args, kwargs):
     if all(i.is_commutative for i in args):
         _addsort(args)
         _eval_numbers(expr, args)
+        args = flatten_args(args, Add, ignore=EvalDerivative)
         return expr.func(*args, evaluate=False)
     else:
         return expr._new_rawargs(*args)
 
 
 @_uxreplace_handle.register(Mul)
-def _(expr, args):
+def _(expr, args, kwargs):
+    # Perform some basic simplifications at least
+    args = [i for i in args if i != 1]
+    if any(i == 0 for i in args):
+        return S.Zero
+
     if all(i.is_commutative for i in args):
         _mulsort(args)
         _eval_numbers(expr, args)
+        args = flatten_args(args, Mul, ignore=EvalDerivative)
         return expr.func(*args, evaluate=False)
     else:
         return expr._new_rawargs(*args)
 
 
-@_uxreplace_handle.register(Eq)
-def _(expr, args):
-    # Preserve properties such as `implicit_dims`
-    return expr.func(*args, subdomain=expr.subdomain, coefficients=expr.substitutions,
-                     implicit_dims=expr.implicit_dims)
+def _uxreplace_handle_reconstructable(expr, args, kwargs):
+    return expr.func(*args, **kwargs)
 
 
-def _eval_numbers(expr, args):
+class UxreplaceRegistry(list):
+
     """
-    Helper function for in-place reduction of the expr arguments.
+    A registry used by `uxreplace` to handle Reconstructable objects. These
+    may differ from canonincal SymPy objects since:
+
+        * They carry one or more fields (the so called "__rkwargs__") in addition
+          to the classic SymPy arguments.
+        * An `__rkwargs__`, in turn, may or may not be a SymPy object.
+
+    The user may then use this registry to register callbacks to be used when
+    one such Reconstructable object is encountered.
     """
-    numbers, others = split(args, lambda i: i.is_Number)
-    if len(numbers) > 1:
-        args[:] = [expr.func(*numbers)] + others
+
+    def register(self, cls, rkwargs_callback_mapper=None):
+        self.append(cls)
+        _uxreplace_handle.register(cls, _uxreplace_handle_reconstructable)
+
+        for kls, callback in (rkwargs_callback_mapper or {}).items():
+            _uxreplace_dispatch.register(kls, callback)
+
+    def dispatchable(self, obj):
+        return isinstance(obj, tuple(self))
+
+
+_uxreplace_registry = UxreplaceRegistry()
+_uxreplace_registry.register(Eq)
+_uxreplace_registry.register(DefFunction)
+_uxreplace_registry.register(ComponentAccess)
 
 
 class Uxmapper(dict):
@@ -169,6 +258,20 @@ class Uxmapper(dict):
             self[base] = self.extracted[base] = make()
 
 
+def subs_if_composite(expr, subs):
+    """
+    Call `expr.subs(subs)` if `subs` contain composite expressions, that is
+    expressions that can be part of larger expressions of the same type (e.g.,
+    `a*b` could be part of `a*b*c`, while `a[1]` cannot be part of a "larger
+    Indexed"). Instead, if `subs` consists of just "primitive" expressions, then
+    resort to the much faster `uxreplace`.
+    """
+    if all(isinstance(i, (Indexed, IndexDerivative)) for i in subs):
+        return uxreplace(expr, subs)
+    else:
+        return expr.subs(subs)
+
+
 def xreplace_indices(exprs, mapper, key=None):
     """
     Replace array indices in expressions.
@@ -195,63 +298,79 @@ def xreplace_indices(exprs, mapper, key=None):
     return replaced if isinstance(exprs, Iterable) else replaced[0]
 
 
+def _eval_numbers(expr, args):
+    """
+    Helper function for in-place reduction of the expr arguments.
+    """
+    numbers, others = split(args, lambda i: i.is_Number)
+    if len(numbers) > 1:
+        if isinstance(expr, UnevaluableMixin):
+            cls = expr.func.__base__
+        else:
+            cls = expr.func
+        args[:] = [cls(*numbers)] + others
+
+
+def flatten_args(args, op, ignore=None):
+    """
+    Flatten the arguments of type `op` in `args`.
+
+    Examples
+    --------
+    * (a+b)+c -> a+b+c
+    * (a*b)*c -> a*b*c
+    * (a+b)*c -> (a+b)*c
+    """
+    if ignore is not None and any(isinstance(a, ignore) for a in args):
+        return args
+
+    key = lambda e: isinstance(e, op)
+    nested, others = split(args, key)
+
+    return flatten(e.args for e in nested) + list(others)
+
+
 def pow_to_mul(expr):
     if q_leaf(expr) or isinstance(expr, Basic):
         return expr
     elif expr.is_Pow:
         base, exp = expr.as_base_exp()
-        if exp > 10 or exp < -10 or int(exp) != exp or exp == 0:
-            # Large and non-integer powers remain untouched
+        try:
+            int(exp)
+        except TypeError:
+            # E.g., a Symbol, or possibly a generic expression
             return expr
-        elif exp == -1:
-            # Reciprocals also remain untouched, but we traverse the base
-            # looking for other Pows
+        if exp > 10 or exp < -10 or exp == 0:
+            # Large powers remain untouched
+            return expr
+        elif (int(exp) - exp != 0):
+            # Fractional powers also remain untouched,
+            # but at least we traverse the base looking for other Pows
             return expr.func(pow_to_mul(base), exp, evaluate=False)
         elif exp > 0:
-            return Mul(*[base]*int(exp), evaluate=False)
+            return UnevalMul(*[pow_to_mul(base)]*int(exp), evaluate=False)
+        elif exp < 0:
+            # Reciprocal powers become inverse of the negative power
+            # for example Pow(expr, -2) becomes Pow(expr * expr, -1)
+            return expr.func(pow_to_mul(base**(-int(exp))), -1, evaluate=False)
         else:
-            # SymPy represents 1/x as Pow(x,-1). Also, it represents
-            # 2/x as Mul(2, Pow(x, -1)). So we shouldn't end up here,
-            # but just in case SymPy changes its internal conventions...
-            posexpr = Mul(*[base]*(-int(exp)), evaluate=False)
-            return Pow(posexpr, -1, evaluate=False)
+            # Default. We should not end up here as all cases are handled
+            return expr
     else:
-        return expr.func(*[pow_to_mul(i) for i in expr.args], evaluate=False)
+        args = [pow_to_mul(i) for i in expr.args]
 
+        # Some SymPy versions will evaluate the two-args case
+        # `(negative integer, mul)` despite the `evaluate=False`. For example,
+        # `Mul(-2, 1/a*a, evaluate=False)` gets evaluated to `-2/a**2`. By swapping
+        # the args, the issue disappears...
+        try:
+            a0, a1 = args
+            if a0.is_Number and a0 < 0 and not q_leaf(a1):
+                args = [a1, a0]
+        except ValueError:
+            pass
 
-AffineFunction = namedtuple("AffineFunction", "var, coeff, shift")
-
-
-def split_affine(expr):
-    """
-    Split an affine scalar function into its three components, namely variable,
-    coefficient, and translation from origin.
-
-    Raises
-    ------
-    ValueError
-        If ``expr`` is non affine.
-    """
-    if expr.is_Number:
-        return AffineFunction(None, None, expr)
-
-    # Handle super-quickly the calls like `split_affine(x+1)`, which are
-    # the majority.
-    if expr.is_Add and len(expr.args) == 2:
-        if expr.args[0].is_Number and expr.args[1].is_Symbol:
-            # SymPy deterministically orders arguments -- first numbers, then symbols
-            return AffineFunction(expr.args[1], 1, expr.args[0])
-
-    # Fallback
-    poly = expr.as_poly()
-    try:
-        if not (poly.is_univariate and poly.is_linear) or not LM(poly).is_Symbol:
-            raise ValueError
-    except AttributeError:
-        # `poly` might be None
-        raise ValueError
-
-    return AffineFunction(LM(poly), LC(poly), poly.TC())
+        return expr.func(*args, evaluate=False)
 
 
 def indexify(expr):
@@ -279,6 +398,22 @@ def subs_op_args(expr, args):
     return expr.subs({i.name: args[i.name] for i in expr.free_symbols if i.name in args})
 
 
+def normalize_args(args):
+    """
+    Produce a new `args` dictionary in which only the actually substitutable
+    arguments are retained. This ensures that none of the subsequent substitutions
+    will throw `SympifyError` exceptions.
+    """
+    retval = {}
+    for k, v in args.items():
+        try:
+            retval[k] = sympify(v, strict=True)
+        except (TypeError, SympifyError):
+            continue
+
+    return retval
+
+
 def reuse_if_untouched(expr, args, evaluate=False):
     """
     Reconstruct `expr` iff any of the provided `args` is different than
@@ -293,8 +428,8 @@ def reuse_if_untouched(expr, args, evaluate=False):
 def evalrel(func=min, input=None, assumptions=None):
     """
     The purpose of this function is two-fold: (i) to reduce the `input` candidates of a
-    for a MIN/MAX expression based on the given `assumptions` and (ii) return the nested
-    MIN/MAX expression of the reduced-size input.
+    for a Min/Max expression based on the given `assumptions` and (ii) return the nested
+    Min/Max expression of the reduced-size input.
 
     Parameters
     ----------
@@ -317,7 +452,7 @@ def evalrel(func=min, input=None, assumptions=None):
     >>> c = Symbol('c')
     >>> d = Symbol('d')
     >>> evalrel(max, [a, b, c, d], [Le(d, a), Ge(c, b)])
-    MAX(a, c)
+    Max(a, c)
     """
     sfunc = (Min if func is min else Max)  # Choose SymPy's Min/Max
 
@@ -363,7 +498,7 @@ def evalrel(func=min, input=None, assumptions=None):
     mapper = transitive_closure(mapper)
     input = [i.subs(mapper) for i in input]
 
-    # Explore simplification opportunities that may have emerged and generate MIN/MAX
+    # Explore simplification opportunities that may have emerged and generate Min/Max
     # expression
     try:
         exp = sfunc(*input)  # Can it be evaluated or simplified?
@@ -377,3 +512,29 @@ def evalrel(func=min, input=None, assumptions=None):
     except TypeError:
         pass
     return rfunc(func, *input)
+
+
+uneval_mapper = {Add: UnevalAdd, Mul: UnevalMul, Pow: UnevalPow}
+
+
+def unevaluate(expr):
+    if q_leaf(expr):
+        return expr
+
+    args = [unevaluate(a) for a in expr.args]
+
+    try:
+        return uneval_mapper[expr.func](*args)
+    except KeyError:
+        return reuse_if_untouched(expr, args)
+
+
+def as_long(expr):
+    """
+    Convert an expression and its symbolic args to a long integer.
+    """
+    try:
+        syms = retrieve_symbols(expr)
+        return expr.subs({s: LONG(s) for s in syms})
+    except AttributeError:
+        return LONG(expr)
