@@ -1,21 +1,19 @@
-import ctypes
 from collections import OrderedDict
 from collections.abc import Iterable
-from functools import reduce
+from functools import reduce, wraps
 from itertools import chain, combinations, groupby, product, zip_longest
 from operator import attrgetter, mul
 import types
 
 import numpy as np
 import sympy
-from cgen import Struct, Value, dtype_to_ctype as cgen_dtype_to_ctype
 
-__all__ = ['prod', 'as_tuple', 'is_integer', 'generator', 'grouper', 'split', 'roundm',
-           'powerset', 'invert', 'flatten', 'single_or', 'filter_ordered', 'as_mapper',
-           'filter_sorted', 'dtype_to_cstr', 'dtype_to_ctype', 'dtype_to_mpitype',
-           'ctypes_to_cstr', 'ctypes_to_cgen', 'pprint', 'sweep', 'all_equal', 'as_list',
+__all__ = ['prod', 'as_tuple', 'is_integer', 'generator', 'grouper', 'split',
+           'roundm', 'powerset', 'invert', 'flatten', 'single_or', 'filter_ordered',
+           'as_mapper', 'filter_sorted', 'pprint', 'sweep', 'all_equal', 'as_list',
            'indices_to_slices', 'indices_to_sections', 'transitive_closure',
-           'humanbytes', 'c_restrict_void_p']
+           'humanbytes', 'contains_val', 'sorted_priority', 'as_set', 'is_number',
+           'smart_lt', 'smart_gt']
 
 
 def prod(iterable, initial=1):
@@ -29,6 +27,13 @@ def as_list(item, type=None, length=None):
     return list(as_tuple(item, type=type, length=length))
 
 
+def as_set(iterable, type=None, length=None):
+    """
+    Force item to a set.
+    """
+    return set(as_tuple(iterable, type=type, length=length))
+
+
 def as_tuple(item, type=None, length=None):
     """
     Force item to a tuple. Passes tuple subclasses through also.
@@ -78,6 +83,20 @@ def is_integer(value):
     return isinstance(value, (int, np.integer, sympy.Integer))
 
 
+def is_number(value):
+    """
+    A thorough instance comparison for all number types.
+    """
+    return isinstance(value, (int, float, np.number, sympy.Number))
+
+
+def contains_val(val, items):
+    try:
+        return val in items
+    except TypeError:
+        return val == items
+
+
 def generator():
     """
     Return a function ``f`` that generates integer numbers starting at 0
@@ -162,23 +181,15 @@ def filter_ordered(elements, key=None):
     key : callable, optional
         Conversion key used during equality comparison.
     """
+    # This method exploits the fact that dictionary keys are unique and ordered
+    # (since Python 3.7). It's concise and often faster for larger lists
     if isinstance(elements, types.GeneratorType):
         elements = list(elements)
-    seen = set()
+
     if key is None:
-        try:
-            unordered, inds = np.unique(elements, return_index=True)
-            return unordered[np.argsort(inds)].tolist()
-        except:
-            return sorted(list(set(elements)), key=elements.index)
+        return list(dict.fromkeys(elements))
     else:
-        ret = []
-        for e in elements:
-            k = key(e)
-            if k not in seen:
-                ret.append(e)
-                seen.add(k)
-        return ret
+        return list(dict(zip([key(i) for i in elements], elements)).values())
 
 
 def filter_sorted(elements, key=None):
@@ -191,134 +202,6 @@ def filter_sorted(elements, key=None):
     return sorted(filter_ordered(elements, key=key), key=key)
 
 
-def dtype_to_cstr(dtype):
-    """Translate numpy.dtype into C string."""
-    return cgen_dtype_to_ctype(dtype)
-
-
-def dtype_to_ctype(dtype):
-    """Translate numpy.dtype into a ctypes type."""
-    if issubclass(dtype, ctypes._SimpleCData):
-        # Bypass np.ctypeslib's normalization rules such as
-        # `np.ctypeslib.as_ctypes_type(ctypes.c_void_p) -> ctypes.c_ulong`
-        return dtype
-    else:
-        return np.ctypeslib.as_ctypes_type(dtype)
-
-
-def dtype_to_mpitype(dtype):
-    """Map numpy types to MPI datatypes."""
-    return {np.ubyte: 'MPI_BYTE',
-            np.ushort: 'MPI_UNSIGNED_SHORT',
-            np.int32: 'MPI_INT',
-            np.float32: 'MPI_FLOAT',
-            np.int64: 'MPI_LONG',
-            np.float64: 'MPI_DOUBLE'}[dtype]
-
-
-def ctypes_to_cstr(ctype, toarray=None, qualifiers=None):
-    """Translate ctypes types into C strings."""
-    if issubclass(ctype, ctypes.Structure):
-        retval = 'struct %s' % ctype.__name__
-    elif issubclass(ctype, ctypes.Union):
-        retval = 'union %s' % ctype.__name__
-    elif issubclass(ctype, ctypes._Pointer):
-        if toarray:
-            retval = ctypes_to_cstr(ctype._type_, '(* %s)' % toarray)
-        else:
-            retval = ctypes_to_cstr(ctype._type_)
-            if issubclass(ctype._type_, ctypes._Pointer):
-                # Look-ahead to avoid extra ugly spaces
-                retval = '%s*' % retval
-            else:
-                retval = '%s *' % retval
-    elif issubclass(ctype, ctypes.Array):
-        retval = '%s[%d]' % (ctypes_to_cstr(ctype._type_, toarray), ctype._length_)
-    elif ctype.__name__.startswith('c_'):
-        name = ctype.__name__[2:]
-
-        # A primitive datatype
-        # FIXME: Is there a better way of extracting the C typename ?
-        # Here, we're following the ctypes convention that each basic type has
-        # either the format c_X_p or c_X, where X is the C typename, for instance
-        # `int` or `float`.
-        if name.endswith('_p'):
-            name = name.split('_')[-2]
-            suffix = '*'
-        elif toarray:
-            suffix = toarray
-        else:
-            suffix = None
-
-        if name.startswith('u'):
-            name = name[1:]
-            prefix = 'unsigned'
-        else:
-            prefix = None
-
-        # Special cases
-        if name == 'byte':
-            name = 'char'
-
-        retval = name
-        if prefix:
-            retval = '%s %s' % (prefix, retval)
-        if suffix:
-            retval = '%s %s' % (retval, suffix)
-    else:
-        # A custom datatype (e.g., a typedef-ed pointer to struct)
-        retval = ctype.__name__
-
-    # Add qualifiers, if any
-    if qualifiers:
-        retval = '%s %s' % (' '.join(qualifiers), retval)
-
-    return retval
-
-
-class c_restrict_void_p(ctypes.c_void_p):
-    pass
-
-
-def ctypes_to_cgen(ctype, fields=None):
-    """
-    Create a cgen.Structure off a ctypes.Struct, or (possibly nested) ctypes.Pointer
-    to ctypes.Struct; return None in all other cases.
-    """
-    if issubclass(ctype, ctypes._Pointer):
-        return ctypes_to_cgen(ctype._type_, fields=fields)
-    elif not issubclass(ctype, ctypes.Structure):
-        return None
-
-    # Sanity check
-    # If a thorough description of the struct fields was required (that is, one
-    # symbol per field, each symbol with its own qualifiers), then the number
-    # of fields much coincide with that of the `ctype`
-    if fields is None:
-        fields = (None,)*len(ctype._fields_)
-    else:
-        assert len(fields) == len(ctype._fields_)
-
-    entries = []
-    for i, (n, ct) in zip(fields, ctype._fields_):
-        try:
-            cstr = i._C_typename
-
-            # Certain qualifiers are to be removed as meaningless within a struct
-            degenerate_quals = ('const',)
-            cstr = ' '.join([j for j in cstr.split() if j not in degenerate_quals])
-        except AttributeError:
-            cstr = ctypes_to_cstr(ct)
-
-        # All struct pointers are by construction restrict
-        if ct is c_restrict_void_p:
-            cstr = '%srestrict' % cstr
-
-        entries.append(Value(cstr, n))
-
-    return Struct(ctype.__name__, entries)
-
-
 def pprint(node, verbose=True):
     """
     Shortcut to pretty print Iteration/Expression trees.
@@ -430,6 +313,69 @@ def humanbytes(B):
     elif MB <= B < GB:
         return '%d MB' % round(B / MB)
     elif GB <= B < TB:
-        return '%d GB' % round(B / GB)
+        return '%.1f GB' % round(B / GB, 1)
     elif TB <= B:
-        return '%d TB' % round(B / TB)
+        return '%.2f TB' % round(B / TB, 1)
+
+
+def sorted_priority(items, priority):
+    """
+    Sort items based on their type priority.
+
+    Rules:
+
+        * Each type has an integer priority.
+        * Types with higher priority precede types with lower priority.
+        * Types with same priority are sorted based on the type name.
+        * Types with unknown priority are given 0-priority.
+
+    Parameters
+    ----------
+    items : iterable
+        The objects to be sorted.
+    priority : dict
+        A dictionary from types to integer values.
+    """
+
+    def key(i):
+        for cls in sorted(priority, key=priority.get, reverse=True):
+            if isinstance(i, cls):
+                v = priority[cls]
+                break
+        else:
+            v = 0
+        return (v, str(type(i)))
+
+    return sorted(items, key=key, reverse=True)
+
+
+def avoid_symbolic(default=None):
+    """
+    Decorator to avoid calling a function where doing so will result in symbolic
+    computation being performed. For use if symbolic computation may be slow. In
+    the case that an arg is symbolic, just give up and return a default value.
+    """
+    def _avoid_symbolic(func):
+        @wraps(func)
+        def wrapper(*args):
+            if any(isinstance(expr, sympy.Basic) for expr in args):
+                # An argument is symbolic, so give up and assume default
+                return default
+
+            return func(*args)
+
+        return wrapper
+
+    return _avoid_symbolic
+
+
+@avoid_symbolic(default=False)
+def smart_lt(a, b):
+    """An Lt that gives up and returns False if supplied a symbolic argument"""
+    return a < b
+
+
+@avoid_symbolic(default=False)
+def smart_gt(a, b):
+    """A Gt that gives up and returns False if supplied a symbolic argument"""
+    return a > b
