@@ -21,7 +21,7 @@ class AugmentedWeakRef(weakref.ref):
         return obj
 
 
-class Uncached(object):
+class Uncached:
 
     """
     Mixin class for unique, and therefore uncached, symbolic objects
@@ -32,7 +32,7 @@ class Uncached(object):
         return id(self)
 
 
-class Cached(object):
+class Cached:
 
     """
     Mixin class for cached symbolic objects.
@@ -138,7 +138,7 @@ class Cached(object):
         return {}
 
 
-class CacheManager(object):
+class CacheManager:
 
     """
     Drop unreferenced objects from the SymPy and Devito caches. The associated
@@ -168,16 +168,22 @@ class CacheManager(object):
 
         # Wipe out the hidden module-private SymPy caches
         sympy.polys.rootoftools.ComplexRootOf.clear_cache()
-        sympy.polys.rings._ring_cache.clear()
-        sympy.polys.fields._field_cache.clear()
         sympy.polys.domains.modularinteger._modular_integer_cache.clear()
+        try:
+            sympy.polys.rings._ring_cache.clear()
+            sympy.polys.fields._field_cache.clear()
+        except AttributeError:
+            # SymPy 1.14 and later
+            pass
 
         # Take a copy of the dictionary so we can safely iterate over it
         # even if another thread is making changes
         cache_copied = safe_dict_copy(_SymbolCache)
 
         # Maybe trigger garbage collection
-        if force is False:
+        if force:
+            gc.collect()
+        else:
             if cls.ncalls_w_force_false + 1 == cls.force_ths:
                 # Case 1: too long since we called gc.collect, let's do it now
                 gc.collect()
@@ -189,8 +195,6 @@ class CacheManager(object):
             else:
                 # We won't call gc.collect() this time
                 cls.ncalls_w_force_false += 1
-        else:
-            gc.collect()
 
         for key in cache_copied:
             obj = _SymbolCache.get(key)
@@ -198,6 +202,10 @@ class CacheManager(object):
                 # deleted by another thread since we took the copy
                 continue
             if obj() is None:
-                # pop(x, None) does not error if already gone
                 # (key could be removed in another thread since get() above)
                 _SymbolCache.pop(key, None)
+
+        # Maybe trigger garbage collection
+        if force:
+            del cache_copied
+            gc.collect()
