@@ -1,22 +1,26 @@
 from abc import ABC
 from collections import namedtuple
-from math import floor
+from functools import cached_property
+from itertools import product
 
 import numpy as np
 from sympy import prod
-from cached_property import cached_property
 
-from devito.data import LEFT, RIGHT
+from devito import configuration
+from devito.data import LEFT, RIGHT, CENTER
 from devito.logger import warning
-from devito.mpi import Distributor, MPI
-from devito.tools import ReducerMap, as_tuple
+from devito.mpi import Distributor, MPI, SubDistributor
+from devito.tools import ReducerMap, as_tuple, frozendict
 from devito.types.args import ArgProvider
 from devito.types.basic import Scalar
 from devito.types.dense import Function
+from devito.types.utils import DimensionTuple
 from devito.types.dimension import (Dimension, SpaceDimension, TimeDimension,
-                                    SteppingDimension, SubDimension)
+                                    Spacing, SteppingDimension, SubDimension,
+                                    MultiSubDimension, DefaultDimension)
+from devito.deprecations import deprecations
 
-__all__ = ['Grid', 'SubDomain', 'SubDomainSet']
+__all__ = ['Grid', 'SubDomain', 'SubDomainSet', 'Border']
 
 
 GlobalLocal = namedtuple('GlobalLocal', 'glb loc')
@@ -29,6 +33,10 @@ class CartesianDiscretization(ABC):
     physical domains by congruent parallelotopes (e.g., "tiles" or "bricks").
     """
 
+    is_Grid = False
+    is_SubDomain = False
+    is_MultiSubDomain = False
+
     def __init__(self, shape=None, dimensions=None, dtype=None):
         self._shape = as_tuple(shape)
         self._dimensions = as_tuple(dimensions)
@@ -37,7 +45,7 @@ class CartesianDiscretization(ABC):
     @property
     def shape(self):
         """Shape of the physical domain."""
-        return self._shape
+        return DimensionTuple(*self._shape, getters=self.dimensions)
 
     @property
     def dimensions(self):
@@ -56,6 +64,10 @@ class CartesianDiscretization(ABC):
         """
         return self._dtype
 
+    @property
+    def root(self):
+        return self
+
 
 class Grid(CartesianDiscretization, ArgProvider):
 
@@ -67,21 +79,19 @@ class Grid(CartesianDiscretization, ArgProvider):
     ----------
     shape : tuple of ints
         Shape of the computational domain in grid points.
-    extent : tuple of floats, optional
-        Physical extent of the domain in m; defaults to a unit box of extent 1m
-        in all dimensions.
-    origin : tuple of floats, optional
-        Physical coordinate of the origin of the domain; defaults to 0.0 in all
-        dimensions.
+    extent : tuple of values interpretable as dtype, default=unit box of extent 1m
+    in all dimensions.
+        Physical extent of the domain in m.
+    origin : tuple of values interpretable as dtype, default=0.0 in all dimensions
+        Physical coordinate of the origin of the domain.
     dimensions : tuple of SpaceDimension, optional
         The dimensions of the computational domain encapsulated by this Grid.
     time_dimension : TimeDimension, optional
         The dimension used to define time in a `TimeFunction` created from
         this Grid.
-    dtype : data-type, optional
+    dtype : data-type, default=np.float32
         Any object that can be interpreted as a numpy data type, used as default
         data type to be inherited by all Functions created from this Grid.
-        Defaults to ``np.float32``.
     subdomains : tuple of SubDomain, optional
         If no subdomains are specified, the Grid only defines the two default
         subdomains ``interior`` and ``domain``.
@@ -135,6 +145,8 @@ class Grid(CartesianDiscretization, ArgProvider):
              v
     """
 
+    is_Grid = True
+
     _default_dimensions = ('x', 'y', 'z')
 
     def __init__(self, shape, extent=None, origin=None, dimensions=None,
@@ -147,7 +159,7 @@ class Grid(CartesianDiscretization, ArgProvider):
             ndim = len(shape)
             assert ndim <= 3
             dim_names = self._default_dimensions[:ndim]
-            dim_spacing = tuple(Scalar(name='h_%s' % n, dtype=dtype, is_const=True)
+            dim_spacing = tuple(Spacing(name='h_%s' % n, dtype=dtype, is_const=True)
                                 for n in dim_names)
             dimensions = tuple(SpaceDimension(name=n, spacing=s)
                                for n, s in zip(dim_names, dim_spacing))
@@ -165,18 +177,26 @@ class Grid(CartesianDiscretization, ArgProvider):
 
         # Create a Distributor, used internally to implement domain decomposition
         # by all Functions defined on this Grid
-        self._distributor = Distributor(shape, dimensions, comm, topology)
+        topology = topology or configuration['topology']
+        if topology:
+            if len(topology) == len(self.shape):
+                self._topology = topology
+            else:
+                warning("Ignoring the provided topology `%s` as it "
+                        "is incompatible with the grid shape `%s`" %
+                        (topology, self.shape))
+                self._topology = None
+        else:
+            self._topology = None
+        self._distributor = Distributor(shape, dimensions, comm, self._topology)
 
         # The physical extent
-        self._extent = as_tuple(extent or tuple(1. for _ in self.shape))
+        extent = as_tuple(extent or tuple(1. for _ in self.shape))
+        self._extent = tuple(dtype(e) for e in extent)
 
-        # Initialize SubDomains
-        subdomains = tuple(i for i in (Domain(), Interior(), *as_tuple(subdomains)))
-        for counter, i in enumerate(subdomains):
-            i.__subdomain_finalize__(self, counter=counter)
-        self._subdomains = subdomains
-
-        self._origin = as_tuple(origin or tuple(0. for _ in self.shape))
+        # The origin of the grid
+        origin = as_tuple(origin or tuple(0. for _ in self.shape))
+        self._origin = tuple(dtype(o) for o in origin)
         self._origin_symbols = tuple(Scalar(name='o_%s' % d.name, dtype=dtype,
                                             is_const=True)
                                      for d in self.dimensions)
@@ -196,6 +216,13 @@ class Grid(CartesianDiscretization, ArgProvider):
         else:
             raise ValueError("`time_dimension` must be None or of type TimeDimension")
 
+        # Initialize SubDomains for legacy interface
+        if subdomains is not None:
+            deprecations.subdomain_warn
+        self._subdomains = tuple(i for i in (Domain(), Interior(), *as_tuple(subdomains)))
+        for i in self._subdomains:
+            i.__subdomain_finalize_legacy__(self)
+
     def __repr__(self):
         return "Grid[extent=%s, shape=%s, dimensions=%s]" % (
             self.extent, self.shape, self.dimensions
@@ -222,11 +249,17 @@ class Grid(CartesianDiscretization, ArgProvider):
         return dict(zip(self.origin_symbols, self.origin))
 
     @property
-    def origin_offset(self):
-        """Offset of the local (per-process) origin from the domain origin."""
+    def origin_ioffset(self):
+        """Offset index of the local (per-process) origin from the domain origin."""
         grid_origin = [min(i) for i in self.distributor.glb_numb]
         assert len(grid_origin) == len(self.spacing)
-        return tuple(i*h for i, h in zip(grid_origin, self.spacing))
+        return DimensionTuple(*grid_origin, getters=self.dimensions)
+
+    @property
+    def origin_offset(self):
+        """Physical offset of the local (per-process) origin from the domain origin."""
+        return DimensionTuple(*[i*h for i, h in zip(self.origin_ioffset, self.spacing)],
+                              getters=self.dimensions)
 
     @property
     def time_dim(self):
@@ -289,14 +322,19 @@ class Grid(CartesianDiscretization, ArgProvider):
         return self._distributor.shape
 
     @property
-    def dimension_map(self):
+    def size_map(self):
         """Map between SpaceDimensions and their global/local size."""
         return {d: GlobalLocal(g, l)
                 for d, g, l in zip(self.dimensions, self.shape, self.shape_local)}
 
+    @property
+    def topology(self):
+        """The topology used for decomposing the CartesianDiscretization."""
+        return self._topology
+
     @property
     def distributor(self):
-        """The Distributor used for MPI-decomposing the CartesianDiscretization."""
+        """The Distributor used for decomposing the CartesianDiscretization."""
         return self._distributor
 
     @property
@@ -329,7 +367,7 @@ class Grid(CartesianDiscretization, ArgProvider):
         args = ReducerMap()
 
         # Dimensions size
-        for k, v in self.dimension_map.items():
+        for k, v in self.size_map.items():
             args.update(k._arg_defaults(_min=0, size=v.loc))
 
         # Dimensions spacing
@@ -372,6 +410,8 @@ class AbstractSubDomain(CartesianDiscretization):
     Abstract base class for subdomains.
     """
 
+    is_SubDomain = True
+
     name = None
     """A unique name for the SubDomain."""
 
@@ -382,7 +422,13 @@ class AbstractSubDomain(CartesianDiscretization):
         # All other attributes get initialized upon `__subdomain_finalize__`
         super().__init__()
 
-    def __subdomain_finalize__(self, grid, **kwargs):
+        self._distributor = None
+
+        self._grid = kwargs.get('grid')
+        if self.grid:
+            self.__subdomain_finalize__()
+
+    def __subdomain_finalize__(self):
         """
         Finalize the subdomain initialization.
 
@@ -392,6 +438,18 @@ class AbstractSubDomain(CartesianDiscretization):
         """
         raise NotImplementedError
 
+    def __subdomain_finalize_legacy__(self, grid):
+        """
+        Finalize the subdomain initialization.
+
+        (Backward-compatible version for legacy SubDomain API)
+
+        Notes
+        -----
+        Must be overridden by subclasses.
+        """
+        raise NotImplementedError
+
     def __eq__(self, other):
         if not isinstance(other, AbstractSubDomain):
             return False
@@ -430,6 +488,55 @@ class AbstractSubDomain(CartesianDiscretization):
     def dimension_map(self):
         return {d.root: d for d in self.dimensions}
 
+    @property
+    def grid(self):
+        return self._grid
+
+    @property
+    def root(self):
+        return self.grid.root
+
+    @property
+    def time_dim(self):
+        if self.grid:
+            return self.grid.time_dim
+        raise AttributeError("SubDomain has no Grid, and thus no time dimension")
+
+    @property
+    def stepping_dim(self):
+        if self.grid:
+            return self.grid.stepping_dim
+        raise AttributeError("SubDomain has no Grid, and thus no stepping dimension")
+
+    @property
+    def distributor(self):
+        """The Distributor used for MPI-decomposing the CartesianDiscretization."""
+        return self._distributor
+
+    def is_distributed(self, dim):
+        """
+        True if `dim` is a distributed Dimension for this CartesianDiscretization,
+        False otherwise.
+        """
+        if self.grid:
+            return any(dim is d for d in self.distributor.dimensions)
+        return False
+
+    @property
+    def comm(self):
+        """The MPI communicator inherited from the distributor."""
+        if self.grid:
+            return self.grid.comm
+        raise ValueError("`SubDomain` %s has no `Grid` attached and thus no `comm`"
+                         % self.name)
+
+    def _arg_values(self, **kwargs):
+        try:
+            return self.grid._arg_values(**kwargs)
+        except AttributeError:
+            raise AttributeError("%s is not attached to a Grid and has no _arg_values"
+                                 % self)
+
 
 class SubDomain(AbstractSubDomain):
 
@@ -475,13 +582,27 @@ class SubDomain(AbstractSubDomain):
     --------
     Domain : An example of preset SubDomain.
     Interior : An example of preset Subdomain.
+
+    Notes
+    -----
+    SubDomains are the only way to harness the benefits of domain decomposition,
+    especially when defining BCs.
     """
 
-    def __subdomain_finalize__(self, grid, **kwargs):
+    def __subdomain_finalize__(self):
+        self.__subdomain_finalize_legacy__(self.grid)
+        self._distributor = SubDistributor(self)
+
+        # Intervals of form Interval(n, n) automatically become FiniteSet
+        # Add one to end as intervals are in terms of indices (inclusive of endpoints)
+        # Empty interval corresponds to a size of zero
+        self._shape_local = tuple(0 if i.is_empty else i.end-i.start + 1 if i.is_Interval
+                                  else 1 for i in self.distributor.intervals)
+
+    def __subdomain_finalize_legacy__(self, grid):
         # Create the SubDomain's SubDimensions
         sub_dimensions = []
         sdshape = []
-        counter = kwargs.get('counter', 0) - 1
         for k, v, s in zip(self.define(grid.dimensions).keys(),
                            self.define(grid.dimensions).values(), grid.shape):
             if isinstance(v, Dimension):
@@ -490,40 +611,40 @@ class SubDomain(AbstractSubDomain):
             else:
                 try:
                     # Case ('middle', int, int)
-                    side, thickness_left, thickness_right = v
+                    side, ltkn, rtkn = v
                     if side != 'middle':
-                        raise ValueError("Expected side 'middle', not `%s`" % side)
-                    sub_dimensions.append(SubDimension.middle('i%d%s' %
-                                                              (counter, k.name),
-                                                              k, thickness_left,
-                                                              thickness_right))
-                    thickness = s-thickness_left-thickness_right
+                        raise ValueError(f"Expected side 'middle', not `{side}`")
+                    sub_dimensions.append(SubDimension.middle(f'i{k.name}',
+                                                              k, ltkn, rtkn))
+                    thickness = s-ltkn-rtkn
                     sdshape.append(thickness)
                 except ValueError:
                     side, thickness = v
-                    if side == 'left':
-                        if s-thickness < 0:
-                            raise ValueError("Maximum thickness of dimension %s "
-                                             "is %d, not %d" % (k.name, s, thickness))
-                        sub_dimensions.append(SubDimension.left('i%d%s' %
-                                                                (counter, k.name),
-                                                                k, thickness))
-                        sdshape.append(thickness)
-                    elif side == 'right':
-                        if s-thickness < 0:
-                            raise ValueError("Maximum thickness of dimension %s "
-                                             "is %d, not %d" % (k.name, s, thickness))
-                        sub_dimensions.append(SubDimension.right('i%d%s' %
-                                                                 (counter, k.name),
-                                                                 k, thickness))
-                        sdshape.append(thickness)
-                    else:
-                        raise ValueError("Expected sides 'left|right', not `%s`" % side)
+                    constructor = {'left': SubDimension.left,
+                                   'right': SubDimension.right}.get(side, None)
+                    if constructor is None:
+                        raise ValueError(f"Expected sides 'left|right', not `{side}`")
+
+                    if s - thickness < 0:
+                        raise ValueError(f"Maximum thickness of dimension {k.name} "
+                                         f"is {s}, not {thickness}")
+                    sub_dimensions.append(constructor(f'i{k.name}', k, thickness))
+                    sdshape.append(thickness)
 
         self._shape = tuple(sdshape)
         self._dimensions = tuple(sub_dimensions)
         self._dtype = grid.dtype
 
+    @property
+    def shape_local(self):
+        return self._shape_local
+
+    @property
+    def size_map(self):
+        """Map between SpaceDimensions and their global/local size."""
+        return {d: GlobalLocal(g, l)
+                for d, g, l in zip(self.dimensions, self.shape, self.shape_local)}
+
     def define(self, dimensions):
         """
         Parametrically describe the SubDomain w.r.t. a generic Grid.
@@ -535,39 +656,33 @@ class SubDomain(AbstractSubDomain):
         """
         raise NotImplementedError
 
-
-class MultiSubDimension(SubDimension):
-
-    """
-    A special SubDimension for graceful lowering of MultiSubDomains.
-    """
-
-    def __init_finalize__(self, name, parent, msd):
-        # NOTE: a MultiSubDimension stashes a reference to the originating MultiSubDomain.
-        # This creates a circular pattern as the `msd` itself carries references to
-        # its MultiSubDimensions. This is currently necessary because during compilation
-        # we drop the MultiSubDomain early, but when the MultiSubDimensions are processed
-        # we still need it to create the implicit equations. Untangling this is
-        # definitely possible, but not straightforward
-        self.msd = msd
-
-        lst, rst = self._symbolic_thickness(name)
-        left = parent.symbolic_min + lst
-        right = parent.symbolic_max - rst
-
-        super().__init_finalize__(name, parent, left, right, ((lst, 0), (rst, 0)), False)
-
-    def __hash__(self):
-        # There is no possibility for two MultiSubDimensions to ever hash the same, since
-        # a MultiSubDimension carries a reference to a MultiSubDomain, which is unique
-        return id(self)
-
     @cached_property
-    def bound_symbols(self):
-        # Unlike a SubDimension, a MultiSubDimension does *not* bind its thickness,
-        # which is rather bound by an Operation (which can alter its value
-        # dynamically so as to implement a MultiSubDomain)
-        return self.parent.bound_symbols
+    def _arg_names(self):
+        try:
+            ret = self.grid._arg_names
+        except AttributeError:
+            msg = f"{self} is not attached to a Grid and has no _arg_names"
+            raise AttributeError(msg)
+
+        # Names for SubDomain thicknesses
+        thickness_names = tuple([k.name for k in d._thickness_map]
+                                for d in self.dimensions if d.is_Sub)
+
+        ret += tuple(thickness_names)
+
+        return ret
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        # A SubDistributor wraps an MPI communicator, which can't and shouldn't be pickled
+        state.pop('_distributor', None)
+        return state
+
+    def __setstate__(self, state):
+        for k, v in state.items():
+            setattr(self, k, v)
+        if self.grid:
+            self._distributor = SubDistributor(self)
 
 
 class MultiSubDomain(AbstractSubDomain):
@@ -576,6 +691,8 @@ class MultiSubDomain(AbstractSubDomain):
     Abstract base class for types representing groups of SubDomains.
     """
 
+    is_MultiSubDomain = True
+
     def __hash__(self):
         # There is no possibility for two MultiSubDomains to ever hash the same since
         # they are by construction unique and different from each other
@@ -702,10 +819,9 @@ class SubDomainSet(MultiSubDomain):
     """
 
     def __init__(self, **kwargs):
-        super().__init__()
-
         self._n_domains = kwargs.get('N', 1)
         self._global_bounds = kwargs.get('bounds', None)
+        super().__init__(**kwargs)
 
         try:
             self.implicit_dimension
@@ -714,15 +830,9 @@ class SubDomainSet(MultiSubDomain):
         except AttributeError:
             pass
 
-    def __subdomain_finalize__(self, grid, counter=0, **kwargs):
-        self._grid = grid
+    def __subdomain_finalize_core__(self, grid):
         self._dtype = grid.dtype
 
-        # Create the SubDomainSet SubDimensions
-        self._dimensions = tuple(
-            MultiSubDimension('%si' % d.name, d, self) for d in grid.dimensions
-        )
-
         # Compute the SubDomainSet shapes
         global_bounds = []
         for i in self._global_bounds:
@@ -753,34 +863,42 @@ class SubDomainSet(MultiSubDomain):
             self._local_bounds = self._global_bounds
 
         # Sanity check
-        if len(self._local_bounds) != 2*len(self.dimensions):
+        if len(self._local_bounds) != 2*len(grid.dimensions):
             raise ValueError("Left and right bounds must be supplied for each dimension")
 
         # Associate the `_local_bounds` to suitable symbolic objects that the
         # compiler can use to generate code
-        n = counter - npresets
-        assert n >= 0
-        self._implicit_dimension = i_dim = Dimension(name='n%d' % n)
-        functions = []
-        for j in range(len(self._local_bounds)):
-            index = floor(j/2)
-            d = self.dimensions[index]
-            if j % 2 == 0:
-                fname = "%s_%s" % (self.name, d.min_name)
-            else:
-                fname = "%s_%s" % (self.name, d.max_name)
-            f = Function(name=fname, grid=self._grid, shape=(self._n_domains,),
-                         dimensions=(i_dim,), dtype=np.int32)
 
+        # Dimensions with identical names hash the same, hence tag them with the
+        # SubDomainSet ID to make them unique so they can be used to key a dictionary
+        # of replacements without risking overwriting.
+        i_dim = Dimension(f'n_{str(id(self))}')
+        d_dim = DefaultDimension(name='d', default_value=2*grid.dim)
+        sd_func = Function(name=self.name, grid=self._grid,
+                           shape=(self._n_domains, 2*grid.dim),
+                           dimensions=(i_dim, d_dim), dtype=np.int32)
+
+        dimensions = []
+        for i, d in enumerate(grid.dimensions):
             # Check if shorthand notation has been provided:
-            if isinstance(self._local_bounds[j], int):
-                f.data[:] = np.full((self._n_domains,), self._local_bounds[j],
-                                    dtype=np.int32)
-            else:
-                f.data[:] = self._local_bounds[j]
+            for j in range(2):
+                idx = 2*i + j
+                sd_func.data[:, idx] = self._local_bounds[idx]
 
-            functions.append(f)
-        self._functions = as_tuple(functions)
+            dimensions.append(MultiSubDimension(
+                f'i{d.name}', d, None, functions=sd_func,
+                bounds_indices=(2*i, 2*i+1), implicit_dimension=i_dim
+            ))
+
+        self._dimensions = tuple(dimensions)
+        self._subfunction = sd_func
+
+    def __subdomain_finalize__(self):
+        self.__subdomain_finalize_core__(self.grid)
+
+    def __subdomain_finalize_legacy__(self, grid):
+        self._grid = grid
+        self.__subdomain_finalize_core__(self.grid)
 
     @property
     def n_domains(self):
@@ -791,6 +909,285 @@ class SubDomainSet(MultiSubDomain):
         return self._local_bounds
 
 
+class Border(SubDomainSet):
+    """
+    A convenience class for constructing a SubDomainSet which covers specified edges
+    of the domain to a thickness of `border`.
+
+    By default, this border covers all sides of the grid. Alternatively, it is possible
+    to add the border selectively to specific sides by supplying, for example,
+    `dims={y: 'left'}` to obtain only a border on the left (from index zero) side of the
+    y dimension, or `dims={x: x, y: 'left'}` to obtain borders on both sides of the x
+    dimension, but only on the left of the y. One can also supply a single dimension on
+    which to construct a border, using `dims=x` or similar.
+
+    Corners can be included, excluded, or overlapped by setting the `corners` kwarg.
+
+    Parameters
+    ----------
+    grid : Grid
+        The computational grid on which the border should be constructed
+    border : int, tuple of int, or tuple of tuple of int
+        The thickness of the border in gridpoints. A tuple with thickness for each
+        dimension can be supplied if different thicknesses are required per-dimension.
+        A tuple of tuples can also be supplied for more granular control of left and
+        right border thicknesses for each dimension.
+    dims : Dimension, dict, or None, optional
+        The dimensions on which a border should be applied. Default is None, corresponding
+        to borders on both sides of all dimensions.
+    inset : int, tuple of int, or tuple of tuple of int, optional
+        Inset the border from the edges by some number of gridpoints. Default is 0.
+    name : str, optional
+        A unique name for the SubDomainSet created. Default is 'border'.
+    corners : str, optional
+        Behaviour at the corners. Can be set to 'overlap' for overlapping subdomains at
+        the corners, 'nooverlap' for non-overlapping corner subdomains, or 'nocorners'
+        to omit the corners entirely. Default is `nooverlap`.
+
+    Examples
+    --------
+    Set up a border surrounding the whole grid:
+
+    >>> from devito import Grid, Border, Function, Eq, Operator
+    >>> grid = Grid(shape=(7, 7))
+    >>> x, y = grid.dimensions
+
+    >>> border = Border(grid, 2)  # Border of thickness 2
+    >>> f = Function(name='f', grid=grid, dtype=np.int32)
+    >>> eq = Eq(f, f+1, subdomain=border)
+    >>> summary = Operator(eq)()
+    >>> f.data
+    Data([[1, 1, 1, 1, 1, 1, 1],
+          [1, 1, 1, 1, 1, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 1, 1, 1, 1, 1],
+          [1, 1, 1, 1, 1, 1, 1]], dtype=int32)
+
+    Set up a border consisting of the right side of the x dimension and both sides
+    of the y dimension:
+
+    >>> border2 = Border(grid, 2, dims={x: 'right', y: y})
+    >>> g = Function(name='g', grid=grid, dtype=np.int32)
+    >>> eq2 = Eq(g, g+1, subdomain=border2)
+    >>> summary = Operator(eq2)()
+    >>> g.data
+    Data([[1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 1, 1, 1, 1, 1],
+          [1, 1, 1, 1, 1, 1, 1]], dtype=int32)
+
+    Set up a border consisting of only the sides in the y dimension:
+
+    >>> border3 = Border(grid, 2, dims=y)
+    >>> h = Function(name='h', grid=grid, dtype=np.int32)
+    >>> eq3 = Eq(h, h+1, subdomain=border3)
+    >>> summary = Operator(eq3)()
+    >>> h.data
+    Data([[1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1]], dtype=int32)
+
+    which is equivalent to:
+
+    >>> border4 = Border(grid, 2, dims={y: y})
+    >>> i = Function(name='i', grid=grid, dtype=np.int32)
+    >>> eq4 = Eq(i, i+1, subdomain=border4)
+    >>> summary = Operator(eq4)()
+    >>> i.data
+    Data([[1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1],
+          [1, 1, 0, 0, 0, 1, 1]], dtype=int32)
+
+    """
+
+    DimSpec = None | dict[Dimension, Dimension | str]
+    ParsedDimSpec = frozendict[Dimension, Dimension | str]
+
+    BorderInt = int | np.integer
+    BorderSpec = BorderInt | tuple[BorderInt] | tuple[tuple[BorderInt]]
+    ParsedBorderSpec = tuple[tuple[BorderInt]]
+
+    def __init__(self, grid: Grid, border: BorderSpec,
+                 dims: DimSpec = None, name: str = 'border',
+                 inset: BorderSpec = 0, corners: str = 'nooverlap') -> None:
+
+        self._name = name
+        self._border = Border._parse_border(border, grid)
+        self._border_dims = Border._parse_dims(dims, grid)
+        self._inset = Border._parse_border(inset, grid, mode='inset')
+
+        if corners not in ('overlap', 'nooverlap', 'nocorners'):
+            raise ValueError(f"Unrecognised corners option: {corners}")
+        self._corners = corners
+
+        ndomains, bounds = self._build_domains(grid)
+        super().__init__(N=ndomains, bounds=bounds, grid=grid)
+
+    @property
+    def border(self):
+        return self._border
+
+    @property
+    def border_dims(self):
+        return self._border_dims
+
+    @property
+    def name(self):
+        return self._name
+
+    @property
+    def corners(self):
+        return self._corners
+
+    @property
+    def inset(self):
+        return self._inset
+
+    def _inset_flat(self, i):
+        """Flattened access into self.inset"""
+        return self.inset[i // len(self.inset)][i % 2]
+
+    @staticmethod
+    def _parse_dims(dims: DimSpec, grid: Grid) -> ParsedDimSpec:
+        if dims is None:
+            _border_dims = {d: d for d in grid.dimensions}
+        elif isinstance(dims, Dimension):
+            _border_dims = {dims: dims}
+        elif isinstance(dims, dict):
+            _border_dims = dims
+        else:
+            raise ValueError("Dimensions should be supplied as a single dimension, or "
+                             "a dict of the form `{x: x, y: 'left'}`")
+
+        return frozendict(_border_dims)
+
+    @staticmethod
+    def _parse_border(border: BorderSpec, grid: Grid,
+                      mode: str = 'border') -> ParsedBorderSpec:
+        if isinstance(border, (int, np.integer)):
+            return ((border, border),)*len(grid.dimensions)
+
+        else:  # Tuple guaranteed by typing
+            if not len(border) == len(grid.dimensions):
+                raise ValueError(f"Length of {mode} specification should "
+                                 "match number of dimensions")
+            retval = []
+            for b, d in zip(border, grid.dimensions):
+                if isinstance(b, tuple):
+                    if not len(b) == 2:
+                        raise ValueError(f"{b}: more than two thicknesses supplied "
+                                         f"for dimension {d}")
+                    retval.append(b)
+                else:
+                    retval.append((b, b))
+
+            return tuple(retval)
+
+    def _build_domains(self, grid: Grid) -> tuple[int, tuple[np.ndarray]]:
+        """
+        Constructs the bounds and ndomains kwargs for the SubDomainSet.
+        """
+        if self.corners == 'overlap':
+            return self._build_domains_overlap(grid)
+        else:
+            return self._build_domains_nooverlap(grid)
+
+    def _build_domains_overlap(self, grid: Grid) -> tuple[int, tuple[np.ndarray]]:
+
+        bounds = []
+        iterations = zip(grid.dimensions, grid.shape, self.border, strict=True)
+        for i, (d, s, b) in enumerate(iterations):
+
+            if d in self.border_dims:
+                # Note: slightly counterintuitive since a left-side boundary only has
+                # right-side thickness
+                bounds_l = [s - b[0] - self.inset[i][0] if j == 2*i+1
+                            else self._inset_flat(j)
+                            for j in range(2*len(grid.dimensions))]
+                bounds_r = [s - b[1] - self.inset[i][1] if j == 2*i
+                            else self._inset_flat(j)
+                            for j in range(2*len(grid.dimensions))]
+
+                side = self.border_dims[d]
+                if isinstance(side, Dimension):
+                    bounds.extend([bounds_l, bounds_r])
+                elif side == 'left':
+                    bounds.append(bounds_l)
+                elif side == 'right':
+                    bounds.append(bounds_r)
+                else:
+                    raise ValueError(f"Unrecognised side value: {side}")
+
+        # Need to transpose array to fit into expected format for SubDomainSet
+        return len(bounds), tuple(np.array(bounds).T)
+
+    def _build_domains_nooverlap(self, grid: Grid) -> tuple[int, tuple[np.ndarray]]:
+        domain_map = {}  # Stores the side
+        interval_map = {}  # Stores the mapping from the side to bounds
+
+        # Unpack the user-provided specification into a set of sides (on which
+        # a cartesian product is taken) and a mapper from those sides to a set of
+        # bounds for each dimension.
+        for d, s, b, i in zip(grid.dimensions, grid.shape, self.border, self.inset):
+            if d in self.border_dims:
+                side = self.border_dims[d]
+
+                if isinstance(side, Dimension):
+                    domain_map[d] = (LEFT, CENTER, RIGHT)
+                    interval_map[d] = {LEFT: (i[0], s - b[0] - i[0]),
+                                       CENTER: (b[0] + i[0], b[1] + i[1]),
+                                       RIGHT: (s - b[1] - i[1], i[1])}
+                elif side == 'left':
+                    domain_map[d] = (LEFT, CENTER)
+                    # For intuitive behaviour, 'nocorners' should always skip corners
+                    centerval = b[1] + i[1] if self.corners == 'nocorners' else i[1]
+                    interval_map[d] = {LEFT: (i[0], s - b[0] - i[0]),
+                                       CENTER: (b[0] + i[0], centerval)}
+                elif side == 'right':
+                    domain_map[d] = (CENTER, RIGHT)
+                    centerval = b[0] + i[0] if self.corners == 'nocorners' else i[0]
+                    interval_map[d] = {CENTER: (centerval, b[1] + i[1]),
+                                       RIGHT: (s - b[1] - i[1], i[1])}
+                else:
+                    raise ValueError(f"Unrecognised side value {side}")
+            else:
+                domain_map[d] = (CENTER,)
+                interval_map[d] = {CENTER: (0, 0)}
+
+        # Get the cartesian product, then select the required domains. The sides are used
+        # to make this step more straightforward.
+        maybe_domains = list(product(*domain_map.values()))
+        domains = []
+        for d in maybe_domains:
+            if not all(i is CENTER for i in d):
+                # Don't add any domains that are completely centered
+                if self.corners != 'nocorners' or any(i is CENTER for i in d):
+                    # Don't add corners if 'no corners' option selected
+                    domains.append([interval_map[dim][dom] for (dim, dom)
+                                    in zip(grid.dimensions, d)])
+
+        domains = np.array(domains)
+
+        # Reshape the bounds into the format expected by the SubDomainSet init
+        shape = (domains.shape[0], domains.shape[1]*domains.shape[2])
+        bounds = np.reshape(domains, shape).T
+
+        return domains.shape[0], tuple(bounds)
+
+
 # Preset SubDomains
 
 
