@@ -7,9 +7,9 @@ identifying a device attached to a node).
 """
 
 import os
-from ctypes import c_void_p
+from ctypes import POINTER, c_void_p
+from functools import cached_property
 
-from cached_property import cached_property
 import numpy as np
 
 from devito.exceptions import InvalidArgument
@@ -18,14 +18,14 @@ from devito.tools import as_list, as_tuple, is_integer
 from devito.types.array import Array, ArrayObject
 from devito.types.basic import Scalar, Symbol
 from devito.types.dimension import CustomDimension
-from devito.types.misc import Pointer, VolatileInt
+from devito.types.misc import Fence, VolatileInt
 
 __all__ = ['NThreads', 'NThreadsNested', 'NThreadsNonaffine', 'NThreadsBase',
-           'DeviceID', 'ThreadID', 'Lock', 'PThreadArray', 'SharedData',
-           'NPThreads', 'DeviceRM', 'DevicePointer', 'QueueID']
+           'DeviceID', 'ThreadID', 'Lock', 'ThreadArray', 'PThreadArray',
+           'SharedData', 'NPThreads', 'DeviceRM', 'QueueID', 'Barrier', 'TBArray']
 
 
-class NThreadsBase(Scalar):
+class NThreadsAbstract(Scalar):
 
     is_Input = True
     is_PerfKnob = True
@@ -39,10 +39,29 @@ class NThreadsBase(Scalar):
     def __dtype_setup__(cls, **kwargs):
         return np.int32
 
+
+class NThreadsBase(NThreadsAbstract):
+
     @cached_property
     def default_value(self):
-        return int(os.environ.get('OMP_NUM_THREADS',
-                                  configuration['platform'].cores_physical))
+        return int(os.environ.get(
+            'OMP_NUM_THREADS',
+            configuration['platform'].cores_physical_per_numa_domain
+        ))
+
+    def _arg_defaults(self, **kwargs):
+        base_nthreads = self.default_value
+
+        try:
+            npthreads = kwargs['metadata']['npthreads']
+        except KeyError:
+            raise InvalidArgument("Cannot determine `npthreads`")
+
+        # If a symbolic object, it must be resolved
+        if isinstance(npthreads, NPThreads):
+            npthreads = kwargs.get(npthreads.name, npthreads.size)
+
+        return {self.name: max(base_nthreads - npthreads, 1)}
 
 
 class NThreads(NThreadsBase):
@@ -64,7 +83,7 @@ class NThreadsNested(NThreadsBase):
         return configuration['platform'].threads_per_core
 
 
-class NPThreads(NThreadsBase):
+class NPThreads(NThreadsAbstract):
 
     name = 'npthreads'
 
@@ -111,6 +130,9 @@ class ThreadID(CustomDimension):
 
 class ThreadArray(ArrayObject):
 
+    # Not a performance-sensitive object
+    _data_alignment = False
+
     @classmethod
     def __indices_setup__(cls, **kwargs):
         try:
@@ -125,6 +147,10 @@ class ThreadArray(ArrayObject):
         assert len(self.dimensions) == 1
         return self.dimensions[0]
 
+    @property
+    def npthreads(self):
+        return self.dim.symbolic_size
+
     @property
     def index(self):
         if self.size == 1:
@@ -132,14 +158,10 @@ class ThreadArray(ArrayObject):
         else:
             return self.dim
 
-    @cached_property
-    def symbolic_base(self):
-        return Symbol(name=self.name, dtype=None)
-
 
 class PThreadArray(ThreadArray):
 
-    dtype = type('pthread_t', (c_void_p,), {})
+    dtype = POINTER(type('pthread_t', (c_void_p,), {}))
 
     @classmethod
     def __dtype_setup__(cls, **kwargs):
@@ -204,6 +226,9 @@ class Lock(Array):
 
     is_volatile = True
 
+    # Not a performance-sensitive object
+    _data_alignment = False
+
     def __init_finalize__(self, *args, **kwargs):
         kwargs.setdefault('scope', 'stack')
 
@@ -254,6 +279,14 @@ class DeviceID(DeviceSymbol):
     def default_value(self):
         return -1
 
+    def _arg_values(self, **kwargs):
+        if self.name in kwargs:
+            return {self.name: kwargs.pop(self.name)}
+        elif configuration['deviceid'] != self.default_value:
+            return {self.name: configuration['deviceid']}
+        else:
+            return {self.name: self.default_value}
+
 
 class DeviceRM(DeviceSymbol):
 
@@ -274,24 +307,27 @@ class DeviceRM(DeviceSymbol):
 class QueueID(Symbol):
 
     def __new__(cls, *args, **kwargs):
-        kwargs.setdefault('name', 'qid')
+        kwargs['name'] = 'qid'
         kwargs.setdefault('is_const', True)
         return super().__new__(cls, *args, **kwargs)
 
 
-class DevicePointer(Pointer):
+class Barrier(Fence):
 
-    __rkwargs__ = Pointer.__rkwargs__ + ('mapped',)
+    """
+    A generic synchronization barrier for threads or processes.
+    """
 
-    def __init_finalize__(self, *args, mapped=None, **kwargs):
-        self._mapped = mapped
+    pass
+
+
+class TBArray(Array):
+
+    """
+    An Array used for performance optimization within a thread block.
+    """
+
+    def __init_finalize__(self, *args, **kwargs):
+        kwargs['liveness'] = 'eager'
 
         super().__init_finalize__(*args, **kwargs)
-
-    @property
-    def mapped(self):
-        return self._mapped
-
-    @property
-    def _C_ctype(self):
-        return self.mapped.indexed._C_ctype
