@@ -1,11 +1,30 @@
 """User API to specify relationals."""
+from functools import singledispatch
 
 import sympy
 
-__all__ = ['Le', 'Lt', 'Ge', 'Gt', 'Ne']
+__all__ = ['Le', 'Lt', 'Ge', 'Gt', 'Ne', 'relational_min']
 
 
-class Le(sympy.Le):
+class AbstractRel:
+    """
+    Abstract mixin class for objects subclassing sympy.Relational.
+    """
+    @property
+    def negated(self):
+        return ops.get(self.func)(*self.args)
+
+    @property
+    def reversed(self):
+        return rev.get(self.func)(self.rhs, self.lhs)
+
+    @property
+    def subdomain(self):
+        """The SubDomain in which self is defined."""
+        return self._subdomain
+
+
+class Le(AbstractRel, sympy.Le):
     """
     A less-than or equal ("<=") relation between two objects, the left-hand side and the
     right-hand side. It can be used to build conditionals but not directly to
@@ -36,19 +55,11 @@ class Le(sympy.Le):
         kwargs.update({'evaluate': False})
         obj = sympy.Le.__new__(cls, lhs, rhs, **kwargs)
         obj._subdomain = subdomain
+
         return obj
 
-    @property
-    def subdomain(self):
-        """The SubDomain in which the Le is defined."""
-        return self._subdomain
 
-    @property
-    def negated(self):
-        return ops.get(self.func)(*self.args)
-
-
-class Lt(sympy.Lt):
+class Lt(AbstractRel, sympy.Lt):
     """
     A less-than ("<") relation between two objects, the left-hand side and the
     right-hand side.It can be used to build conditionals but not directly to
@@ -79,19 +90,11 @@ class Lt(sympy.Lt):
         kwargs.update({'evaluate': False})
         obj = sympy.Lt.__new__(cls, lhs, rhs, **kwargs)
         obj._subdomain = subdomain
+
         return obj
 
-    @property
-    def subdomain(self):
-        """The SubDomain in which the Lt is defined."""
-        return self._subdomain
 
-    @property
-    def negated(self):
-        return ops.get(self.func)(*self.args)
-
-
-class Ge(sympy.Ge):
+class Ge(AbstractRel, sympy.Ge):
     """
     A greater-than or equal (">=") relation between two objects, the left-hand side and
     the right-hand side. It can be used to build conditionals but not directly to
@@ -125,17 +128,8 @@ class Ge(sympy.Ge):
 
         return obj
 
-    @property
-    def subdomain(self):
-        """The SubDomain in which the Ge is defined."""
-        return self._subdomain
 
-    @property
-    def negated(self):
-        return ops.get(self.func)(*self.args)
-
-
-class Gt(sympy.Gt):
+class Gt(AbstractRel, sympy.Gt):
     """
     A greater-than (">") relation between two objects, the left-hand side and the
     right-hand side. It can be used to build conditionals but not directly to
@@ -169,17 +163,8 @@ class Gt(sympy.Gt):
 
         return obj
 
-    @property
-    def subdomain(self):
-        """The SubDomain in which the Gt is defined."""
-        return self._subdomain
 
-    @property
-    def negated(self):
-        return ops.get(self.func)(*self.args)
-
-
-class Ne(sympy.Ne):
+class Ne(AbstractRel, sympy.Ne):
     """
     A not-equal ("!=") relation between two objects (see Notes below), the left-hand side
     and the right-hand side. It can be used to build conditionals but not directly to
@@ -221,10 +206,47 @@ class Ne(sympy.Ne):
 
         return obj
 
-    @property
-    def subdomain(self):
-        """The SubDomain in which the Ne is defined."""
-        return self._subdomain
-
 
 ops = {Ge: Lt, Gt: Le, Le: Gt, Lt: Ge}
+rev = {Ge: Le, Gt: Lt, Lt: Gt, Le: Ge}
+
+
+def relational_min(expr, s):
+    """
+    Infer the minimum valid value for symbol `s` in the expression `expr`.
+    For example
+        - if `expr` is `s < 10`, then the minimum valid value for `s` is 0
+        - if `expr` is `s >= 10`, then the minimum valid value for `s` is 10
+    """
+    if not expr.has(s):
+        return 0
+
+    return _relational_min(expr, s)
+
+
+@singledispatch
+def _relational_min(s, expr):
+    return 0
+
+
+@_relational_min.register(sympy.And)
+def _(expr, s):
+    return max([_relational_min(e, s) for e in expr.args])
+
+
+@_relational_min.register(Gt)
+@_relational_min.register(Lt)
+def _(expr, s):
+    if s == expr.gts:
+        return expr.lts + 1
+    else:
+        return 0
+
+
+@_relational_min.register(Ge)
+@_relational_min.register(Le)
+def _(expr, s):
+    if s == expr.gts:
+        return expr.lts
+    else:
+        return 0
