@@ -1,7 +1,12 @@
 from collections import OrderedDict
-from cached_property import cached_property
+from functools import cached_property
 
 import numpy as np
+try:
+    from sympy.matrices.matrixbase import MatrixBase
+except ImportError:
+    # Before 1.13
+    from sympy.matrices.matrices import MatrixBase
 from sympy.core.sympify import converter as sympify_converter
 
 from devito.finite_differences import Differentiable
@@ -29,8 +34,8 @@ class TensorFunction(AbstractTensor):
         Carries shape, dimensions, and dtype of the TensorFunction. When grid is not
         provided, shape and dimensions must be given. For MPI execution, a
         Grid is compulsory.
-    space_order : int or 3-tuple of ints, optional
-        Discretisation order for space derivatives. Defaults to 1. ``space_order`` also
+    space_order : int or 3-tuple of ints, optional, default=1
+        Discretisation order for space derivatives. ``space_order`` also
         impacts the number of points available around a generic point of interest.  By
         default, ``space_order`` points are available on both sides of a generic point of
         interest, including those nearby the grid boundary. Sometimes, fewer points
@@ -42,11 +47,11 @@ class TensorFunction(AbstractTensor):
         Shape of the domain region in grid points. Only necessary if ``grid`` isn't given.
     dimensions : tuple of Dimension, optional
         Dimensions associated with the object. Only necessary if ``grid`` isn't given.
-    dtype : data-type, optional
-        Any object that can be interpreted as a numpy data type. Defaults
-        to ``np.float32``.
+    dtype : data-type, optional, default=np.float32
+        Any object that can be interpreted as a numpy data type.
     staggered : Dimension or tuple of Dimension or Stagger, optional
-        Define how the TensorFunction is staggered.
+        Staggering of each component, needs to have the size of the tensor. Defaults
+        to the Dimensions.
     initializer : callable or any object exposing the buffer interface, optional
         Data initializer. If a callable is provided, data is allocated lazily.
     allocator : MemoryAllocator, optional
@@ -54,16 +59,12 @@ class TensorFunction(AbstractTensor):
         to take advantage of the memory hierarchy in a NUMA architecture. Refer to
         `default_allocator.__doc__` for more information.
     padding : int or tuple of ints, optional
-        .. deprecated:: shouldn't be used; padding is now automatically inserted.
         Allocate extra grid points to maximize data access alignment. When a tuple
         of ints, one int per Dimension should be provided.
-    symmetric : bool, optional
-        Whether the tensor is symmetric or not. Defaults to True.
-    diagonal : Bool, optional
-        Whether the tensor is diagonal or not. Defaults to False.
-    staggered: tuple of Dimension, optional
-        Staggering of each component, needs to have the size of the tensor. Defaults
-        to the Dimensions.
+    symmetric : bool, optional, default=True
+        Whether the tensor is symmetric or not.
+    diagonal : Bool, optional, default=False
+        Whether the tensor is diagonal or not.
     """
 
     _is_TimeDependent = False
@@ -73,13 +74,29 @@ class TensorFunction(AbstractTensor):
     _class_priority = 10
     _op_priority = Differentiable._op_priority + 1.
 
+    __rkwargs__ = AbstractTensor.__rkwargs__ + ('dimensions', 'space_order')
+
     def __init_finalize__(self, *args, **kwargs):
+        super().__init_finalize__(*args, **kwargs)
         grid = kwargs.get('grid')
         dimensions = kwargs.get('dimensions')
-        inds, _ = Function.__indices_setup__(grid=grid,
-                                             dimensions=dimensions)
+        inds, _ = Function.__indices_setup__(grid=grid, dimensions=dimensions)
         self._space_dimensions = inds
 
+    @classmethod
+    def _component_kwargs(cls, inds, **kwargs):
+        """
+        Get the kwargs for a single component
+        from the kwargs of the TensorFunction.
+        """
+        kw = {}
+        for k, v in kwargs.items():
+            if isinstance(v, MatrixBase):
+                kw[k] = v[inds]
+            else:
+                kw[k] = v
+        return kw
+
     @classmethod
     def __subfunc_setup__(cls, *args, **kwargs):
         """
@@ -109,10 +126,12 @@ class TensorFunction(AbstractTensor):
             start = i if (symm or diag) else 0
             stop = i + 1 if diag else len(dims)
             for j in range(start, stop):
-                kwargs["name"] = "%s_%s%s" % (name, d.name, dims[j].name)
-                kwargs["staggered"] = (stagg[i][j] if stagg is not None
-                                       else (NODE if i == j else (d, dims[j])))
-                funcs2[j] = cls._sub_type(**kwargs)
+                staggj = (stagg[i][j] if stagg is not None
+                          else (NODE if i == j else (d, dims[j])))
+                sub_kwargs = cls._component_kwargs((i, j), **kwargs)
+                sub_kwargs.update({'name': f"{name}_{d.name}{dims[j].name}",
+                                   'staggered': staggj})
+                funcs2[j] = cls._sub_type(**sub_kwargs)
             funcs.append(funcs2)
 
         # Symmetrize and fill diagonal if symmetric
@@ -129,6 +148,8 @@ class TensorFunction(AbstractTensor):
         -----
         This method acts as a fallback for __getattribute__
         """
+        if name in ['_sympystr', '_pretty', '_latex']:
+            return super().__getattr__(self, name)
         try:
             return self.applyfunc(lambda x: x if x == 0 else getattr(x, name))
         except:
@@ -161,10 +182,19 @@ class TensorFunction(AbstractTensor):
         """Spatial dimensions."""
         return self._space_dimensions
 
+    @cached_property
+    def root_dimensions(self):
+        """Tuple of root Dimensions of the physical space Dimensions."""
+        return tuple(d.root for d in self.space_dimensions)
+
     @cached_property
     def space_order(self):
         """The space order for all components."""
-        return ({a.space_order for a in self} - {None}).pop()
+        orders = self.applyfunc(lambda x: x.space_order)
+        if len(set(orders)) > 1:
+            return orders
+        else:
+            return orders[0]
 
     @property
     def is_diagonal(self):
@@ -172,31 +202,53 @@ class TensorFunction(AbstractTensor):
         return np.all([self[i, j] == 0 for j in range(self.cols)
                        for i in range(self.rows) if i != j])
 
-    @property
-    def evaluate(self):
-        return self.applyfunc(lambda x: getattr(x, 'evaluate', x))
+    def _evaluate(self, **kwargs):
+        def _do_evaluate(x):
+            try:
+                expand = kwargs.get('expand', True)
+                return x._evaluate(expand=expand)
+            except AttributeError:
+                return x
+        return self.applyfunc(_do_evaluate)
 
     def values(self):
         if self.is_diagonal:
             return [self[i, i] for i in range(self.shape[0])]
         elif self.is_symmetric:
-            val = super(TensorFunction, self).values()
+            val = super().values()
             return list(OrderedDict.fromkeys(val))
         else:
-            return super(TensorFunction, self).values()
+            return super().values()
 
-    def div(self, shift=None):
+    def div(self, shift=None, order=None, method='FD', **kwargs):
         """
         Divergence of the TensorFunction (is a VectorFunction).
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite differences.
         """
+        w = kwargs.get('weights', kwargs.get('w'))
         comps = []
         func = vec_func(self)
         ndim = len(self.space_dimensions)
+        space_dims = self.root_dimensions
         shift_x0 = make_shift_x0(shift, (ndim, ndim))
+        order = order or self.space_order
         for i in range(len(self.space_dimensions)):
             comps.append(sum([getattr(self[j, i], 'd%s' % d.name)
-                              (x0=shift_x0(shift, d, i, j))
-                              for j, d in enumerate(self.space_dimensions)]))
+                              (x0=shift_x0(shift, d, i, j), fd_order=order,
+                               method=method, w=w)
+                              for j, d in enumerate(space_dims)]))
         return func._new(comps)
 
     @property
@@ -204,25 +256,53 @@ class TensorFunction(AbstractTensor):
         """
         Laplacian of the TensorFunction.
         """
+        return self.laplacian()
+
+    def laplacian(self, shift=None, order=None, method='FD', **kwargs):
+        """
+        Laplacian of the TensorFunction with shifted derivatives and custom
+        FD order.
+
+        Each second derivative is left-right (i.e D^T D with D the first derivative ):
+        `(self.dx(x0=dim+shift*dim.spacing,
+                  fd_order=order)).dx(x0=dim-shift*dim.spacing, fd_order=order)`
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite
+        """
+        w = kwargs.get('weights', kwargs.get('w'))
         comps = []
         func = vec_func(self)
-        for j, d in enumerate(self.space_dimensions):
+        order = order or self.space_order
+        space_dims = self.root_dimensions
+        ndim = len(self.space_dimensions)
+        shift_x0 = make_shift_x0(shift, (ndim, ndim))
+        for j in range(ndim):
             comps.append(sum([getattr(self[j, i], 'd%s2' % d.name)
-                              for i, d in enumerate(self.space_dimensions)]))
+                              (x0=shift_x0(shift, d, j, i), fd_order=order,
+                               method=method, w=w)
+                              for i, d in enumerate(space_dims)]))
         return func._new(comps)
 
-    def grad(self, shift=None):
+    def grad(self, shift=None, order=None, method=None, **kwargs):
         raise AttributeError("Gradient of a second order tensor not supported")
 
     def new_from_mat(self, mat):
         func = tens_func(self)
         return func._new(self.rows, self.cols, mat)
 
-    def classof_prod(self, other, mat):
-        try:
-            is_mat = len(mat[0]) > 1
-        except TypeError:
-            is_mat = False
+    def classof_prod(self, other, cols):
+        is_mat = cols > 1
         is_time = (getattr(self, '_is_TimeDependent', False) or
                    getattr(other, '_is_TimeDependent', False))
         return mat_time_dict[(is_time, is_mat)]
@@ -264,9 +344,10 @@ class VectorFunction(TensorFunction):
         stagg = kwargs.get("staggered", None)
         name = kwargs.get("name")
         for i, d in enumerate(dims):
-            kwargs["name"] = "%s_%s" % (name, d.name)
-            kwargs["staggered"] = stagg[i] if stagg is not None else d
-            funcs.append(cls._sub_type(**kwargs))
+            sub_kwargs = cls._component_kwargs(i, **kwargs)
+            sub_kwargs.update({'name': f"{name}_{d.name}",
+                               'staggered': stagg[i] if stagg is not None else d})
+            funcs.append(cls._sub_type(**sub_kwargs))
 
         return funcs
 
@@ -277,50 +358,132 @@ class VectorFunction(TensorFunction):
 
     __repr__ = __str__
 
-    def div(self, shift=None):
+    def div(self, shift=None, order=None, method='FD', **kwargs):
         """
         Divergence of the VectorFunction, creates the divergence Function.
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite difference coefficients.
         """
+        w = kwargs.get('weights', kwargs.get('w'))
         shift_x0 = make_shift_x0(shift, (len(self.space_dimensions),))
-        return sum([getattr(self[i], 'd%s' % d.name)(x0=shift_x0(shift, d, None, i))
-                    for i, d in enumerate(self.space_dimensions)])
+        order = order or self.space_order
+        space_dims = self.root_dimensions
+        return sum([getattr(self[i], 'd%s' % d.name)(x0=shift_x0(shift, d, None, i),
+                                                     fd_order=order, method=method, w=w)
+                    for i, d in enumerate(space_dims)])
 
     @property
     def laplace(self):
         """
         Laplacian of the VectorFunction, creates the Laplacian VectorFunction.
         """
+        return self.laplacian()
+
+    def laplacian(self, shift=None, order=None, method='FD', **kwargs):
+        """
+        Laplacian of the VectorFunction, creates the Laplacian VectorFunction.
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite
+        """
+        w = kwargs.get('weights', kwargs.get('w'))
         func = vec_func(self)
-        comps = [sum([getattr(s, 'd%s2' % d.name) for d in self.space_dimensions])
+        shift_x0 = make_shift_x0(shift, (len(self.space_dimensions),))
+        order = order or self.space_order
+        space_dims = self.root_dimensions
+        comps = [sum([getattr(s, 'd%s2' % d.name)(x0=shift_x0(shift, d, None, i),
+                                                  fd_order=order, w=w, method=method)
+                      for i, d in enumerate(space_dims)])
                  for s in self]
         return func._new(comps)
 
-    @property
-    def curl(self):
+    def curl(self, shift=None, order=None, method='FD', **kwargs):
         """
         Gradient of the (3D) VectorFunction, creates the curl VectorFunction.
-        """
 
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite difference coefficients.
+        """
         if len(self.space_dimensions) != 3:
             raise AttributeError("Curl only supported for 3D VectorFunction")
         # The curl of a VectorFunction is a VectorFunction
-        derivs = ['d%s' % d.name for d in self.space_dimensions]
-        comp1 = getattr(self[2], derivs[1]) - getattr(self[1], derivs[2])
-        comp2 = getattr(self[0], derivs[2]) - getattr(self[2], derivs[0])
-        comp3 = getattr(self[1], derivs[0]) - getattr(self[0], derivs[1])
-
+        w = kwargs.get('weights', kwargs.get('w'))
+        dims = self.root_dimensions
+        derivs = ['d%s' % d.name for d in dims]
+        shift_x0 = make_shift_x0(shift, (len(dims), len(dims)))
+        order = order or self.space_order
+        comp1 = (getattr(self[2], derivs[1])(x0=shift_x0(shift, dims[1], 2, 1),
+                                             fd_order=order, method=method, w=w) -
+                 getattr(self[1], derivs[2])(x0=shift_x0(shift, dims[2], 1, 2),
+                                             fd_order=order, method=method, w=w))
+        comp2 = (getattr(self[0], derivs[2])(x0=shift_x0(shift, dims[2], 0, 2),
+                                             fd_order=order, method=method, w=w) -
+                 getattr(self[2], derivs[0])(x0=shift_x0(shift, dims[0], 2, 0),
+                                             fd_order=order, method=method, w=w))
+        comp3 = (getattr(self[1], derivs[0])(x0=shift_x0(shift, dims[0], 1, 0),
+                                             fd_order=order, method=method, w=w) -
+                 getattr(self[0], derivs[1])(x0=shift_x0(shift, dims[1], 0, 1),
+                                             fd_order=order, method=method, w=w))
         func = vec_func(self)
         return func._new(3, 1, [comp1, comp2, comp3])
 
-    def grad(self, shift=None):
+    def grad(self, shift=None, order=None, method='FD', **kwargs):
         """
         Gradient of the VectorFunction, creates the gradient TensorFunction.
+
+        Parameters
+        ----------
+        shift: Number, optional, default=None
+            Shift for the center point of the derivative in number of gridpoints
+        order: int, optional, default=None
+            Discretization order for the finite differences.
+            Uses `func.space_order` when not specified
+        method: str, optional, default='FD'
+            Discretization method. Options are 'FD' (default) and
+            'RSFD' (rotated staggered grid finite-difference).
+        weights/w: list, tuple, or dict, optional, default=None
+            Custom weights for the finite difference coefficients.
         """
+        w = kwargs.get('weights', kwargs.get('w'))
         func = tens_func(self)
         ndim = len(self.space_dimensions)
         shift_x0 = make_shift_x0(shift, (ndim, ndim))
-        comps = [[getattr(f, 'd%s' % d.name)(x0=shift_x0(shift, d, i, j))
-                  for j, d in enumerate(self.space_dimensions)]
+        order = order or self.space_order
+        space_dims = self.root_dimensions
+        comps = [[getattr(f, 'd%s' % d.name)(x0=shift_x0(shift, d, i, j), w=w,
+                                             fd_order=order, method=method)
+                  for j, d in enumerate(space_dims)]
                  for i, f in enumerate(self)]
         return func._new(comps)
 
