@@ -68,7 +68,10 @@ def plot_velocity(model, source=None, receiver=None, colorbar=True, cmap="jet"):
               model.origin[1] + domain_size[1], model.origin[1]]
 
     slices = tuple(slice(model.nbl, -model.nbl) for _ in range(2))
-    field = (getattr(model, 'vp', None) or getattr(model, 'lam')).data[slices]
+    if getattr(model, 'vp', None) is not None:
+        field = model.vp.data[slices]
+    else:
+        field = model.lam.data[slices]
     plot = plt.imshow(np.transpose(field), animated=True, cmap=cmap,
                       vmin=np.min(field), vmax=np.max(field),
                       extent=extent)
