@@ -1,6 +1,9 @@
 from scipy.special import hankel2
 import numpy as np
-import pytest
+try:
+    import pytest
+except:
+    pass
 from devito import Grid, Function, Eq, Operator, info
 from examples.seismic import RickerSource, TimeAxis, Model, AcquisitionGeometry
 from examples.seismic.self_adjoint import (acoustic_sa_setup, setup_w_over_q,
@@ -12,7 +15,7 @@ dtypes = [np.float64, ]
 space_orders = [8, ]
 
 
-class TestWavesolver(object):
+class TestWavesolver:
 
     @pytest.mark.parametrize('shape', shapes)
     @pytest.mark.parametrize('dtype', dtypes)
@@ -125,12 +128,12 @@ class TestWavesolver(object):
         ns = 2 * size + 1
         if len(shape) == 2:
             nx2, nz2 = shape[0]//2, shape[1]//2
-            dm.data[nx2-size:nx2+size, nz2-size:nz2+size] = \
+            dm.data[nx2-size:nx2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns)
         else:
             nx2, ny2, nz2 = shape[0]//2, shape[1]//2, shape[2]//2
             nx, ny, nz = shape
-            dm.data[nx2-size:nx2+size, ny2-size:ny2+size, nz2-size:nz2+size] = \
+            dm.data[nx2-size:nx2+size+1, ny2-size:ny2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns, ns)
 
         # Compute F(m + dm)
@@ -191,12 +194,12 @@ class TestWavesolver(object):
         ns = 2 * size + 1
         if len(shape) == 2:
             nx2, nz2 = shape[0]//2, shape[1]//2
-            m1.data[nx2-size:nx2+size, nz2-size:nz2+size] = \
+            m1.data[nx2-size:nx2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns)
         else:
             nx2, ny2, nz2 = shape[0]//2, shape[1]//2, shape[2]//2
             nx, ny, nz = shape
-            m1.data[nx2-size:nx2+size, ny2-size:ny2+size, nz2-size:nz2+size] = \
+            m1.data[nx2-size:nx2+size+1, ny2-size:ny2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns, ns)
 
         a = np.random.rand()
@@ -239,12 +242,12 @@ class TestWavesolver(object):
         ns = 2 * size + 1
         if len(shape) == 2:
             nx2, nz2 = shape[0]//2, shape[1]//2
-            m1.data[nx2-size:nx2+size, nz2-size:nz2+size] = \
+            m1.data[nx2-size:nx2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns)
         else:
             nx2, ny2, nz2 = shape[0]//2, shape[1]//2, shape[2]//2
             nx, ny, nz = shape
-            m1.data[nx2-size:nx2+size, ny2-size:ny2+size, nz2-size:nz2+size] = \
+            m1.data[nx2-size:nx2+size+1, ny2-size:ny2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns, ns)
 
         a = np.random.rand()
@@ -286,12 +289,12 @@ class TestWavesolver(object):
         ns = 2 * size + 1
         if len(shape) == 2:
             nx2, nz2 = shape[0]//2, shape[1]//2
-            dm1.data[nx2-size:nx2+size, nz2-size:nz2+size] = \
+            dm1.data[nx2-size:nx2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns)
         else:
             nx2, ny2, nz2 = shape[0]//2, shape[1]//2, shape[2]//2
             nx, ny, nz = shape
-            dm1.data[nx2-size:nx2+size, ny2-size:ny2+size, nz2-size:nz2+size] = \
+            dm1.data[nx2-size:nx2+size+1, ny2-size:ny2+size+1, nz2-size:nz2+size+1] = \
                 -1 + 2 * np.random.rand(ns, ns, ns)
 
         # Data perturbation
