@@ -1,6 +1,13 @@
 from devito import (Eq, Operator, Function, TimeFunction, NODE, Inc, solve,
                     cos, sin, sqrt)
-from examples.seismic import PointSource, Receiver
+from examples.seismic.acoustic.operators import freesurface
+
+
+def _subs(func, d, d0):
+    try:
+        return func._subs(d, d0)
+    except AttributeError:
+        return func
 
 
 def second_order_stencil(model, u, v, H0, Hz, qu, qv, forward=True):
@@ -20,10 +27,16 @@ def second_order_stencil(model, u, v, H0, Hz, qu, qv, forward=True):
     stencilp = solve(m * u.dt2 - H0 - qu + damp * udt, unext)
     stencilr = solve(m * v.dt2 - Hz - qv + damp * vdt, vnext)
 
-    first_stencil = Eq(unext, stencilp)
-    second_stencil = Eq(vnext, stencilr)
+    first_stencil = Eq(unext, stencilp, subdomain=model.grid.subdomains['physdomain'])
+    second_stencil = Eq(vnext, stencilr, subdomain=model.grid.subdomains['physdomain'])
 
     stencils = [first_stencil, second_stencil]
+
+    # Add free surface
+    if model.fs:
+        stencils.append(freesurface(model, Eq(unext, stencilp)))
+        stencils.append(freesurface(model, Eq(vnext, stencilr)))
+
     return stencils
 
 
@@ -50,78 +63,78 @@ def trig_func(model):
     return costheta, sintheta
 
 
-def Gzz_centered(model, field, costheta, sintheta, cosphi, sinphi, space_order):
+def Gzz_centered(model, field):
     """
     3D rotated second order derivative in the direction z.
 
     Parameters
     ----------
+    model : Model
+        Physical parameters model structure.
     field : Function
         Input for which the derivative is computed.
-    costheta : Function or float
-        Cosine of the tilt angle.
-    sintheta : Function or float
-        Sine of the tilt angle.
-    cosphi : Function or float
-        Cosine of the azymuth angle.
-    sinphi : Function or float
-        Sine of the azymuth angle.
-    space_order : int
-        Space discretization order.
 
     Returns
     -------
     Rotated second order derivative w.r.t. z.
     """
-    order1 = space_order // 2
-    Gz = -(sintheta * cosphi * field.dx(fd_order=order1) +
-           sintheta * sinphi * field.dy(fd_order=order1) +
-           costheta * field.dz(fd_order=order1))
+    b = getattr(model, 'b', 1)
+    costheta, sintheta, cosphi, sinphi = trig_func(model)
 
-    Gzz = (Gz * costheta).dz(fd_order=order1).T
+    order1 = field.space_order // 2
+    x, y, z = field.grid.dimensions
+    dx, dy, dz = x.spacing/2, y.spacing/2, z.spacing/2
+
+    Gz = (_subs(b, x, x+dx) * sintheta * cosphi * field.dx(fd_order=order1, x0=x+dx) +
+          _subs(b, y, y+dy) * sintheta * sinphi * field.dy(fd_order=order1, x0=y+dy) +
+          _subs(b, z, z+dz) * costheta * field.dz(fd_order=order1, x0=z+dz))
+
+    Gzz = (Gz * costheta).dz(fd_order=order1, x0=z-dz)
     # Add rotated derivative if angles are not zero. If angles are
     # zeros then `0*Gz = 0` and doesn't have any `.dy` ....
     if sintheta != 0:
-        Gzz += (Gz * sintheta * cosphi).dx(fd_order=order1).T
+        Gzz += (Gz * sintheta * cosphi).dx(fd_order=order1, x0=x-dx)
     if sinphi != 0:
-        Gzz += (Gz * sintheta * sinphi).dy(fd_order=order1).T
+        Gzz += (Gz * sintheta * sinphi).dy(fd_order=order1, x0=y-dy)
 
     return Gzz
 
 
-def Gzz_centered_2d(model, field, costheta, sintheta, space_order):
+def Gzz_centered_2d(model, field):
     """
     2D rotated second order derivative in the direction z.
 
     Parameters
     ----------
+    model : Model
+        Physical parameters model structure.
     field : Function
         Input for which the derivative is computed.
-    costheta : Function or float
-        Cosine of the tilt angle.
-    sintheta : Function or float
-        Sine of the tilt angle.
-    space_order : int
-        Space discretization order.
 
     Returns
     -------
     Rotated second order derivative w.r.t. z.
     """
-    order1 = space_order // 2
-    Gz = -(sintheta * field.dx(fd_order=order1) +
-           costheta * field.dy(fd_order=order1))
-    Gzz = (Gz * costheta).dy(fd_order=order1).T
+    b = getattr(model, 'b', 1)
+    costheta, sintheta = trig_func(model)
+
+    order1 = field.space_order // 2
+    x, y = field.grid.dimensions
+    dx, dy = x.spacing/2, y.spacing/2
+
+    Gz = (_subs(b, x, x+dx) * sintheta * field.dx(fd_order=order1, x0=x+dx) +
+          _subs(b, y, y+dy) * costheta * field.dy(fd_order=order1, x0=y+dy))
+    Gzz = (Gz * costheta).dy(fd_order=order1, x0=y-dy)
 
     # Add rotated derivative if angles are not zero. If angles are
     # zeros then `0*Gz = 0` and doesn't have any `.dy` ....
     if sintheta != 0:
-        Gzz += (Gz * sintheta).dx(fd_order=order1).T
+        Gzz += (Gz * sintheta).dx(fd_order=order1, x0=x-dx)
     return Gzz
 
 
 # Centered case produces directly Gxx + Gyy
-def Gxxyy_centered(model, field, costheta, sintheta, cosphi, sinphi, space_order):
+def Gh_centered(model, field):
     """
     Sum of the 3D rotated second order derivative in the direction x and y.
     As the Laplacian is rotation invariant, it is computed as the conventional
@@ -130,57 +143,35 @@ def Gxxyy_centered(model, field, costheta, sintheta, cosphi, sinphi, space_order
 
     Parameters
     ----------
+    model : Model
+        Physical parameters model structure.
     field : Function
         Input field.
-    costheta : Function or float
-        Cosine of the tilt angle.
-    sintheta : Function or float
-        Sine of the tilt angle.
-    cosphi : Function or float
-        Cosine of the azymuth angle.
-    sinphi : Function or float
-        Sine of the azymuth angle.
-    space_order : int
-        Space discretization order.
 
     Returns
     -------
     Sum of the 3D rotated second order derivative in the direction x and y.
     """
-    Gzz = Gzz_centered(model, field, costheta, sintheta, cosphi, sinphi, space_order)
-    return field.laplace - Gzz
+    if model.dim == 3:
+        Gzz = Gzz_centered(model, field)
+    else:
+        Gzz = Gzz_centered_2d(model, field)
+    b = getattr(model, 'b', None)
+    if b is not None:
+        _diff = lambda f, d: getattr(f, f'd{d.name}')
+        so = field.space_order // 2
+        lap = 0
+        for d in model.space_dimensions:
+            x0 = d + d.spacing / 2
+            x0m = d - d.spacing / 2
+            lap += _diff(_subs(b, d, x0) * _diff(field, d)(x0=x0, fd_order=so),
+                         d)(x0=x0m, fd_order=so)
+    else:
+        lap = field.laplace
+    return lap - Gzz
 
 
-def Gxx_centered_2d(model, field, costheta, sintheta, space_order):
-    """
-    2D rotated second order derivative in the direction x.
-    As the Laplacian is rotation invariant, it is computed as the conventional
-    Laplacian minus the second order rotated second order derivative in the direction z
-    Gxx = field.laplace - Gzz
-
-    Parameters
-    ----------
-    field : TimeFunction
-        Input field.
-    costheta : Function or float
-        Cosine of the tilt angle.
-    sintheta : Function or float
-        Sine of the tilt angle.
-    cosphi : Function or float
-        Cosine of the azymuth angle.
-    sinphi : Function or float
-        Sine of the azymuth angle.
-    space_order : int
-        Space discretization order.
-
-    Returns
-    -------
-    Sum of the 3D rotated second order derivative in the direction x.
-    """
-    return field.laplace - Gzz_centered_2d(model, field, costheta, sintheta, space_order)
-
-
-def kernel_centered_2d(model, u, v, space_order, **kwargs):
+def kernel_centered(model, u, v, **kwargs):
     """
     TTI finite difference kernel. The equation solved is:
 
@@ -222,9 +213,6 @@ def kernel_centered_2d(model, u, v, space_order, **kwargs):
     # Forward or backward
     forward = kwargs.get('forward', True)
 
-    # Tilt and azymuth setup
-    costheta, sintheta = trig_func(model)
-
     delta, epsilon = model.delta, model.epsilon
     epsilon = 1 + 2*epsilon
     delta = sqrt(1 + 2*delta)
@@ -233,82 +221,17 @@ def kernel_centered_2d(model, u, v, space_order, **kwargs):
     qu = kwargs.get('qu', 0)
     qv = kwargs.get('qv', 0)
 
+    Gzz = Gzz_centered_2d if model.dim == 2 else Gzz_centered
+
     if forward:
-        Gxx = Gxx_centered_2d(model, u, costheta, sintheta, space_order)
-        Gzz = Gzz_centered_2d(model, v, costheta, sintheta, space_order)
+        Gxx = Gh_centered(model, u)
+        Gzz = Gzz(model, v)
         H0 = epsilon*Gxx + delta*Gzz
         Hz = delta*Gxx + Gzz
         return second_order_stencil(model, u, v, H0, Hz, qu, qv)
     else:
-        H0 = Gxx_centered_2d(model, (epsilon*u + delta*v), costheta,
-                             sintheta, space_order)
-        Hz = Gzz_centered_2d(model, (delta*u + v), costheta, sintheta, space_order)
-        return second_order_stencil(model, u, v, H0, Hz, qu, qv, forward=forward)
-
-
-def kernel_centered_3d(model, u, v, space_order, **kwargs):
-    """
-    TTI finite difference kernel. The equation solved is:
-
-    u.dt2 = H0
-    v.dt2 = Hz
-
-    where H0 and Hz are defined as:
-    H0 = (1+2 *epsilon) (Gxx(u)+Gyy(u)) + sqrt(1+ 2*delta) Gzz(v)
-    Hz = sqrt(1+ 2*delta) (Gxx(u)+Gyy(u)) +  Gzz(v)
-
-    and
-
-    H0 = (Gxx+Gyy)((1+2 *epsilon)*u + sqrt(1+ 2*delta)*v)
-    Hz = Gzz(sqrt(1+ 2*delta)*u + v)
-
-    for the forward and adjoint cases, respectively. Epsilon and delta are the Thomsen
-    parameters. This function computes H0 and Hz.
-
-    References:
-        * Zhang, Yu, Houzhu Zhang, and Guanquan Zhang. "A stable TTI reverse
-          time migration and its implementation." Geophysics 76.3 (2011): WA3-WA11.
-        * Louboutin, Mathias, Philipp Witte, and Felix J. Herrmann. "Effects of
-          wrong adjoints for RTM in TTI media." SEG Technical Program Expanded
-          Abstracts 2018. Society of Exploration Geophysicists, 2018. 331-335.
-
-    Parameters
-    ----------
-    u : TimeFunction
-        First TTI field.
-    v : TimeFunction
-        Second TTI field.
-    space_order : int
-        Space discretization order.
-
-    Returns
-    -------
-    u and v component of the rotated Laplacian in 3D.
-    """
-    # Forward or backward
-    forward = kwargs.get('forward', True)
-
-    costheta, sintheta, cosphi, sinphi = trig_func(model)
-
-    delta, epsilon = model.delta, model.epsilon
-    epsilon = 1 + 2*epsilon
-    delta = sqrt(1 + 2*delta)
-
-    # Get source
-    qu = kwargs.get('qu', 0)
-    qv = kwargs.get('qv', 0)
-
-    if forward:
-        Gxx = Gxxyy_centered(model, u, costheta, sintheta, cosphi, sinphi, space_order)
-        Gzz = Gzz_centered(model, v, costheta, sintheta, cosphi, sinphi, space_order)
-        H0 = epsilon*Gxx + delta*Gzz
-        Hz = delta*Gxx + Gzz
-        return second_order_stencil(model, u, v, H0, Hz, qu, qv)
-    else:
-        H0 = Gxxyy_centered(model, (epsilon*u + delta*v), costheta, sintheta,
-                            cosphi, sinphi, space_order)
-        Hz = Gzz_centered(model, (delta*u + v), costheta, sintheta, cosphi,
-                          sinphi, space_order)
+        H0 = Gh_centered(model, (epsilon*u + delta*v))
+        Hz = Gzz(model, (delta*u + v))
         return second_order_stencil(model, u, v, H0, Hz, qu, qv, forward=forward)
 
 
@@ -320,7 +243,6 @@ def particle_velocity_fields(model, space_order):
         x, z = model.space_dimensions
         stagg_x = x
         stagg_z = z
-        x, z = model.grid.dimensions
         # Create symbols for forward wavefield, source and receivers
         vx = TimeFunction(name='vx', grid=model.grid, staggered=stagg_x,
                           time_order=1, space_order=space_order)
@@ -332,7 +254,6 @@ def particle_velocity_fields(model, space_order):
         stagg_x = x
         stagg_y = y
         stagg_z = z
-        x, y, z = model.grid.dimensions
         # Create symbols for forward wavefield, source and receivers
         vx = TimeFunction(name='vx', grid=model.grid, staggered=stagg_x,
                           time_order=1, space_order=space_order)
@@ -344,7 +265,7 @@ def particle_velocity_fields(model, space_order):
     return vx, vz, vy
 
 
-def kernel_staggered_2d(model, u, v, space_order, **kwargs):
+def kernel_staggered_2d(model, u, v, **kwargs):
     """
     TTI finite difference. The equation solved is:
 
@@ -353,6 +274,7 @@ def kernel_staggered_2d(model, u, v, space_order, **kwargs):
     m * v.dt = - sqrt(1 + 2 delta) vx.dx - vz.dz + Fh
     m * u.dt = - (1 + 2 epsilon) vx.dx - sqrt(1 + 2 delta) vz.dz + Fv
     """
+    space_order = u.space_order
     # Forward or backward
     forward = kwargs.get('forward', True)
 
@@ -373,14 +295,14 @@ def kernel_staggered_2d(model, u, v, space_order, **kwargs):
 
     if forward:
         # Stencils
-        phdx = costheta * u.dx - sintheta * u.dy
+        phdx = costheta * u.dx - sintheta * u.dyc
         u_vx = Eq(vx.forward, dampl * vx - dampl * s * phdx)
 
-        pvdz = sintheta * v.dx + costheta * v.dy
+        pvdz = sintheta * v.dxc + costheta * v.dy
         u_vz = Eq(vz.forward, dampl * vz - dampl * s * pvdz)
 
-        dvx = costheta * vx.forward.dx - sintheta * vx.forward.dy
-        dvz = sintheta * vz.forward.dx + costheta * vz.forward.dy
+        dvx = costheta * vx.forward.dx - sintheta * vx.forward.dyc
+        dvz = sintheta * vz.forward.dxc + costheta * vz.forward.dy
 
         # u and v equations
         pv_eq = Eq(v.forward, dampl * (v - s / m * (delta * dvx + dvz)) + s / m * qv)
@@ -388,16 +310,16 @@ def kernel_staggered_2d(model, u, v, space_order, **kwargs):
                    s / m * qu)
     else:
         # Stencils
-        phdx = ((costheta*epsilon*u).dx - (sintheta*epsilon*u).dy +
-                (costheta*delta*v).dx - (sintheta*delta*v).dy)
+        phdx = ((costheta*epsilon*u).dx - (sintheta*epsilon*u).dyc +
+                (costheta*delta*v).dx - (sintheta*delta*v).dyc)
         u_vx = Eq(vx.backward, dampl * vx + dampl * s * phdx)
 
-        pvdz = ((sintheta*delta*u).dx + (costheta*delta*u).dy +
-                (sintheta*v).dx + (costheta*v).dy)
+        pvdz = ((sintheta*delta*u).dxc + (costheta*delta*u).dy +
+                (sintheta*v).dxc + (costheta*v).dy)
         u_vz = Eq(vz.backward, dampl * vz + dampl * s * pvdz)
 
-        dvx = (costheta * vx.backward).dx - (sintheta * vx.backward).dy
-        dvz = (sintheta * vz.backward).dx + (costheta * vz.backward).dy
+        dvx = (costheta * vx.backward).dx - (sintheta * vx.backward).dyc
+        dvz = (sintheta * vz.backward).dxc + (costheta * vz.backward).dy
 
         # u and v equations
         pv_eq = Eq(v.backward, dampl * (v + s / m * dvz))
@@ -406,7 +328,7 @@ def kernel_staggered_2d(model, u, v, space_order, **kwargs):
     return [u_vx, u_vz] + [pv_eq, ph_eq]
 
 
-def kernel_staggered_3d(model, u, v, space_order, **kwargs):
+def kernel_staggered_3d(model, u, v, **kwargs):
     """
     TTI finite difference. The equation solved is:
 
@@ -416,6 +338,7 @@ def kernel_staggered_3d(model, u, v, space_order, **kwargs):
     m * v.dt = - sqrt(1 + 2 delta) (vx.dx + vy.dy) - vz.dz + Fh
     m * u.dt = - (1 + 2 epsilon) (vx.dx + vy.dy) - sqrt(1 + 2 delta) vz.dz + Fv
     """
+    space_order = u.space_order
     # Forward or backward
     forward = kwargs.get('forward', True)
 
@@ -533,19 +456,16 @@ def ForwardOperator(model, geometry, space_order=4,
     v = TimeFunction(name='v', grid=model.grid, staggered=stagg_v,
                      save=geometry.nt if save else None,
                      time_order=time_order, space_order=space_order)
-    src = PointSource(name='src', grid=model.grid, time_range=geometry.time_axis,
-                      npoint=geometry.nsrc)
-    rec = Receiver(name='rec', grid=model.grid, time_range=geometry.time_axis,
-                   npoint=geometry.nrec)
+    src = geometry.src
+    rec = geometry.rec
 
     # FD kernels of the PDE
     FD_kernel = kernels[(kernel, len(model.shape))]
-    stencils = FD_kernel(model, u, v, space_order)
+    stencils = FD_kernel(model, u, v)
 
     # Source and receivers
     expr = src * dt / m if kernel == 'staggered' else src * dt**2 / m
-    stencils += src.inject(field=u.forward, expr=expr)
-    stencils += src.inject(field=v.forward, expr=expr)
+    stencils += src.inject(field=(u.forward, v.forward), expr=expr)
     stencils += rec.interpolate(expr=u + v)
 
     # Substitute spacing terms to reduce flops
@@ -583,19 +503,16 @@ def AdjointOperator(model, geometry, space_order=4,
                      time_order=time_order, space_order=space_order)
     r = TimeFunction(name='r', grid=model.grid, staggered=stagg_r,
                      time_order=time_order, space_order=space_order)
-    srca = PointSource(name='srca', grid=model.grid, time_range=geometry.time_axis,
-                       npoint=geometry.nsrc)
-    rec = Receiver(name='rec', grid=model.grid, time_range=geometry.time_axis,
-                   npoint=geometry.nrec)
+    srca = geometry.new_src(name='srca', src_type=None)
+    rec = geometry.rec
 
     # FD kernels of the PDE
     FD_kernel = kernels[(kernel, len(model.shape))]
-    stencils = FD_kernel(model, p, r, space_order, forward=False)
+    stencils = FD_kernel(model, p, r, forward=False)
 
     # Construct expression to inject receiver values
     expr = rec * dt / m if kernel == 'staggered' else rec * dt**2 / m
-    stencils += rec.inject(field=p.backward, expr=expr)
-    stencils += rec.inject(field=r.backward, expr=expr)
+    stencils += rec.inject(field=(p.backward, r.backward), expr=expr)
 
     # Create interpolation expression for the adjoint-source
     stencils += srca.interpolate(expr=p + r)
@@ -626,11 +543,8 @@ def JacobianOperator(model, geometry, space_order=4,
     time_order = 2
 
     # Create source and receiver symbols
-    src = Receiver(name='src', grid=model.grid, time_range=geometry.time_axis,
-                   npoint=geometry.nsrc)
-
-    rec = Receiver(name='rec', grid=model.grid, time_range=geometry.time_axis,
-                   npoint=geometry.nrec)
+    src = geometry.src
+    rec = geometry.rec
 
     # Create wavefields and a dm field
     u0 = TimeFunction(name='u0', grid=model.grid, save=None, time_order=time_order,
@@ -645,17 +559,16 @@ def JacobianOperator(model, geometry, space_order=4,
 
     # FD kernels of the PDE
     FD_kernel = kernels[('centered', len(model.shape))]
-    eqn1 = FD_kernel(model, u0, v0, space_order)
+    eqn1 = FD_kernel(model, u0, v0)
 
     # Linearized source and stencil
     lin_usrc = -dm * u0.dt2
     lin_vsrc = -dm * v0.dt2
 
-    eqn2 = FD_kernel(model, du, dv, space_order, qu=lin_usrc, qv=lin_vsrc)
+    eqn2 = FD_kernel(model, du, dv, qu=lin_usrc, qv=lin_vsrc)
 
     # Construct expression to inject source values, injecting at u0(t+dt)/v0(t+dt)
-    src_term = src.inject(field=u0.forward, expr=src * dt**2 / m)
-    src_term += src.inject(field=v0.forward, expr=src * dt**2 / m)
+    src_term = src.inject(field=(u0.forward, v0.forward), expr=src * dt**2 / m)
 
     # Create interpolation expression for receivers, extracting at du(t)+dv(t)
     rec_term = rec.interpolate(expr=du + dv)
@@ -699,23 +612,21 @@ def JacobianAdjOperator(model, geometry, space_order=4,
 
     dm = Function(name="dm", grid=model.grid)
 
-    rec = Receiver(name='rec', grid=model.grid, time_range=geometry.time_axis,
-                   npoint=geometry.nrec)
+    rec = geometry.rec
 
     # FD kernels of the PDE
     FD_kernel = kernels[('centered', len(model.shape))]
-    eqn = FD_kernel(model, du, dv, space_order, forward=False)
+    eqn = FD_kernel(model, du, dv, forward=False)
 
     dm_update = Inc(dm, - (u0 * du.dt2 + v0 * dv.dt2))
 
     # Add expression for receiver injection
-    rec_term = rec.inject(field=du.backward, expr=rec * dt**2 / m)
-    rec_term += rec.inject(field=dv.backward, expr=rec * dt**2 / m)
+    rec_term = rec.inject(field=(du.backward, dv.backward), expr=rec * dt**2 / m)
 
     # Substitute spacing terms to reduce flops
     return Operator(eqn + rec_term + [dm_update], subs=model.spacing_map,
                     name='GradientTTI', **kwargs)
 
 
-kernels = {('centered', 3): kernel_centered_3d, ('centered', 2): kernel_centered_2d,
+kernels = {('centered', 3): kernel_centered, ('centered', 2): kernel_centered,
            ('staggered', 3): kernel_staggered_3d, ('staggered', 2): kernel_staggered_2d}
